Package: piar
Title: Price Index Aggregation
Version: 0.10.1
Authors@R: c(
    person("Steve", "Martin", role = c("aut", "cre", "cph"),
           email = "marberts@protonmail.com",
           comment = c(ORCID = "0000-0003-2544-9480"))
    )
Description: Most price indexes are made with a two-step procedure, where
    period-over-period elementary indexes are first calculated for a collection
    of elementary aggregates at each point in time, and then aggregated according
    to a price index aggregation structure. These indexes can then be chained
    together to form a time series that gives the evolution of prices with
    respect to a fixed base period. This package contains a collection of
    functions that revolve around this work flow, making it easy to build
    standard price indexes, and implement the methods described by
    Balk (2008, <doi:10.1017/CBO9780511720758>), von der Lippe (2007,
    <doi:10.3726/978-3-653-01120-3>), and the CPI manual (2020,
    <doi:10.5089/9781484354841.069>) and PPI manual
    (2004, <doi:10.5089/9781589063044.069>) for bilateral price indexes.
Depends: R (>= 4.1)
Imports: stats, utils, Matrix (>= 1.5-0)
Suggests: data.tree, litedown, rsmatrix, spelling, sps, tinytest,
        treemap
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://marberts.github.io/piar/, https://github.com/marberts/piar
BugReports: https://github.com/marberts/piar/issues
LazyData: true
VignetteBuilder: litedown
Config/roxygen2/version: 8.0.0
Language: en-US
NeedsCompilation: no
Packaged: 2026-07-30 19:00:47 UTC; steve
Author: Steve Martin [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-2544-9480>)
Maintainer: Steve Martin <marberts@protonmail.com>
Repository: CRAN
Date/Publication: 2026-07-30 23:30:02 UTC
Built: R 4.6.1; ; 2026-07-31 03:55:32 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",piar_index)
S3method("[<-",chainable_piar_index)
S3method("[<-",direct_piar_index)
S3method("[<-",piar_index)
S3method("levels<-",piar_aggregation_structure)
S3method("levels<-",piar_index)
S3method("split<-",piar_index)
S3method("time<-",piar_index)
S3method("weights<-",piar_aggregation_structure)
S3method("window<-",piar_index)
S3method(Math,piar_index)
S3method(Ops,piar_index)
S3method(Summary,piar_index)
S3method(aggregate,chainable_piar_index)
S3method(aggregate,direct_piar_index)
S3method(anyNA,piar_index)
S3method(as.data.frame,piar_aggregation_structure)
S3method(as.data.frame,piar_index)
S3method(as.double,piar_index)
S3method(as.list,piar_aggregation_structure)
S3method(as.matrix,piar_aggregation_structure)
S3method(as.matrix,piar_index)
S3method(as.ts,piar_index)
S3method(as_aggregation_structure,data.frame)
S3method(as_aggregation_structure,default)
S3method(as_aggregation_structure,matrix)
S3method(as_aggregation_structure,piar_aggregation_structure)
S3method(as_index,chainable_piar_index)
S3method(as_index,data.frame)
S3method(as_index,default)
S3method(as_index,direct_piar_index)
S3method(as_index,matrix)
S3method(as_index,mts)
S3method(chain,chainable_piar_index)
S3method(chain,default)
S3method(chain,direct_piar_index)
S3method(cut,piar_aggregation_structure)
S3method(elementary_index,data.frame)
S3method(elementary_index,default)
S3method(elementary_index,numeric)
S3method(end,piar_index)
S3method(head,piar_index)
S3method(impute_prices,data.frame)
S3method(impute_prices,default)
S3method(impute_prices,matrix)
S3method(impute_prices,numeric)
S3method(is.na,piar_index)
S3method(levels,piar_aggregation_structure)
S3method(levels,piar_index)
S3method(mean,chainable_piar_index)
S3method(mean,direct_piar_index)
S3method(merge,chainable_piar_index)
S3method(merge,direct_piar_index)
S3method(merge,piar_index)
S3method(price_relative,data.frame)
S3method(price_relative,default)
S3method(print,piar_aggregation_structure)
S3method(print,piar_index)
S3method(rebase,chainable_piar_index)
S3method(rebase,default)
S3method(rebase,direct_piar_index)
S3method(split,piar_index)
S3method(stack,chainable_piar_index)
S3method(stack,direct_piar_index)
S3method(stack,piar_index)
S3method(start,piar_index)
S3method(str,piar_aggregation_structure)
S3method(str,piar_index)
S3method(summary,chainable_piar_index)
S3method(summary,direct_piar_index)
S3method(summary,piar_aggregation_structure)
S3method(tail,piar_index)
S3method(time,piar_index)
S3method(unchain,chainable_piar_index)
S3method(unchain,default)
S3method(unchain,direct_piar_index)
S3method(unstack,chainable_piar_index)
S3method(unstack,direct_piar_index)
S3method(unstack,piar_index)
S3method(update,piar_aggregation_structure)
S3method(weights,piar_aggregation_structure)
S3method(window,piar_index)
export("contrib<-")
export("time<-")
export("weights<-")
export(aggregation_structure)
export(as_aggregation_structure)
export(as_index)
export(back_period)
export(carry_backward)
export(carry_forward)
export(chain)
export(combine_classifications)
export(contrib)
export(contrib2DF)
export(elemental_index)
export(elementary_index)
export(emean)
export(expand_classification)
export(geks_index)
export(gmean)
export(hb_transform)
export(impute_prices)
export(interact_classifications)
export(is_aggregation_structure)
export(is_chainable_index)
export(is_direct_index)
export(is_index)
export(nested_gmean)
export(ntime)
export(outliers)
export(price_relative)
export(rebase)
export(scale_weights)
export(set_contrib)
export(set_contrib_from_index)
export(set_levels)
export(set_time)
export(set_weights)
export(shadow_price)
export(splice_index)
export(split_classification)
export(transmute_weights)
export(transmute_weights2)
export(unchain)
export(update_weights)
importFrom(stats,"window<-")
importFrom(stats,aggregate)
importFrom(stats,as.ts)
importFrom(stats,end)
importFrom(stats,start)
importFrom(stats,time)
importFrom(stats,ts)
importFrom(stats,update)
importFrom(stats,weights)
importFrom(stats,window)
importFrom(utils,head)
importFrom(utils,stack)
importFrom(utils,str)
importFrom(utils,tail)
importFrom(utils,unstack)

<!-- README.md is generated from README.Rmd. Please edit that file. -->

# Price Index Aggregation in R <a href="https://marberts.github.io/piar/"><img src="man/figures/logo.png" align="right" height="139" alt="piar website" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/piar)](https://cran.r-project.org/package=piar)
[![piar status
badge](https://marberts.r-universe.dev/badges/piar)](https://marberts.r-universe.dev/piar)
[![Conda
Version](https://img.shields.io/conda/vn/conda-forge/r-piar.svg)](https://anaconda.org/conda-forge/r-piar)
[![R-CMD-check](https://github.com/marberts/piar/workflows/R-CMD-check/badge.svg)](https://github.com/marberts/piar/actions)
[![codecov](https://codecov.io/gh/marberts/piar/graph/badge.svg?token=lHDHsGHsLd)](https://app.codecov.io/gh/marberts/piar)
[![DOI](https://zenodo.org/badge/370889275.svg)](https://zenodo.org/doi/10.5281/zenodo.10110046)
[![Mentioned in Awesome Official
Statistics](https://awesome.re/mentioned-badge.svg)](https://github.com/SNStatComp/awesome-official-statistics-software)
[![DOI](https://joss.theoj.org/papers/10.21105/joss.06781/status.svg)](https://doi.org/10.21105/joss.06781)
<!-- badges: end -->

Most price indexes are made with a two-step procedure, where
period-over-period *elementary indexes* are first calculated for a
collection of *elementary aggregates* at each point in time, and then
aggregated according to a *price index aggregation structure*. These
indexes can then be chained together to form a time series that gives
the evolution of prices with respect to a fixed base period. This
package contains a collection of functions that revolve around this work
flow, making it easy to build standard price indexes, and implement the
methods described by Balk (2008), von der Lippe (2007), and the CPI
manual (2020) / PPI manual (2004) for bilateral price indexes.

The tools in this package are designed to be useful for both researching
new sources of data and methods to construct price indexes, and the
regular production of price statistics. It is targeted towards
economists, statisticians, and data scientists working at national
statistical agencies, central banks, financial institutions, and in
academia that want to measure and study the evolution of prices over
time.

## Installation

Get the stable version from CRAN.

``` r
install.packages("piar")
```

The development version can be installed from R-Universe

``` r
install.packages(
  "piar",
  repos = c("https://marberts.r-universe.dev", "https://cloud.r-project.org")
)
```

or directly from Github.

``` r
pak::pak("marberts/piar")
```

## Usage

There are several detailed vignette showing how to use **piar**:
`browseVignettes("piar")`. But the basic work flow is fairly simple.

The starting point is to make period-over-period elementary price
indexes with the `elementary_index()` function.

``` r
library(piar)

# Make Jevons business-level elementary indexes

head(ms_prices)
#>   period business product price
#> 1 202001       B1       1  1.14
#> 2 202001       B1       2    NA
#> 3 202001       B1       3  6.09
#> 4 202001       B2       4  6.23
#> 5 202001       B2       5  8.61
#> 6 202001       B2       6  6.40

elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)

elementals
#> Period-over-period price index for 4 levels over 4 time periods 
#>       time
#> levels 202001    202002    202003   202004
#>     B1      1 0.8949097 0.3342939      NaN
#>     B2      1       NaN       NaN 2.770456
#>     B3      1 2.0200036 1.6353355 0.537996
#>     B4    NaN       NaN       NaN 4.576286
```

And an aggregation structure.

``` r
# Make an aggregation structure from businesses to higher-level
# industrial classifications

ms_weights
#>   business classification weight
#> 1       B1             11    553
#> 2       B2             11    646
#> 3       B3             11    312
#> 4       B4             12    622
#> 5       B5             12    330

ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")]

pias
#>   level1 level2 business weight
#> 1      1     11       B1    553
#> 2      1     11       B2    646
#> 3      1     11       B3    312
#> 4      1     12       B4    622
#> 5      1     12       B5    330
```

The `aggregate()` method can then be used to aggregate the elementary
indexes according to the aggregation structure (the first three rows
below) and fill in missing elementary indexes while maintaining
consistency in aggregation. There are a variety of methods to work with
these index objects, such as chaining them over time.

``` r
# Aggregate elementary indexes with an arithmetic index

index <- aggregate(elementals, pias, na.rm = TRUE)

# Chain them to get a time series

chain(index)
#> Fixed-base price index for 8 levels over 4 time periods 
#>       time
#> levels 202001    202002    202003    202004
#>     1       1 1.3007239 1.3827662 3.7815355
#>     11      1 1.3007239 1.3827662 2.1771866
#>     12      1 1.3007239 1.3827662 6.3279338
#>     B1      1 0.8949097 0.2991629 0.4710366
#>     B2      1 1.3007239 1.3827662 3.8308934
#>     B3      1 2.0200036 3.3033836 1.7772072
#>     B4      1 1.3007239 1.3827662 6.3279338
#>     B5      1 1.3007239 1.3827662 6.3279338
```

## Contributing

All contributions are welcome. Please start by opening an issue on
GitHub to report any bugs or suggest improvements and new features. See
the contribution guidelines for this project for more information.

## References

Balk, B. M. (2008). *Price and Quantity Index Numbers*. Cambridge
University Press. <https://doi.org/10.1017/CBO9780511720758>.

ILO, IMF, UNECE, OECD, and World Bank. (2004). *Producer Price Index
Manual: Theory and Practice*. International Monetary Fund.
<https://doi.org/10.5089/9781589063044.069>.

IMF, ILO, Eurostat, UNECE, OECD, and World Bank. (2020). *Consumer Price
Index Manual: Concepts and Methods*. International Monetary Fund.
<https://doi.org/10.5089/9781484354841.069>.

von der Lippe, P. (2007). *Index Theory and Price Statistics*. Peter
Lang. <https://doi.org/10.3726/978-3-653-01120-3>

## -----------------------------------------------------------------------------
library(piar)

elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)

## -----------------------------------------------------------------------------
ms_weights$stratum <- c("TS", "TA", "TS", "TS", "TS")

ms_weights

## -----------------------------------------------------------------------------
classification_sps <- paste0(ms_weights$classification, ms_weights$stratum) |>
  expand_classification(width = c(1, 1, 2))

pias_sps <- aggregation_structure(
  c(classification_sps, list(ms_weights$business)),
  ms_weights$weight
)

pias_sps

## -----------------------------------------------------------------------------
index_sps <- aggregate(elementals, pias_sps, na.rm = TRUE)

index_sps

## -----------------------------------------------------------------------------
interacted_hierarchy <- interact_classifications(
  expand_classification(ms_weights$classification),
  expand_classification(ms_weights$stratum)
)

pias_sps2 <- lapply(
  interacted_hierarchy,
  \(x) aggregation_structure(c(x, list(ms_weights$business)), ms_weights$weight)
)

index_sps2 <- lapply(pias_sps2, \(x) {
  aggregate(index_sps, x, include_ea = FALSE)
})

## -----------------------------------------------------------------------------
Reduce(merge, index_sps2)

library(piar)

# Make an aggregation structure.
ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

# Make a fixed-base index.
elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product),
    business = factor(business, levels = ms_weights$business)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)

index <- elementals |>
  aggregate(pias, na.rm = TRUE) |>
  chain()

index

pias_matrix <- as.matrix(pias)

pias_matrix

pias_matrix %*% as.matrix(index[levels(pias)$business])

pias_matrix <- as.matrix(pias) > 0

pias_matrix %*% is.na(elementals) / rowSums(pias_matrix)

as.matrix(pias, sparse = TRUE)

library(piar)

set.seed(12345)

# Make 6 elementary indexes over 4 time periods.
elementals <- matrix(c(rep(1, 6), runif(6 * 3, 0.8, 1.2)), nrow = 6) |>
  as_index() |>
  set_levels(paste0("B", 1:6))

head(elementals)

# Make aggregation weights over 4 time periods.
#            1
#      |-----+-----|
#      11          12
#  |---+---|   |---+---|
#  B1  B2  B3  B4  B5  B6

weights <- data.frame(
  level1 = 1,
  level2 = rep(11:12, each = 3),
  ea = levels(elementals),
  weights = runif(4 * 6, 100, 200),
  period = rep(1:4, each = 6)
)

head(weights)

elementals <- unstack(elementals)

paasche_pias <- split(
  weights[c("level1", "level2", "ea", "weights")],
  weights[["period"]]
) |>
  lapply(as_aggregation_structure)

paasche <- Map(
  aggregate,
  elementals,
  paasche_pias,
  na.rm = TRUE,
  include_ea = FALSE,
  order = -1
) |>
  Reduce(stack, x = _)

paasche

laspeyres_pias <- paasche_pias[c(1, 1, 2, 3)]

fisher <- Map(
  aggregate,
  elementals,
  pias = laspeyres_pias,
  pias2 = paasche_pias,
  na.rm = TRUE,
  include_ea = FALSE
) |>
  Reduce(stack, x = _)

fisher

laspeyres <- Map(
  aggregate,
  elementals,
  pias = laspeyres_pias,
  na.rm = TRUE,
  include_ea = FALSE
) |>
  Reduce(stack, x = _)

sqrt(as.matrix(laspeyres) * as.matrix(paasche))

set.seed(54321)

library(piar)

# Make an aggregation structure.
#                  1
#      |-----------|-----------|
#      11          12          13
#  |---+---|   |---+---|   |---+---|
# 111     121 121     122 131     132

pias <- data.frame(
  level1 = rep(1, 12),
  level2 = rep(c(11, 12, 13), each = 4),
  level3 = rep(c(111, 112, 121, 122, 131, 132), each = 2),
  ea = sprintf("B%02d", 1:12),
  weight = 1:12
) |>
  as_aggregation_structure()

pias

# Make elementary indexes over 4 quarters.
elementals <- matrix(
  runif(12 * 4, 0.4, 1.2),
  nrow = 12,
  dimnames = list(sprintf("B%02d", 1:12), paste0("Q", 1:4))
) |>
  as_index()

elementals

weights(pias) / rowMeans(as.matrix(chain(elementals)))

# Function to adjust annual weights.
adjust_weights <- function(
  index,
  pias,
  tol = .Machine$double.eps^0.5,
  max_iter = 100
) {
  adj_pias <- pias
  for (i in seq_len(max_iter)) {
    # Parentally impute missing elementary indexes.
    agg_index <- aggregate(index, adj_pias, na.rm = TRUE, contrib = FALSE)
    elementals <- chain(agg_index[levels(pias)[[nlevels(pias)]]])
    # Compute annual elementary indexes.
    pb <- rowMeans(as.matrix(elementals))
    # Stop if average price-update weights are within tolerance of original
    # weights; adjust otherwise.
    if (max(abs(pb * weights(adj_pias) - weights(pias))) < tol) {
      message(gettextf("Converged after %d iterations", i - 1))
      return(adj_pias)
    } else {
      weights(adj_pias) <- weights(pias) / pb
    }
  }
  warning("weights adjustment did not converge")
  adj_pias
}

elementals[11:12] <- NA

adjust_weights(elementals, pias)

library(piar)

prices <- data.frame(
  period = rep(1:3, each = 6),
  product = paste0("P", 1:6),
  business = rep(c("B1", "B2"), each = 3),
  price = 1:18,
  quantity = 18:1
)

prices[c("back_price", "back_quantity")] <-
  prices[back_period(prices$period, prices$product), c("price", "quantity")]

head(prices)

prices |>
  elementary_index(price / back_price ~ period + business, order = 1)

prices |>
  elementary_index(price / back_price ~ period + business, order = -1)

prices |>
  elementary_index(
    price / back_price ~ period + business,
    weights = back_price * back_quantity
  )

period_by_business <- interaction(prices$period, prices$business)

tornqvist_weights <- split(prices, ~period_by_business) |>
  lapply(\(df) {
    0.5 * scale_weights(df$price * df$quantity) +
      0.5 * scale_weights(df$back_price * df$back_quantity)
  }) |>
  unsplit(period_by_business)

prices |>
  elementary_index(
    price / back_price ~ period + business,
    weights = tornqvist_weights
  )

fisher_weights <- split(prices, ~period_by_business) |>
  lapply(\(df) {
    transmute_weights2(
      df$price / df$back_price,
      list(df$back_price * df$back_quantity, df$price * df$quantity),
      to = 0
    )
  }) |>
  unsplit(period_by_business)

prices |>
  elementary_index(
    price / back_price ~ period + business,
    weights = fisher_weights
  )

fisher_index <- prices |>
  elementary_index(
    price / back_price ~ period + business,
    weights = fisher_weights,
    contrib = TRUE
  )

contrib(fisher_index, "B1")

diewert_contributions <- function(p1, p0, q1, q0) {
  rel <- p1 / p0
  v0 <- scale_weights(p0 * q0)
  v1 <- scale_weights(p0 * q1)
  laspeyres <- gmean(rel, v0)
  fisher <- nested_gmean(rel, list(v0, v1), order = c(1, 1))

  (v0 + laspeyres * v1) / (1 + fisher) * (rel - 1)
}

contrib(fisher_index, "B1") <- subset(prices, business == "B1") |>
  split(~period) |>
  sapply(
    \(df) {
      diewert_contributions(
        df$price,
        df$back_price,
        df$quantity,
        df$back_quantity
      )
    }
  )

contrib(fisher_index, "B1")

# ---- Arithmetic mean ----
# Carli
carli <- \(p1, p0, q1, q0) gmean(p1 / p0)

# Dutot
dutot <- \(p1, p0, q1, q0) gmean(p1 / p0, p0)

# Laspeyres
laspeyres <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * q0)

# Palgrave
palgrave <- \(p1, p0, q1, q0) gmean(p1 / p0, p1 * q1)

# Walsh-1
walsh1 <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * sqrt(q0 * q1))

# Marshall-Edgeworth
marshall_edgeworth <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * (q0 * q1))

# Geay-Khamis
geary_khamis <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 / (1 / q0 + 1 / q1))

# Hybrid CSWD
hybrid_cswd <- \(p1, p0, q1, q0) gmean(p1 / p0, sqrt(p0 / p1))

# Martini
martini <- \(p1, p0, q1, q0, a) gmean(p1 / p0, p0 * q0 * (q1 / p0)^a)

# ---- Geometric mean ----
# Jevons
jevons <- \(p1, p0, q1, q0) gmean(p1 / p0, order = 0)

# Geometric Laspeyres (Jöhr)
geo_laspeyres <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * q0, order = 0)

# Geometric Paasche
geo_paasche <- \(p1, p0, q1, q0) gmean(p1 / p0, p1 * q1, order = 0)

# Walsh-2
walsh2 <- \(p1, p0, q1, q0) gmean(p1 / p0, sqrt(p0 * p1 * q0 * q1), order = 0)

# Theil
theil <- \(p1, p0, q1, q0) {
  w0 <- scale_weights(p0 * q0)
  w1 <- scale_weights(p1 * q1)
  gmean(p1 / p0, ((w0 + w1) / 2 * w0 * w1)^(1 / 3), order = 0)
}

# Rao
rao <- \(p1, p0, q1, q0) {
  w0 <- scale_weights(p0 * q0)
  w1 <- scale_weights(p1 * q1)
  gmean(p1 / p0, w0 * w1 / (w0 + w1), order = 0)
}

# Sato-Vartia
sato_vartia <- function(p1, p0, q1, q0) {
  v0 <- scale_weights(p0 * q0)
  v1 <- scale_weights(p1 * q1)
  gmean(p1 / p0, emean(v0, v1), order = 0)
}

# ---- Harmonic mean ----
# Coggeshall
coggeshall <- \(p1, p0, q1, q0) gmean(p1 / p0, order = -1)

# Paasche
paasche <- \(p1, p0, q1, q0) gmean(p1 / p0, p1 * q1, order = -1)

# Harmonic Laspeyres
harm_laspeyres <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * q0, order = -1)

# ---- Generalized mean ----
# Lloyd-Moulton
lloyd_moulton <- \(p1, p0, q1, q0, sigma) {
  gmean(p1 / p0, p0 * q0, order = 1 - sigma)
}

# ---- Nested means ----
# Drobisch (Sidgwick)
drobisch <- \(p1, p0, q1, q0) {
  nested_gmean(p1 / p0, list(p0 * q0, p1 * q1), outer_order = 1)
}

# Unnamed
unnamed <- \(p1, p0, q1, q0) {
  nested_gmean(
    p1 / p0,
    list(p0 * q0, p1 * q1),
    order = c(1, 1),
    outer_order = 1
  )
}

# Törnqvist-Theil
tornqvist <- \(p1, p0, q1, q0) {
  nested_gmean(
    p1 / p0,
    list(p0 * q0, p1 * q1),
    order = c(0, 0)
  )
}

# Fisher
fisher <- \(p1, p0, q1, q0) nested_gmean(p1 / p0, list(p0 * q0, p1 * q1))

# CSWD
cswd <- \(p1, p0, q1, q0) nested_gmean(p1 / p0)

# Balk-Walsh
balk_walsh <- \(p1, p0, q1, q0) nested_gmean(p1 / p0, order = c(0.5, -0.5))

# Geometric AG mean
ag_mean <- \(p1, p0, q1, q0, elasticity) {
  nested_gmean(
    p1 / p0,
    list(p0 * q0, p0 * q0),
    order = c(0, 1),
    outer_weights = c(elasticity, 1 - elasticity)
  )
}

library(piar)

# Make an aggregation structure.
ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

# Make elementary index.
elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)

elementals

set.seed(12345)

scanner_prices <- data.frame(
  period = rep(c("201904", time(elementals)), each = 200),
  product = 1:200,
  price = round(rlnorm(5 * 200) * 10, 1),
  quantity = round(runif(5 * 200, 100, 1000))
)

head(scanner_prices)

geks_elementals <- with(
  scanner_prices,
  geks_index(price, quantity, period, product, window = 3)
) |>
  splice_index() |>
  t() |>
  as_index(chainable = FALSE) |>
  set_levels("B5") |>
  rebase(base = "202001")

geks_elementals

merge(elementals, geks_elementals) |>
  aggregate(pias, na.rm = TRUE)

library(piar)

elementals <- ms_prices |>
  transform(
    imputed_price = impute_prices(
      price,
      period = period,
      product = product,
      method = "carry-forward"
    )
  ) |>
  elementary_index(
    price_relative(imputed_price, period = period, product = product) ~
      period + business,
    na.rm = TRUE
  )

elementals

elementals2 <- elementals
elementals2["B4", 1:3] <- 1

elementals2

ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

aggregate(elementals2, pias, na.rm = TRUE)

impute <- function(x, pias) {
  if (is.na(x["B4"])) x["B4"] <- 1
  x
}

aggregate(elementals, pias, na.rm = TRUE, impute_rules = impute)

library(piar)

set.seed(12345)

# Make quarterly elementary indexes for two baskets with a 1 year overlap.

elementals1 <- matrix(
  runif(5 * 8, 0.8, 1.2),
  nrow = 5,
  dimnames = list(paste0("B", 1:5), NULL)
) |>
  as_index()

elementals2 <- matrix(
  runif(6 * 8, 0.8, 1.2),
  nrow = 6,
  dimnames = list(paste0("B", 1:6), 5:12)
) |>
  as_index()

# Make aggregation weights for basket 1.
#            1
#      |-----+-----|
#      11          12
#  |---+---|   |---+---|
#  B1  B2  B3  B4      B5

weights1 <- data.frame(
  level1 = 1,
  level2 = c(11, 11, 11, 12, 12),
  ea = levels(elementals1),
  weights = runif(5, 100, 200)
)

# Make aggregation weights for basket 2.
#            1
#      |-----+-----|
#      11          12
#  |---+---|   |---+---|
#  B1  B2  B3  B4  B5  B6

weights2 <- data.frame(
  level1 = 1,
  level2 = c(11, 11, 11, 12, 12, 12),
  ea = levels(elementals2),
  weights = runif(6, 100, 200)
)

index <- Map(
  aggregate,
  list(elementals1, elementals2),
  list(weights1, weights2),
  include_ea = FALSE
)

stack(index[[1]], window(index[[2]], start = "9")) |>
  chain()

link_factor <- chain(index[[1]]) |>
  window(start = end(index[[1]])) |>
  as.numeric()

stack(
  chain(index[[1]]),
  chain(window(index[[2]], start = "9"), link = link_factor)
)

index <- index |>
  lapply(chain) |>
  lapply(\(x) rebase(x, base = mean(x[, 1:4])))

link_factor <- window(index[[1]], start = "5") |>
  mean() |>
  as.numeric()

stack(
  index[[1]],
  rebase(window(index[[2]], start = "9"), base = 1 / link_factor)
)

index[[1]] <- rebase(index[[1]], base = mean(window(index[[1]], start = "5")))

link_factor <- as.numeric(index[[1]][, "8"]) / as.numeric(index[[2]][, "8"])

stack(
  index[[1]],
  rebase(window(index[[2]], start = "9"), base = 1 / link_factor)
)

library(piar)

p2 <- c(1.2, 3, 1.3, 0.7, 1.4, 0.8)
p1 <- 1
q2 <- c(0.8, 0.9, 1.9, 1.3, 4.7, 0.6)
q1 <- c(1, 1, 2, 1, 4.5, 0.5)

rel <- p2 / p1

s1 <- scale_weights(p1 * q1)
s2 <- scale_weights(p2 * q2)

v <- transmute_weights(rel, s1, order = 2)

all.equal(gmean(rel, s1, order = 2), gmean(rel, v))

(v - s1)[order(rel)]

all.equal(
  transmute_weights(rel, v, order = 1, to = 0),
  transmute_weights(rel, s1, order = 2, to = 0)
)

v1 <- transmute_weights2(rel, list(s1, s2))

all.equal(nested_gmean(rel, list(s1, s2)), gmean(rel, v1))

all.equal(
  v1,
  transmute_weights(
    rel,
    transmute_weights2(rel, list(s1, s2), to = 2),
    order = 2
  )
)

v2 <- transmute_weights2(rel, list(s1, s2), pivot = 1)

all.equal(nested_gmean(rel, list(s1, s2)), gmean(rel, v2))

all.equal(
  v2,
  transmute_weights(
    rel,
    transmute_weights2(rel, list(s1, s2), to = 2, pivot = 1),
    order = 2
  )
)

summary(v1 - v2)

group <- rep(c("a", "b"), each = 3)

s1_by_group <- split(s1, group)
rel_by_group <- split(rel, group)

index_a <- gmean(rel_by_group$a, s1_by_group$a, order = 2)
index_b <- gmean(rel_by_group$b, order = 0)

gmean(c(index_a, index_b), sapply(s1_by_group, sum), order = 2)

decomp_a <- transmute_weights(rel_by_group$a, s1_by_group$a, order = 2)
decomp_b <- transmute_weights(rel_by_group$b, order = 0)

v <- Map(
  `*`,
  transmute_weights(c(index_a, index_b), sapply(s1_by_group, sum), order = 2),
  list(decomp_a, decomp_b)
) |>
  unlist()

gmean(rel, v)

V <- sum(p2 * q2) / sum(p1 * q1)

v <- transmute_weights2(rel, list(s1, s2), to = -1)

all.equal(
  gmean(V / rel, v),
  nested_gmean(q2 / q1, list(s1, s2))
)

V / rel * v

contraharmonic_mean <- function(x, weights, order) {
  gmean(x, weights * x)
}

# Arithmetic hybrid index
all.equal(
  gmean(p2 / p1, p2 * q1),
  contraharmonic_mean(p2 / p1, p1 * q1)
)

# Palgrave index
all.equal(
  gmean(p2 / p1, p2 * q2),
  contraharmonic_mean(p2 / p1, p1 * q2)
)

library(piar)

head(ms_prices)

ms_weights

elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)

elementals

elementals[, "202004"]

elementals[c("B1", "B3"), ]

ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

index <- aggregate(elementals, pias, na.rm = TRUE)

index

chained_index <- chain(index)

chained_index

rebase(chained_index, base = chained_index[, "202004"])

as.matrix(chained_index)

as.data.frame(chained_index)

update(pias, index) |>
  as.data.frame()

library(piar)

set.seed(12345)

# Make indexes for 6 basic headings for 4 countries.
bh_index <- matrix(
  c(rep(1, 6), runif(6 * 3, 0.8, 1.2)),
  nrow = 6,
  dimnames = list(
    paste0("BH", 1:6),
    paste("Country", 1:4)
  )
) |>
  as_index(chainable = FALSE)

head(bh_index)

# Make fixed aggregation weights.
#            1
#      |-----+-----|
#      11          12
#  |---+---|   |---+---|
#  B1  B2  B3  B4  B5  B6

weights <- data.frame(
  level1 = 1,
  level2 = rep(11:12, each = 3),
  bh = levels(bh_index),
  weights = runif(6, 100, 200)
)

head(weights)

index <- aggregate(bh_index, weights)

as.matrix(index[1, -1])

update_factors <- runif(3, 0.8, 1.2)

as.matrix(index[1, -1]) * update_factors

set.seed(54321)

library(piar)

# Make an aggregation structure.
#                  1
#      |-----------|-----------|
#      11          12          13
#  |---+---|   |---+---|   |---+---|
# 111     121 121     122 131     132

pias <- data.frame(
  level1 = rep(1, 12),
  level2 = rep(c(11, 12, 13), each = 4),
  level3 = rep(c(111, 112, 121, 122, 131, 132), each = 2),
  ea = sprintf("B%02d", 1:12),
  weight = 1:12
) |>
  as_aggregation_structure()

pias

# Make elementary indexes over 3 years and aggregate.
quarterly_index <- matrix(
  runif(12 * 12, 0.4, 1.2),
  nrow = 12,
  dimnames = list(sprintf("B%02d", 1:12), paste0("Q", 1:12))
) |>
  as_index() |>
  aggregate(pias)

head(quarterly_index)

annual_index <- chain(quarterly_index) |>
  mean(window = 4)

annual_index |>
  rebase(base = "Q1") |>
  head()

annual_pias <- pias |>
  update(annual_index, period = "Q1")

annual_index <- annual_index |>
  rebase(base = "Q1")

annual_index |>
  aggregate(annual_pias) |>
  all.equal(annual_index)

annual_index |>
  unchain() |>
  set_contrib_from_index() |>
  aggregate(cut(annual_pias, 2)) |>
  contrib()

library(piar)

# Make an aggregation structure.
ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

# Make elementary index with contributions.
elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(
    relative ~ period + business,
    product = product,
    na.rm = TRUE,
    contrib = TRUE
  )

contrib(elementals, level = "B1")

contrib2DF(elementals, level = "B1")

index <- aggregate(elementals, pias, na.rm = TRUE)

contrib(index)

index <- as_index(as.matrix(index), contrib = TRUE)

set_contrib_from_index(index) |>
  aggregate(cut(pias, 2)) |>
  contrib()

chain(index) |>
  set_contrib_from_index() |>
  aggregate(cut(pias, 2)) |>
  contrib()

#---- Class generator ----
new_piar_aggregation_structure <- function(child, parent, levels, weights) {
  stopifnot(is.list(child))
  stopifnot(is.list(parent))
  stopifnot(is.list(levels))
  stopifnot(is.double(weights))
  res <- list(
    child = child,
    parent = parent,
    levels = levels,
    weights = weights
  )
  structure(res, class = "piar_aggregation_structure")
}

piar_aggregation_structure <- function(child, parent, levels, weights) {
  levels <- lapply(as.list(levels), as.character)
  weights <- as.numeric(weights)
  validate_piar_aggregation_structure(
    new_piar_aggregation_structure(child, parent, levels, weights)
  )
}

#---- Validator ----
# The first two validators are not extensively used and duplicate checking
# already done by aggregation_structure(). They're here in case a validator
# is needed in the future.
validate_pias_levels <- function(x) {
  lev <- unlist(x$levels, use.names = FALSE)
  if (missing_names(lev)) {
    stop(
      "cannot make an aggregation structure with missing or length-zero levels"
    )
  }
  if (anyDuplicated(lev)) {
    stop("cannot make an aggregation structure with duplicated levels")
  }
  x
}

validate_pias_structure <- function(x) {
  height <- length(x$levels)
  if (
    height != length(x$child) + 1L ||
      height != length(x$parent) + 1L ||
      anyNA(x$child, recursive = TRUE) ||
      anyNA(x$parent, recursive = TRUE) ||
      any(vapply(x$child, \(x) any(lengths(x) == 0L), logical(1L)))
  ) {
    stop(
      "invalid aggregation structure; the input is likely not a nested ",
      "hierachy"
    )
  }
  x
}

validate_pias_weights <- function(x) {
  if (length(x$weights) != length(last(x$levels))) {
    stop(
      "cannot make an aggregation structure with a different number of ",
      "weights and elementary aggregates"
    )
  }
  if (any(x$weights < 0, na.rm = TRUE)) {
    stop("cannot make an aggregation structure with negative weights")
  }
  x
}

validate_piar_aggregation_structure <- function(x) {
  validate_pias_levels(x)
  validate_pias_structure(x)
  validate_pias_weights(x)
}

#---- Undocumented methods ----
#' @export
summary.piar_aggregation_structure <- function(object, ...) {
  chkDots(...)
  cat(
    "Aggregation structure for",
    length(last(object$levels)),
    "elementary aggregates with",
    length(object$levels) - 1L,
    "levels above the elementary aggregates",
    "\n"
  )
  invisible()
}

#' @export
print.piar_aggregation_structure <- function(x, ...) {
  summary(x)
  print(as.data.frame(x, fix.empty.names = FALSE), ...)
  invisible(x)
}

#' @export
str.piar_aggregation_structure <- function(object, ...) {
  str(unclass(object), ...)
}

#' Test if an object is an aggregation structure
#'
#' Test if an object is a price index aggregation structure.
#'
#' @param x `[object]` An object to test.
#'
#' @returns
#' Returns `TRUE` if `x` inherits from [`piar_aggregation_structure`].
#'
#' @export
is_aggregation_structure <- function(x) {
  inherits(x, "piar_aggregation_structure")
}

#' Price data
#'
#' Sample price and weight data for both a match sample and fixed sample type
#' index.
#'
#' @name price_data
#' @aliases ms_prices ms_weights fs_prices fs_weights
#' @docType data
NULL

#' Price index objects
#'
#' There are several classes to represent price indexes.
#' - All indexes inherit from the `piar_index` virtual class.
#' - Period-over-period indexes that can be chained over time inherit from
#' `chainable_piar_index`.
#' - Fixed-base indexes inherit from `direct_piar_index`.
#'
#' The `piar_index` object is a list-S3 class with the following
#' components:
#' \describe{
#' \item{index}{A matrix of index values with a column for each period in `time`
#' and a row for each level in `levels`.}
#' \item{contrib}{A list-matrix containing named vectors that give the
#' percent-change contributions for each price relative with a column for each
#' time period in `time` and a row for each level in `levels`, or `NULL`.}
#' \item{levels}{A character vector giving the levels of the index.}
#' \item{time}{A character vector giving the time periods for the index.}
#' }
#'
#' The `chainable_piar_index` and `direct_piar_index` subclasses have
#' the same structure as the `piar_index` class, but differ in the methods
#' used to manipulate the indexes.
#'
#' @name piar_index
#' @aliases piar_index chainable_piar_index direct_piar_index
#'
NULL

#---- Class generator ----
new_piar_index <- function(index, contrib, levels, time, chainable) {
  stopifnot(is.double(index) && is.matrix(index))
  stopifnot(is.list(contrib) && is.matrix(contrib) || is.null(contrib))
  stopifnot(is.character(levels))
  stopifnot(is.character(time))
  res <- list(index = index, contrib = contrib, levels = levels, time = time)
  type <- if (chainable) "chainable_piar_index" else "direct_piar_index"
  structure(res, class = c(type, "piar_index"))
}

piar_index <- function(index, contrib, levels, time, chainable) {
  index <- structure(as.numeric(index), dim = dim(index))
  if (!is.null(contrib)) {
    contrib <- structure(as.list(contrib), dim = dim(contrib))
  }
  levels <- as.character(levels)
  time <- as.character(time)
  validate_piar_index(
    new_piar_index(index, contrib, levels, time, chainable)
  )
}

#---- Validator ----
validate_levels <- function(x) {
  if (length(x$levels) == 0L) {
    stop("cannot make an index with no levels")
  }
  if (missing_names(x$levels)) {
    stop("cannot make an index with missing or length-zero levels")
  }
  if (anyDuplicated(x$levels)) {
    stop("cannot make an index with duplicate levels")
  }
  if (nrow(x$index) != length(x$levels)) {
    stop("number of levels does not agree with number of index values")
  }
  x
}

validate_time <- function(x) {
  if (length(x$time) == 0L) {
    stop("cannot make an index with no time periods")
  }
  if (missing_names(x$time)) {
    stop("cannot make an index with missing or length-zero time periods")
  }
  if (anyDuplicated(x$time)) {
    stop("cannot make an index with duplicate time periods")
  }
  if (ncol(x$index) != length(x$time)) {
    stop("number of time periods does not agree with number of index values")
  }
  x
}

validate_index_values <- function(x) {
  if (any(x$index <= 0, na.rm = TRUE)) {
    stop("cannot make an index with non-positive values")
  }
  x
}

validate_contrib <- function(x) {
  if (!is.null(x$contrib)) {
    if (ncol(x$contrib) != length(x$time)) {
      stop("number of time periods does not agree with number of contributions")
    }
    if (nrow(x$contrib) != length(x$levels)) {
      stop("number of levels does not agree with number of contributions")
    }
  }
  x
}

validate_index_structure <- function(x) {
  validate_levels(x)
  validate_time(x)
  validate_contrib(x)
}

validate_piar_index <- function(x) {
  validate_index_values(x)
  validate_index_structure(x)
}

#---- Undocumented methods ----
#' @importFrom utils str
#' @export
str.piar_index <- function(object, ...) {
  str(unclass(object), ...)
}

#' @export
summary.chainable_piar_index <- function(object, ...) {
  chkDots(...)
  cat(
    "Period-over-period price index",
    "for",
    length(object$levels),
    "levels over",
    length(object$time),
    "time periods",
    "\n"
  )
  invisible()
}

#' @export
summary.direct_piar_index <- function(object, ...) {
  chkDots(...)
  cat(
    "Fixed-base price index",
    "for",
    length(object$levels),
    "levels over",
    length(object$time),
    "time periods",
    "\n"
  )
  invisible()
}

#' @export
print.piar_index <- function(x, ...) {
  summary(x)
  print(as.matrix(x), ...)
  invisible(x)
}

#' Test if an object is a price index
#'
#' Test if an object is a index object or a subclass of an index object.
#'
#' @param x `[object]` An object to test.
#'
#' @returns
#' `is_index()` returns `TRUE` if `x` inherits from [`piar_index`].
#'
#' `is_chainable_index()` returns `TRUE` if `x` inherits from
#' [`chainable_piar_index`].
#'
#' `is_direct_index()` returns `TRUE` if `x` inherits from
#' [`direct_piar_index`].
#'
#' @export
is_index <- function(x) {
  inherits(x, "piar_index")
}

#' @rdname is_index
#' @export
is_chainable_index <- function(x) {
  inherits(x, "chainable_piar_index")
}

#' @rdname is_index
#' @export
is_direct_index <- function(x) {
  inherits(x, "direct_piar_index")
}

#' Group generics
#' @noRd
#' @export
Math.piar_index <- function(x, ...) {
  stop(gettextf("'%s' not meaningful for index objects", .Generic))
}

#' @export
Ops.piar_index <- function(e1, e2) {
  boolean <- switch(
    .Generic,
    `<` = ,
    `>` = ,
    `==` = ,
    `!=` = ,
    `<=` = ,
    `>=` = TRUE,
    FALSE
  )
  if (!boolean) {
    stop(gettextf("'%s' not meaningful for index objects", .Generic))
  }
  if (is_index(e1)) {
    e1 <- as.matrix(e1)
  }
  if (is_index(e2)) {
    e2 <- as.matrix(e2)
  }
  NextMethod(.Generic)
}

#' @export
Summary.piar_index <- function(..., na.rm) {
  stop(gettextf("'%s' not meaningful for index objects", .Generic))
}

epr1 <- elementary_index(
  1:8,
  period = rep(1:2, each = 4),
  ea = rep(letters[1:4], 2),
  contrib = TRUE
)
epr2 <- as_index(matrix(1:8, 4, 2, dimnames = list(letters[5:8], 1:2)))
pias <- as_aggregation_structure(
  data.frame(level1 = "0", level2 = letters[1:4], weight = 1:4)
)
index1 <- aggregate(epr1, pias)

# Merge returns the correct type of index.
local({
  expect_true(is_chainable_index(merge(epr1, epr2)))
  expect_true(is_direct_index(merge(chain(epr1), chain(epr2))))
  expect_true(is_chainable_index(merge(epr1, chain(epr2))))
  expect_true(is_direct_index(merge(chain(epr1), epr2)))
})

# merge doesn't work when it shouldn't.
local({
  expect_error(merge(epr1, epr1))
  expect_error(merge(epr1, mtcars))
  epr3 <- as_index(matrix(1:8, 4, 2, dimnames = list(letters[5:8], 2:3)))
  expect_error(merge(epr1, epr3))
})

# merge returns the correct result.
local({
  epr3 <- merge(epr1, epr2)
  res <- rbind(
    matrix(1:8, 4, 2, dimnames = list(letters[1:4], 1:2)),
    matrix(1:8, 4, 2, dimnames = list(letters[5:8], 1:2))
  )
  names(dimnames(res)) <- c("levels", "time")
  expect_equal(as.matrix(epr3), res)
  expect_identical(levels(epr3), letters[1:8])
  expect_identical(time(epr3), time(epr1))
  expect_equal(
    contrib(epr3, "a"),
    matrix(c(0, 4), 1, 2, dimnames = list(product = "a.1", time = 1:2))
  )
  expect_equal(
    contrib(epr3, "e"),
    matrix(numeric(0), 0, 2, dimnames = list(product = NULL, time = 1:2))
  )
  res2 <- rbind(
    matrix(1:8, 4, 2, dimnames = list(levels = letters[5:8], 1:2)),
    matrix(1:8, 4, 2, dimnames = list(levels = letters[1:4], 1:2))
  )
  names(dimnames(res2)) <- c("levels", "time")
  expect_equal(as.matrix(merge(epr2, epr1)), res2)
})

# merge works with [.
local({
  epr3 <- merge(epr1, epr2)
  expect_equal(epr3[1:4], epr1)
  contrib(epr2) <- numeric(0)
  expect_equal(epr3[5:8], epr2)
})

# Coercion works as expected.
local({
  expect_equal(merge(epr1, epr2), merge(epr1, chain(epr2)))
  expect_equal(merge(chain(epr1), epr2), merge(chain(epr1), chain(epr2)))
})

epr1 <- elementary_index(
  1:8,
  period = rep(1:2, each = 4),
  ea = rep(letters[1:4], 2),
  contrib = TRUE
)
epr2 <- as_index(matrix(1:8, 4, 2, dimnames = list(letters[1:4], 3:4)))
pias1 <- as_aggregation_structure(
  data.frame(level1 = "0", level2 = letters[1:4], weight = 1:4)
)
pias2 <- as_aggregation_structure(
  data.frame(level1 = "0", level2 = letters[1:4], weight = 4:1)
)
index1 <- aggregate(epr1, pias1)
index2 <- aggregate(epr2, pias2)

# stack returns the correct type of index.
local({
  expect_true(is_chainable_index(stack(epr1, epr2)))
  expect_true(is_chainable_index(stack(index1, index2)))
  expect_true(is_chainable_index(stack(epr1, chain(epr2))))
  expect_true(is_direct_index(stack(chain(epr1), epr2)))
  expect_error(suppressWarnings(stack(index1, epr1)))
  time(epr2) <- time(epr1)
  expect_error(stack(epr1, epr2))
  expect_error(stack(index1, aggregate(epr2, pias2, order = 0)))
})

# Stacking and unstacking are opposite operations.
local({
  expect_equal(epr1, Reduce(stack, unstack(epr1)))
  expect_equal(chain(epr1), Reduce(stack, unstack(chain(epr1))))
  expect_equal(index2, Reduce(stack, unstack(index2)))
  expect_equal(chain(index2), Reduce(stack, unstack(chain(index2))))
})

# Stacking returns the correct result.
local({
  epr3 <- stack(epr1, epr2)
  expect_equal(
    as.matrix(epr3),
    matrix(1:8, 4, 4, dimnames = list(levels = letters[1:4], time = 1:4))
  )
  expect_identical(levels(epr3), levels(epr1))
  expect_identical(time(epr3), as.character(1:4))
  expect_equal(
    contrib(epr3),
    matrix(c(0, 4, 0, 0), 1, 4, dimnames = list(product = "a.1", time = 1:4))
  )
  expect_equal(
    as.matrix(stack(epr2, epr1)),
    matrix(
      1:8,
      4,
      4,
      dimnames = list(levels = letters[1:4], time = c(3:4, 1:2))
    )
  )
})

# stack works with [.
local({
  epr3 <- stack(epr1, epr2)
  expect_equal(epr3[, 1:2], epr1)
  contrib(epr2) <- numeric(0)
  expect_equal(epr3[, 3:4], epr2)
})

# Coercion works as expected.
local({
  expect_equal(stack(epr1, epr2), stack(epr1, chain(epr2)))
  expect_equal(stack(chain(epr1), chain(epr2)), stack(chain(epr1), chain(epr2)))
})

# Price updating works with a stacked index.
local({
  index <- stack(index1, index2)
  expect_equal(
    aggregate(index[, 2], update(pias1, index, period = 1)),
    index[, 2]
  )
  expect_equal(
    aggregate(index[, 4], update(pias2, index[, 3:4], period = 3)),
    index[, 4]
  )
})

# stack matches y to x.
local({
  expect_equal(stack(epr1, epr2), stack(epr1, epr2[4:1]))
})

epr1 <- elementary_index(
  1:8,
  period = rep(1:2, each = 4),
  ea = rep(letters[1:4], 2),
  contrib = TRUE
)
epr2 <- as_index(matrix(1:8, 4, 2, dimnames = list(letters[1:4], 3:4)))
pias1 <- as_aggregation_structure(
  data.frame(level1 = "0", level2 = letters[1:4], weight = 1:4)
)
pias2 <- as_aggregation_structure(
  data.frame(level1 = "0", level2 = letters[1:4], weight = 4:1)
)
index1 <- aggregate(epr1, pias1)
index2 <- aggregate(epr2, pias2)

test_that("stack returns the correct type of index", {
  expect_true(is_chainable_index(stack(epr1, epr2)))
  expect_true(is_chainable_index(stack(index1, index2)))
  expect_true(is_chainable_index(stack(epr1, chain(epr2))))
  expect_true(is_direct_index(stack(chain(epr1), epr2)))
  expect_error(suppressWarnings(stack(index1, epr1)))
  time(epr2) <- time(epr1)
  expect_error(stack(epr1, epr2))
  expect_error(stack(index1, aggregate(epr2, pias2, r = 0)))
})

test_that("stacking and unstacking are opposite operations", {
  expect_equal(epr1, Reduce(stack, unstack(epr1)))
  expect_equal(chain(epr1), Reduce(stack, unstack(chain(epr1))))
  expect_equal(index2, Reduce(stack, unstack(index2)))
  expect_equal(chain(index2), Reduce(stack, unstack(chain(index2))))
})

test_that("stacking returns the correct result", {
  epr3 <- stack(epr1, epr2)
  expect_equal(
    as.matrix(epr3),
    matrix(1:8, 4, 4, dimnames = list(levels = letters[1:4], time = 1:4))
  )
  expect_identical(levels(epr3), levels(epr1))
  expect_identical(time(epr3), as.character(1:4))
  expect_equal(
    contrib(epr3),
    matrix(c(0, 4, 0, 0), 1, 4, dimnames = list(product = "a.1", time = 1:4))
  )
  expect_equal(
    as.matrix(stack(epr2, epr1)),
    matrix(
      1:8,
      4,
      4,
      dimnames = list(levels = letters[1:4], time = c(3:4, 1:2))
    )
  )
})

test_that("stack works with [", {
  epr3 <- stack(epr1, epr2)
  expect_equal(epr3[, 1:2], epr1)
  contrib(epr2) <- numeric(0)
  expect_equal(epr3[, 3:4], epr2)
})

test_that("coercion works as expected", {
  expect_equal(stack(epr1, epr2), stack(epr1, chain(epr2)))
  expect_equal(stack(chain(epr1), chain(epr2)), stack(chain(epr1), chain(epr2)))
})

test_that("price updating works with a stacked index", {
  index <- stack(index1, index2)
  expect_equal(
    aggregate(index[, 2], update(pias1, index, period = 1)),
    index[, 2]
  )
  expect_equal(
    aggregate(index[, 4], update(pias2, index[, 3:4], period = 3)),
    index[, 4]
  )
})

test_that("stack matches y to x", {
  expect_equal(stack(epr1, epr2), stack(epr1, epr2[4:1]))
})

epr1 <- elementary_index(
  1:8,
  period = rep(1:2, each = 4),
  ea = rep(letters[1:4], 2),
  contrib = TRUE
)
epr2 <- as_index(matrix(1:8, 4, 2, dimnames = list(letters[5:8], 1:2)))
pias <- as_aggregation_structure(
  data.frame(level1 = "0", level2 = letters[1:4], weight = 1:4)
)
index1 <- aggregate(epr1, pias)

test_that("merge returns the correct type of index", {
  expect_true(is_chainable_index(merge(epr1, epr2)))
  expect_true(is_direct_index(merge(chain(epr1), chain(epr2))))
  expect_true(is_chainable_index(merge(epr1, chain(epr2))))
  expect_true(is_direct_index(merge(chain(epr1), epr2)))
})

test_that("merge doesn't work when it shouldn't", {
  expect_error(merge(epr1, epr1))
  expect_error(merge(epr1, mtcars))
  epr3 <- as_index(matrix(1:8, 4, 2, dimnames = list(letters[5:8], 2:3)))
  expect_error(merge(epr1, epr3))
})

test_that("merge returns the correct result", {
  epr3 <- merge(epr1, epr2)
  res <- rbind(
    matrix(1:8, 4, 2, dimnames = list(letters[1:4], 1:2)),
    matrix(1:8, 4, 2, dimnames = list(letters[5:8], 1:2))
  )
  names(dimnames(res)) <- c("levels", "time")
  expect_equal(as.matrix(epr3), res)
  expect_identical(levels(epr3), letters[1:8])
  expect_identical(time(epr3), time(epr1))
  expect_equal(
    contrib(epr3, "a"),
    matrix(c(0, 4), 1, 2, dimnames = list(product = "a.1", time = 1:2))
  )
  expect_equal(
    contrib(epr3, "e"),
    matrix(numeric(0), 0, 2, dimnames = list(product = NULL, time = 1:2))
  )
  res2 <- rbind(
    matrix(1:8, 4, 2, dimnames = list(levels = letters[5:8], 1:2)),
    matrix(1:8, 4, 2, dimnames = list(levels = letters[1:4], 1:2))
  )
  names(dimnames(res2)) <- c("levels", "time")
  expect_equal(as.matrix(merge(epr2, epr1)), res2)
})

test_that("merge works with [", {
  epr3 <- merge(epr1, epr2)
  expect_equal(epr3[1:4], epr1)
  contrib(epr2) <- numeric(0)
  expect_equal(epr3[5:8], epr2)
})

test_that("coercion works as expected", {
  expect_equal(merge(epr1, epr2), merge(epr1, chain(epr2)))
  expect_equal(merge(chain(epr1), epr2), merge(chain(epr1), chain(epr2)))
})

#' Aggregate a price index over subperiods
#'
#' Aggregate an index over subperiods by taking the (usually arithmetic) mean
#' of index values over consecutive windows of subperiods.
#'
#' The `mean()` method constructs a set of non-overlapping windows of
#' length `window`, starting in the first period of the index, and takes
#' the mean of each index value in these windows for each level of the index.
#' The last window is discarded if it is incomplete (with a warning), so that
#' index values are
#' always averaged over `window` periods. The names for the first time
#' period in each window form the new names for the aggregated time periods.
#'
#' Percent-change contributions are aggregated if `contrib = TRUE` following the
#' same approach as [`aggregate()`][aggregate.piar_index].
#'
#' An optional vector of weights can be specified when aggregating index values
#' over subperiods, which is often useful when aggregating a Paasche index; see
#' section 4.3 of Balk (2008) for details.
#'
#' @name mean.piar_index
#' @aliases mean.piar_index
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param weights `[numeric >= 0]` A numeric vector of weights for the index
#'   values in `x`, or
#'   something that can be coerced into one. The
#'   default is equal weights. It is usually easiest to specify these weights as
#'   a matrix with a row for each index value in `x` and a column for each
#'   time period.
#' @param window `[integer(1) > 0]` A positive integer giving the size of the
#'   window used to
#'   average index values across subperiods. The default averages over all
#'   periods in `x`. Non-integers are truncated towards 0.
#' @param na.rm `[logical(1)]` Should missing values be removed? By default,
#'   missing values
#'   are not removed. Setting `na.rm = TRUE` is equivalent to overall-mean
#'   imputation.
#' @param order `[numeric(1)]` Order of the generalized mean to aggregate
#'   index values. 0 for a
#'   geometric index (the default for making elementary indexes), 1 for an
#'   arithmetic index (the default for aggregating elementary indexes and
#'   averaging indexes over subperiods), or -1 for a harmonic index (usually for
#'   a Paasche index). Other values are possible; see
#'   [gmean()] for details.
#' @param contrib `[logical(1)]` Aggregate percent-change contributions in `x`?
#'   By default contributions are aggregated.
#' @param ... Not currently used.
#' @param duplicate_contrib `[character(1)]` The method to deal with duplicate
#'   product contributions. Either `"make.unique"` to make duplicate product
#'   names unique with [make.unique()] or `"sum"` to add contributions for the
#'   same products across subperiods (the default).
#' @param r Deprecated.
#'
#' @returns
#' A price index, averaged over subperiods, that inherits from the same
#' class as `x`.
#'
#' @references
#' Balk, B. M. (2008). *Price and Quantity Index Numbers*.
#' Cambridge University Press.
#'
#' @examples
#' index <- as_index(matrix(c(1:12, 12:1), 2, byrow = TRUE), chainable = FALSE)
#'
#' # Turn a monthly index into a quarterly index.
#' mean(index, window = 3)
mean.chainable_piar_index <- function(
  x,
  ...,
  weights = NULL,
  window = NULL,
  na.rm = FALSE,
  contrib = TRUE,
  order = 1,
  r = order,
  duplicate_contrib = c("sum", "make.unique")
) {
  chkDots(...)
  if ("r" %in% names(sys.call())) {
    warning("`r` is deprecated and will be removed; use `order` instead")
  }
  mean_index(
    x,
    weights,
    window = window,
    na.rm = na.rm,
    contrib = contrib,
    order = r,
    chainable = TRUE,
    duplicate_contrib = match.arg(duplicate_contrib)
  )
}

#' @rdname mean.piar_index
#' @export
mean.direct_piar_index <- function(
  x,
  ...,
  weights = NULL,
  window = NULL,
  na.rm = FALSE,
  contrib = TRUE,
  order = 1,
  r = order,
  duplicate_contrib = c("sum", "make.unique")
) {
  chkDots(...)
  if ("r" %in% names(sys.call())) {
    warning("`r` is deprecated and will be removed; use `order` instead")
  }
  mean_index(
    x,
    weights,
    window = window,
    na.rm = na.rm,
    contrib = contrib,
    order = r,
    chainable = FALSE,
    duplicate_contrib = match.arg(duplicate_contrib)
  )
}

#' Internal function to aggregate over subperiods
#' @noRd
mean_index <- function(
  x,
  weights,
  window,
  na.rm,
  contrib,
  order,
  chainable,
  duplicate_contrib
) {
  if (!is.null(weights)) {
    weights <- as.numeric(weights)
    if (any(weights < 0, na.rm = TRUE)) {
      stop("all elements of `weights` must be non-negative")
    }
    if (length(weights) != ntime(x) * nlevels(x)) {
      stop("`weights` must have a value for each index value in 'x'")
    }
    dim(weights) <- c(nlevels(x), ntime(x))
  }

  order <- as.numeric(order)
  window <- as.integer(window %||% ntime(x))
  if (window < 1L) {
    stop("`window` must be a positive integer")
  }
  if (window > ntime(x)) {
    stop("`x` must have at least `window` time periods")
  }

  # Get the starting location for each window.
  if (ntime(x) %% window != 0) {
    warning("`window` is not a multiple of the number of time periods in `x`")
  }
  len <- ntime(x) %/% window
  loc <- seq.int(1L, by = window, length.out = len)
  periods <- x$time[loc]

  has_contrib <- !is.null(x$contrib) && contrib
  # Loop over each window and calculate the mean for each level.
  res <- contrib <- vector("list", length(periods))
  rows <- seq_len(nlevels(x))
  w <- list(NULL)
  for (i in seq_along(loc)) {
    j <- seq(loc[i], length.out = window)
    rel <- split_rows(x$index[, j, drop = FALSE], rows)
    if (!is.null(weights)) {
      w <- split_rows(weights[, j, drop = FALSE], rows)
    }
    res[[i]] <- mapply(gmean, rel, w, order, na.rm = na.rm, USE.NAMES = FALSE)
    if (has_contrib) {
      con <- split_rows(x$contrib[, j, drop = FALSE], rows)
      contrib[[i]] <- mapply(
        aggregate_contrib,
        con,
        rel,
        w,
        order,
        res[[i]],
        duplicate_contrib,
        SIMPLIFY = FALSE,
        USE.NAMES = FALSE
      )
    }
  }

  piar_index(
    do.call(cbind, res),
    do.call(cbind, contrib),
    x$levels,
    periods,
    chainable = chainable
  )
}

#' Splice an index series
#'
#' Splice a collection of index series computed over a rolling window into one
#' index series. Splicing on multiple points combines the results with a
#' geometric mean.
#'
#' @family price index functions
#' @export
#'
#' @param x `[list]` A list of equal-length numeric vectors giving the
#'   period-over-period indexes for each window.
#' @param periods `[integer]` An integer vector giving the splice points for
#'   each window. The default splices on each point in the window.
#' @param initial `[numeric]` A numeric vector giving an initial
#'   period-over-period index
#'   series onto which the elements of `x` are spliced. The default uses the
#'   first element of `x`.
#' @param published `[logical(1)]` Should the splice be done against the
#'   published series? The default splices using the recalculated index series.
#'
#' @returns
#' A numeric vector giving the spliced (fixed-base) index series.
#'
#' @references
#' Chessa, A. G. (2019).
#' *A Comparison of Index Extension Methods for Multilateral Methods.* Paper
#' presented at the 16th Meeting of the Ottawa Group on Price Indices,
#' 8-10 May 2019, Rio de Janeiro, Brazil.
#'
#' Krsinich, F. (2016). The FEWS index: Fixed effects with a window splice.
#' *Journal of Official Statistics*, 32(2), 375-404.
#'
#' @examples
#' # Make an index series over a rolling window.
#' x <- list(c(1.1, 0.9, 1.2), c(0.8, 1.3, 1.4), c(1.3, 1.3, 0.8))
#'
#' # Mean splice.
#' splice_index(x)
#'
#' # Movement splice.
#' splice_index(x, 3)
#'
#' # Window splice.
#' splice_index(x, 1)
#'
#' # Splicing on the published series preserves the within-window
#' # movement of the index series.
#' splice_index(x, 1, published = TRUE)
splice_index <- function(x, periods = NULL, initial = NULL, published = FALSE) {
  x <- as.list(x)
  if (do.call(different_length, x)) {
    stop("all elements of `x` must be the same length")
  }

  if (is.null(initial) && length(x) > 0L) {
    initial <- x[[1L]]
    x <- x[-1L]
  }
  initial <- cumprod(initial)
  if (length(x) == 0L) {
    return(initial)
  }

  n <- length(x[[1L]])
  offset <- length(initial)
  if (offset < n) {
    stop("`initial` must be at least as long as each element of `x`")
  }

  if (is.null(periods)) {
    periods <- seq_len(n)
  } else {
    periods <- as.integer(periods)
  }

  y <- lapply(x, \(z) rev(cumprod(rev(z)))[periods])
  # Use recursive = TRUE to keep names.
  res <- c(initial, lapply(x, \(x) x[length(x)]), recursive = TRUE)

  iw <- seq.int(to = offset - 1L, length.out = n)[periods]
  if (published) {
    for (i in seq_along(x)) {
      res[i + offset] <- gmean(y[[i]] * res[iw + i], order = 0)
    }
  } else {
    links <- c(
      list(rep.int(1, length(periods))),
      lapply(x[-length(x)], `[`, periods)
    )
    links <- Reduce(`*`, links, accumulate = TRUE)
    base <- res[iw + 1L]
    for (i in seq_along(x)) {
      res[i + offset] <- gmean(y[[i]] * links[[i]] * base, order = 0)
    }
  }
  res
}

x <- list(1:3, 3:5, 5:7)

# Key splice methods works.
local({
  # Movement splice.
  expect_equal(splice_index(x, 3), c(1, 2, 6, 30, 210))
  expect_equal(splice_index(x, 3, published = TRUE), c(1, 2, 6, 30, 210))
  # Window splice.
  expect_equal(splice_index(x, 1), c(1, 2, 6, 60, 630))
  expect_equal(splice_index(x, 1, published = TRUE), c(1, 2, 6, 60, 420))
  # Mean splice.
  expect_equal(
    splice_index(x),
    c(
      1,
      2,
      6,
      gmean(c(60, 40, 30), order = 0),
      gmean(rev(cumprod(rev(5:7))) * 3:5 * c(1, 2, 6), order = 0)
    )
  )
  expect_equal(
    splice_index(x, published = TRUE),
    c(
      1,
      2,
      6,
      gmean(c(60, 40, 30), order = 0),
      gmean(c(420, 252, 7 * gmean(c(60, 40, 30), order = 0)), order = 0)
    )
  )
})

# Result length is correct.
local({
  expect_equal(
    splice_index(x[-1], 3, c(1, 1, 1, 1, 2, 3)),
    c(1, 1, 1, 1, 2, 6, 30, 210)
  )
  expect_equal(
    splice_index(x[-1], 1, c(1, 1, 1, 1, 2, 3)),
    c(1, 1, 1, 1, 2, 6, 60, 630)
  )
  expect_equal(
    splice_index(x[-1], 1, c(1, 1, 1, 1, 2, 3), published = TRUE),
    c(1, 1, 1, 1, 2, 6, 60, 420)
  )
})

# Period subscripting works.
local({
  expect_equal(
    splice_index(x[-1], rep(3, 10), c(1, 1, 1, 1, 2, 3)),
    c(1, 1, 1, 1, 2, 6, 30, 210)
  )
  expect_equal(
    splice_index(x[-1], 4, c(1, 1, 1, 1, 2, 3)),
    c(1, 1, 1, 1, 2, 6, NA, NA)
  )
  expect_equal(
    splice_index(x[-1], -(2:3), c(1, 1, 1, 1, 2, 3), published = TRUE),
    c(1, 1, 1, 1, 2, 6, 60, 420)
  )
})

# Splicing is invariant.
local({
  # Movement
  expect_equal(splice_index(x, 3), splice_index(x[-1], 3, x[[1]]))
  expect_equal(splice_index(x, 3), splice_index(x[-(1:2)], 3, c(1, 2, 3, 5)))
  # Window
  expect_equal(splice_index(x, 1), splice_index(x[-1], 1, x[[1]]))
  expect_equal(
    splice_index(x, 1, published = TRUE),
    splice_index(x[-(1:2)], 1, c(1, 2, 3, 10), published = TRUE)
  )
  # Mean
  expect_equal(splice_index(x), splice_index(x[-1], initial = x[[1]]))
  expect_equal(
    splice_index(x, published = TRUE),
    splice_index(
      x[-(1:2)],
      initial = c(1, 2, 3, gmean(c(10, 40 / 6, 5), order = 0)),
      published = TRUE
    )
  )
})

# NAs return NA.
local({
  expect_equal(splice_index(x, c(1, NA, 4)), c(1, 2, 6, NA, NA))
  x[[2]][2] <- NA
  expect_equal(splice_index(x), c(1, 2, 6, NA, NA))
  expect_equal(splice_index(x, 3), c(1, 2, 6, 30, 210))
})

# Corner cases.
local({
  expect_equal(splice_index(list()), numeric(0))
  expect_equal(splice_index(list(), initial = 1:5), cumprod(1:5))
  expect_equal(splice_index(list(1:5)), cumprod(1:5))
  expect_equal(splice_index(list(1, 2, 3, 4, 5)), cumprod(1:5))
})

# Errors.
local(
  {
    expect_error(splice_index(list(1:3, 1:2)))
    expect_error(splice_index(list(1:3, 1:3), initial = 1:2))
  }
)

# Splicing is the same as chaining.
local({
  x <- as.list(1:5)
  expect_equal(splice_index(x), cumprod(1:5))
  expect_equal(splice_index(x, initial = 1:3), cumprod(c(1:3, 1:5)))
  expect_equal(splice_index(x, published = TRUE), cumprod(1:5))
})

# Splicing keeps names.
local({
  x <- list(1:3, c(a = 1, b = 2, c = 3), z = c(1:2, c = 3))
  expect_identical(names(splice_index(x)), c("", "", "", "c", "z.c"))
})

#' Split an index into groups
#'
#' Split an index into groups of indexes according to a factor, along either
#' the levels or time periods of the index.
#'
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param f `[factor | list]` A factor or list of factors to group elements
#'   of `x`.
#' @param drop `[logical(1)]` Should levels that do not occur in `f` be dropped?
#'   By default all levels are kept.
#' @param along `[character(1)]` Either `"levels"` to split over the levels
#'   of `x` (the default), or `"time"` to split over the time periods of `x`.
#' @param value `[list]` A list of values compatible with the splitting of `x`,
#'   or something that can be coerced into one, recycled if necessary.
#' @param ... Further arguments passed to [`split.default()`].
#'
#' @returns
#' `split()` returns a list of index objects for each level in `f`. The
#' replacement method replaces these values with the corresponding element of
#' `value`.
#'
#' @examples
#' index <- as_index(matrix(1:6, 2))
#'
#' split(index, 1:2)
#'
#' split(index, c(1, 1, 2), along = "time")
split.piar_index <- function(
  x,
  f,
  drop = FALSE,
  ...,
  along = c("levels", "time")
) {
  along <- match.arg(along)
  ix <- split(x = seq_along(x[[along]]), f = f, drop = drop, ...)
  if (along == "levels") {
    lapply(ix, \(i) x[i, ])
  } else {
    lapply(ix, \(i) x[, i])
  }
}

#' @rdname split.piar_index
#' @export
`split<-.piar_index` <- function(
  x,
  f,
  drop = FALSE,
  ...,
  along = c("levels", "time"),
  value
) {
  value <- as.list(value)
  along <- match.arg(along)
  ix <- split(x = seq_along(x[[along]]), f = f, drop = drop, ...)
  n <- length(value)
  if (n > 0L && length(ix) %% n != 0) {
    warning(
      "number of items to replace is not a multiple of replacement length"
    )
  }
  j <- 0
  if (along == "levels") {
    for (i in ix) {
      j <- j %% n + 1
      x[i, ] <- value[[j]]
    }
  } else {
    for (i in ix) {
      j <- j %% n + 1
      x[, i] <- value[[j]]
    }
  }
  x
}

#' Generalized mean
#'
#' Calculated a weighted generalized mean.
#'
#' The generalized mean is also called the power mean, Hölder mean, or \eqn{l_p}
#' mean; see Bullen (2003, p. 175) for details.
#'
#' Both `x` and `weights` are usually strictly positive
#' (and finite), especially for the purpose of making a price index. This is not
#' enforced, but the results may not make sense if the generalized mean is not
#' defined. There are two exceptions to this.
#' 1. The convention by Hardy et al. (1952, p. 13) is used in cases where `x`
#' has zeros: the generalized mean is 0 whenever the weights are strictly
#' positive and `order < 0`. The analogous convention holds whenever at least
#' one element of `x` is `Inf`: the generalized mean is `Inf` whenever the
#' weights are strictly positive and `order > 0`.
#'
#' 2. Some authors let the weights be non-negative and sum to 1. If there are
#' zero weights then the corresponding element
#' of `x` has no impact on the result whenever `x` is strictly
#' positive. Unlike [weighted.mean()], however,
#' zero weights are not strong zeros, so infinite values in `x` will
#' propagate.
#'
#' The weights are scaled to sum to 1 to satisfy the definition of a
#' generalized mean.
#'
#' @family math functions
#' @export
#'
#' @param x `[numeric > 0]` A strictly positive numeric vector.
#' @param weights `[numeric >= 0]` A positive numeric vector of weights, the
#'   same length as `x`. The default is to equally weight each element of `x`.
#' @param order `[numeric(1)]` A finite number giving the order (or exponent) of
#'   the generalized mean. The default calculates an arithmetic mean.
#' @param na.rm `[logical(1)]` Should missing values be removed? By default,
#'   missing values are not removed.
#'
#' @returns
#' A numeric value for the generalized mean.
#'
#' @note
#' The generalized mean can be defined on the extended real line, so
#' that `order = -Inf / Inf` returns [min()]/[max()], to agree with the
#' definition by Bullen (2003). This is not implemented, and the order of the
#' generalized mean must be finite.
#'
#' @references
#' Bullen, P. S. (2003). *Handbook of Means and Their Inequalities*.
#' Springer Science+Business Media.
#'
#' Hardy, G., Littlewood, J. E., and Polya, G. (1952). *Inequalities* (2nd
#' edition). Cambridge University Press.
#'
#' @examples
#' x <- 1:3
#' w <- c(0.25, 0.25, 0.5)
#'
#' # Arithmetic mean.
#' gmean(x, w)
#'
#' # Geometric mean.
#' gmean(x, w, order = 0)
#'
#' # The Lehmer mean is a generalized mean with specific weights.
#' gmean(x, w * x)
gmean <- function(x, weights = NULL, order = 1, na.rm = FALSE) {
  if (not_finite_scalar(order)) {
    stop("`order` must be a finite number")
  }
  if (!is.null(weights) && length(x) != length(weights)) {
    stop("`x` and `weights` must be the same length")
  }
  na_mask <- if (na.rm && (anyNA(x) || anyNA(weights))) {
    stats::complete.cases(x, weights)
  }
  .gmean(x, weights, order, na_mask)
}

#' Nested generalized means
#'
#' Calculate a weighted (outer) generalized mean of two (inner) generalized
#' means (i.e., crossing means).
#'
#' @family math functions
#' @export
#'
#' @param x `[numeric > 0]` A strictly positive numeric vector.
#' @param weights `[list]` A list of positive numeric vector of
#'   weights, each
#'   the same length as `x`, for both of the inner generalized means. `NULL`
#'   elements of `weights` equally weight each element of `x`. The default
#'   uses equal weights for both inner generalized mean.
#' @param order `[numeric(2)]` A finite numeric vector giving the order of each
#'   of the inner
#'   generalized means. The default computes an arithmetic mean and a harmonic
#'   mean.
#' @param outer_weights `[numeric(2)]` A strictly positive numeric vector
#'   weights for each of
#'   the inner generalized means as used in the outer generalized mean. The
#'   default weights each inner generalized mean equally.
#' @param outer_order `[numeric(1)]` A finite number giving the order of the
#'   outer generalized mean. The default uses a geometric mean.
#' @param na.rm `[logical(1)]` Should missing values in `x` and `weights` be
#'   removed? By
#'   default missing values are not removed. Note that removal of missing values
#'   is balanced across `x` and both elements of `weights`.
#'
#' @returns
#' A numeric value for the nested generalized mean.
#'
#' @examples
#' x <- 1:3
#' w1 <- c(0.25, 0.25, 0.5)
#' w2 <- c(0.3, 0.3, 0.4)
#' # Calculate the geometric mean of the arithmetic and harmonic means
#' # to make a Fisher index.
#' nested_gmean(x, list(w1, w2))
nested_gmean <- function(
  x,
  weights = list(NULL, NULL),
  order = c(1, -1),
  outer_weights = NULL,
  outer_order = 0,
  na.rm = FALSE
) {
  if (different_length(x, weights[[1]], weights[[2]])) {
    stop("`x` and non-NULL components of `weights` must be the same length")
  }
  if (not_finite_pair(order)) {
    stop("`order` must be a pair of finite numbers")
  }
  if (length(weights) != length(order)) {
    stop("`weights` and `order` must be the same length")
  }
  na_mask <- if (na.rm && (anyNA(x) || anyNA(weights, recursive = TRUE))) {
    stats::complete.cases(x, weights[[1]], weights[[2]])
  }
  inner_mean <- c(
    .gmean(x, weights[[1]], order[1], na_mask),
    .gmean(x, weights[[2]], order[2], na_mask)
  )
  gmean(inner_mean, outer_weights, outer_order, na.rm = na.rm)
}

#' Internal generalized mean
#' @noRd
.gmean <- function(x, weights, r, na_mask) {
  if (!is.null(na_mask)) {
    x <- x[na_mask]
    weights <- weights[na_mask]
  }
  if (is.null(weights)) {
    if (r == 0) {
      exp(sum(log(x)) / length(x))
    } else if (r == 1) {
      # The arithmetic case is important enough for the optimization.
      sum(x) / length(x)
    } else if (r == -1) {
      # Same with the harmonic.
      length(x) / sum(1 / x)
    } else {
      (sum(x^r) / length(x))^(1 / r)
    }
  } else {
    if (r == 0) {
      exp(sum(log(x) * weights) / sum(weights))
    } else if (r == 1) {
      sum(x * weights) / sum(weights)
    } else if (r == -1) {
      sum(weights) / sum(weights / x)
    } else {
      (sum(x^r * weights) / sum(weights))^(1 / r)
    }
  }
}

# Generalized mean works.
local({
  # Corner cases.
  expect_equal(gmean(c(0, 1, 2), order = -1), 0)
  expect_equal(gmean(c(Inf, 1, 2), order = 2), Inf)

  # No NAs.
  x <- 1:5
  w <- 5:1

  expect_equal(gmean(x, w), weighted.mean(x, w))
  expect_equal(gmean(x, w, order = -1), 1 / weighted.mean(1 / x, w))
  expect_equal(gmean(x, w, order = 0), exp(weighted.mean(log(x), w)))
  expect_equal(gmean(x, w, order = 2), sqrt(weighted.mean(x^2, w)))

  expect_equal(gmean(x), weighted.mean(x))
  expect_equal(gmean(x, order = -1), 1 / weighted.mean(1 / x))
  expect_equal(gmean(x, order = 0), exp(weighted.mean(log(x))))
  expect_equal(gmean(x, order = 2), sqrt(weighted.mean(x^2)))

  # With NAs.
  x[1] <- NA
  w[2] <- NA

  xx <- x[3:5]
  ww <- w[3:5]

  expect_equal(gmean(x, w, na.rm = TRUE), weighted.mean(xx, ww))
  expect_equal(
    gmean(x, w, order = -1, na.rm = TRUE),
    1 / weighted.mean(1 / xx, ww)
  )
  expect_equal(
    gmean(x, w, order = 0, na.rm = TRUE),
    exp(weighted.mean(log(xx), ww))
  )
  expect_equal(
    gmean(x, w, order = 2, na.rm = TRUE),
    sqrt(weighted.mean(xx^2, ww))
  )

  expect_equal(gmean(x, na.rm = TRUE), weighted.mean(x, na.rm = TRUE))
  expect_equal(
    gmean(x, order = -1, na.rm = TRUE),
    1 / weighted.mean(1 / x, na.rm = TRUE)
  )
  expect_equal(
    gmean(x, order = 0, na.rm = TRUE),
    exp(weighted.mean(log(x), na.rm = TRUE))
  )
  expect_equal(
    gmean(x, order = 2, na.rm = TRUE),
    sqrt(weighted.mean(x^2, na.rm = TRUE))
  )

  # Errors.
  expect_error(gmean(x, order = NA), "`order` must be a finite number")
  expect_error(
    gmean(x, weights = 1),
    "`x` and `weights` must be the same length"
  )
})

# Nested generalized means works.
local({
  # No NAs.
  x <- 5:1
  w1 <- c(1, 2, 1, 2, 1)
  w2 <- c(1, 1, 2, 2, 3)

  expect_equal(nested_gmean(x), sqrt(weighted.mean(x) / weighted.mean(1 / x)))
  expect_equal(
    nested_gmean(x, list(w1, NULL)),
    sqrt(weighted.mean(x, w1) / weighted.mean(1 / x))
  )
  expect_equal(
    nested_gmean(x, list(NULL, w2)),
    sqrt(weighted.mean(x) / weighted.mean(1 / x, w2))
  )
  expect_equal(
    nested_gmean(x, list(w1, w2)),
    sqrt(weighted.mean(x, w1) / weighted.mean(1 / x, w2))
  )
  expect_equal(
    nested_gmean(x, outer_weights = 1:2, outer_order = 1),
    weighted.mean(c(weighted.mean(x), 1 / weighted.mean(1 / x)), 1:2)
  )

  # With NAs.
  x[1] <- NA
  w1[2] <- NA

  expect_equal(
    nested_gmean(x, list(NULL, w2), na.rm = TRUE),
    sqrt(weighted.mean(x[-1]) / weighted.mean(1 / x[-1], w2[-1]))
  )
  expect_equal(
    nested_gmean(x, list(w1, w2), na.rm = TRUE),
    sqrt(
      weighted.mean(x[-(1:2)], w1[-(1:2)]) /
        weighted.mean(1 / x[-(1:2)], w2[-(1:2)])
    )
  )
  expect_equal(
    nested_gmean(x, list(w1, w2), outer_weights = c(NA, 1), na.rm = TRUE),
    gmean(x[-(1:2)], w2[-(1:2)], order = -1)
  )

  # Errors.
  expect_error(
    nested_gmean(1:5, list(NULL, 1:4)),
    "`x` and non-NULL components of `weights` must be the same length"
  )

  expect_error(
    nested_gmean(1:5, order = NA),
    "`order` must be a pair of finite numbers"
  )
})

# Transmuting weights works.
local({
  # No NAs.
  x <- 1:5
  w <- 5:1
  expect_equal(transmute_weights(x, order = 2, to = 2), rep(0.2, 5))
  expect_equal(
    gmean(x, transmute_weights(x, mean = gmean(x, order = 0))),
    gmean(x, order = 0)
  )
  expect_equal(
    gmean(x, transmute_weights(x, order = -2.5, to = 3), 3),
    gmean(x, order = -2.5)
  )

  expect_equal(transmute_weights(x, w, order = -1, to = -1), scale_weights(w))
  expect_equal(gmean(x, transmute_weights(x, w)), gmean(x, w, order = 0))
  expect_equal(
    gmean(x, transmute_weights(x, w, order = 3.25, to = -1), -1),
    gmean(x, w, order = 3.25)
  )

  expect_equal(
    transmute_weights(c(1, 1, 1), 1:3, order = 2, to = 0),
    scale_weights(1:3)
  )

  # With NAs.
  x[1] <- NA
  w[2] <- NA

  expect_equal(transmute_weights(x, order = 2, to = 2), c(NA, rep(0.25, 4)))
  expect_equal(
    gmean(x, transmute_weights(x), na.rm = TRUE),
    gmean(x, order = 0, na.rm = TRUE)
  )
  expect_equal(
    gmean(x, transmute_weights(x, order = -2.5, to = 3), 3, na.rm = TRUE),
    gmean(x, order = -2.5, na.rm = TRUE)
  )

  expect_equal(
    transmute_weights(x, w, order = -1, to = -1),
    scale_weights(replace(w, 1, NA))
  )
  expect_equal(
    gmean(x, transmute_weights(x, w), na.rm = TRUE),
    gmean(x, w, order = 0, na.rm = TRUE)
  )
  expect_equal(
    gmean(x, transmute_weights(x, w, order = 3.25, to = -1), -1, na.rm = TRUE),
    gmean(x, w, order = 3.25, na.rm = TRUE)
  )

  # Errors.
  expect_error(
    transmute_weights(x, 1:4),
    "`x` and `weights` must be the same length"
  )
})

# Nested transmute works.
local({
  # No NAs.
  x <- 1:5
  w1 <- 1:5
  w2 <- c(5, 1, 4, 2, 3)

  expect_equal(
    gmean(x, transmute_weights2(x)),
    nested_gmean(x)
  )
  expect_equal(
    gmean(x, transmute_weights2(x, list(w1, NULL))),
    nested_gmean(x, list(w1, NULL))
  )
  expect_equal(
    gmean(x, transmute_weights2(x, list(NULL, w2))),
    nested_gmean(x, list(NULL, w2))
  )
  expect_equal(
    gmean(x, transmute_weights2(x, list(w1, w2))),
    nested_gmean(x, list(w1, w2))
  )
  expect_equal(
    gmean(x, transmute_weights2(x, list(w1, w2), pivot = 3)),
    nested_gmean(x, list(w1, w2))
  )

  expect_equal(
    transmute_weights2(x, list(w1, w2), outer_weights = c(NA, 1)),
    transmute_weights(x, w2, -1, 1)
  )
  expect_equal(
    transmute_weights2(x, list(w1, w1), order = c(2, 2), to = -1),
    transmute_weights(x, w1, 2, -1)
  )

  # With NAs.
  x[1] <- NA
  w1[2] <- NA

  expect_equal(
    gmean(x, transmute_weights2(x), na.rm = TRUE),
    nested_gmean(x, na.rm = TRUE)
  )
  expect_equal(
    gmean(x, transmute_weights2(x, list(w1, NULL)), na.rm = TRUE),
    nested_gmean(x, list(w1, NULL), na.rm = TRUE)
  )
  expect_equal(
    gmean(x, transmute_weights2(x, list(NULL, w2)), na.rm = TRUE),
    nested_gmean(x, list(NULL, w2), na.rm = TRUE)
  )
  expect_equal(
    gmean(x, transmute_weights2(x, list(w1, w2)), na.rm = TRUE),
    nested_gmean(x, list(w1, w2), na.rm = TRUE)
  )
  expect_equal(
    gmean(x, transmute_weights2(x, list(w1, w2), pivot = 3), na.rm = TRUE),
    nested_gmean(x, list(w1, w2), na.rm = TRUE)
  )

  expect_equal(
    transmute_weights2(x, list(w1, w2), outer_weights = c(NA, 1)),
    transmute_weights(x, replace(w2, 2, NA), -1, 1)
  )
  expect_equal(
    transmute_weights2(x, list(w1, w1), order = c(2, 2), to = -1),
    transmute_weights(x, w1, 2, -1)
  )

  # Errors.
  expect_error(
    transmute_weights2(x, list(1:4, NULL)),
    "`x` and non-NULL components of `weights` must be the same length"
  )
})

# Extended mean satisfies key identities.
a <- 1:5
b <- c(1, 5, 4, 1, 2)

# Symmetry.
expect_equal(emean(a, b), emean(b, a))
expect_equal(
  emean(a, b, order = c(-0.1, 2.5)),
  emean(b, a, order = c(2.5, -0.1))
)
expect_equal(
  emean(a, b, order = c(0, 2)),
  emean(b, a, order = c(2, 0))
)
expect_equal(
  emean(a, b, order = c(0, 0)),
  emean(b, a, order = c(0, 0))
)
expect_equal(
  emean(a, b, order = c(1, 1)),
  emean(b, a, order = c(1, 1))
)

# Identities.
expect_equal(
  emean(a, b, order = c(-1, 1)),
  apply(matrix(c(a, b), ncol = 2), 1, \(x) gmean(x, order = 0))
)
expect_equal(
  emean(a, b, order = c(2, 1)),
  apply(matrix(c(a, b), ncol = 2), 1, mean)
)
expect_equal(
  emean(a, b, order = c(-2, 1)),
  apply(
    matrix(c(a, b), ncol = 2),
    1,
    \(x) {
      (gmean(x, order = -1) * gmean(x, order = 0)^2)^(1 / 3)
    }
  )
)
expect_equal(
  emean(a, b, order = c(0.5, 1)),
  apply(
    matrix(c(a, b), ncol = 2),
    1,
    \(x) nested_gmean(x, order = c(1, 0), outer_order = 1)
  )
)
expect_equal(
  emean(a, b),
  apply(matrix(c(a, b), ncol = 2), 1, \(x) gmean(x, order = 0))^2 *
    emean(1 / a, 1 / b)
)
expect_equal(
  emean(a, b, order = c(-2, -1)),
  apply(matrix(c(a, b), ncol = 2), 1, \(x) gmean(x, order = -1))
)
expect_equal(
  emean(a, b, order = c(-2, 2)),
  apply(matrix(c(a, b), ncol = 2), 1, \(x) gmean(x, order = 0))
)
expect_equal(
  emean(a, b, order = c(2, 2)),
  c(1, ((a^a^2 / b^b^2)^(1 / (a^2 - b^2)) / exp(1)^(1 / 2))[-1])
)

# Errors when expected.
expect_error(emean(a, b, tol = 1:10), "`tol` cannot be longer than `x` or `y`")
expect_error(
  emean(a, b, order = 1:3),
  "`order` must be a pair of finite numbers"
)

# Recycling works
expect_warning(
  emean(1:3, 1:5),
  "length of `y` is not a multiple of length of `x`"
)
expect_equal(
  suppressWarnings(emean(1:3, 1:5)),
  emean(c(1:3, 1:2), 1:5)
)
expect_equal(emean(1:5, numeric(0)), numeric(0))

expect_warning(
  emean(1:5, 1:3),
  "length of `x` is not a multiple of length of `y`"
)
expect_equal(
  suppressWarnings(emean(1:5, 1:3)),
  emean(1:5, c(1:3, 1:2))
)
expect_equal(emean(numeric(0), 1:5), numeric(0))
expect_equal(emean(numeric(0), numeric(0)), numeric(0))

pias <- with(
  ms_weights,
  aggregation_structure(
    c(expand_classification(classification), list(business)),
    weight
  )
)

sp <- impute_prices(
  ms_prices,
  price ~ period + product,
  ea = business,
  pias = pias,
  method = "overall-mean"
)

# A length 0 inputs returns a length 0 output.
local({
  expect_length(
    impute_prices(
      integer(0),
      period = factor(integer(0), 1:5),
      product = integer(0),
      method = "carry-forward"
    ),
    0
  )
  expect_length(
    impute_prices(
      integer(0),
      period = integer(0),
      product = integer(0),
      ea = integer(0),
      pias = pias,
      method = "overall-mean"
    ),
    0
  )
  expect_equal(
    impute_prices(
      cbind(integer(0), integer(0)),
      period = integer(0),
      product = integer(0),
      ea = integer(0),
      pias = pias,
      method = "overall-mean"
    ),
    cbind(integer(0), integer(0))
  )
})

# Imputing imputed prices does noting.
local({
  expect_equal(
    sp,
    impute_prices(
      ms_prices,
      sp ~ period + product,
      ea = business,
      pias = pias,
      method = "overall-mean"
    )
  )

  expect_equal(
    sp,
    impute_prices(
      ms_prices,
      sp ~ period + product,
      ea = business,
      pias = pias,
      weights = price,
      method = "overall-mean"
    )
  )
})

# Same are regular parental in periods 1, 2.
local({
  epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    na.rm = TRUE
  )

  epr2 <- elementary_index(
    ms_prices,
    price_relative(sp, period = period, product = product) ~ period + business,
    na.rm = TRUE
  )

  # B2 is imputed from above the EA level
  expect_equal(epr[c(1, 3:4), 1:3], epr2[c(1, 3:4), 1:3])

  expect_equal(
    aggregate(epr, pias, na.rm = TRUE)[, 1:2],
    aggregate(epr2, pias, na.rm = TRUE)[, 1:2]
  )
})

# Imputing with an improper pias does nothing.
local({
  # Append a '1' to each business label to make a garbage pias
  pias2 <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(paste0(business, 1))),
      weight
    )
  )

  expect_equal(
    ms_prices$price,
    impute_prices(
      ms_prices,
      price ~ period + product,
      ea = business,
      pias = pias2,
      method = "overall-mean"
    )
  )
})

# Jumbling prices does nothing.
local({
  jumble <- c(
    27,
    29,
    10,
    15,
    24,
    5,
    38,
    34,
    40,
    13,
    4,
    30,
    19,
    28,
    22,
    23,
    2,
    35,
    6,
    16,
    17,
    25,
    3,
    39,
    11,
    20,
    1,
    33,
    9,
    18,
    14,
    8,
    32,
    26,
    31,
    12,
    21,
    36,
    7,
    37
  )
  ms_prices <- ms_prices[jumble, ]
  expect_equal(
    impute_prices(
      ms_prices,
      price ~ period + product,
      ea = business,
      pias = pias,
      method = "overall-mean"
    ),
    sp[jumble]
  )
})

# Carrying forward/backwards imputation works.
local({
  df <- data.frame(
    price = c(NA, 1, 2, NA, 3),
    period = gl(5, 1),
    product = gl(1, 5)
  )
  expect_equal(
    impute_prices(df, price ~ period + product, method = "carry-forward"),
    c(NA, 1, 2, 2, 3)
  )
  expect_equal(
    impute_prices(df, price ~ period + product, method = "carry-backward"),
    c(1, 1, 2, 3, 3)
  )
})

# Imputing with a matrix works.
local({
  ms_prices$back_price <- with(
    ms_prices,
    price[back_period(period, product)]
  )
  sp2 <- impute_prices(
    ms_prices,
    cbind(price, back_price) ~ period + product,
    ea = business,
    pias = pias,
    method = "overall-mean"
  )
  expect_equal(sp2[, 1L], sp)
  expect_equal(
    impute_prices(
      matrix(c(1, NA, NA, 4, 5, 6, 3, 2, 1, 1, NA, 3), ncol = 2),
      rep(1:2, each = 3),
      rep(1:3, 2),
      method = "carry-forward"
    ),
    matrix(c(1, 2, 1, 4, 5, 6, 3, 2, 1, 1, 2, 3), ncol = 2)
  )

  expect_error(
    impute_prices(matrix(1:9, 3), period = 1:3, product = 1:3),
    "`x` must be a two-column matrix"
  )
})

# NA products don't get imputed.
local({
  df <- data.frame(
    period = c(1, 1, 1, 2, 2, 2, 3, 3, 3),
    product = c(1, 2, 3, 1, 2, NA, 1, NA, NA),
    price = c(1, 2, 3, 4, 5, NA, 7, NA, NA),
    back_price = c(1, 2, 3, 1, 2, 3, 4, 5, NA)
  )

  expect_equal(
    impute_prices(
      df,
      cbind(price, back_price) ~ period + product,
      method = "carry-forward"
    ),
    cbind(
      price = replace(df$price, c(6, 8), c(3, 5)),
      back_price = df$back_price
    )
  )
})

# Impute rules work.
local({
  prices <- data.frame(
    period = rep(1:3, each = 3),
    product = 1:3,
    price = c(2:3, NA, 5:6, NA, 8:9, NA),
    back_price = c(1:3, 2:3, NA, 5:6, NA),
    ea = 1:3
  )

  rules <- function(x, pias) {
    if (time(x) == "1") {
      if (is.na(x["3"])) x["3"] <- x["1"]
    }
    if (time(x) == "2") {
      if (is.na(x["3"])) x["3"] <- 1
    }
    x
  }

  imputed <- impute_prices(
    prices,
    cbind(price, back_price) ~ period + product,
    ea = ea,
    pias = list(c(0, 0, 0), 1:3),
    impute_rules = rules
  )

  res <- as.matrix(prices[c("price", "back_price")])
  res[3, 1] <- 6
  res[6, ] <- c(6, 6)
  res[9, ] <- c(gmean(c(8 / 5, 9 / 6), c(5, 3)) * 6, 6)

  expect_equal(imputed, res)

  prices_long <- prices[-4]
  prices_long[10:12, ] <- prices[1:3, c(1, 2, 4, 5)]
  prices_long$period[10:12] <- 0

  imputed2 <- impute_prices(
    prices_long,
    price ~ period + product,
    ea = ea,
    pias = list(c(0, 0, 0), 1:3),
    impute_rules = rules
  )

  expect_equal(imputed2[1:9], imputed[1:9, 1])
})

#' Extract percent-change contributions
#'
#' Extract a matrix or data frame of percent-change contributions from a price
#' index.
#'
#' @export contrib
#' @family index methods
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param level,levels `[character]` The level of an index for which
#'   percent-change contributions
#'   are desired, defaulting to the first level (usually the top-level for an
#'   aggregate index). `contrib2DF()` can accept multiple levels.
#' @param period `[character]` The time periods for which percent-change
#'   contributions are desired, defaulting to all time periods.
#' @param pad `[numeric(1)]` A numeric value to pad contributions so that they
#'   fit into a
#'   rectangular array when products differ over time. The default is 0.
#' @param value `[matrix]` A numeric matrix of replacement contributions with a
#'   row for each product and a column for each time period. Recycling occurs
#'   along time periods.
#'
#' @returns
#' `contrib()` returns a matrix of percent-change contributions with a column
#' for each `period` and a row for each product (sorted) for which there are
#' contributions in `level`. Contributions are padded with `pad` to fit into a
#' rectangular array when products differ over time. The replacement methods
#' returns a copy of `x` with contributions given by the matrix `value`.
#' (`set_contrib()` is an alias that's easier to use with pipes.)
#' `set_contrib_from_index()` is a helper to return a copy of `x` with all
#' contributions set to the corresponding index value minus 1.
#'
#' `contrib2DF()` returns a data frame of contributions with four
#' columns: `period`, `level`, `product`, and `value`.
#'
#' @examples
#' prices <- data.frame(
#'   rel = 1:8,
#'   period = rep(1:2, each = 4),
#'   ea = rep(letters[1:2], 4)
#' )
#'
#' index <- elementary_index(prices, rel ~ period + ea, contrib = TRUE)
#'
#' pias <- aggregation_structure(
#'   list(c("top", "top", "top"), c("a", "b", "c")),
#'   weights = 1:3
#' )
#'
#' index <- aggregate(index, pias, na.rm = TRUE)
#'
#' # Percent-change contributions for the top-level index.
#' contrib(index)
#'
#' contrib2DF(index)
#'
#' # Calculate EA contributions for the chained index.
#' arithmetic_contributions <- function(x, w, order = 1) {
#'   (x - 1) * transmute_weights(x, w, order, to = 1)
#' }
#'
#' arithmetic_contributions(
#'   as.matrix(chain(index))[c("a", "b", "c"), 2],
#'   weights(pias)
#' )
contrib <- function(x, level = NULL, period = NULL, pad = 0) {
  level <- if (!is.null(level)) match_levels(as.character(level), x) else 1L
  period <- if (!is.null(period)) {
    match_time(as.character(period), x, several = TRUE)
  } else {
    seq_along(x$time)
  }
  pad <- as.numeric(pad)
  if (length(pad) != 1L) {
    stop("`pad` must be a length 1 numeric value")
  }
  if (is.null(x$contrib)) {
    return(
      matrix(
        numeric(0L),
        nrow = 0L,
        ncol = length(period),
        dimnames = list(product = character(0L), time = period)
      )
    )
  }
  con <- x$contrib[level, period]

  con_names <- lapply(con, names)
  products <- sort.int(unique(unlist(con_names, use.names = FALSE)))

  out <- vector("list", length(con))
  names(out) <- x$time[period]

  # Initialize `pad` contributions for all products in all time periods, then
  # replace with the actual values so products that didn't sell have `pad` and
  # not NA contributions.
  out[] <- list(structure(rep.int(pad, length(products)), names = products))
  res <- do.call(cbind, Map(replace, out, con_names, con))
  names(dimnames(res)) <- c("product", "time")
  res
}

#' @rdname contrib
#' @export
contrib2DF <- function(x, levels = NULL, period = NULL) {
  level <- if (!is.null(levels)) {
    match_levels(as.character(levels), x, several = TRUE)
  } else {
    1L
  }
  period <- if (!is.null(period)) {
    match_time(as.character(period), x, several = TRUE)
  } else {
    seq_along(x$time)
  }
  if (is.null(x$contrib)) {
    return(
      data.frame(
        period = character(0L),
        level = character(0L),
        product = character(0L),
        value = numeric(0L)
      )
    )
  }

  con <- x$contrib[level, period, drop = FALSE]

  products <- apply(con, 2L, lengths, simplify = FALSE)

  levels <- x$levels[level]
  levels <- unlist(
    lapply(products, \(z) rep.int(levels, z)),
    use.names = FALSE
  )

  periods <- rep.int(x$time[period], vapply(products, sum, numeric(1L)))

  contributions <- unlist(con)
  data.frame(
    period = periods,
    level = levels,
    # NULL if there are no contributions.
    product = as.character(names(contributions)),
    value = unname(contributions)
  )
}

#' @rdname contrib
#' @export
`contrib<-` <- function(x, level = NULL, period = NULL, value) {
  level <- if (!is.null(level)) match_levels(as.character(level), x) else 1L
  period <- if (!is.null(period)) {
    match_time(as.character(period), x, several = TRUE)
  } else {
    seq_along(x$time)
  }

  if (is.null(x$contrib)) {
    x$contrib <- contrib_skeleton(x$levels, x$time)
  }

  value <- as.matrix(value)
  if (ncol(value) == 0L) {
    stop("replacement has length zero")
  } else if (length(period) %% ncol(value) != 0) {
    warning(
      "number of items to replace is not a multiple of replacement length"
    )
  }
  value <- matrix(
    as.numeric(value),
    nrow = nrow(value),
    ncol = ncol(value),
    dimnames = dimnames(value)
  )

  products <- if (nrow(value) > 0L) {
    if (is.null(rownames(value))) {
      as.character(seq_len(nrow(value)))
    } else {
      valid_product_names(rownames(value))
    }
  }

  j <- 0
  for (t in period) {
    j <- j %% ncol(value) + 1
    con <- value[, j]
    names(con) <- products
    x$contrib[level, t] <- list(con)
  }
  validate_contrib(x)
}

#' @rdname contrib
#' @export
set_contrib <- `contrib<-`

#' @rdname contrib
#' @export
set_contrib_from_index <- function(x) {
  x$contrib <- index2contrib(x$index, x$levels, x$time)
  validate_contrib(x)
}

dat <- data.frame(
  rel = c(1:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

pias <- as_aggregation_structure(
  data.frame(level1 = 1, level2 = c(11, 12, 13, 14), weight = 1)
)

epr <- elementary_index(
  dat,
  setNames(rel, c(1:5, 1, 3, 2, 6)) ~ period + ea,
  contrib = TRUE
)
index <- aggregate(epr, pias, na.rm = TRUE)
epr2 <- elementary_index(dat, rel ~ period + ea)

# contrib works.
local({
  expect_equal(
    contrib(epr),
    matrix(
      c(0, 0.414213562373095, 2.5962965079607, 2.88444419044716),
      2,
      2,
      dimnames = list(product = c("1", "2"), time = 1:2)
    )
  )
  expect_equal(
    contrib(epr, "14"),
    matrix(c(0, 7), 1, 2, dimnames = list(product = "6", time = 1:2))
  )
  expect_equal(
    contrib(epr2),
    matrix(numeric(0), 0, 2, dimnames = list(product = NULL, time = 1:2))
  )
  expect_equal(
    contrib(epr, 12, 2),
    matrix(NA_real_, 1, 1, dimnames = list(product = 3, time = 2))
  )
})

# contrib2DF works.
local({
  expect_equal(
    contrib2DF(epr),
    data.frame(
      period = c("1", "1", "2", "2"),
      level = c("11", "11", "11", "11"),
      product = c("1", "2", "1", "2"),
      value = as.numeric(contrib(epr))
    )
  )

  expect_equal(
    contrib2DF(epr, levels(epr), 2),
    data.frame(
      period = c("2", "2", "2", "2"),
      level = c("11", "11", "12", "14"),
      product = c("1", "2", "3", "6"),
      value = c(
        contrib(epr, "11", 2),
        contrib(epr, "12", 2),
        contrib(epr, "14", 2)
      )
    )
  )

  expect_equal(
    contrib2DF(epr2),
    data.frame(
      period = character(0),
      level = character(0),
      product = character(0),
      value = numeric(0)
    )
  )
})

# Padding works.
local({
  expect_equal(
    contrib(epr, 12, pad = -99),
    matrix(
      c(1.07179676972449, 1.39230484541326, NA, -99),
      2,
      2,
      dimnames = list(product = 3:4, time = 1:2)
    )
  )
  expect_error(contrib(epr, 12, pad = 1:2))
})

# Aggregate contributions have the right form.
local({
  contributions <- contrib(index)
  expect_equal(
    rownames(contributions),
    c("1", "2", "3", "4", "5", "6")
  )
  expect_equal(colnames(contributions), c("1", "2"))
})

# Product names are correct.
local({
  expect_equal(
    suppressWarnings(
      contrib(elementary_index(
        c(a = 1, b = 2, c = 3, a = 4, a = 5),
        period = c(1, 1, 1, 2, 2),
        ea = gl(1, 5),
        contrib = TRUE,
        r = 1
      ))
    ),
    matrix(
      c(0, 0, 1 / 3, 2 / 3, 1.5, 2, 0, 0),
      4,
      2,
      dimnames = list(product = c("a", "a.1", "b", "c"), time = 1:2)
    )
  )

  w <- setNames(rep(1, 9), letters[1:9])
  expect_equal(
    contrib(epr),
    contrib(
      elementary_index(
        dat,
        rel ~ period + ea,
        weights = w,
        contrib = TRUE,
        product = c(1:5, 1, 3, 2, 6)
      )
    )
  )
})

# Replacement.
local({
  # No op.
  epr2 <- epr
  contrib(epr) <- contrib(epr)
  expect_identical(epr, epr2)

  # Replacing with change in index.
  contrib(epr) <- matrix(as.numeric(epr[1]) - 1, 1)
  expect_equal(
    contrib(epr),
    matrix(as.numeric(epr[1]) - 1, 1, dimnames = list(product = 1, time = 1:2))
  )

  # Deleting contributions, used to be give an error.
  contrib(epr, 12) <- numeric(0)
  expect_equal(
    contrib2DF(epr, 12),
    data.frame(
      period = character(0),
      level = character(0),
      product = character(0),
      value = numeric(0)
    )
  )

  # Adding contributions.
  contrib(epr, 14) <- matrix(
    c(NA, 2:3, 6, 1, NA),
    3,
    dimnames = list(product = letters[1:3], time = NULL)
  )
  expect_equal(
    contrib(epr, 14),
    matrix(
      c(NA, 2:3, 6, 1, NA),
      3,
      dimnames = list(product = letters[1:3], time = 1:2)
    )
  )

  # Errors work.
  expect_error(contrib(epr) <- matrix(0, 0, 0))
  expect_warning(contrib(epr) <- matrix(c(as.numeric(epr[1]), 3) - 1, 1, 3))
  expect_warning(contrib(epr, 14, 1) <- c(a = 1, a = NA))
})

# Setting as index works.
local({
  reset_contrib <- function(index) {
    contrib_matrix <- as.matrix(index) - 1
    for (l in levels(index)) {
      contrib(index, l) <- contrib_matrix[l, , drop = FALSE]
    }
    index
  }
  expect_equal(reset_contrib(epr), set_contrib_from_index(epr2))
})

# as.matrix works.
local({
  expect_equal(
    as.matrix(as_index(matrix(1:4, 2))),
    matrix(1:4, 2, dimnames = list(levels = 1:2, time = 1:2))
  )
  expect_equal(
    as.matrix(as_index(matrix(1:4, 1))),
    matrix(1:4, 1, dimnames = list(levels = 1, time = 1:4))
  )
  expect_equal(
    as.matrix(as_index(matrix(1:4, 4))),
    matrix(1:4, 4, dimnames = list(levels = 1:4, time = 1))
  )
})

# as.data.frame works.
local({
  expect_equal(
    as.data.frame(as_index(matrix(1:4, 2)), row.names = 4:1),
    data.frame(
      period = as.character(rep(1:2, each = 2)),
      level = as.character(1:2),
      value = 1:4,
      row.names = 4:1
    )
  )
  expect_equal(
    as.data.frame(as_index(matrix(1:4, 1))),
    data.frame(
      period = as.character(1:4),
      level = as.character(1),
      value = 1:4
    )
  )
  expect_equal(
    as.data.frame(as_index(matrix(1:4, 4))),
    data.frame(
      period = as.character(1),
      level = as.character(1:4),
      value = 1:4
    )
  )
  # Check that factors are ordered correctly
  x <- matrix(1:6, 2, dimnames = list(c("b", "a"), c(3, 1, 2)))
  expect_equal(
    as.data.frame(as_index(x), stringsAsFactors = TRUE),
    data.frame(
      period = factor(rep(c(3, 1, 2), each = 2), c(3, 1, 2)),
      level = factor(c("b", "a"), c("b", "a")),
      value = 1:6
    )
  )
  expect_equal(
    as.data.frame(as_index(x), stringsAsFactors = FALSE),
    data.frame(
      period = as.character(rep(c(3, 1, 2), each = 2)),
      level = c("b", "a"),
      value = 1:6
    )
  )
})

# as.data.frame works with contribs.
local({
  x <- as_index(matrix(1:4, 2), contrib = TRUE)
  contrib(x, 1, 1) <- c(a = -1, b = 1)
  contrib(x, 2, 2) <- numeric(0)
  res <- data.frame(
    period = as.character(c(1, 1, 2, 2)),
    level = as.character(c(1, 2, 1, 2)),
    value = 1:4
  )
  res$contrib <- c(
    list(c(a = -1, b = 1)),
    list(c("2" = 1)),
    list(c("1" = 2)),
    list(numeric(0))
  )
  expect_equal(as.data.frame(x, contrib = TRUE), res)
})

dat <- data.frame(
  rel = c(1:2, NA, 4:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

epr <- elementary_index(dat, rel ~ period + ea, contrib = TRUE, na.rm = TRUE)

# is.na works.
local({
  expect_equal(
    is.na(epr),
    matrix(
      c(FALSE, FALSE, FALSE, TRUE, FALSE, TRUE, TRUE, FALSE),
      4,
      dimnames = list(levels = levels(epr), time = time(epr))
    )
  )
  expect_true(anyNA(epr))
  expect_false(anyNA(epr[1:3, 1]))
  expect_true(anyNA(epr[1:3, 1], recursive = TRUE))

  epr <- as_index(as.matrix(epr))
  expect_false(anyNA(epr[1:3, 1]))
  expect_false(anyNA(epr[1:3, 1], recursive = TRUE))
})

# Replacement method works.
local({
  epr2 <- epr
  epr2[is.na(epr2)] <- 1:3
  m <- as.matrix(epr)
  expect_equal(as.matrix(epr2), replace(m, is.na(m), 1:3))
  expect_equal(contrib(epr, "14"), contrib(epr2, "14"))

  epr2[is.na(epr2)] <- numeric(0)
})

dat <- data.frame(
  rel = c(1:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

pias <- as_aggregation_structure(
  data.frame(level1 = 1, level2 = c(11, 12, 13, 14), weight = 1)
)

epr <- elementary_index(dat, rel ~ period + ea, contrib = TRUE)
epr2 <- elementary_index(dat, rel ~ period + ea, contrib = FALSE)
index <- aggregate(epr, pias)

# as_index makes a valid index.
local({
  res <- as_index(matrix(1:6, 2))
  expect_identical(levels(res), as.character(1:2))
  expect_identical(time(res), as.character(1:3))
  expect_error(as_index(numeric(0)))
  expect_error(as_index(matrix(1:2, 2, dimnames = list(c(1, 1), 2))))
  expect_error(as_index(matrix(1:2, 1, dimnames = list(1, c(2, 2)))))
  expect_error(as_index(matrix(1:2, 1, dimnames = list(1, c(NA, 2)))))
})

# as_index works with matrices.
local({
  expect_equal(as_index(as.matrix(epr)), epr2)
  expect_equivalent(
    contrib(as_index(as.matrix(epr), contrib = TRUE)),
    as.matrix(epr)[1, , drop = FALSE] - 1,
  )

  # A character vector used to get pass through without coercion.
  mat <- as.matrix(epr)
  mat[] <- as.character(mat)
  expect_equal(as_index(mat), epr2)
  expect_equal(as_index(as.matrix(chain(epr)), chainable = FALSE), chain(epr))
})

# as_index works for data frames.
local({
  expect_equal(as_index(as.data.frame(epr)), epr2)
  expect_equivalent(
    contrib(as_index(as.data.frame(epr), contrib = TRUE)),
    as.matrix(epr)[1, , drop = FALSE] - 1
  )
  df <- as.data.frame(epr)
  df[[1]] <- factor(df[[1]], levels = 2:1)
  expect_equal(
    as_index(df),
    with(
      dat,
      elementary_index(rel, period = factor(period, levels = 2:1), ea = ea)
    )
  )

  expect_equal(
    as_index(data.frame(1:5, 1:5, 1:5)),
    as_index(matrix(replace(NA, c(1, 7, 13, 19, 25), 1:5), 5))
  )

  expect_error(as_index(df[1:2]))
})

# as_index works for ts.
local({
  expect_identical(
    as_index(ts(matrix(1:6, 2))),
    as_index(matrix(
      c(1, 3, 5, 2, 4, 6),
      3,
      dimnames = list(paste("Series", 1:3), 1:2)
    ))
  )
})

# as_index works with contribs.
local({
  expect_equal(
    as_index(as.data.frame(epr, contrib = TRUE), contrib = TRUE),
    epr
  )

  contrib(epr2) <- numeric(0)
  expect_equal(
    as_index(as.data.frame(epr2, contrib = TRUE), contrib = TRUE),
    epr2
  )

  index2 <- aggregate(epr, pias, contrib = FALSE)
  index2df <- as.data.frame(index2, contrib = TRUE)
  contrib(index2) <- numeric(0)
  expect_equal(
    as_index(index2df, contrib = TRUE),
    index2
  )

  index2df[1, 4][[1]] <- list(c(a = 0))
  expect_equal(
    as.data.frame(as_index(index2df, contrib = TRUE), contrib = TRUE),
    index2df
  )

  index2df[1, 4][[1]] <- list(c(a = 2, a = 1, b = NA))
  expect_warning(index2 <- as_index(index2df, contrib = TRUE))
  expect_identical(rownames(contrib(index2)), c("a", "a.1", "b"))
})

# as_index works for index.
local({
  expect_equal(as_index(epr), epr)
  expect_equal(as_index(index), index)
  expect_equal(as_index(epr, chainable = FALSE), chain(epr))
  expect_equal(as_index(index, chainable = FALSE), chain(index))
  expect_equal(
    as_index(chain(aggregate(epr, pias, na.rm = TRUE)), chainable = TRUE),
    aggregate(epr2, pias, na.rm = TRUE)
  )
})

test_that("as.matrix work", {
  expect_equal(
    as.matrix(as_index(matrix(1:4, 2))),
    matrix(1:4, 2, dimnames = list(levels = 1:2, time = 1:2))
  )
  expect_equal(
    as.matrix(as_index(matrix(1:4, 1))),
    matrix(1:4, 1, dimnames = list(levels = 1, time = 1:4))
  )
  expect_equal(
    as.matrix(as_index(matrix(1:4, 4))),
    matrix(1:4, 4, dimnames = list(levels = 1:4, time = 1))
  )
})

test_that("as.data.frame works", {
  expect_equal(
    as.data.frame(as_index(matrix(1:4, 2)), row.names = 4:1),
    data.frame(
      period = as.character(rep(1:2, each = 2)),
      level = as.character(1:2),
      value = 1:4,
      row.names = 4:1
    )
  )
  expect_equal(
    as.data.frame(as_index(matrix(1:4, 1))),
    data.frame(
      period = as.character(1:4),
      level = as.character(1),
      value = 1:4
    )
  )
  expect_equal(
    as.data.frame(as_index(matrix(1:4, 4))),
    data.frame(
      period = as.character(1),
      level = as.character(1:4),
      value = 1:4
    )
  )
  # Check that factors are ordered correctly
  x <- matrix(1:6, 2, dimnames = list(c("b", "a"), c(3, 1, 2)))
  expect_equal(
    as.data.frame(as_index(x), stringsAsFactors = TRUE),
    data.frame(
      period = factor(rep(c(3, 1, 2), each = 2), c(3, 1, 2)),
      level = factor(c("b", "a"), c("b", "a")),
      value = 1:6
    )
  )
  expect_equal(
    as.data.frame(as_index(x), stringsAsFactors = FALSE),
    data.frame(
      period = as.character(rep(c(3, 1, 2), each = 2)),
      level = c("b", "a"),
      value = 1:6
    )
  )
})

test_that("as.data.frame works with contribs", {
  x <- as_index(matrix(1:4, 2), contrib = TRUE)
  contrib(x, 1, 1) <- c(a = -1, b = 1)
  contrib(x, 2, 2) <- numeric(0)
  res <- data.frame(
    period = as.character(c(1, 1, 2, 2)),
    level = as.character(c(1, 2, 1, 2)),
    value = 1:4
  )
  res$contrib <- c(
    list(c(a = -1, b = 1)),
    list(c("2" = 1)),
    list(c("1" = 2)),
    list(numeric(0))
  )
  expect_equal(as.data.frame(x, contrib = TRUE), res)
})

dat <- data.frame(
  rel = c(1:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

pias <- as_aggregation_structure(
  data.frame(level1 = 1, level2 = c(11, 12, 13, 14), weight = 1)
)

epr <- elementary_index(dat, rel ~ period + ea, contrib = TRUE)
epr2 <- elementary_index(dat, rel ~ period + ea, contrib = FALSE)
index <- aggregate(epr, pias)

test_that("as_index makes a valid index", {
  res <- as_index(matrix(1:6, 2))
  expect_identical(levels(res), as.character(1:2))
  expect_identical(time(res), as.character(1:3))
  expect_error(as_index(numeric(0)))
  expect_error(as_index(matrix(1:2, 2, dimnames = list(c(1, 1), 2))))
  expect_error(as_index(matrix(1:2, 1, dimnames = list(1, c(2, 2)))))
  expect_error(as_index(matrix(1:2, 1, dimnames = list(1, c(NA, 2)))))
})

test_that("as_index works with matrices", {
  expect_equal(as_index(as.matrix(epr)), epr2)
  expect_equal(
    contrib(as_index(as.matrix(epr), contrib = TRUE)),
    as.matrix(epr)[1, , drop = FALSE] - 1,
    ignore_attr = TRUE
  )

  # A character vector used to get pass through without coercion.
  mat <- as.matrix(epr)
  mat[] <- as.character(mat)
  expect_equal(as_index(mat), epr2)
  expect_equal(as_index(as.matrix(chain(epr)), chainable = FALSE), chain(epr))
})

test_that("as_index works for data frames", {
  expect_equal(as_index(as.data.frame(epr)), epr2)
  expect_equal(
    contrib(as_index(as.data.frame(epr), contrib = TRUE)),
    as.matrix(epr)[1, , drop = FALSE] - 1,
    ignore_attr = TRUE
  )
  df <- as.data.frame(epr)
  df[[1]] <- factor(df[[1]], levels = 2:1)
  expect_equal(
    as_index(df),
    with(
      dat,
      elementary_index(rel, period = factor(period, levels = 2:1), ea = ea)
    )
  )

  expect_equal(
    as_index(data.frame(1:5, 1:5, 1:5)),
    as_index(matrix(replace(NA, c(1, 7, 13, 19, 25), 1:5), 5))
  )

  expect_error(as_index(df[1:2]))
})

test_that("as_index works for ts", {
  expect_identical(
    as_index(ts(matrix(1:6, 2))),
    as_index(matrix(
      c(1, 3, 5, 2, 4, 6),
      3,
      dimnames = list(paste("Series", 1:3), 1:2)
    ))
  )
})

test_that("as_index works with contribs", {
  expect_equal(
    as_index(as.data.frame(epr, contrib = TRUE), contrib = TRUE),
    epr
  )

  contrib(epr2) <- numeric(0)
  expect_equal(
    as_index(as.data.frame(epr2, contrib = TRUE), contrib = TRUE),
    epr2
  )

  index2 <- aggregate(epr, pias, contrib = FALSE)
  index2df <- as.data.frame(index2, contrib = TRUE)
  contrib(index2) <- numeric(0)
  expect_equal(
    as_index(index2df, contrib = TRUE),
    index2
  )

  index2df[1, 4][[1]] <- list(c(a = 0))
  expect_equal(
    as.data.frame(as_index(index2df, contrib = TRUE), contrib = TRUE),
    index2df
  )

  index2df[1, 4][[1]] <- list(c(a = 2, a = 1, b = NA))
  expect_warning(index2 <- as_index(index2df, contrib = TRUE))
  expect_identical(rownames(contrib(index2)), c("a", "a.1", "b"))
})

test_that("as_index works for indexes", {
  expect_equal(as_index(epr), epr)
  expect_equal(as_index(index), index)
  expect_equal(as_index(epr, chainable = FALSE), chain(epr))
  expect_equal(as_index(index, chainable = FALSE), chain(index))
  expect_equal(
    as_index(chain(aggregate(epr, pias, na.rm = TRUE)), chainable = TRUE),
    aggregate(epr2, pias, na.rm = TRUE)
  )
})

dat <- data.frame(
  rel = c(1:2, NA, 4:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

epr <- elementary_index(dat, rel ~ period + ea, contrib = TRUE, na.rm = TRUE)

test_that("is.na works", {
  expect_equal(
    is.na(epr),
    matrix(
      c(FALSE, FALSE, FALSE, TRUE, FALSE, TRUE, TRUE, FALSE),
      4,
      dimnames = list(levels = levels(epr), time = time(epr))
    )
  )
  expect_true(anyNA(epr))
  expect_false(anyNA(epr[1:3, 1]))
  expect_true(anyNA(epr[1:3, 1], recursive = TRUE))

  epr <- as_index(as.matrix(epr))
  expect_false(anyNA(epr[1:3, 1]))
  expect_false(anyNA(epr[1:3, 1], recursive = TRUE))
})

test_that("replacement method works", {
  epr2 <- epr
  epr2[is.na(epr2)] <- 1:3
  m <- as.matrix(epr)
  expect_equal(as.matrix(epr2), replace(m, is.na(m), 1:3))
  expect_equal(contrib(epr, "14"), contrib(epr2, "14"))

  epr2[is.na(epr2)] <- numeric(0)
})

dat <- data.frame(
  rel = c(1:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

pias <- as_aggregation_structure(
  data.frame(level1 = 1, level2 = c(11, 12, 13, 14), weight = 1)
)

epr <- elementary_index(
  dat,
  setNames(rel, c(1:5, 1, 3, 2, 6)) ~ period + ea,
  contrib = TRUE
)
index <- aggregate(epr, pias, na.rm = TRUE)
epr2 <- elementary_index(dat, rel ~ period + ea)

test_that("contrib works", {
  expect_equal(
    contrib(epr),
    matrix(
      c(0, 0.414213562373095, 2.5962965079607, 2.88444419044716),
      2,
      2,
      dimnames = list(product = c("1", "2"), time = 1:2)
    )
  )
  expect_equal(
    contrib(epr, "14"),
    matrix(c(0, 7), 1, 2, dimnames = list(product = "6", time = 1:2))
  )
  expect_equal(
    contrib(epr2),
    matrix(numeric(0), 0, 2, dimnames = list(product = NULL, time = 1:2))
  )
  expect_equal(
    contrib(epr, 12, 2),
    matrix(NA_real_, 1, 1, dimnames = list(product = 3, time = 2))
  )
})

test_that("contrib2DF works", {
  expect_equal(
    contrib2DF(epr),
    data.frame(
      period = c("1", "1", "2", "2"),
      level = c("11", "11", "11", "11"),
      product = c("1", "2", "1", "2"),
      value = as.numeric(contrib(epr))
    )
  )

  expect_equal(
    contrib2DF(epr, levels(epr), 2),
    data.frame(
      period = c("2", "2", "2", "2"),
      level = c("11", "11", "12", "14"),
      product = c("1", "2", "3", "6"),
      value = c(
        contrib(epr, "11", 2),
        contrib(epr, "12", 2),
        contrib(epr, "14", 2)
      )
    )
  )

  expect_equal(
    contrib2DF(epr2),
    data.frame(
      period = character(0),
      level = character(0),
      product = character(0),
      value = numeric(0)
    )
  )
})

test_that("padding work", {
  expect_equal(
    contrib(epr, 12, pad = -99),
    matrix(
      c(1.07179676972449, 1.39230484541326, NA, -99),
      2,
      2,
      dimnames = list(product = 3:4, time = 1:2)
    )
  )
  expect_error(contrib(epr, 12, pad = 1:2))
})

test_that("aggregate contributions have the right form", {
  contributions <- contrib(index)
  expect_equal(
    rownames(contributions),
    c("1", "2", "3", "4", "5", "6")
  )
  expect_equal(colnames(contributions), c("1", "2"))
})

test_that("product names are correct", {
  expect_equal(
    suppressWarnings(
      contrib(elementary_index(
        c(a = 1, b = 2, c = 3, a = 4, a = 5),
        period = c(1, 1, 1, 2, 2),
        ea = gl(1, 5),
        contrib = TRUE,
        r = 1
      ))
    ),
    matrix(
      c(0, 0, 1 / 3, 2 / 3, 1.5, 2, 0, 0),
      4,
      2,
      dimnames = list(product = c("a", "a.1", "b", "c"), time = 1:2)
    )
  )

  w <- setNames(rep(1, 9), letters[1:9])
  expect_equal(
    contrib(epr),
    contrib(
      elementary_index(
        dat,
        rel ~ period + ea,
        weights = w,
        contrib = TRUE,
        product = c(1:5, 1, 3, 2, 6)
      )
    )
  )
})

test_that("replacement works", {
  # No op.
  epr2 <- epr
  contrib(epr) <- contrib(epr)
  expect_identical(epr, epr2)

  # Replacing with change in index.
  contrib(epr) <- matrix(as.numeric(epr[1]) - 1, 1)
  expect_equal(
    contrib(epr),
    matrix(as.numeric(epr[1]) - 1, 1, dimnames = list(product = 1, time = 1:2))
  )

  # Deleting contributions, used to be give an error.
  contrib(epr, 12) <- numeric(0)
  expect_equal(
    contrib2DF(epr, 12),
    data.frame(
      period = character(0),
      level = character(0),
      product = character(0),
      value = numeric(0)
    )
  )

  # Adding contributions.
  contrib(epr, 14) <- matrix(
    c(NA, 2:3, 6, 1, NA),
    3,
    dimnames = list(product = letters[1:3], time = NULL)
  )
  expect_equal(
    contrib(epr, 14),
    matrix(
      c(NA, 2:3, 6, 1, NA),
      3,
      dimnames = list(product = letters[1:3], time = 1:2)
    )
  )

  # Errors work.
  expect_error(contrib(epr) <- matrix(0, 0, 0))
  expect_warning(contrib(epr) <- matrix(c(as.numeric(epr[1]), 3) - 1, 1, 3))
  expect_warning(contrib(epr, 14, 1) <- c(a = 1, a = NA))
})

test_that("setting as index works", {
  reset_contrib <- function(index) {
    contrib_matrix <- as.matrix(index) - 1
    for (l in levels(index)) {
      contrib(index, l) <- contrib_matrix[l, , drop = FALSE]
    }
    index
  }
  expect_equal(reset_contrib(epr), set_contrib_from_index(epr2))
})

#' Extended mean
#'
#' Calculate the component-wise extended mean.
#'
#' The extended mean is also called the difference mean, Stolarsky mean, or
#' extended mean-value mean; see Bullen (2003, p. 393) for details.
#'
#' Both `x` and `y` should be strictly positive. This is not enforced, but the
#' results may not make sense when the extended mean is not defined. The usual
#' recycling rules apply when `x` and `y` are not the same length.
#'
#' By definition, the extended mean of `x`
#' and `y` is `x` when `x == y`. The `tol` argument is used
#' to test equality by checking if `abs(x - y) <= tol`.  In some cases it's
#' useful to multiply
#' `tol` by a scale factor, such as `max(abs(x), abs(y))`. This often
#' doesn't matter when making price indexes, however, as `x` and `y`
#' are usually around 1.
#'
#' @family math functions
#' @export
#'
#' @param x,y `[numeric > 0]` A strictly positive numeric vector.
#' @param order `[numeric(2)]` A pair of finite numbers giving the order of the
#'   extended mean.
#'   The default calculates the ordinary logarithmic mean. Setting either the
#'   first or second element to 1 gives the generalized logarithmic mean.
#' @param tol `[numeric > 0]` The tolerance used to determine if `x == y`.
#'   The default value is the same as [all.equal()].
#'
#' @returns
#' A numeric vector, the same length as
#' `max(length(x), length(y))`, giving the component-wise extended mean
#' of `x` and `y`.
#'
#' @references
#' Bullen, P. S. (2003). *Handbook of Means and Their Inequalities*.
#' Springer Science+Business Media.
#'
#' @examples
#' x <- 8:5
#' y <- 1:4
#'
#' # The arithmetic and geometric means are special cases of the
#' # generalized logarithmic mean.
#' all.equal(emean(x, y, c(2, 1)), (x + y) / 2)
#' all.equal(emean(x, y, c(-1, 1)), sqrt(x * y))
#'
#' # The harmonic mean cannot be expressed as a logarithmic mean, but can
#' # be expressed as an extended mean.
#' all.equal(emean(x, y, c(-2, -1)), 2 / (1 / x + 1 / y))
#'
#' # The quadratic mean is also a type of extended mean.
#' all.equal(emean(x, y, c(2, 4)), sqrt(x^2 / 2 + y^2 / 2))
#'
#' # As are heronian and centroidal means.
#' all.equal(
#'   emean(x, y, c(0.5, 1.5)),
#'   (x + sqrt(x * y) + y) / 3
#' )
#' all.equal(
#'   emean(x, y, c(2, 3)),
#'   2 / 3 * (x^2 + x * y + y^2) / (x + y)
#' )
emean <- function(
  x,
  y,
  order = c(0, 1),
  tol = .Machine$double.eps^0.5
) {
  if (not_finite_pair(order)) {
    stop("`order` must be a pair of finite numbers")
  }
  r <- order[[1]]
  s <- order[[2]]

  # Recycling x and y here avoids multiple warnings if one is not a multiple
  # length of the other.
  if (length(x) > length(y)) {
    if (length(y) > 0 && length(x) %% length(y) != 0) {
      warning("length of `x` is not a multiple of length of `y`")
      y <- rep_len(y, length(x))
    }
  } else if (length(x) < length(y)) {
    if (length(x) > 0 && length(y) %% length(x) != 0) {
      warning("length of `y` is not a multiple of length of `x`")
      x <- rep_len(x, length(y))
    }
  }

  max_len <- max(length(x), length(y))
  if (max_len > 0 && length(tol) > max_len) {
    stop("`tol` cannot be longer than `x` or `y`")
  }

  if (r == 0 && s == 0) {
    res <- sqrt(x * y)
  } else if (r == 0) {
    res <- ((x^s - y^s) / log(x / y) / s)^(1 / s)
  } else if (s == 0) {
    res <- ((x^r - y^r) / log(x / y) / r)^(1 / r)
  } else if (r == s) {
    res <- exp((x^r * log(x) - y^r * log(y)) / (x^r - y^r) - 1 / r)
  } else {
    res <- ((x^s - y^s) / (x^r - y^r) * (r / s))^(1 / (s - r))
  }
  # Set output to `x` when `x` == `y`.
  i <- which(abs(x - y) <= tol)
  res[i] <- x[(i - 1L) %% length(x) + 1L]
  res
}

#' Update weights
#'
#' Factor weights to turn the generalized mean of a product into the product
#' of generalized means. Useful for price-updating the weights in a
#' generalized-mean index.
#'
#' This function generalizes the result in section C.5 of Chapter 9 of the PPI
#' Manual for chaining the Young index, and gives a way to chain
#' generalized-mean price indexes over time. It returns a value such that
#'
#' \preformatted{gmean(x * y, w) ==
#'     gmean(x, w) * gmean(y, update_weights(x, w))}
#'
#' Factoring weights returns a value that is the same length as `x`,
#' so any missing values in `x` or `weights` will return `NA`.
#' Unless all values are `NA`, however, the result will still satisfy
#' the above identity when `na.rm = TRUE`.
#'
#' @family math functions
#' @export
#'
#' @inheritParams gmean
#'
#' @returns
#' A numeric vector the same length as `x`.
#'
#' @references
#' ILO, IMF, OECD, UNECE, and World Bank. (2004).
#' *Producer Price Index Manual: Theory and Practice*.
#' International Monetary Fund.
#'
#' @examples
#' x <- 1:3
#' y <- 4:6
#' w <- 3:1
#'
#' # Factor the arithmetic mean by chaining the calculation.
#' gmean(x * y, w)
#' gmean(x, w) * gmean(y, update_weights(x, w))
#'
#' # In cases where x and y have the same order, Chebyshev's
#' # inequality implies that the chained calculation is too small.
#' gmean(x * y, w) > gmean(x, w) * gmean(y, w)
update_weights <- function(x, weights = NULL, order = 1) {
  if (not_finite_scalar(order)) {
    stop("`order` must be a finite number")
  }
  if (!is.null(weights) && length(x) != length(weights)) {
    stop("`x` and `weights` must be the same length")
  }
  if (order == 0) {
    if (is.null(weights)) {
      weights <- rep.int(1, length(x))
    }
    if (anyNA(x)) {
      weights[is.na(x)] <- NA_real_
    }
    weights
  } else {
    if (is.null(weights)) x^order else weights * x^order
  }
}

local({
  x <- 1:5
  w <- 5:1

  # No NAs.
  expect_equal(update_weights(x, w), w * x)
  expect_equal(update_weights(x, w, order = -1), w / x)
  expect_equal(update_weights(x, w, order = 0), w)

  expect_equal(update_weights(x), x)
  expect_equal(update_weights(x, order = -1), 1 / x)
  expect_equal(update_weights(x, order = 0), rep(1, 5))

  # With NAs.
  x[1] <- NA
  w[2] <- NA

  expect_equal(update_weights(x, w), w * x)
  expect_equal(update_weights(x, w, order = -1), w / x)
  expect_equal(update_weights(x, w, order = 0), replace(w, 1, NA))

  expect_equal(update_weights(x), x)
  expect_equal(update_weights(x, order = -1), 1 / x)
  expect_equal(update_weights(x, order = 0), c(NA, rep(1, 4)))

  # Errors.
  expect_error(update_weights(x, order = NA), "`order` must be a finite number")
  expect_error(
    update_weights(x, weights = 1),
    "`x` and `weights` must be the same length"
  )
})

#---- Useful functions ----
padded_extract <- function(x, i, pad) {
  pad <- as.character(pad)
  if (length(pad) != 1L) {
    stop("'pad' must be a length 1 character")
  }
  res <- x[seq_len(i)]
  res[seq.int(to = i, length.out = max(i - length(x), 0L))] <- pad
  res
}

missing_weights <- function(x) {
  is.na(x) | x == 0
}

missing_names <- function(x) {
  anyNA(x) || !all(nzchar(x))
}

last <- function(x) {
  x[[length(x)]]
}

drop_last <- function(x) {
  x[-length(x)]
}

split_rows <- function(x, rows) {
  lapply(rows, \(i) x[i, ])
}

match_eas <- function(pias, index) {
  match(last(pias$levels), index$levels)
}

same_hierarchy <- function(x, y) {
  identical(x[1:2], x[1:2])
}

#---- Replacing contributions ----
valid_contrib <- function(contrib) {
  if (is.null(names(contrib))) {
    products <- if (length(contrib) > 0L) {
      as.character(seq_along(contrib))
    }
  } else {
    products <- valid_product_names(names(contrib))
  }
  contrib <- as.numeric(contrib)
  names(contrib) <- products
  contrib
}

index2contrib <- function(index, levels, time) {
  contrib <- Map(stats::setNames, index - 1, levels)
  dim(contrib) <- c(length(levels), length(time))
  contrib
}

#---- Product names ----
which_duplicate_products <- function(x) {
  vapply(x, anyDuplicated, integer(1L), incomparables = NA) > 0L
}

duplicate_products <- function(x) {
  any(which_duplicate_products(x))
}

sequential_names <- function(...) {
  f <- interaction(...)
  unsplit(Map(seq_len, tabulate(f)), f)
}

valid_product_names <- function(x, period = NULL) {
  x <- as.character(x)
  if (missing_names(x)) {
    stop("each product must have a non-missing or length-zero name")
  }
  if (is.null(period)) {
    if (anyDuplicated(x)) {
      warning("product names are not unique")
      x <- make.unique(x)
    } else {
      x
    }
  } else {
    period <- as.factor(period)
    xs <- split(x, period)
    dups <- which_duplicate_products(xs)
    if (any(dups)) {
      warning("product names are not unique in each time period")
      xs[dups] <- lapply(xs[dups], make.unique)
      unsplit(xs, period)
    } else {
      x
    }
  }
}

#---- Validate vector inputs ----
different_length <- function(...) {
  res <- lengths(Filter(Negate(is.null), list(...)))
  any(res != res[1L])
}

formula_vars <- function(formula, x, n = 2L) {
  formula <- stats::as.formula(formula)
  if (length(formula) != 3L) {
    stop("'formula' must have a left-hand and right-hand side")
  }
  fterms <- stats::terms(formula, data = x)
  x <- eval(attr(fterms, "variables"), x, environment(formula))
  if (length(x) != n + 1L) {
    stop(gettextf("right-hand side of 'formula' must have exactly %s terms", n))
  }
  x
}

#---- Subscript indexes ----
subscript_index <- function(x, i) {
  if (missing(i)) {
    return(seq_along(x))
  }
  if (anyNA(i)) {
    stop("cannot subscript an index with missing values")
  }
  if (is.character(i)) {
    res <- match(i, x)
  } else {
    if (is.logical(i)) {
      if (length(i) > length(x)) {
        stop("logical subscript too long")
      } else if (length(x) %% length(i) != 0) {
        warning("logical subscript is not a multiple of dimension length")
      }
    }
    res <- match(x[i], x)
  }
  if (anyNA(res)) {
    stop("subscript out of bounds for index")
  }
  res
}

subscript_index_matrix <- function(x, i) {
  if (is.logical(i)) {
    if (nrow(i) != nlevels(x) || ncol(i) != ntime(x)) {
      stop(
        "'i' must have a row for each level and a column for each ",
        "time period in 'x'"
      )
    }
    if (anyNA(i)) {
      stop("cannot subscript an index with missing values")
    }
    i <- which(i, arr.ind = TRUE)
  }

  if (ncol(i) != 2L) {
    stop("'i' must have exactly two columns")
  }
  if (is.numeric(i) && any(i < 0L, na.rm = TRUE)) {
    stop("cannot subscript an index using a matrix with negative values")
  }

  cbind(
    subscript_index(x$levels, i[, 1L]),
    subscript_index(x$time, i[, 2L])
  )
}

match_dim <- function(what = c("time", "levels")) {
  what <- match.arg(what)
  dim <- switch(what, time = "time period", levels = "index level")
  function(x, index, several = FALSE) {
    if (!several && length(x) != 1L) {
      stop(gettextf("must supply exactly one %s", dim))
    } else if (several && length(x) == 0L) {
      stop(gettextf("must supply at least one %s", dim))
    }
    i <- match(x, index[[what]])
    no_match <- is.na(i)
    if (any(no_match)) {
      stop(gettextf("'%s' is not a %s", x[no_match][1L], dim))
    }
    i
  }
}

match_levels <- match_dim("levels")

match_time <- match_dim("time")

#---- Generate index ----
index_skeleton <- function(levels, time) {
  matrix(NA_real_, length(levels), length(time))
}

contrib_skeleton <- function(levels, time) {
  matrix(list(numeric(0L)), length(levels), length(time))
}

# Backport Reduce and %||%
# TODO: Remove once min R version gets bumped.
if (getRversion() < "4.4.0") {
  Reduce <- function(
    f,
    x,
    init,
    right = FALSE,
    accumulate = FALSE,
    simplify = TRUE
  ) {
    mis <- missing(init)
    len <- length(x)
    if (len == 0L) {
      return(if (mis) NULL else init)
    }
    f <- match.fun(f)
    if (!is.vector(x) || is.object(x)) {
      x <- as.list(x)
    }
    ind <- seq_len(len)
    if (mis) {
      if (right) {
        init <- x[[len]]
        ind <- ind[-len]
      } else {
        init <- x[[1L]]
        ind <- ind[-1L]
      }
    }
    if (!accumulate) {
      if (right) {
        for (i in rev(ind)) {
          init <- forceAndCall(
            2,
            f,
            x[[i]],
            init
          )
        }
      } else {
        for (i in ind) {
          init <- forceAndCall(2, f, init, x[[i]])
        }
      }
      init
    } else {
      len <- length(ind) + 1L
      out <- vector("list", len)
      if (mis) {
        if (right) {
          out[[len]] <- init
          for (i in rev(ind)) {
            init <- forceAndCall(2, f, x[[i]], init)
            out[[i]] <- init
          }
        } else {
          out[[1L]] <- init
          for (i in ind) {
            init <- forceAndCall(2, f, init, x[[i]])
            out[[i]] <- init
          }
        }
      } else {
        if (right) {
          out[[len]] <- init
          for (i in rev(ind)) {
            init <- forceAndCall(2, f, x[[i]], init)
            out[[i]] <- init
          }
        } else {
          for (i in ind) {
            out[[i]] <- init
            init <- forceAndCall(2, f, init, x[[i]])
          }
          out[[len]] <- init
        }
      }
      if (all(lengths(out) == 1L) && simplify) {
        out <- unlist(out, recursive = FALSE)
      }
      out
    }
  }

  `%||%` <- function(x, y) {
    if (is.null(x)) y else x
  }
}

#' Get the levels for an aggregation structure
#'
#' Get the hierarchical list of levels for an aggregation structure. It is
#' an error to try and replace these values.
#'
#' @family aggregation structure methods
#' @export
#'
#' @param x `[piar_aggregation_structure]` A price index aggregation structure,
#'   as made by [aggregation_structure()].
#'
#' @returns
#' A list of character vectors giving the levels for each position in the
#' aggregation structure.
levels.piar_aggregation_structure <- function(x) {
  x$levels
}

#' @export
`levels<-.piar_aggregation_structure` <- function(x, value) {
  stop("cannot replace levels attribute for an aggregation structure")
}

#' Cut an aggregation structure
#'
#' Keep only the part of an aggregation structure above or below a certain
#' level.
#'
#' @family aggregation structure methods
#' @export
#'
#' @param x `[piar_aggregation_structure]` A price index aggregation structure,
#'   as made by [aggregation_structure()].
#' @param ... Not currently used.
#' @param level `[integer(1) > 0]` A positive integer, or something that can be
#'   coerced into one, giving the level at which to cut `x`.
#' @param na.rm `[logical(1)]` Should missing values be removed when aggregating
#'   the weights? By default, missing values are not removed.
#' @param upper `[logical(1)]` Keep only the part of `x` above `level`
#'   (the default); otherwise, return the part of `x` below `level`.
#'
#' @returns
#' A price index aggregation structure of class [`piar_aggregation_structure`].
#'
#' @examples
#' # A simple aggregation structure.
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #  |---+---|       |
#' #  111     112     121
#' #  (1)     (3)     (4)
#' aggregation_weights <- data.frame(
#'   level1 = c("1", "1", "1"),
#'   level2 = c("11", "11", "12"),
#'   ea     = c("111", "112", "121"),
#'   weight = c(1, 3, 4)
#' )
#'
#' pias <- aggregation_structure(
#'   aggregation_weights[1:3],
#'   weights = aggregation_weights[[4]]
#' )
#'
#' # Turn it into
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #     (4)         (4)
#' cut(pias, 2)
cut.piar_aggregation_structure <- function(
  x,
  level,
  ...,
  na.rm = FALSE,
  upper = TRUE
) {
  chkDots(...)
  level <- as.integer(level)
  n <- nlevels(x)
  if (level < 1L) {
    stop("`level` must be greater than or equal to 1")
  }
  if (level > n) {
    stop("`level` must be smaller than the number of levels in `x`")
  }
  if ((!upper && level == 1L) || (upper && level == n)) {
    return(x)
  }

  i <- seq.int(to = n - 1L, length.out = level - 1L)
  if (upper) {
    weights <- weights(x, ea_only = FALSE, na.rm = na.rm)
    piar_aggregation_structure(
      x$child[i],
      x$parent[i],
      x$levels[seq_len(level)],
      weights[[level]]
    )
  } else {
    piar_aggregation_structure(
      x$child[-i],
      x$parent[-i],
      x$levels[-seq_len(level - 1L)],
      x$weights
    )
  }
}

#' Get the time periods for a price index
#'
#' Methods to get and set the time periods for a price index.
#'
#' @importFrom stats time
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param value `[character]` A character vector, or something that can be
#'   coerced into one, giving the replacement time periods for `x`.
#' @param ... Not currently used.
#'
#' @returns
#' `time()` returns a character vector with the time periods for a price index.
#' `start()` and `end()` return the first and last time period.
#'
#' `ntime()` returns the number of time periods, analogous to `nlevels()`.
#'
#' The replacement method returns a copy of `x` with the time periods in
#' `value`. (`set_time()` is an alias that's easier to use with pipes.)
time.piar_index <- function(x, ...) {
  chkDots(...)
  x$time
}

#' @rdname time.piar_index
#' @export
`time<-` <- function(x, value) {
  UseMethod("time<-")
}

#' @rdname time.piar_index
#' @export
`time<-.piar_index` <- function(x, value) {
  x$time <- as.character(value)
  validate_time(x)
}

#' @rdname time.piar_index
#' @export
set_time <- `time<-`

#' @rdname time.piar_index
#' @importFrom stats start
#' @export
start.piar_index <- function(x, ...) {
  chkDots(...)
  x$time[[1L]]
}

#' @rdname time.piar_index
#' @importFrom stats end
#' @export
end.piar_index <- function(x, ...) {
  chkDots(...)
  last(x$time)
}

#' @rdname time.piar_index
#' @export
ntime <- function(x) {
  length(time(x))
}

#' Get the levels for a price index
#'
#' Methods to get and set the levels for a price index.
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param value `[character]` A character vector, or something that can be
#'   coerced into one, giving the replacement levels for `x`.
#'
#' @returns
#' `levels()` returns a character vector with the levels for a price index.
#'
#' The replacement method returns a copy of `x` with the levels in `value`.
#' (`set_levels()` is an alias that's easier to use with pipes.)
#'
#' @family index methods
#' @export
levels.piar_index <- function(x) {
  x$levels
}

#' @rdname levels.piar_index
#' @export
`levels<-.piar_index` <- function(x, value) {
  x$levels <- as.character(value)
  validate_levels(x)
}

#' @rdname levels.piar_index
#' @export
set_levels <- function(x, value) {
  levels(x) <- value
  x
}

#' Stack price indexes
#'
#' @description
#' `stack()` combines two price indexes with common levels, stacking index
#' values and percent-change contributions for one index after the other.
#'
#' `unstack()` breaks up a price index into a list of indexes for each
#' time period.
#'
#' These methods can be used in a map-reduce to make an index with multiple
#' aggregation structures (like a Paasche index).
#'
#' @name stack.piar_index
#' @aliases stack.piar_index
#' @importFrom utils stack
#' @family index methods
#' @export
#'
#' @inheritParams merge.piar_index
#'
#' @returns
#' `stack()` returns a combined price index that inherits from the same class
#' as `x`.
#'
#' `unstack()` returns a list of price indexes with the same class as `x`.
#'
#' @note
#' It may be necessary to use `rebase()` prior to stacking fixed-based price
#' indexes to ensure they have the same base period.
#'
#' @examples
#' index1 <- as_index(matrix(1:6, 2))
#'
#' index2 <- index1
#' time(index2) <- 4:6
#'
#' stack(index1, index2)
#'
#' # Unstack does the reverse.
#' all.equal(
#'   c(unstack(index1), unstack(index2)),
#'   unstack(stack(index1, index2))
#' )
stack.chainable_piar_index <- function(x, y, ...) {
  y <- as_index(y, chainable = TRUE)
  res <- NextMethod("stack")
  new_piar_index(res$index, res$contrib, res$levels, res$time, chainable = TRUE)
}

#' @rdname stack.piar_index
#' @export
stack.direct_piar_index <- function(x, y, ...) {
  y <- as_index(y, chainable = FALSE)
  res <- NextMethod("stack")
  new_piar_index(
    res$index,
    res$contrib,
    res$levels,
    res$time,
    chainable = FALSE
  )
}

#' @export
stack.piar_index <- function(x, y, ...) {
  chkDots(...)
  if (nlevels(x) != nlevels(y) || !setequal(x$levels, y$levels)) {
    stop("`x` and `y` must be indexes for the same levels")
  }
  if (any(x$time %in% y$time)) {
    stop("the same time periods cannot appear in both `x` and `y`")
  }
  if (any(x$levels != y$levels)) {
    y <- y[x$levels]
  }
  x$index <- cbind(x$index, y$index)
  if (is.null(x$contrib) && !is.null(y$contrib)) {
    x$contrib <- contrib_skeleton(x$levels, x$time)
  } else if (!is.null(x$contrib) && is.null(y$contrib)) {
    y$contrib <- contrib_skeleton(y$levels, y$time)
  }
  if (!is.null(x$contrib)) {
    x$contrib <- cbind(x$contrib, y$contrib)
  }
  x$time <- c(x$time, y$time)
  # Alternatively call validate_piar_index(x), but this is not necessary.
  x
}

#' @rdname stack.piar_index
#' @importFrom utils unstack
#' @export
unstack.chainable_piar_index <- function(x, ...) {
  NextMethod("unstack", chainable = TRUE)
}

#' @rdname stack.piar_index
#' @export
unstack.direct_piar_index <- function(x, ...) {
  NextMethod("unstack", chainable = FALSE)
}

#' @export
unstack.piar_index <- function(x, ..., chainable) {
  chkDots(...)
  res <- vector("list", ntime(x))
  names(res) <- x$time
  for (t in seq_along(res)) {
    res[[t]] <- new_piar_index(
      x$index[, t, drop = FALSE],
      x$contrib[, t, drop = FALSE],
      x$levels,
      x$time[t],
      chainable = chainable
    )
  }
  res
}

#---- Useful functions ----
padded_extract <- function(x, i, pad) {
  pad <- as.character(pad)
  if (length(pad) != 1L) {
    stop("`pad` must be a length 1 character")
  }
  res <- x[seq_len(i)]
  res[seq.int(to = i, length.out = max(i - length(x), 0L))] <- pad
  res
}

missing_weights <- function(x) {
  is.na(x) | x == 0
}

missing_names <- function(x) {
  anyNA(x) || !all(nzchar(x))
}

last <- function(x) {
  x[[length(x)]]
}

drop_last <- function(x) {
  x[-length(x)]
}

split_rows <- function(x, rows) {
  lapply(rows, \(i) x[i, ])
}

match_eas <- function(pias, index) {
  match(last(pias$levels), index$levels)
}

same_hierarchy <- function(x, y) {
  identical(x[1:2], x[1:2])
}

not_finite_scalar <- function(x) {
  length(x) != 1L || !is.finite(x)
}

not_finite_pair <- function(x) {
  length(x) != 2L || !all(is.finite(x))
}

#---- Replacing contributions ----
valid_contrib <- function(contrib) {
  if (is.null(names(contrib))) {
    products <- if (length(contrib) > 0L) {
      as.character(seq_along(contrib))
    }
  } else {
    products <- valid_product_names(names(contrib))
  }
  contrib <- as.numeric(contrib)
  names(contrib) <- products
  contrib
}

index2contrib <- function(index, levels, time) {
  contrib <- Map(stats::setNames, index - 1, levels)
  dim(contrib) <- c(length(levels), length(time))
  contrib
}

#---- Product names ----
which_duplicate_products <- function(x) {
  vapply(x, anyDuplicated, integer(1L), incomparables = NA) > 0L
}

duplicate_products <- function(x) {
  any(which_duplicate_products(x))
}

sequential_names <- function(...) {
  f <- interaction(...)
  unsplit(Map(seq_len, tabulate(f)), f)
}

valid_product_names <- function(x, period = NULL) {
  x <- as.character(x)
  if (missing_names(x)) {
    stop("each product must have a non-missing or length-zero name")
  }
  if (is.null(period)) {
    if (anyDuplicated(x)) {
      warning("product names are not unique")
      x <- make.unique(x)
    } else {
      x
    }
  } else {
    period <- as.factor(period)
    xs <- split(x, period)
    dups <- which_duplicate_products(xs)
    if (any(dups)) {
      warning("product names are not unique in each time period")
      xs[dups] <- lapply(xs[dups], make.unique)
      unsplit(xs, period)
    } else {
      x
    }
  }
}

#---- Validate vector inputs ----
different_length <- function(...) {
  res <- lengths(Filter(Negate(is.null), list(...)))
  any(res != res[1L])
}

formula_vars <- function(formula, x, n = 2L) {
  formula <- stats::as.formula(formula)
  if (length(formula) != 3L) {
    stop("`formula` must have a left-hand and right-hand side")
  }
  fterms <- stats::terms(formula, data = x)
  x <- eval(attr(fterms, "variables"), x, environment(formula))
  if (length(x) != n + 1L) {
    stop(gettextf("right-hand side of `formula` must have exactly %s terms", n))
  }
  x
}

#---- Subscript indexes ----
subscript_index <- function(x, i) {
  if (missing(i)) {
    return(seq_along(x))
  }
  if (anyNA(i)) {
    stop("cannot subscript an index with missing values")
  }
  if (is.character(i)) {
    res <- match(i, x)
  } else {
    if (is.logical(i)) {
      if (length(i) > length(x)) {
        stop("logical subscript too long")
      } else if (length(x) %% length(i) != 0) {
        warning("logical subscript is not a multiple of dimension length")
      }
    }
    res <- match(x[i], x)
  }
  if (anyNA(res)) {
    stop("subscript out of bounds for index")
  }
  res
}

subscript_index_matrix <- function(x, i) {
  if (is.logical(i)) {
    if (nrow(i) != nlevels(x) || ncol(i) != ntime(x)) {
      stop(
        "`i` must have a row for each level and a column for each ",
        "time period in `x`"
      )
    }
    if (anyNA(i)) {
      stop("cannot subscript an index with missing values")
    }
    i <- which(i, arr.ind = TRUE)
  }

  if (ncol(i) != 2L) {
    stop("`i` must have exactly two columns")
  }
  if (is.numeric(i) && any(i < 0L, na.rm = TRUE)) {
    stop("cannot subscript an index using a matrix with negative values")
  }

  cbind(
    subscript_index(x$levels, i[, 1L]),
    subscript_index(x$time, i[, 2L])
  )
}

match_dim <- function(what = c("time", "levels")) {
  what <- match.arg(what)
  dim <- switch(what, time = "time period", levels = "index level")
  function(x, index, several = FALSE) {
    if (!several && length(x) != 1L) {
      stop(gettextf("must supply exactly one %s", dim))
    } else if (several && length(x) == 0L) {
      stop(gettextf("must supply at least one %s", dim))
    }
    i <- match(x, index[[what]])
    no_match <- is.na(i)
    if (any(no_match)) {
      stop(gettextf("'%s' is not a %s", x[no_match][1L], dim))
    }
    i
  }
}

match_levels <- match_dim("levels")

match_time <- match_dim("time")

#---- Generate index ----
index_skeleton <- function(levels, time) {
  matrix(NA_real_, length(levels), length(time))
}

contrib_skeleton <- function(levels, time) {
  matrix(list(numeric(0L)), length(levels), length(time))
}

#---- Multilaterals ----
balance_products <- function(product, period) {
  ux <- unique(product)
  product <- split(product, period)
  if (duplicate_products(product)) {
    warning("there are duplicated period-product pairs")
  }
  lapply(product, \(x) match(ux, x, incomparables = NA))
}

# Backport Reduce and %||%
# TODO: Remove once min R version gets bumped.
if (getRversion() < "4.4.0") {
  Reduce <- function(
    f,
    x,
    init,
    right = FALSE,
    accumulate = FALSE,
    simplify = TRUE
  ) {
    mis <- missing(init)
    len <- length(x)
    if (len == 0L) {
      return(if (mis) NULL else init)
    }
    f <- match.fun(f)
    if (!is.vector(x) || is.object(x)) {
      x <- as.list(x)
    }
    ind <- seq_len(len)
    if (mis) {
      if (right) {
        init <- x[[len]]
        ind <- ind[-len]
      } else {
        init <- x[[1L]]
        ind <- ind[-1L]
      }
    }
    if (!accumulate) {
      if (right) {
        for (i in rev(ind)) {
          init <- forceAndCall(
            2,
            f,
            x[[i]],
            init
          )
        }
      } else {
        for (i in ind) {
          init <- forceAndCall(2, f, init, x[[i]])
        }
      }
      init
    } else {
      len <- length(ind) + 1L
      out <- vector("list", len)
      if (mis) {
        if (right) {
          out[[len]] <- init
          for (i in rev(ind)) {
            init <- forceAndCall(2, f, x[[i]], init)
            out[[i]] <- init
          }
        } else {
          out[[1L]] <- init
          for (i in ind) {
            init <- forceAndCall(2, f, init, x[[i]])
            out[[i]] <- init
          }
        }
      } else {
        if (right) {
          out[[len]] <- init
          for (i in rev(ind)) {
            init <- forceAndCall(2, f, x[[i]], init)
            out[[i]] <- init
          }
        } else {
          for (i in ind) {
            out[[i]] <- init
            init <- forceAndCall(2, f, init, x[[i]])
          }
          out[[len]] <- init
        }
      }
      if (all(lengths(out) == 1L) && simplify) {
        out <- unlist(out, recursive = FALSE)
      }
      out
    }
  }

  `%||%` <- function(x, y) {
    if (is.null(x)) y else x
  }
}

#' Expand a hierarchical classification
#'
#' Expand a character representation of a hierarchical classification to make a
#' price index aggregation structure. Expanded classifications be interacted
#' together to get all combinations of aggregation structures.
#'
#' @export
#'
#' @param x `[character]` A character vector, or something that can be coerced
#'   into one, of
#'   codes/labels for a specific level in a classification (e.g., 5-digit
#'   COICOP, 5-digit NAICS, 4-digit SIC).
#' @param width `[integer > 0]` An integer vector that gives the width of each
#'   digit in
#'   `x`. A single value is recycled to span the longest element in
#'   `x`. This cannot contain NAs. The default assumes each digit has a
#'   width of 1, as in the NAICS, NAPCS, and SIC classifications.
#' @param ... `[list]` Lists of character vectors that give the codes/labels for
#'   each
#'   level of the classification, ordered so that moving down the list goes down
#'   the hierarchy (as made by `expand_classification()`).
#' @param sep `[character(1)]` A character used to combine codes/labels across
#'   elements of `...`. The default uses `":"`.
#' @param pad `[character(1)]` A string used to pad the shorter labels for an
#'   unbalanced classification. The default pads with NA.
#'
#' @returns
#' `expand_classification()` returns a list with a entry for each level
#' in `x` giving the "digits" that represent each level in the hierarchy.
#'
#' `interact_classfications()` returns a list of lists with the same structure
#' as `expand_classification()`.
#'
#' @seealso
#' [aggregation_structure()] to make a price-index aggregation structure.
#'
#' [split_classification()] to expand a classification by splitting along
#' a delimiter.
#'
#' [combine_classifications()] for combining multiple hierarchical
#' classifications.
#'
#' `csh_from_digits()` in the \pkg{accumulate} package for different handling
#' of unbalanced classifications.
#'
#' @examples
#' # A simple classification structure.
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #  |---+---|       |
#' #  111     112     121
#'
#' expand_classification(c("111", "112", "121"))
#'
#' # Expanding more complex classifications:
#' # ... if last 'digit' is either TA or TS.
#' expand_classification(
#'   c("111TA", "112TA", "121TS"),
#'   width = c(1, 1, 1, 2)
#' )
#'
#' # ... if first 'digit' is either 11 or 12.
#' expand_classification(c("111", "112", "121"), width = c(2, 1))
#'
#' # ...if there are delimiters in the classification (like COICOP).
#' expand_classification(c("01.1.1", "01.1.2", "01.2.1"), width = 2)
expand_classification <- function(x, width = 1L, pad = NA) {
  x <- as.character(x)
  if (length(x) == 0L) {
    return(list())
  }
  width <- as.integer(width)
  if (anyNA(width)) {
    stop("`width` cannot contain NAs")
  }
  if (any(width <= 0L)) {
    stop("`width` must be at least 1")
  }

  if (length(width) == 1L) {
    longest <- max(nchar(x), 1L, na.rm = TRUE)
    width <- rep.int(width, ceiling(longest / width))
  }
  w <- cumsum(width)
  x <- strsplit(x, character(0L), fixed = TRUE)
  lapply(w, \(i) {
    vapply(
      x,
      \(x) paste(padded_extract(x, i, pad), collapse = ""),
      character(1L)
    )
  })
}

#' @rdname expand_classification
#' @export
interact_classifications <- function(..., sep = ":") {
  dots <- list(...)
  if (length(dots) == 0L || all(lengths(dots) == 0L)) {
    return(list())
  }
  if (any(lengths(dots) == 0L)) {
    stop("each element in `...` must be a non-empty list")
  }
  len <- unlist(lapply(dots, lengths), use.names = FALSE)
  n <- len[1L]
  atomics <- unlist(lapply(dots, \(x) lapply(x, is.atomic)), use.names = FALSE)
  if (any(len != n) || n == 0L || !all(atomics)) {
    stop(
      "each element in `...` must contain a list representing an ",
      "aggregation structure"
    )
  }
  interact <- function(x, y) {
    rapply(x, \(ix) lapply(y, \(iy) paste(ix, iy, sep = sep)), how = "replace")
  }
  res <- unlist(Reduce(interact, dots), use.names = FALSE)
  m <- length(last(dots))
  k <- m * n
  res <- unname(split(res, gl(length(res) / k, k)))
  lapply(res, \(x) unname(split(x, gl(m, n))))
}

# Combining classifications works.
local({
  x <- c("111", "121", "122")
  y <- c("01", "02", "01")
  expect_identical(
    combine_classifications(
      expand_classification(x),
      expand_classification(y)
    ),
    expand_classification(paste(x, y, sep = "."), c(1, 1, 1, 2, 1))
  )
})

# Corner cases work.
local({
  x <- list(c("11", "12"), c("111", "121"))
  expect_identical(combine_classifications(x, list()), x)
  expect_identical(combine_classifications(list(), x), x)
  expect_identical(combine_classifications(list(), list()), list())
})

# Combining classifications errors with unequal length.
local({
  expect_error(
    combine_classifications(
      list(c("1", "1"), c("11", "12")),
      list(c("01", "02"), c("011"))
    )
  )
})

# Split agrees with expand.
local({
  expect_identical(
    split_classification(c("111", "112", "121"), "", sep = ""),
    expand_classification(c("111", "112", "121"))
  )
  expect_identical(
    split_classification(c("01.1.1", "01.1.2", "01.2.1"), ".", fixed = TRUE),
    expand_classification(c("01.1.1", "01.1.2", "01.2.1"), width = 2)
  )
  expect_identical(
    split_classification(
      c("01.1.1", "01.1.2.0", "01.2.1.0"),
      ".",
      fixed = TRUE,
      pad = "00"
    ),
    expand_classification(
      c("01.1.1.", "01.1.2.0", "01.2.1.0"),
      width = 2,
      pad = "00"
    )
  )
})

# Missing levels are padded correctly.
local({
  expect_identical(
    split_classification(c("ab.c.d", "a.bc..e"), ".", fixed = TRUE),
    list(
      c("ab", "a"),
      c("ab.c", "a.bc"),
      c("ab.c.d", "a.bc."),
      c("ab.c.d.NA", "a.bc..e")
    )
  )
  expect_identical(
    split_classification(
      c("ab.c.d", "a.bc..e"),
      ".",
      fixed = TRUE,
      pad = "zz"
    ),
    list(
      c("ab", "a"),
      c("ab.c", "a.bc"),
      c("ab.c.d", "a.bc."),
      c("ab.c.d.zz", "a.bc..e")
    )
  )
})

# Works with empty strings.
local({
  expect_identical(split_classification(character(0), "1"), list())
  expect_identical(split_classification(c("", ""), "1"), list(c("", "")))
  expect_identical(
    split_classification(c("", "a"), "a"),
    list(c(NA, ""))
  )
})

test_that("split agrees with expand", {
  expect_identical(
    split_classification(c("111", "112", "121"), "", sep = ""),
    expand_classification(c("111", "112", "121"))
  )
  expect_identical(
    split_classification(c("01.1.1", "01.1.2", "01.2.1"), ".", fixed = TRUE),
    expand_classification(c("01.1.1", "01.1.2", "01.2.1"), width = 2)
  )
  expect_identical(
    split_classification(
      c("01.1.1", "01.1.2.0", "01.2.1.0"),
      ".",
      fixed = TRUE,
      pad = "00"
    ),
    expand_classification(
      c("01.1.1.", "01.1.2.0", "01.2.1.0"),
      width = 2,
      pad = "00"
    )
  )
})

test_that("missing levels are padded correctly", {
  expect_identical(
    split_classification(c("ab.c.d", "a.bc..e"), ".", fixed = TRUE),
    list(
      c("ab", "a"),
      c("ab.c", "a.bc"),
      c("ab.c.d", "a.bc."),
      c("ab.c.d.NA", "a.bc..e")
    )
  )
  expect_identical(
    split_classification(
      c("ab.c.d", "a.bc..e"),
      ".",
      fixed = TRUE,
      pad = "zz"
    ),
    list(
      c("ab", "a"),
      c("ab.c", "a.bc"),
      c("ab.c.d", "a.bc."),
      c("ab.c.d.zz", "a.bc..e")
    )
  )
})

test_that("works with empty strings", {
  expect_identical(split_classification(character(0), "1"), list())
  expect_identical(split_classification(c("", ""), "1"), list(c("", "")))
  expect_identical(
    split_classification(c("", "a"), "a"),
    list(c(NA, ""))
  )
})

test_that("combining classifications works", {
  x <- c("111", "121", "122")
  y <- c("01", "02", "01")
  expect_identical(
    combine_classifications(
      expand_classification(x),
      expand_classification(y)
    ),
    expand_classification(paste(x, y, sep = "."), c(1, 1, 1, 2, 1))
  )
})

test_that("corner cases work", {
  x <- list(c("11", "12"), c("111", "121"))
  expect_identical(combine_classifications(x, list()), x)
  expect_identical(combine_classifications(list(), x), x)
  expect_identical(combine_classifications(list(), list()), list())
})

test_that("combining classifications errors with unequal length", {
  expect_error(
    combine_classifications(
      list(c("1", "1"), c("11", "12")),
      list(c("01", "02"), c("011"))
    )
  )
})

pias <- with(
  ms_weights,
  aggregation_structure(
    c(expand_classification(classification), list(business)),
    weight
  )
)

sp <- impute_prices(
  ms_prices,
  price ~ period + product,
  ea = business,
  pias = pias,
  method = "overall-mean"
)

test_that("a length 0 inputs returns a length 0 output", {
  expect_length(
    impute_prices(
      integer(0),
      period = factor(integer(0), 1:5),
      product = integer(0),
      method = "carry-forward"
    ),
    0
  )
  expect_length(
    impute_prices(
      integer(0),
      period = integer(0),
      product = integer(0),
      ea = integer(0),
      pias = pias,
      method = "overall-mean"
    ),
    0
  )
  expect_equal(
    impute_prices(
      cbind(integer(0), integer(0)),
      period = integer(0),
      product = integer(0),
      ea = integer(0),
      pias = pias,
      method = "overall-mean"
    ),
    matrix(numeric(0), 0, 2)
  )
})

test_that("imputing shadow prices does noting", {
  expect_equal(
    sp,
    impute_prices(
      ms_prices,
      sp ~ period + product,
      ea = business,
      pias = pias,
      method = "overall-mean"
    )
  )

  expect_equal(
    sp,
    impute_prices(
      ms_prices,
      sp ~ period + product,
      ea = business,
      pias = pias,
      weights = price,
      method = "overall-mean"
    )
  )
})

test_that("sp imputation is the same are regular parental in periods 1, 2", {
  epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    na.rm = TRUE
  )

  epr2 <- elementary_index(
    ms_prices,
    price_relative(sp, period = period, product = product) ~ period + business,
    na.rm = TRUE
  )

  # B2 is imputed from above the EA level
  expect_equal(epr[c(1, 3:4), 1:3], epr2[c(1, 3:4), 1:3])

  expect_equal(
    aggregate(epr, pias, na.rm = TRUE)[, 1:2],
    aggregate(epr2, pias, na.rm = TRUE)[, 1:2]
  )
})

test_that("imputing with an improper pias does nothing", {
  # Append a '1' to each business label to make a garbage pias
  pias2 <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(paste0(business, 1))),
      weight
    )
  )

  expect_equal(
    ms_prices$price,
    impute_prices(
      ms_prices,
      price ~ period + product,
      ea = business,
      pias = pias2,
      method = "overall-mean"
    )
  )
})

test_that("jumbling prices does nothing", {
  jumble <- c(
    27,
    29,
    10,
    15,
    24,
    5,
    38,
    34,
    40,
    13,
    4,
    30,
    19,
    28,
    22,
    23,
    2,
    35,
    6,
    16,
    17,
    25,
    3,
    39,
    11,
    20,
    1,
    33,
    9,
    18,
    14,
    8,
    32,
    26,
    31,
    12,
    21,
    36,
    7,
    37
  )
  ms_prices <- ms_prices[jumble, ]
  expect_equal(
    impute_prices(
      ms_prices,
      price ~ period + product,
      ea = business,
      pias = pias,
      method = "overall-mean"
    ),
    sp[jumble]
  )
})

test_that("carrying forward/backwards imputation works", {
  df <- data.frame(
    price = c(NA, 1, 2, NA, 3),
    period = gl(5, 1),
    product = gl(1, 5)
  )
  expect_equal(
    impute_prices(df, price ~ period + product, method = "carry-forward"),
    c(NA, 1, 2, 2, 3)
  )
  expect_equal(
    impute_prices(df, price ~ period + product, method = "carry-backward"),
    c(1, 1, 2, 3, 3)
  )
})

test_that("imputing with a matrix works", {
  ms_prices$back_price <- with(
    ms_prices,
    price[gpindex::back_period(period, product)]
  )
  sp2 <- impute_prices(
    ms_prices,
    cbind(price, back_price) ~ period + product,
    ea = business,
    pias = pias,
    method = "overall-mean"
  )
  expect_equal(sp2[, 1L], sp)
  expect_equal(
    impute_prices(
      matrix(c(1, NA, NA, 4, 5, 6, 3, 2, 1, 1, NA, 3), ncol = 2),
      rep(1:2, each = 3),
      rep(1:3, 2),
      method = "carry-forward"
    ),
    matrix(c(1, 2, 1, 4, 5, 6, 3, 2, 1, 1, 2, 3), ncol = 2)
  )
})

#' Missing values in a price index
#'
#' Identify missing values in a price index.
#'
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param recursive `[logical(1)]` Check if `x` also has missing percent-change
#'   contributions. By default only index values are checked for missingness.
#'
#' @returns
#' `is.na()` returns a logical matrix, with a row for each level of `x` and a
#' columns for each time period, that indicates which index values are missing.
#'
#' `anyNA()` returns `TRUE` if any index values are missing, or percent-change
#' contributions (if `recursive = TRUE`).
#'
#' @examples
#' index <- as_index(matrix(c(1, 2, 3, NA, 5, NA), 2))
#'
#' anyNA(index)
#' is.na(index)
#'
#' # Carry forward imputation.
#' index[is.na(index)] <- 1
#' index
is.na.piar_index <- function(x) {
  is.na(as.matrix(x))
}

#' @rdname is.na.piar_index
#' @export
anyNA.piar_index <- function(x, recursive = FALSE) {
  res <- anyNA(x$index)
  if (recursive && !res) {
    anyNA(x$contrib, recursive = TRUE)
  } else {
    res
  }
}

#' Get the weights for an aggregation structure
#'
#' Get and set the weights for a price index aggregation structure.
#'
#' @importFrom stats weights
#' @family aggregation structure methods
#' @export
#'
#' @param object `[piar_aggregation_structure]` A price index aggregation
#'   structure, as made by [aggregation_structure()].
#' @param ea_only `[logical(1)]` Should weights be returned for only the
#'   elementary aggregates
#'   (the default)? Setting to `FALSE` gives the weights for the entire
#'   aggregation structure.
#' @param na.rm `[logical(1)]` Should missing values be removed when aggregating
#'   the weights (i.e., when `ea_only = FALSE`)? By default, missing values are
#'   not removed.
#' @param value `[numeric >= 0]` A numeric vector of weights for the elementary
#'   aggregates of `object`.
#' @param ... Not currently used.
#'
#' @returns
#' `weights()` returns a named vector of weights for the elementary aggregates.
#' The replacement method replaces these values without changing the
#' aggregation structure. (`set_weights()` is an alias that's easier to use with
#' pipes.)
#'
#' If `ea_only = FALSE` then the return value is a list
#' with a named vector of weights for each level in the aggregation structure.
#'
#' @examples
#' # A simple aggregation structure.
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #  |---+---|       |
#' #  111     112     121
#' #  (1)     (3)     (4)
#' aggregation_weights <- data.frame(
#'   level1 = c("1", "1", "1"),
#'   level2 = c("11", "11", "12"),
#'   ea     = c("111", "112", "121"),
#'   weight = c(1, 3, 4)
#' )
#'
#' pias <- as_aggregation_structure(aggregation_weights)
#'
#' # Extract the weights.
#' weights(pias)
#'
#' # ... or update them.
#' weights(pias) <- 1:3
#' weights(pias)
weights.piar_aggregation_structure <- function(
  object,
  ...,
  ea_only = TRUE,
  na.rm = FALSE
) {
  chkDots(...)
  names(object$weights) <- last(object$levels)
  if (ea_only) {
    return(object$weights)
  }
  res <- vector("list", nlevels(object))
  names(res) <- rev(names(object$levels))
  res[[1L]] <- object$weights

  for (i in seq_along(res)[-1L]) {
    res[[i]] <- vapply(
      object$child[[i - 1L]],
      \(z) sum(res[[i - 1L]][z], na.rm = na.rm),
      numeric(1L)
    )
  }
  rev(res)
}

#' @rdname weights.piar_aggregation_structure
#' @export
`weights<-` <- function(object, value) {
  UseMethod("weights<-")
}

#' @rdname weights.piar_aggregation_structure
#' @export
`weights<-.piar_aggregation_structure` <- function(object, value) {
  object$weights[] <- as.numeric(value)
  validate_pias_weights(object)
}

#' @rdname weights.piar_aggregation_structure
#' @export
set_weights <- `weights<-`

#' Chain and rebase a price index
#'
#' @description
#' Chain a period-over-period index by taking the cumulative product of its
#' values to turn it into a fixed-base (direct) index.
#'
#' Unchain a fixed-base index by dividing its values for successive periods to
#' get a period-over-period index.
#'
#' Rebase a fixed-base index by dividing its values with the value of the index
#' in the new base period.
#'
#' @details
#' The default methods attempt to coerce `x` into an index with
#' [as_index()] prior to chaining/unchaining/rebasing.
#'
#' Chaining an index takes the cumulative product of the index values for each
#' level; this is roughly the same as
#' `t(apply(as.matrix(x), 1, cumprod)) * link`. Unchaining does the opposite,
#' so these are inverse operations. Note that unchaining a period-over-period
#' index does nothing, as does chaining a fixed-base index.
#'
#' Rebasing a fixed-base index divides the values for each level of this index
#' by the corresponding values for each level in the new base period. It's
#' roughly the same as `as.matrix(x) / base`. Like unchaining, rebasing a
#' period-over-period index does nothing.
#'
#' Percent-change contributions are removed when chaining/unchaining/rebasing
#' an index as it's not usually possible to update them correctly.
#'
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param link `[numeric > 0]` A numeric vector, or something that can coerced
#'   into one, of
#'   link values for each level in `x`. The default is equivalent to a vector
#'   of 1s so that no linking is done.
#' @param base `[numeric > 0 | character(1)]` A numeric vector, or something
#'   that can coerced into one, of
#'   base-period index values for each level in `x`. The default is a equivalent
#'   to a vector
#'   of 1s so that the base period remains the same. If `base` is a length-one
#'   character vector giving a time period of `x` then the index values for this
#'   time period are used as the base-period values.
#' @param ... Further arguments passed to or used by methods.
#'
#' @returns
#' `chain()` and `rebase()` return a fixed-base index that inherits
#' from [`direct_piar_index`].
#'
#' `unchain()` returns a period-over-period index that inherits from
#' [`chainable_piar_index`].
#'
#' @examples
#' index <- as_index(matrix(1:9, 3))
#'
#' # Make period 0 the fixed base period.
#' chain(index)
#'
#' # Chaining and unchaining reverse each other.
#' all.equal(index, unchain(chain(index)))
#'
#' # Change the base period to period 2 (note the
#' # loss of information for period 0).
#' index <- chain(index)
#' rebase(index, base = index[, 2])
chain <- function(x, ...) {
  UseMethod("chain")
}

#' @rdname chain
#' @export
chain.default <- function(x, ...) {
  chain(as_index(x), ...)
}

#' @rdname chain
#' @export
chain.chainable_piar_index <- function(x, ..., link = NULL) {
  chkDots(...)
  if (!is.null(link)) {
    link <- as.numeric(link)
    if (any(link <= 0, na.rm = TRUE)) {
      stop("`link` must be strictly positive")
    }
    if (length(link) != nlevels(x)) {
      stop("`link` must have a value for each level of `x`")
    }
    x$index[, 1L] <- x$index[, 1L] * link
  }
  x$index[] <- do.call(rbind, apply(x$index, 1L, cumprod, simplify = FALSE))
  # Contributions are difficult to chain, so remove them.
  new_piar_index(x$index, NULL, x$levels, x$time, chainable = FALSE)
}

#' @export
chain.direct_piar_index <- function(x, ...) {
  chkDots(...)
  x
}

#' @rdname chain
#' @export
unchain <- function(x, ...) {
  UseMethod("unchain")
}

#' @rdname chain
#' @export
unchain.default <- function(x, ...) {
  unchain(as_index(x, chainable = FALSE), ...)
}

#' @export
unchain.chainable_piar_index <- function(x, ...) {
  chkDots(...)
  x
}

#' @rdname chain
#' @export
unchain.direct_piar_index <- function(x, ..., base = NULL) {
  chkDots(...)
  if (!is.null(base)) {
    if (length(base) == 1L && is.character(base)) {
      base <- x$index[, match_time(base, x)] / x$index[, 1L]
    } else {
      base <- as.numeric(base)
      if (any(base <= 0, na.rm = TRUE)) {
        stop("`base` must be strictly positive")
      }
      if (length(base) != nlevels(x)) {
        stop("`base` must have a value for each level of `x`")
      }
    }
  }
  x$index[, -1L] <- x$index[, -1L] / x$index[, -ncol(x$index)]
  if (!is.null(base)) {
    x$index[, 1L] <- x$index[, 1L] * base
  }
  # Contributions are difficult to unchain, so remove them.
  new_piar_index(x$index, NULL, x$levels, x$time, chainable = TRUE)
}

#' @rdname chain
#' @export
rebase <- function(x, ...) {
  UseMethod("rebase")
}

#' @rdname chain
#' @export
rebase.default <- function(x, ...) {
  rebase(as_index(x, chainable = FALSE), ...)
}

#' @export
rebase.chainable_piar_index <- function(x, ...) {
  chkDots(...)
  x
}

#' @rdname chain
#' @export
rebase.direct_piar_index <- function(x, ..., base = NULL) {
  chkDots(...)
  if (!is.null(base)) {
    if (length(base) == 1L && is.character(base)) {
      base <- x$index[, match_time(base, x)]
    } else {
      base <- as.numeric(base)
      if (any(base <= 0, na.rm = TRUE)) {
        stop("`base` must be strictly positive")
      }
      if (length(base) != nlevels(x)) {
        stop("`base` must have a value for each level of `x`")
      }
    }
    x$index[] <- x$index / base
  }
  # Contributions are difficult to rebase, so remove them.
  new_piar_index(x$index, NULL, x$levels, x$time, chainable = FALSE)
}

# Matched-sample index aggregates correctly.
local({
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    na.rm = TRUE
  )

  ms_pias <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index <- aggregate(ms_epr, ms_pias, na.rm = TRUE)

  res <- c(
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1.30072391107879,
    1.30072391107879,
    1.30072391107879,
    0.894909688013136,
    1.30072391107879,
    2.02000360773041,
    1.30072391107879,
    1.30072391107879,
    1.06307432720167,
    1.06307432720167,
    1.06307432720167,
    0.334293948126801,
    1.06307432720167,
    1.63533549062897,
    1.06307432720167,
    1.06307432720167,
    2.73476132776126,
    1.57451535678038,
    4.57628618296306,
    1.57451535678038,
    2.77045633390764,
    0.537995998322608,
    4.57628618296306,
    4.57628618296306
  )
  res <- matrix(
    res,
    8,
    4,
    dimnames = list(
      levels = c("1", "11", "12", paste0("B", 1:5)),
      time = sprintf("2020%02d", 1:4)
    )
  )
  expect_equal(as.matrix(ms_index), res)

  # Same as matrix calculation
  expect_equal(
    as.matrix(ms_pias) %*% as.matrix(chain(ms_index[paste0("B", 1:5)])),
    as.matrix(chain(ms_index[1:3, ]))
  )

  # Lower levels add up
  expect_equal(
    apply(
      as.matrix(chain(ms_index)[4:8, ]),
      2,
      weighted.mean,
      weights(ms_pias, ea_only = FALSE)[[3]]
    ),
    as.matrix(chain(ms_index))[1, ]
  )
  expect_equal(
    apply(
      as.matrix(chain(ms_index)[2:3, ]),
      2,
      weighted.mean,
      weights(ms_pias, ea_only = FALSE)[[2]]
    ),
    as.matrix(chain(ms_index))[1, ]
  )

  # Re aggregating does nothing to the index values
  expect_equal(as.matrix(aggregate(ms_index, ms_pias)), as.matrix(ms_index))
  expect_equal(aggregate(chain(ms_index), ms_pias), chain(ms_index))
  expect_equal(
    as.matrix(aggregate(ms_index, ms_pias, na.rm = TRUE)),
    as.matrix(ms_index)
  )

  # Re aggregating doesn't breaks contributions for imputed indexes
  expect_equal(
    colSums(contrib(ms_index), na.rm = TRUE),
    colSums(contrib(aggregate(ms_index, ms_pias)), na.rm = TRUE)
  )
  expect_equal(
    ms_index,
    aggregate(ms_index, ms_pias, duplicate_contrib = "sum")
  )
  expect_equal(
    contrib(aggregate(ms_index, ms_pias)),
    contrib(aggregate(aggregate(ms_index, ms_pias), ms_pias))
  )

  # Two step aggregation gives the same result
  pias2 <- aggregation_structure(
    list(c(1, 1), c(11, 12)),
    weights(ms_pias, ea_only = FALSE)[[2]]
  )
  expect_equal(as.matrix(aggregate(ms_index, pias2)), as.matrix(ms_index[1:3]))
  expect_equal(
    contrib(aggregate(aggregate(ms_index, ms_pias), pias2)),
    contrib(aggregate(ms_index, ms_pias))
  )

  # Re-arranging the index shouldn't do anything
  s1 <- c(
    14,
    16,
    26,
    28,
    24,
    29,
    11,
    32,
    36,
    2,
    22,
    34,
    6,
    7,
    10,
    17,
    8,
    27,
    37,
    1,
    12,
    33,
    20,
    3,
    9,
    40,
    13,
    4,
    38,
    23,
    31,
    15,
    25,
    39,
    21,
    30,
    35,
    19,
    18,
    5
  )
  s2 <- c(5, 3, 4, 1, 2)
  ms_epr <- elementary_index(
    ms_prices[s1, ],
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    na.rm = TRUE
  )

  ms_pias <- with(
    ms_weights[s2, ],
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index2 <- aggregate(ms_epr, ms_pias, na.rm = TRUE)
  expect_equal(
    as.matrix(ms_index2)[levels(ms_index), ],
    as.matrix(ms_index)
  )
  expect_equal(
    contrib(ms_index),
    contrib(ms_index2)
  )
  expect_equal(
    as.matrix(ms_pias) %*% as.matrix(chain(ms_index2[paste0("B", s2)])),
    as.matrix(chain(ms_index[c(1, 3, 2), ]))
  )

  # Aggregated contributions should add up
  expect_equal(
    as.matrix(ms_index)[1, ],
    colSums(contrib(ms_index), na.rm = TRUE) + 1
  )
  expect_equal(
    as.matrix(ms_index)[2, ],
    colSums(contrib(ms_index, "11"), na.rm = TRUE) + 1
  )

  # Check that weights are getting price updated correctly
  w <- apply(as.matrix(chain(ms_index)[4:8, ]), 2, `*`, ms_weights$weight)

  expect_equal(weights(update(ms_pias, ms_index), ea_only = TRUE), w[s2, 4])

  expect_equal(
    weights(update(ms_pias, ms_index, period = "202003"), ea_only = TRUE),
    w[s2, 3]
  )
})

# A weird index aggregates correctly.
local({
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    order = 0.2
  )

  ms_pias <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index <- aggregate(ms_epr, ms_pias, order = -1.7, na.rm = TRUE)

  res <- c(
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.31335391535665,
    1.00564223748248,
    5.47029871913538,
    1.00564223748248,
    3.15295124015891,
    0.551843416514059,
    5.47029871913538,
    5.47029871913538
  )
  res <- matrix(
    res,
    8,
    4,
    dimnames = list(
      levels = c("1", "11", "12", paste0("B", 1:5)),
      time = sprintf("2020%02d", 1:4)
    )
  )
  expect_equal(as.matrix(ms_index), res)

  expect_equal(
    as.matrix(aggregate(ms_index, ms_pias, order = -1.7, na.rm = TRUE)),
    as.matrix(ms_index)
  )

  expect_equal(
    aggregate(chain(ms_index), ms_pias, order = -1.7),
    chain(ms_index)
  )

  expect_equal(
    as.matrix(ms_index)[1, ],
    colSums(contrib(ms_index), na.rm = TRUE) + 1
  )
  expect_equal(
    as.matrix(ms_index)[1, ],
    colSums(contrib(aggregate(ms_index, ms_pias, order = -1.7)), na.rm = TRUE) +
      1
  )
  expect_equal(
    aggregate(ms_index, ms_pias, order = -1.7, duplicate_contrib = "sum"),
    ms_index
  )

  expect_equal(
    apply(
      as.matrix(chain(ms_index)[2:3, ]),
      2,
      \(x, w) gmean(x, w, order = -1.7),
      weights(ms_pias, ea_only = FALSE)[[2]]
    ),
    as.matrix(chain(ms_index))[1, ]
  )

  expect_equal(
    as.matrix(
      as.matrix(ms_pias, sparse = TRUE) %*%
        as.matrix(chain(ms_index[paste0("B", 1:5)]))^(-1.7)
    )^(1 / -1.7),
    as.matrix(chain(ms_index[1:3, ]))
  )

  ms_index <- aggregate(ms_epr, ms_pias, order = -1.7)

  expect_equal(aggregate(ms_index, ms_pias, order = -1.7), ms_index)

  expect_equal(as.matrix(ms_index)[1, ], colSums(contrib(ms_index)) + 1)
  expect_equal(
    as.matrix(ms_index)["B2", ],
    colSums(contrib(ms_index, "B2")) + 1
  )
})

# Fixed-sample index aggregates correctly.
local({
  fs_epr <- elementary_index(
    fs_prices,
    price_relative(price, period = period, product = business) ~
      period + classification,
    weights = weight,
    contrib = TRUE
  )

  fs_pias <- with(
    fs_weights,
    aggregation_structure(expand_classification(classification), weight)
  )

  fs_index <- aggregate(fs_epr, fs_pias, na.rm = TRUE)

  res <- c(
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    0.687039919074495,
    0.935841069334005,
    0.612229588412764,
    0.611111111111111,
    0.935841069334005,
    0.935841069334005,
    0.612229588412764,
    0.612229588412764,
    0.611111111111111,
    2.45961303155833,
    2.45961303155833,
    2.48533513976043,
    1.68506493506494,
    2.45961303155833,
    2.45961303155833,
    2.48533513976043,
    2.48533513976043,
    1.68506493506494,
    1.08265987174601,
    1.08265987174601,
    1.08265987174601,
    1.08265987174601,
    1.08265987174601,
    1.08265987174601,
    1.61372443582219,
    0.860786536350297,
    1.08265987174601
  )
  res <- matrix(
    res,
    9,
    4,
    dimnames = list(
      levels = c(1, 11:13, 111, 112, 121, 122, 131),
      time = sprintf("2020%02d", 1:4)
    )
  )
  expect_equal(as.matrix(fs_index), res)

  # Re-aggregating the index shouldn't do anything
  expect_equal(as.matrix(aggregate(fs_index, fs_pias)), as.matrix(fs_index))

  # Contributions should add up
  expect_equal(
    as.matrix(fs_index)[1, ],
    colSums(contrib(fs_index), na.rm = TRUE) + 1
  )

  # Check adding up of lower level indexes
  expect_equal(
    apply(
      as.matrix(chain(fs_index)[5:9, ]),
      2,
      weighted.mean,
      weights(fs_pias, ea_only = FALSE)[[3]]
    ),
    as.matrix(chain(fs_index))[1, ]
  )

  expect_equal(
    apply(
      as.matrix(chain(fs_index)[2:4, ]),
      2,
      weighted.mean,
      weights(fs_pias, ea_only = FALSE)[[2]]
    ),
    as.matrix(chain(fs_index))[1, ]
  )

  # Non-missing indexes should be the same when missing values are not removed
  fs_index2 <- aggregate(fs_epr, fs_pias)
  expect_equal(
    as.matrix(fs_index2)[!is.na(as.matrix(fs_index2))],
    as.matrix(fs_index)[!is.na(as.matrix(fs_index2))]
  )

  expect_equal(
    as.matrix(fs_index2)["121", ],
    colSums(contrib(fs_index2, "121"), na.rm = TRUE) + 1
  )

  expect_equal(
    as.matrix(fs_index2)["13", 1:3],
    contrib(fs_index2, "13")[, 1:3] + 1
  )
})

# Fixed-based index aggregates correctly.
local({
  prices <- data.frame(
    price = 1:15,
    period = letters[1:3],
    product = rep(1:5, each = 3),
    ea = rep(c("f1", "f2"), c(6, 9))
  )
  prices$pop_rel <- price_relative(prices, price ~ period + product)
  prices$fx_rel <- with(
    prices,
    price / price[back_period(period, product, offset = 3)]
  )

  pias <- aggregation_structure(list(c("1", "1"), c("f1", "f2")), 1:2)

  epr_pop <- elementary_index(prices, pop_rel ~ period + ea)
  epr_fx <- elementary_index(prices, fx_rel ~ period + ea, chainable = FALSE)

  index_pop <- aggregate(epr_pop, pias)
  index_fx <- aggregate(epr_fx, pias)

  # Chained calculation and fixed-base calculation should be the same
  expect_equal(index_fx, chain(index_pop))
  expect_equal(
    chain(index_pop[, -1], link = as.matrix(index_fx[, 1])),
    index_fx[, -1]
  )

  # Should work for a non-arithmetic index
  expect_equal(
    chain(aggregate(epr_pop, pias, order = 3)),
    aggregate(epr_fx, pias, order = 3)
  )

  # Consistency in aggregation holds with a change in base period
  expect_equal(
    rebase(index_fx, base = index_fx[, 2]),
    aggregate(
      rebase(index_fx, base = index_fx[, 2]),
      update(pias, index_fx, period = "b")
    )
  )
})

# Corner cases.
local({
  expect_equal(
    as.matrix(aggregate(as_index(matrix(1:10, 5)), list(1:5))),
    matrix(1:10, 5, dimnames = list(levels = 1:5, time = 1:2))
  )
  expect_equal(
    as.matrix(aggregate(as_index(1:5), 6)),
    matrix(NA_real_, dimnames = list(levels = 6, time = 1))
  )
})

# Partial contributions are correct.
local({
  prices <- data.frame(
    rel = 1:8,
    period = rep(1:2, each = 4),
    ea = rep(letters[1:2], 4)
  )

  epr <- elementary_index(prices, rel ~ period + ea, contrib = TRUE)

  epr <- merge(epr, matrix(9:10, 1, dimnames = list("c", 1:2)))

  pias <- aggregation_structure(
    list(c("top", "top", "top"), c("a", "b", "c")),
    1:3
  )

  arithmetic_contributions <- function(x, w, order = 1) {
    (x - 1) * transmute_weights(x, w, order, to = 1)
  }
  index <- aggregate(epr, pias)
  expect_equal(
    sum(contrib(index[, 1])),
    sum(arithmetic_contributions(
      as.numeric(index[letters[1:3], 1]),
      weights(pias, ea_only = TRUE)
    )[1:2])
  )

  pias2 <- update(pias, index[, 1])
  expect_equal(
    sum(contrib(index[, 2])),
    sum(arithmetic_contributions(
      as.numeric(index[letters[1:3], 2]),
      weights(pias2, ea_only = TRUE)
    )[1:2])
  )
})

# Duplicate products get unique names during aggregation.
local({
  epr1 <- elementary_index(
    setNames(1:4, 1:4),
    ea = gl(2, 2),
    period = gl(1, 4),
    contrib = TRUE,
    order = 1
  )
  epr2 <- epr1
  levels(epr2) <- 3:4
  index <- aggregate(
    merge(epr1, epr2),
    list(c(0, 0, 0, 0), c("a", "a", "b", "b"), 1:4),
    duplicate_contrib = "make.unique"
  )
  expect_equal(
    contrib(index, "a"),
    matrix(c(0, 0.25, 0.5, 0.75), 4, dimnames = list(product = 1:4, time = 1))
  )
  expect_equal(
    contrib(index, "a"),
    contrib(index, "b")
  )
  expect_equal(
    contrib(index),
    matrix(
      rep(c(0, 0.125, 0.25, 0.375), each = 2),
      8,
      dimnames = list(
        product = c(1, "1.1", 2, "2.1", 3, "3.1", 4, "4.1"),
        time = 1
      )
    )
  )
})

# Aggregating in parallel works.
local({
  pias <- aggregation_structure(
    list(
      c(1, 1, 1, 2, 2, 2),
      c(11, 11, 12, 21, 22, 22),
      c(111, 112, 121, 211, 221, 222)
    ),
    c(1, 2, 3, 3, 2, 1)
  )

  pias1 <- aggregation_structure(
    list(
      c(1, 1, 1),
      c(11, 11, 12),
      c(111, 112, 121)
    ),
    c(1, 2, 3)
  )

  pias2 <- aggregation_structure(
    list(
      c(2, 2, 2),
      c(21, 22, 22),
      c(211, 221, 222)
    ),
    c(3, 2, 1)
  )

  epr1 <- elementary_index(
    1:18,
    period = rep(1:2, 9),
    ea = rep(c(111, 112, 121), each = 6),
    contrib = TRUE
  )
  epr2 <- elementary_index(
    18:1,
    period = rep(1:2, 9),
    ea = rep(c(111, 112, 121), each = 6),
    contrib = TRUE
  )
  levels(epr2) <- c(211, 221, 222)

  index <- aggregate(merge(epr1, epr2), pias, order = 0.5)
  index1 <- aggregate(epr1, pias, order = 0.5)
  index2 <- aggregate(epr2, pias, order = 0.5)

  expect_equal(index[1], index1[1])
  expect_equal(index[2], index2[2])
})

# Aggregating with a dead branch does nothing.
local({
  epr <- elementary_index(
    1:18,
    period = rep(1:3, each = 6),
    ea = rep(1:2, 9),
    contrib = TRUE
  )
  pias <- data.frame(
    l1 = c(0, 0, 0),
    l2 = c("01", "01", "02"),
    l3 = c("011", "012", "021"),
    ea = 1:3,
    weights = 3:1
  )
  index1 <- aggregate(epr, pias)
  index1na <- aggregate(epr, pias, na.rm = TRUE)
  index2 <- aggregate(epr, pias[1:2, ])

  expect_equal(index1["01"], index2["01"])
  expect_equal(index1na["0"], index2["0"])
  expect_equal(contrib(index1, "01"), contrib(index2, "01"))
  expect_equal(contrib(index1na), contrib(index2))
})

# Reaggregating doesn't introduce incorrect contributions.
local({
  epr <- elementary_index(
    c(1:7, NA),
    period = rep(1:2, each = 4),
    ea = rep(1:2, 4),
    contrib = TRUE
  )
  pias <- as_aggregation_structure(list(c(0, 0), c(1, 2)))
  index <- aggregate(epr, pias, na.rm = TRUE)
  expect_equal(
    contrib(index, period = 1),
    contrib(aggregate(index, pias), period = 1)
  )
  expect_equal(
    sum(contrib(index, period = 2), na.rm = TRUE),
    sum(contrib(aggregate(index, pias), period = 2), na.rm = TRUE)
  )
})

# Skipping time periods works.
local({
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    na.rm = TRUE
  )

  ms_pias <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index <- chain(aggregate(ms_epr, ms_pias, na.rm = TRUE, order = 2))

  ms_index2 <- unchain(ms_index[, -3])

  expect_equal(
    chain(aggregate(ms_index2, ms_pias, order = 2))[, 3],
    ms_index[, 4]
  )
})

# Skipping eas works.
local({
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    na.rm = TRUE
  )

  ms_pias <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index <- aggregate(ms_epr, ms_pias, na.rm = TRUE)
  expect_equal(
    ms_index[1:3],
    aggregate(ms_epr, ms_pias, na.rm = TRUE, include_ea = FALSE)
  )

  expect_equal(
    aggregate(
      ms_epr,
      aggregation_structure(ms_weights["business"]),
      include_ea = FALSE
    ),
    aggregate(
      ms_epr,
      aggregation_structure(ms_weights["business"]),
      include_ea = TRUE
    )
  )
})

# Missing weights ignores those index values.
local({
  prices <- data.frame(
    rel = 1:8,
    period = rep(1:2, each = 4),
    ea = rep(letters[1:2], 4)
  )

  pias <- aggregation_structure(
    list(c("top", "top"), c("a", "b")),
    1:2
  )

  elementary <- elementary_index(prices, rel ~ period + ea, contrib = TRUE)

  index <- aggregate(elementary, pias)

  weights(pias)[1] <- NA
  index2 <- aggregate(elementary, pias, na.rm = TRUE)

  expect_equal(as.numeric(index2[1]), as.numeric(index2[3]))
  expect_equal(
    as.numeric(colSums(contrib(index2), na.rm = TRUE) + 1),
    as.numeric(index2[1])
  )
  expect_equal(contrib(index2), contrib(aggregate(elementary, pias)))
})

# Superlative index aggregates correctly.
local({
  epr <- as_index(matrix(c(1:4, NA, 6:9), 3), contrib = TRUE)
  contrib(epr, 1, 1) <- c(-2, -1, 1, 1, 0.25, 0.75)
  pias <- aggregation_structure(
    list(c("top", "top", "top"), 1:3),
    1:3
  )
  pias2 <- aggregation_structure(
    list(c("top", "top", "top"), 1:3),
    c(3, 1, 2)
  )
  res <- aggregate(epr, pias, pias2 = pias2, na.rm = TRUE)
  res1 <- aggregate(epr, pias, na.rm = TRUE)
  res2 <- aggregate(epr, pias2, order = -1, na.rm = TRUE)

  expect_equal(
    sqrt(as.matrix(res1) * as.matrix(res2)),
    as.matrix(res)
  )

  expect_equal(
    colSums(contrib(res), na.rm = TRUE),
    as.matrix(res)[1, ] - 1
  )
  expect_equal(
    colSums(contrib(res, 1)),
    as.matrix(res)[2, ] - 1
  )

  # Example from vignette.
  w <- mapply(
    \(x, y) transmute_weights(c(x, y)),
    as.numeric(res1[1]),
    as.numeric(res2[1])
  )

  laspeyres_contrib <- contrib(res1)
  paasche_contrib <- contrib(res2)

  fisher_contrib <- w[1, col(laspeyres_contrib)] *
    laspeyres_contrib +
    w[2, col(paasche_contrib)] * paasche_contrib

  expect_equal(contrib(res), fisher_contrib)

  # Tornqvist.
  expect_equal(
    aggregate(epr, pias, order = 0, na.rm = TRUE),
    aggregate(epr, pias, pias2 = pias, order = 0, na.rm = TRUE)
  )
})

# Duplicate product methods work with NAs.
local({
  x <- as_index(c(1, 2, NA))
  contrib(x, 1) <- c(0.5, -0.25, NA, -0.25)
  contrib(x, 3) <- c(1, NA, NA)
  pias <- list(c(0, 0, 0), 1:3)
  expect_equal(
    contrib(aggregate(x, pias, duplicate_contrib = "sum")),
    matrix(
      c(0.25, -0.125, NA, -0.125),
      dimnames = list(product = 1:4, time = 1)
    )
  )
  expect_equal(
    contrib(aggregate(x, pias, duplicate_contrib = "sum", na.rm = TRUE)),
    matrix(
      c(0.25, -0.125, NA, -0.125),
      dimnames = list(product = 1:4, time = 1)
    )
  )
})

# Superlative contributions work with NA weights.
local({
  x <- as_index(c("3" = 1, "4" = 2))
  contrib(x, 3) <- 1:5
  contrib(x, 4) <- 1:2
  pias1 <- aggregation_structure(list(c(0, 0), 1:2, 3:4), c(1, 1))
  weights(pias1)[1] <- NA
  pias2 <- aggregation_structure(list(c(0, 0), 1:2, 3:4), c(2, 2))

  expect_error(aggregate(x, pias1, pias2 = pias2, na.rm = TRUE))
  weights(pias2)[1] <- 0
  res <- aggregate(
    x,
    pias1,
    pias2 = pias2,
    na.rm = TRUE,
    duplicate_contrib = "sum"
  )

  expect_equal(
    contrib(res),
    matrix(
      c(1, 2, NA, NA, NA),
      nrow = 5,
      dimnames = list(product = 1:5, time = 1)
    )
  )
})

# Impute rules.
local({
  x <- as_index(matrix(1:9, 3))
  x[1, 1:2] <- NA

  rules <- function(x, pias) {
    if (time(x) == "2") {
      x[is.na(x)] <- 1
    }
    x[3:1]
  }

  res <- aggregate(x, list(c(0, 0, 0), 1:3), na.rm = TRUE, impute_rules = rules)
  expect_equal(
    as.numeric(res[1]),
    c(
      2.5,
      gmean(c(1, 5, 6), c(2.5, 2, 3)),
      gmean(7:9, c(2.5, 10, 18))
    )
  )
})

dat <- data.frame(
  rel = c(1:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

pias <- as_aggregation_structure(
  data.frame(level1 = 1, level2 = c(11, 12, 13, 14), weight = 1)
)

epr <- elementary_index(dat, rel ~ period + ea, contrib = TRUE)
epr2 <- elementary_index(dat, rel ~ period + ea)

index <- aggregate(epr, pias)

res <- c(
  1.41421356237309,
  3.46410161513775,
  5,
  NaN,
  6.48074069840786,
  NaN,
  NaN,
  8
)

# head and tail work.
local({
  expect_equal(head(epr, 1), epr[1])
  expect_equal(head(epr, 5), epr)
  expect_equal(head(epr, -1), epr[1:3])
  expect_equal(tail(epr, 1), epr[4])
  expect_equal(tail(epr, 5), epr)
  expect_equal(tail(epr, -1), epr[2:4])
  expect_equal(head(epr, 2:1), epr[1:2, 1])
  expect_equal(tail(epr, 2:1), epr[3:4, 2])
  expect_equal(head(epr, c(NA, -1)), epr[, 1])
  expect_equal(tail(epr, c(NA, -1)), epr[, 2])
  # #72
  expect_equal(head(epr2, c(1, 1)), epr2[1, 1])
  expect_equal(tail(epr2, c(1, 1)), epr2[4, 2])
})

# Subscripting methods work.
local({
  expect_equal(epr[], epr)
  expect_equal(index[1:5], index)
  expect_equal(
    epr[c(TRUE, FALSE, TRUE, TRUE), 2:1],
    elementary_index(
      dat,
      rel ~ factor(period, 2:1) + factor(ea, c(11, 13:14)),
      contrib = TRUE
    )
  )
  expect_equal(
    epr[c("14", "12"), TRUE],
    elementary_index(dat, rel ~ period + factor(ea, c(14, 12)), contrib = TRUE)
  )
  expect_equal(epr[matrix(c(1, 2, 1, 2), 2)], list(epr[1, 1], epr[2, 2]))
  expect_equal(epr[matrix(0, 2, 2)], list())
})

# Subscripting methods give errors where expected.
local({
  expect_error(epr[1, NA])
  expect_error(epr[1, 3])
  expect_error(epr[c(1, 2, 1), ])
  expect_error(epr[, NULL])
  expect_error(epr[integer(0)])
  expect_error(epr[NA])
  expect_error(epr[matrix(c(-1, 2), 1, 2)])
})

# Replacement methods work with an index.
local({
  # No op
  expect_equal(replace(epr, 1:4, epr), epr)
  expect_equal(replace(epr, integer(0), epr[1]), epr)
  expect_equal(replace(epr, matrix(0, 2, 2), epr[1]), epr)
  # Adds contributions
  expect_equal(replace(epr2, 1:4, epr), epr)
  # Removes contributions
  expect_equal(replace(epr, 1:4, epr2), set_contrib(epr2, value = numeric(0)))

  epr3 <- epr2
  epr3[c(TRUE, FALSE), 2] <- epr[1, 1]
  expect_equal(
    as.matrix(epr3),
    replace(as.matrix(epr), matrix(c(1, 3, 2, 2), 2), as.numeric(epr[1, 1]))
  )
  expect_equal(contrib(epr3, "13")[, 2], contrib(epr)[, 1])

  # Index doesn't recycle like a matrix
  m <- as.matrix(epr3)
  epr3[c(1, 1, 2, 2)] <- epr[1:2]
  m[1, ] <- as.matrix(epr)[2, ]
  expect_equal(as.matrix(epr3), m)
  expect_equal(contrib(epr3), contrib(epr, "12"))
  epr3[c(1, 1, 2, 2)] <- as.matrix(epr[1:2])
  expect_false(isTRUE(all.equal(as.matrix(epr3), m)))

  # Errors
  expect_error(epr[1:2] <- epr[, 1])
  expect_error(epr[1:3] <- epr[1:2])
  expect_error(epr[NA] <- epr[1])
  expect_error(epr[, NA] <- epr[1])
  expect_error(epr[1] <- chain(epr[1]))
  epr <- chain(epr)
  expect_error(epr[1] <- epr2[1])
  expect_error(epr[matrix(c(1, 2), 1)] <- unchain(epr[1, 1]))
})

# Replacement methods work with a vector.
local({
  epr[, 1] <- 1
  expect_equal(
    as.matrix(epr),
    matrix(
      c(1, 1, 1, 1, res[5:8]),
      4,
      2,
      dimnames = list(levels = 11:14, time = 1:2)
    )
  )
  expect_equal(
    contrib(epr),
    matrix(
      c(0, 0, 2.5962965079607, 2.88444419044716),
      2,
      2,
      dimnames = list(product = c("11.1", "11.2"), time = 1:2)
    )
  )

  epr[1, c(FALSE, TRUE)] <- "2"
  expect_equal(
    as.matrix(epr),
    matrix(
      c(1, 1, 1, 1, 2, res[6:8]),
      4,
      2,
      dimnames = list(levels = 11:14, time = 1:2)
    )
  )
  expect_equal(
    contrib(epr),
    matrix(0, 0, 2, dimnames = list(product = NULL, time = 1:2))
  )

  # recycling should still happen
  epr2[1:3, ] <- 8:9
  expect_equal(
    epr2[1:3, ],
    as_index(matrix(8:9, 3, 2, dimnames = list(levels = 11:13, time = 1:2)))
  )

  epr[1, c(1, 2, 1)] <- setNames(1:3, letters[1:3])
  expect_equal(
    epr[1, ],
    set_contrib(
      as_index(matrix(3:2, 1, dimnames = list(levels = "11", time = 1:2))),
      value = numeric(0)
    )
  )

  epr["14"] <- 1
  expect_equal(
    as.matrix(epr),
    matrix(
      c(3, 1, 1, 1, 2, res[6:7], 1),
      4,
      2,
      dimnames = list(levels = 11:14, time = 1:2)
    )
  )
  expect_equal(
    contrib(epr, "14"),
    matrix(0, 0, 2, dimnames = list(product = NULL, time = 1:2))
  )

  # Replacing nothing
  epr2 <- epr
  suppressWarnings(epr[FALSE] <- "a")
  expect_identical(epr, epr2)

  suppressWarnings(epr[1, 0] <- "a")
  expect_identical(epr, epr2)

  # Errors
  expect_error(epr[1] <- 1:3)
  expect_error(epr[1:2] <- 1:3)
  expect_error(epr[NA] <- 1)
  expect_error(epr[1, NA] <- 1)
  expect_error(epr[1, c(TRUE, FALSE, FALSE)] <- 1)
  expect_error(epr[5] <- 1)
  expect_error(epr[1] <- numeric(0))
  expect_warning(epr[c(TRUE, FALSE, TRUE), ])
})

# Replacement method works with a matrix.
local({
  epr3 <- replace(epr, is.na(epr), 1:3)
  expect_equal(
    as.matrix(epr3),
    replace(as.matrix(epr), c(4, 6, 7), c(1, 2, 3))
  )

  epr4 <- replace(
    epr3,
    matrix(c("11", "11", "12", "12", "2", "2", "2", "2"), 4),
    1:2
  )
  expect_equal(as.matrix(epr4), replace(as.matrix(epr3), 5:6, 2))

  expect_identical(replace(epr, matrix(FALSE, 4, 2), 0), epr)

  epr5 <- replace(epr3, is.na(epr), epr[1, 1])
  expect_equal(
    as.matrix(epr5),
    replace(as.matrix(epr), c(4, 6, 7), sqrt(2))
  )
  expect_equal(contrib(epr5[4, 1]), contrib(epr[1, 1]))

  # Works when there are no contributions.
  epr5 <- replace(epr2, is.na(epr), epr[1, 1])
  expect_equal(
    as.matrix(epr5),
    replace(as.matrix(epr2), c(4, 6, 7), sqrt(2))
  )
  expect_equal(contrib(epr5[4, 1]), contrib(epr[1, 1]))

  expect_error(replace(epr, matrix(NA, 4, 2), 0))
  expect_error(epr[is.na(epr)] <- numeric(0))
  expect_error(epr[matrix(TRUE, 3, 3)] <- 1)
  expect_error(epr[matrix(1, nrow = 1, ncol = 3)] <- 1)
  expect_warning(epr[is.na(epr)] <- 1:2)
  expect_error(epr[is.na(epr)] <- list(1))
  expect_error(epr[!is.na(epr)] <- epr[1])
})

epr1 <- elementary_index(
  ms_prices,
  price_relative(price, period = period, product = product) ~ period + business,
  contrib = TRUE,
  na.rm = TRUE
)

epr2 <- elementary_index(
  ms_prices,
  price_relative(price, period = period, product = product) ~ period + business,
  na.rm = TRUE
)

pias <- with(
  ms_weights,
  aggregation_structure(
    c(expand_classification(classification), list(business)),
    weight
  )
)

index1 <- aggregate(epr1, pias, na.rm = TRUE)
index2 <- aggregate(epr2, pias, na.rm = TRUE)

# Chain is the same as apply.
local({
  expect_equal(
    as.matrix(chain(epr1)),
    t(apply(as.matrix(epr1), 1, cumprod))
  )
  expect_equal(
    as.matrix(chain(epr1, link = 1:4)),
    t(apply(as.matrix(epr1), 1, cumprod)) * 1:4
  )
  expect_equal(
    as.matrix(chain(index1)),
    t(apply(as.matrix(index1), 1, cumprod))
  )
})

# Unchain and chain are inverses with no NAs.
local({
  expect_equal(unchain(chain(index2)), index2)
  expect_equal(unchain(rebase(chain(index2), base = 1:8), base = 1:8), index2)
  expect_equal(unchain(chain(index2, link = 1:8), base = 1 / 1:8), index2)
  expect_false(isTRUE(all.equal(unchain(chain(epr2)), epr2)))
})

# Unchaining/rebasing a chainable index does nothing.
local({
  expect_equal(unchain(epr1), epr1)
  expect_equal(unchain(index1), index1)
  expect_equal(rebase(epr1), epr1)
  expect_equal(rebase(index1), index1)
})

# Chaining a fixed-base index does nothing.
local({
  expect_equal(chain(epr1), chain(chain(epr1)))
  expect_equal(chain(index1), chain(chain(index1)))
})

# Rebase should be the same as division.
local({
  expect_equal(
    as.matrix(rebase(chain(epr2), base = 1:4)),
    as.matrix(chain(epr2)) / 1:4
  )
  expect_equal(
    as.matrix(rebase(chain(index2), base = 1:8)),
    as.matrix(chain(index2)) / 1:8
  )
})

# Rebase works with mean.
local({
  index2 <- chain(index2)
  expect_equal(
    rebase(index2, base = mean(index2, window = 2)[, 1]),
    rebase(index2, base = rowMeans(as.matrix(index2[, 1:2])))
  )

  expect_equal(
    rebase(index2, base = mean(index2, window = 4)),
    rebase(index2, base = rowMeans(as.matrix(index2)))
  )
})

# Chaining returns the correct type of index.
local({
  expect_true(is_chainable_index(epr1))
  expect_true(is_chainable_index(unchain(chain(epr1))))
  expect_true(is_direct_index(chain(epr1)))
  expect_true(is_chainable_index(index1))
  expect_true(is_chainable_index(unchain(chain(index1))))
  expect_true(is_direct_index(chain(index1)))
})

# Chaining keeps EA names.
local({
  expect_equal(
    as.matrix(chain(as_index(matrix(1:5, 1)))),
    matrix(cumprod(1:5), 1, dimnames = list(levels = 1, time = 1:5))
  )
})

# Link and base values are the right length.
local({
  expect_error(chain(epr1, link = 1))
  expect_error(rebase(chain(epr1), base = 1))
})

# Rebasing with a character vector works.
local({
  index1_chain <- chain(index1)
  index1_rebase <- rebase(
    index1_chain,
    base = index1_chain[, end(index1_chain)]
  )
  expect_equal(index1_rebase, rebase(index1_chain, base = end(index1_chain)))
  expect_equal(index2, unchain(index1_rebase, base = end(index1_rebase)))
})

# It used to be possible to get negative indexes values.
local({
  x <- as_index(matrix(1:9, 3))
  expect_error(chain(x, link = c(-1, 2, 3)), "`link` must be strictly positive")

  x <- chain(x)
  expect_error(
    unchain(x, base = c(-1, 2, 3)),
    "`base` must be strictly positive"
  )
  expect_error(
    rebase(x, base = c(-1, 2, 3)),
    "`base` must be strictly positive"
  )
})

# Aggregating over subperiods works.
local({
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    na.rm = TRUE
  )

  epr2 <- mean(ms_epr, window = 2, duplicate_contrib = "make.unique")
  expect_identical(levels(epr2), levels(ms_epr))
  expect_identical(time(epr2), c("202001", "202003"))

  expect_equal(as.matrix(epr2)[, 1], rowMeans(as.matrix(ms_epr)[, 1:2]))
  expect_equal(as.matrix(epr2)[, 2], rowMeans(as.matrix(ms_epr)[, 3:4]))

  expect_true(is_chainable_index(epr2))
  expect_false(is_chainable_index(mean(chain(ms_epr), window = 2)))

  w <- matrix(seq_len(4 * 4), 4)
  expect_equivalent(
    as.matrix(mean(ms_epr, weights = w, window = 2))[, 1],
    diag(
      as.matrix(ms_epr)[, 1:2] %*% apply(w[, 1:2], 1, scale_weights)
    )
  )
  expect_equivalent(
    as.matrix(mean(ms_epr, weights = w, window = 2))[, 2],
    diag(
      as.matrix(ms_epr)[, 3:4] %*% apply(w[, 3:4], 1, scale_weights)
    )
  )

  # Test contributions
  con <- matrix(0, 5, 2)
  dimnames(con) <- list(
    product = c("1", "2", "2.1", "3", "3.1"),
    time = c("202001", "202003")
  )
  con[c(2, 3, 9, 10)] <- NA
  con[c(5, 7)] <- contrib(ms_epr)[c(6, 8)] / 2:1
  expect_equal(contrib(epr2), con)
  expect_equal(
    contrib(mean(ms_epr, window = 2, duplicate_contrib = "sum")),
    structure(
      rbind(
        "1" = con[1, ],
        "2" = colSums(con[2:3, ]),
        "3" = colSums(con[4:5, ])
      ),
      dimnames = list(product = 1:3, time = time(ms_epr)[c(1, 3)])
    )
  )

  expect_equal(
    as.numeric(t(contrib(ms_epr, "B3")[, 3:4])) / 2,
    as.numeric(contrib(epr2, "B3")[, 2])
  )
  expect_equal(colSums(contrib(epr2, "B3")), as.matrix(epr2)["B3", ] - 1)

  epr3 <- mean(
    ms_epr,
    weights = w,
    window = 2,
    order = 2.5,
    na.rm = TRUE,
    duplicate_contrib = "sum"
  )
  expect_equivalent(
    colSums(contrib(epr3, "B1"), na.rm = TRUE),
    as.matrix(epr3)["B1", ] - 1
  )
  expect_equal(colSums(contrib(epr3, "B3")), as.matrix(epr3)["B3", ] - 1)
})

# Aggregating with a window of 1 does nothing.
local({
  x <- as_index(matrix(1:9, 3))
  expect_equal(mean(x, window = 1), x)
  expect_equal(mean(x, weights = 9:1, window = 1, order = 0), x)
  expect_equal(mean(window(x, end(x))), window(x, end(x)))
})

# mean requires a suitable window.
local({
  index <- as_index(matrix(1:9, 3))
  expect_error(mean(index, window = 4))
  expect_error(mean(index, window = 0))
  expect_error(mean(index, window = 1:3))
  expect_warning(mean(index, window = 2))
})

# Mean is consistent in aggregation.
local({
  epr <- as_index(matrix(1:54 / 10, 6))
  pias <- aggregation_structure(
    list(rep("a", 6), rep(c("b", "c"), each = 3), 1:6),
    6:1
  )

  index <- aggregate(epr, pias)

  index <- unchain(mean(chain(index), window = 3))
  expect_equal(aggregate(index, pias), index)
})

test_that("aggregating over subperiods works", {
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    na.rm = TRUE
  )

  epr2 <- mean(ms_epr, window = 2, duplicate_contrib = "make.unique")
  expect_identical(levels(epr2), levels(ms_epr))
  expect_identical(time(epr2), c("202001", "202003"))

  expect_equal(as.matrix(epr2)[, 1], rowMeans(as.matrix(ms_epr)[, 1:2]))
  expect_equal(as.matrix(epr2)[, 2], rowMeans(as.matrix(ms_epr)[, 3:4]))

  expect_true(is_chainable_index(epr2))
  expect_false(is_chainable_index(mean(chain(ms_epr), window = 2)))

  w <- matrix(seq_len(4 * 4), 4)
  expect_equal(
    as.matrix(mean(ms_epr, weights = w, window = 2))[, 1],
    diag(
      as.matrix(ms_epr)[, 1:2] %*% apply(w[, 1:2], 1, gpindex::scale_weights)
    ),
    ignore_attr = TRUE
  )
  expect_equal(
    as.matrix(mean(ms_epr, weights = w, window = 2))[, 2],
    diag(
      as.matrix(ms_epr)[, 3:4] %*% apply(w[, 3:4], 1, gpindex::scale_weights)
    ),
    ignore_attr = TRUE
  )

  # Test contributions
  con <- matrix(0, 5, 2)
  dimnames(con) <- list(
    product = c("1", "2", "2.1", "3", "3.1"),
    time = c("202001", "202003")
  )
  con[c(2, 3, 9, 10)] <- NA
  con[c(5, 7)] <- contrib(ms_epr)[c(6, 8)] / 2:1
  expect_equal(contrib(epr2), con)
  expect_equal(
    contrib(mean(ms_epr, window = 2, duplicate_contrib = "sum")),
    structure(
      rbind(
        "1" = con[1, ],
        "2" = colSums(con[2:3, ]),
        "3" = colSums(con[4:5, ])
      ),
      dimnames = list(product = 1:3, time = time(ms_epr)[c(1, 3)])
    )
  )

  expect_equal(
    as.numeric(t(contrib(ms_epr, "B3")[, 3:4])) / 2,
    as.numeric(contrib(epr2, "B3")[, 2])
  )
  expect_equal(colSums(contrib(epr2, "B3")), as.matrix(epr2)["B3", ] - 1)

  epr3 <- mean(
    ms_epr,
    weights = w,
    window = 2,
    r = 2.5,
    na.rm = TRUE,
    duplicate_contrib = "sum"
  )
  expect_equal(
    colSums(contrib(epr3, "B1"), na.rm = TRUE),
    as.matrix(epr3)["B1", ] - 1,
    ignore_attr = TRUE
  )
  expect_equal(colSums(contrib(epr3, "B3")), as.matrix(epr3)["B3", ] - 1)
})

test_that("aggregating with a window of 1 does nothing", {
  x <- as_index(matrix(1:9, 3))
  expect_equal(mean(x, window = 1), x)
  expect_equal(mean(x, weights = 9:1, window = 1, r = 0), x)
  expect_equal(mean(window(x, end(x))), window(x, end(x)))
})

test_that("mean requires a suitable window", {
  index <- as_index(matrix(1:9, 3))
  expect_error(mean(index, window = 4))
  expect_error(mean(index, window = 0))
  expect_error(mean(index, window = 1:3))
  expect_warning(mean(index, window = 2))
})

test_that("mean is consistent in aggregation", {
  epr <- as_index(matrix(1:54 / 10, 6))
  pias <- aggregation_structure(
    list(rep("a", 6), rep(c("b", "c"), each = 3), 1:6),
    6:1
  )

  index <- aggregate(epr, pias)

  index <- unchain(mean(chain(index), window = 3))
  expect_equal(aggregate(index, pias), index)
})

test_that("a matched-sample index aggregates correctly", {
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    na.rm = TRUE
  )

  ms_pias <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index <- aggregate(ms_epr, ms_pias, na.rm = TRUE)

  res <- c(
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1.30072391107879,
    1.30072391107879,
    1.30072391107879,
    0.894909688013136,
    1.30072391107879,
    2.02000360773041,
    1.30072391107879,
    1.30072391107879,
    1.06307432720167,
    1.06307432720167,
    1.06307432720167,
    0.334293948126801,
    1.06307432720167,
    1.63533549062897,
    1.06307432720167,
    1.06307432720167,
    2.73476132776126,
    1.57451535678038,
    4.57628618296306,
    1.57451535678038,
    2.77045633390764,
    0.537995998322608,
    4.57628618296306,
    4.57628618296306
  )
  res <- matrix(
    res,
    8,
    4,
    dimnames = list(
      levels = c("1", "11", "12", paste0("B", 1:5)),
      time = sprintf("2020%02d", 1:4)
    )
  )
  expect_equal(as.matrix(ms_index), res)

  # Same as matrix calculation
  expect_equal(
    as.matrix(ms_pias) %*% as.matrix(chain(ms_index[paste0("B", 1:5)])),
    as.matrix(chain(ms_index[1:3, ]))
  )

  # Lower levels add up
  expect_equal(
    apply(
      as.matrix(chain(ms_index)[4:8, ]),
      2,
      weighted.mean,
      weights(ms_pias, ea_only = FALSE)[[3]]
    ),
    as.matrix(chain(ms_index))[1, ]
  )
  expect_equal(
    apply(
      as.matrix(chain(ms_index)[2:3, ]),
      2,
      weighted.mean,
      weights(ms_pias, ea_only = FALSE)[[2]]
    ),
    as.matrix(chain(ms_index))[1, ]
  )

  # Re aggregating does nothing to the index values
  expect_equal(as.matrix(aggregate(ms_index, ms_pias)), as.matrix(ms_index))
  expect_equal(aggregate(chain(ms_index), ms_pias), chain(ms_index))
  expect_equal(
    as.matrix(aggregate(ms_index, ms_pias, na.rm = TRUE)),
    as.matrix(ms_index)
  )

  # Re aggregating doesn't breaks contributions for imputed indexes
  expect_equal(
    colSums(contrib(ms_index), na.rm = TRUE),
    colSums(contrib(aggregate(ms_index, ms_pias)), na.rm = TRUE)
  )
  expect_equal(
    ms_index,
    aggregate(ms_index, ms_pias, duplicate_contrib = "sum")
  )
  expect_equal(
    contrib(aggregate(ms_index, ms_pias)),
    contrib(aggregate(aggregate(ms_index, ms_pias), ms_pias))
  )

  # Two step aggregation gives the same result
  pias2 <- aggregation_structure(
    list(c(1, 1), c(11, 12)),
    weights(ms_pias, ea_only = FALSE)[[2]]
  )
  expect_equal(as.matrix(aggregate(ms_index, pias2)), as.matrix(ms_index[1:3]))
  expect_equal(
    contrib(aggregate(aggregate(ms_index, ms_pias), pias2)),
    contrib(aggregate(ms_index, ms_pias))
  )

  # Re-arranging the index shouldn't do anything
  s1 <- c(
    14,
    16,
    26,
    28,
    24,
    29,
    11,
    32,
    36,
    2,
    22,
    34,
    6,
    7,
    10,
    17,
    8,
    27,
    37,
    1,
    12,
    33,
    20,
    3,
    9,
    40,
    13,
    4,
    38,
    23,
    31,
    15,
    25,
    39,
    21,
    30,
    35,
    19,
    18,
    5
  )
  s2 <- c(5, 3, 4, 1, 2)
  ms_epr <- elementary_index(
    ms_prices[s1, ],
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    na.rm = TRUE
  )

  ms_pias <- with(
    ms_weights[s2, ],
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index2 <- aggregate(ms_epr, ms_pias, na.rm = TRUE)
  expect_equal(
    as.matrix(ms_index2)[levels(ms_index), ],
    as.matrix(ms_index)
  )
  expect_equal(
    contrib(ms_index),
    contrib(ms_index2)
  )
  expect_equal(
    as.matrix(ms_pias) %*% as.matrix(chain(ms_index2[paste0("B", s2)])),
    as.matrix(chain(ms_index[c(1, 3, 2), ]))
  )

  # Aggregated contributions should add up
  expect_equal(
    as.matrix(ms_index)[1, ],
    colSums(contrib(ms_index), na.rm = TRUE) + 1
  )
  expect_equal(
    as.matrix(ms_index)[2, ],
    colSums(contrib(ms_index, "11"), na.rm = TRUE) + 1
  )

  # Check that weights are getting price updated correctly
  w <- apply(as.matrix(chain(ms_index)[4:8, ]), 2, `*`, ms_weights$weight)

  expect_equal(weights(update(ms_pias, ms_index), ea_only = TRUE), w[s2, 4])

  expect_equal(
    weights(update(ms_pias, ms_index, period = "202003"), ea_only = TRUE),
    w[s2, 3]
  )
})

test_that("a weird index aggregates correctly", {
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    r = 0.2
  )

  ms_pias <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index <- aggregate(ms_epr, ms_pias, r = -1.7, na.rm = TRUE)

  res <- c(
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    3.59790686240372,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.83344068593548,
    1.31335391535665,
    1.00564223748248,
    5.47029871913538,
    1.00564223748248,
    3.15295124015891,
    0.551843416514059,
    5.47029871913538,
    5.47029871913538
  )
  res <- matrix(
    res,
    8,
    4,
    dimnames = list(
      levels = c("1", "11", "12", paste0("B", 1:5)),
      time = sprintf("2020%02d", 1:4)
    )
  )
  expect_equal(as.matrix(ms_index), res)

  expect_equal(
    as.matrix(aggregate(ms_index, ms_pias, r = -1.7, na.rm = TRUE)),
    as.matrix(ms_index)
  )

  expect_equal(aggregate(chain(ms_index), ms_pias, r = -1.7), chain(ms_index))

  expect_equal(
    as.matrix(ms_index)[1, ],
    colSums(contrib(ms_index), na.rm = TRUE) + 1
  )
  expect_equal(
    as.matrix(ms_index)[1, ],
    colSums(contrib(aggregate(ms_index, ms_pias, r = -1.7)), na.rm = TRUE) + 1
  )
  expect_equal(
    aggregate(ms_index, ms_pias, r = -1.7, duplicate_contrib = "sum"),
    ms_index
  )

  expect_equal(
    apply(
      as.matrix(chain(ms_index)[2:3, ]),
      2,
      gpindex::generalized_mean(-1.7),
      weights(ms_pias, ea_only = FALSE)[[2]]
    ),
    as.matrix(chain(ms_index))[1, ]
  )

  expect_equal(
    as.matrix(
      as.matrix(ms_pias, sparse = TRUE) %*%
        as.matrix(chain(ms_index[paste0("B", 1:5)]))^(-1.7)
    )^(1 / -1.7),
    as.matrix(chain(ms_index[1:3, ]))
  )

  ms_index <- aggregate(ms_epr, ms_pias, r = -1.7)

  expect_equal(aggregate(ms_index, ms_pias, r = -1.7), ms_index)

  expect_equal(as.matrix(ms_index)[1, ], colSums(contrib(ms_index)) + 1)
  expect_equal(
    as.matrix(ms_index)["B2", ],
    colSums(contrib(ms_index, "B2")) + 1
  )
})

test_that("a fixed-sample index aggregates correctly", {
  fs_epr <- elementary_index(
    fs_prices,
    price_relative(price, period = period, product = business) ~
      period + classification,
    weights = weight,
    contrib = TRUE
  )

  fs_pias <- with(
    fs_weights,
    aggregation_structure(expand_classification(classification), weight)
  )

  fs_index <- aggregate(fs_epr, fs_pias, na.rm = TRUE)

  res <- c(
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    0.687039919074495,
    0.935841069334005,
    0.612229588412764,
    0.611111111111111,
    0.935841069334005,
    0.935841069334005,
    0.612229588412764,
    0.612229588412764,
    0.611111111111111,
    2.45961303155833,
    2.45961303155833,
    2.48533513976043,
    1.68506493506494,
    2.45961303155833,
    2.45961303155833,
    2.48533513976043,
    2.48533513976043,
    1.68506493506494,
    1.08265987174601,
    1.08265987174601,
    1.08265987174601,
    1.08265987174601,
    1.08265987174601,
    1.08265987174601,
    1.61372443582219,
    0.860786536350297,
    1.08265987174601
  )
  res <- matrix(
    res,
    9,
    4,
    dimnames = list(
      levels = c(1, 11:13, 111, 112, 121, 122, 131),
      time = sprintf("2020%02d", 1:4)
    )
  )
  expect_equal(as.matrix(fs_index), res)

  # Re-aggregating the index shouldn't do anything
  expect_equal(as.matrix(aggregate(fs_index, fs_pias)), as.matrix(fs_index))

  # Contributions should add up
  expect_equal(
    as.matrix(fs_index)[1, ],
    colSums(contrib(fs_index), na.rm = TRUE) + 1
  )

  # Check adding up of lower level indexes
  expect_equal(
    apply(
      as.matrix(chain(fs_index)[5:9, ]),
      2,
      weighted.mean,
      weights(fs_pias, ea_only = FALSE)[[3]]
    ),
    as.matrix(chain(fs_index))[1, ]
  )

  expect_equal(
    apply(
      as.matrix(chain(fs_index)[2:4, ]),
      2,
      weighted.mean,
      weights(fs_pias, ea_only = FALSE)[[2]]
    ),
    as.matrix(chain(fs_index))[1, ]
  )

  # Non-missing indexes should be the same when missing values are not removed
  fs_index2 <- aggregate(fs_epr, fs_pias)
  expect_equal(
    as.matrix(fs_index2)[!is.na(as.matrix(fs_index2))],
    as.matrix(fs_index)[!is.na(as.matrix(fs_index2))]
  )

  expect_equal(
    as.matrix(fs_index2)["121", ],
    colSums(contrib(fs_index2, "121"), na.rm = TRUE) + 1
  )

  expect_equal(
    as.matrix(fs_index2)["13", 1:3],
    contrib(fs_index2, "13")[, 1:3] + 1
  )
})

test_that("a fixed-based index aggregates correctly", {
  prices <- data.frame(
    price = 1:15,
    period = letters[1:3],
    product = rep(1:5, each = 3),
    ea = rep(c("f1", "f2"), c(6, 9))
  )
  prices$pop_rel <- price_relative(prices, price ~ period + product)
  prices$fx_rel <- with(
    prices,
    price / price[gpindex::base_period(period, product)]
  )

  pias <- aggregation_structure(list(c("1", "1"), c("f1", "f2")), 1:2)

  epr_pop <- elementary_index(prices, pop_rel ~ period + ea)
  epr_fx <- elementary_index(prices, fx_rel ~ period + ea, chainable = FALSE)

  index_pop <- aggregate(epr_pop, pias)
  index_fx <- aggregate(epr_fx, pias)

  # Chained calculation and fixed-base calculation should be the same
  expect_equal(index_fx, chain(index_pop))
  expect_equal(
    chain(index_pop[, -1], link = as.matrix(index_fx[, 1])),
    index_fx[, -1]
  )

  # Should work for a non-arithmetic index
  expect_equal(
    chain(aggregate(epr_pop, pias, r = 3)),
    aggregate(epr_fx, pias, r = 3)
  )

  # Consistency in aggregation holds with a change in base period
  expect_equal(
    rebase(index_fx, base = index_fx[, 2]),
    aggregate(
      rebase(index_fx, base = index_fx[, 2]),
      update(pias, index_fx, period = "b")
    )
  )
})

test_that("corner cases work", {
  expect_equal(
    as.matrix(aggregate(as_index(matrix(1:10, 5)), list(1:5))),
    matrix(1:10, 5, dimnames = list(levels = 1:5, time = 1:2))
  )
  expect_equal(
    as.matrix(aggregate(as_index(1:5), 6)),
    matrix(NA_real_, dimnames = list(levels = 6, time = 1))
  )
})

test_that("partial contributions are correct", {
  prices <- data.frame(
    rel = 1:8,
    period = rep(1:2, each = 4),
    ea = rep(letters[1:2], 4)
  )

  epr <- elementary_index(prices, rel ~ period + ea, contrib = TRUE)

  epr <- merge(epr, matrix(9:10, 1, dimnames = list("c", 1:2)))

  pias <- aggregation_structure(
    list(c("top", "top", "top"), c("a", "b", "c")),
    1:3
  )

  index <- aggregate(epr, pias)
  expect_equal(
    sum(contrib(index[, 1])),
    sum(gpindex::arithmetic_contributions(
      as.numeric(index[letters[1:3], 1]),
      weights(pias, ea_only = TRUE)
    )[1:2])
  )

  pias2 <- update(pias, index[, 1])
  expect_equal(
    sum(contrib(index[, 2])),
    sum(gpindex::arithmetic_contributions(
      as.numeric(index[letters[1:3], 2]),
      weights(pias2, ea_only = TRUE)
    )[1:2])
  )
})

test_that("duplicate products get unique names during aggregation", {
  epr1 <- elementary_index(
    setNames(1:4, 1:4),
    ea = gl(2, 2),
    period = gl(1, 4),
    contrib = TRUE,
    r = 1
  )
  epr2 <- epr1
  levels(epr2) <- 3:4
  index <- aggregate(
    merge(epr1, epr2),
    list(c(0, 0, 0, 0), c("a", "a", "b", "b"), 1:4),
    duplicate_contrib = "make.unique"
  )
  expect_equal(
    contrib(index, "a"),
    matrix(c(0, 0.25, 0.5, 0.75), 4, dimnames = list(product = 1:4, time = 1))
  )
  expect_equal(
    contrib(index, "a"),
    contrib(index, "b")
  )
  expect_equal(
    contrib(index),
    matrix(
      rep(c(0, 0.125, 0.25, 0.375), each = 2),
      8,
      dimnames = list(
        product = c(1, "1.1", 2, "2.1", 3, "3.1", 4, "4.1"),
        time = 1
      )
    )
  )
})

test_that("aggregating in parallel works", {
  pias <- aggregation_structure(
    list(
      c(1, 1, 1, 2, 2, 2),
      c(11, 11, 12, 21, 22, 22),
      c(111, 112, 121, 211, 221, 222)
    ),
    c(1, 2, 3, 3, 2, 1)
  )

  pias1 <- aggregation_structure(
    list(
      c(1, 1, 1),
      c(11, 11, 12),
      c(111, 112, 121)
    ),
    c(1, 2, 3)
  )

  pias2 <- aggregation_structure(
    list(
      c(2, 2, 2),
      c(21, 22, 22),
      c(211, 221, 222)
    ),
    c(3, 2, 1)
  )

  epr1 <- elementary_index(
    1:18,
    period = rep(1:2, 9),
    ea = rep(c(111, 112, 121), each = 6),
    contrib = TRUE
  )
  epr2 <- elementary_index(
    18:1,
    period = rep(1:2, 9),
    ea = rep(c(111, 112, 121), each = 6),
    contrib = TRUE
  )
  levels(epr2) <- c(211, 221, 222)

  index <- aggregate(merge(epr1, epr2), pias, r = 0.5)
  index1 <- aggregate(epr1, pias, r = 0.5)
  index2 <- aggregate(epr2, pias, r = 0.5)

  expect_equal(index[1], index1[1])
  expect_equal(index[2], index2[2])
})

test_that("aggregating with a dead branch does nothing", {
  epr <- elementary_index(
    1:18,
    period = rep(1:3, each = 6),
    ea = rep(1:2, 9),
    contrib = TRUE
  )
  pias <- data.frame(
    l1 = c(0, 0, 0),
    l2 = c("01", "01", "02"),
    l3 = c("011", "012", "021"),
    ea = 1:3,
    weights = 3:1
  )
  index1 <- aggregate(epr, pias)
  index1na <- aggregate(epr, pias, na.rm = TRUE)
  index2 <- aggregate(epr, pias[1:2, ])

  expect_equal(index1["01"], index2["01"])
  expect_equal(index1na["0"], index2["0"])
  expect_equal(contrib(index1, "01"), contrib(index2, "01"))
  expect_equal(contrib(index1na), contrib(index2))
})

test_that("reaggregating doesn't introduce incorrect contributions", {
  epr <- elementary_index(
    c(1:7, NA),
    period = rep(1:2, each = 4),
    ea = rep(1:2, 4),
    contrib = TRUE
  )
  pias <- as_aggregation_structure(list(c(0, 0), c(1, 2)))
  index <- aggregate(epr, pias, na.rm = TRUE)
  expect_equal(
    contrib(index, period = 1),
    contrib(aggregate(index, pias), period = 1)
  )
  expect_equal(
    sum(contrib(index, period = 2), na.rm = TRUE),
    sum(contrib(aggregate(index, pias), period = 2), na.rm = TRUE)
  )
})

test_that("skipping time periods works", {
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    na.rm = TRUE
  )

  ms_pias <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index <- chain(aggregate(ms_epr, ms_pias, na.rm = TRUE, r = 2))

  ms_index2 <- unchain(ms_index[, -3])

  expect_equal(chain(aggregate(ms_index2, ms_pias, r = 2))[, 3], ms_index[, 4])
})

test_that("skipping eas works", {
  ms_epr <- elementary_index(
    ms_prices,
    price_relative(price, period = period, product = product) ~
      period + business,
    contrib = TRUE,
    na.rm = TRUE
  )

  ms_pias <- with(
    ms_weights,
    aggregation_structure(
      c(expand_classification(classification), list(business)),
      weight
    )
  )

  ms_index <- aggregate(ms_epr, ms_pias, na.rm = TRUE)
  expect_equal(
    ms_index[1:3],
    aggregate(ms_epr, ms_pias, na.rm = TRUE, include_ea = FALSE)
  )

  expect_error(
    aggregate(
      ms_epr,
      aggregation_structure(ms_weights["business"]),
      include_ea = FALSE
    )
  )
})

test_that("missing weights ignores those index values", {
  prices <- data.frame(
    rel = 1:8,
    period = rep(1:2, each = 4),
    ea = rep(letters[1:2], 4)
  )

  pias <- aggregation_structure(
    list(c("top", "top"), c("a", "b")),
    1:2
  )

  elementary <- elementary_index(prices, rel ~ period + ea, contrib = TRUE)

  index <- aggregate(elementary, pias)

  weights(pias)[1] <- NA
  index2 <- aggregate(elementary, pias, na.rm = TRUE)

  expect_equal(as.numeric(index2[1]), as.numeric(index2[3]))
  expect_equal(
    as.numeric(colSums(contrib(index2), na.rm = TRUE) + 1),
    as.numeric(index2[1])
  )
  expect_equal(contrib(index2), contrib(aggregate(elementary, pias)))
})

test_that("superlative index aggregates correctly", {
  epr <- as_index(matrix(c(1:4, NA, 6:9), 3), contrib = TRUE)
  contrib(epr, 1, 1) <- c(-2, -1, 1, 1, 0.25, 0.75)
  pias <- aggregation_structure(
    list(c("top", "top", "top"), 1:3),
    1:3
  )
  pias2 <- aggregation_structure(
    list(c("top", "top", "top"), 1:3),
    c(3, 1, 2)
  )
  res <- aggregate(epr, pias, pias2 = pias2, na.rm = TRUE)
  res1 <- aggregate(epr, pias, na.rm = TRUE)
  res2 <- aggregate(epr, pias2, r = -1, na.rm = TRUE)

  expect_equal(
    sqrt(as.matrix(res1) * as.matrix(res2)),
    as.matrix(res)
  )

  expect_equal(
    colSums(contrib(res), na.rm = TRUE),
    as.matrix(res)[1, ] - 1
  )
  expect_equal(
    colSums(contrib(res, 1)),
    as.matrix(res)[2, ] - 1
  )

  # Example from vignette.
  geometric_weights <- gpindex::transmute_weights(0, 1)

  w <- mapply(
    \(x, y) geometric_weights(c(x, y)),
    as.numeric(res1[1]),
    as.numeric(res2[1])
  )

  laspeyres_contrib <- contrib(res1)
  paasche_contrib <- contrib(res2)

  fisher_contrib <- w[1, col(laspeyres_contrib)] *
    laspeyres_contrib +
    w[2, col(paasche_contrib)] * paasche_contrib

  expect_equal(contrib(res), fisher_contrib)

  # Tornqvist.
  expect_equal(
    aggregate(epr, pias, r = 0, na.rm = TRUE),
    aggregate(epr, pias, pias2 = pias, r = 0, na.rm = TRUE)
  )
})

test_that("duplicate product methods work with NAs", {
  x <- as_index(c(1, 2, NA))
  contrib(x, 1) <- c(0.5, -0.25, NA, -0.25)
  contrib(x, 3) <- c(1, NA, NA)
  pias <- list(c(0, 0, 0), 1:3)
  expect_equal(
    contrib(aggregate(x, pias, duplicate_contrib = "sum")),
    matrix(
      c(0.25, -0.125, NA, -0.125),
      dimnames = list(product = 1:4, time = 1)
    )
  )
  expect_equal(
    contrib(aggregate(x, pias, duplicate_contrib = "sum", na.rm = TRUE)),
    matrix(
      c(0.25, -0.125, NA, -0.125),
      dimnames = list(product = 1:4, time = 1)
    )
  )
})

test_that("superlative contributions work with NA weights", {
  x <- as_index(c("3" = 1, "4" = 2))
  contrib(x, 3) <- 1:5
  contrib(x, 4) <- 1:2
  pias1 <- aggregation_structure(list(c(0, 0), 1:2, 3:4), c(1, 1))
  weights(pias1)[1] <- NA
  pias2 <- aggregation_structure(list(c(0, 0), 1:2, 3:4), c(2, 2))

  expect_error(aggregate(x, pias1, pias2 = pias2, na.rm = TRUE))
  weights(pias2)[1] <- 0
  res <- aggregate(
    x,
    pias1,
    pias2 = pias2,
    na.rm = TRUE,
    duplicate_contrib = "sum"
  )

  expect_equal(
    contrib(res),
    matrix(
      c(1, 2, NA, NA, NA),
      nrow = 5,
      dimnames = list(product = 1:5, time = 1)
    )
  )
})

epr1 <- elementary_index(
  ms_prices,
  price_relative(price, period = period, product = product) ~ period + business,
  contrib = TRUE,
  na.rm = TRUE
)

epr2 <- elementary_index(
  ms_prices,
  price_relative(price, period = period, product = product) ~ period + business,
  na.rm = TRUE
)

pias <- with(
  ms_weights,
  aggregation_structure(
    c(expand_classification(classification), list(business)),
    weight
  )
)

index1 <- aggregate(epr1, pias, na.rm = TRUE)
index2 <- aggregate(epr2, pias, na.rm = TRUE)

test_that("chain is the same as apply", {
  expect_equal(
    as.matrix(chain(epr1)),
    t(apply(as.matrix(epr1), 1, cumprod))
  )
  expect_equal(
    as.matrix(chain(epr1, link = 1:4)),
    t(apply(as.matrix(epr1), 1, cumprod)) * 1:4
  )
  expect_equal(
    as.matrix(chain(index1)),
    t(apply(as.matrix(index1), 1, cumprod))
  )
})

test_that("unchain and chain are inverses with no NAs", {
  expect_equal(unchain(chain(index2)), index2)
  expect_equal(unchain(rebase(chain(index2), base = 1:8), base = 1:8), index2)
  expect_equal(unchain(chain(index2, link = 1:8), base = 1 / 1:8), index2)
  expect_failure(expect_equal(unchain(chain(epr2)), epr2))
})

test_that("unchaining/rebasing a chainable index does nothing", {
  expect_equal(unchain(epr1), epr1)
  expect_equal(unchain(index1), index1)
  expect_equal(rebase(epr1), epr1)
  expect_equal(rebase(index1), index1)
})

test_that("chaining a fixed-base index does nothing", {
  expect_equal(chain(epr1), chain(chain(epr1)))
  expect_equal(chain(index1), chain(chain(index1)))
})

test_that("rebase should be the same as division", {
  expect_equal(
    as.matrix(rebase(chain(epr2), base = 1:4)),
    as.matrix(chain(epr2)) / 1:4
  )
  expect_equal(
    as.matrix(rebase(chain(index2), base = 1:8)),
    as.matrix(chain(index2)) / 1:8
  )
})

test_that("rebase works with mean", {
  index2 <- chain(index2)
  expect_equal(
    rebase(index2, base = mean(index2, window = 2)[, 1]),
    rebase(index2, base = rowMeans(as.matrix(index2[, 1:2])))
  )

  expect_equal(
    rebase(index2, base = mean(index2, window = 4)),
    rebase(index2, base = rowMeans(as.matrix(index2)))
  )
})


test_that("chaining returns the correct type of index", {
  expect_true(is_chainable_index(epr1))
  expect_true(is_chainable_index(unchain(chain(epr1))))
  expect_true(is_direct_index(chain(epr1)))
  expect_true(is_chainable_index(index1))
  expect_true(is_chainable_index(unchain(chain(index1))))
  expect_true(is_direct_index(chain(index1)))
})

test_that("chaining keeps EA names", {
  expect_equal(
    as.matrix(chain(as_index(matrix(1:5, 1)))),
    matrix(cumprod(1:5), 1, dimnames = list(levels = 1, time = 1:5))
  )
})

test_that("link and base values are the right length", {
  expect_error(chain(epr1, link = 1))
  expect_error(rebase(chain(epr1), base = 1))
})

test_that("rebasing with a character vector works", {
  index1_chain <- chain(index1)
  index1_rebase <- rebase(
    index1_chain,
    base = index1_chain[, end(index1_chain)]
  )
  expect_equal(index1_rebase, rebase(index1_chain, base = end(index1_chain)))
  expect_equal(index2, unchain(index1_rebase, base = end(index1_rebase)))
})

dat <- data.frame(
  rel = c(1:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

pias <- as_aggregation_structure(
  data.frame(level1 = 1, level2 = c(11, 12, 13, 14), weight = 1)
)

epr <- elementary_index(dat, rel ~ period + ea, contrib = TRUE)
epr2 <- elementary_index(dat, rel ~ period + ea)

index <- aggregate(epr, pias)

res <- c(
  1.41421356237309,
  3.46410161513775,
  5,
  NaN,
  6.48074069840786,
  NaN,
  NaN,
  8
)

test_that("head and tail work", {
  expect_equal(head(epr, 1), epr[1])
  expect_equal(head(epr, 5), epr)
  expect_equal(head(epr, -1), epr[1:3])
  expect_equal(tail(epr, 1), epr[4])
  expect_equal(tail(epr, 5), epr)
  expect_equal(tail(epr, -1), epr[2:4])
  expect_equal(head(epr, 2:1), epr[1:2, 1])
  expect_equal(tail(epr, 2:1), epr[3:4, 2])
  expect_equal(head(epr, c(NA, -1)), epr[, 1])
  expect_equal(tail(epr, c(NA, -1)), epr[, 2])
})

test_that("subscripting methods work", {
  expect_equal(epr[], epr)
  expect_equal(index[1:5], index)
  expect_equal(
    epr[c(TRUE, FALSE, TRUE, TRUE), 2:1],
    elementary_index(
      dat,
      rel ~ factor(period, 2:1) + factor(ea, c(11, 13:14)),
      contrib = TRUE
    )
  )
  expect_equal(
    epr[c("14", "12"), TRUE],
    elementary_index(dat, rel ~ period + factor(ea, c(14, 12)), contrib = TRUE)
  )
  expect_equal(epr[matrix(c(1, 2, 1, 2), 2)], list(epr[1, 1], epr[2, 2]))
  expect_equal(epr[matrix(0, 2, 2)], list())
})

test_that("subscripting methods give errors where expected", {
  expect_error(epr[1, NA])
  expect_error(epr[1, 3])
  expect_error(epr[c(1, 2, 1), ])
  expect_error(epr[, NULL])
  expect_error(epr[integer(0)])
  expect_error(epr[NA])
  expect_error(epr[matrix(c(-1, 2), 1, 2)])
})

test_that("replacement methods work with an index", {
  # No op
  expect_equal(replace(epr, 1:4, epr), epr)
  expect_equal(replace(epr, integer(0), epr[1]), epr)
  expect_equal(replace(epr, matrix(0, 2, 2), epr[1]), epr)
  # Adds contributions
  expect_equal(replace(epr2, 1:4, epr), epr)
  # Removes contributions
  expect_equal(replace(epr, 1:4, epr2), set_contrib(epr2, value = numeric(0)))

  epr3 <- epr2
  epr3[c(TRUE, FALSE), 2] <- epr[1, 1]
  expect_equal(
    as.matrix(epr3),
    replace(as.matrix(epr), matrix(c(1, 3, 2, 2), 2), as.numeric(epr[1, 1]))
  )
  expect_equal(contrib(epr3, "13")[, 2], contrib(epr)[, 1])

  # Index doesn't recycle like a matrix
  m <- as.matrix(epr3)
  epr3[c(1, 1, 2, 2)] <- epr[1:2]
  m[1, ] <- as.matrix(epr)[2, ]
  expect_equal(as.matrix(epr3), m)
  expect_equal(contrib(epr3), contrib(epr, "12"))
  epr3[c(1, 1, 2, 2)] <- as.matrix(epr[1:2])
  expect_error(expect_equal(as.matrix(epr3), m))

  # Errors
  expect_error(epr[1:2] <- epr[, 1])
  expect_error(epr[1:3] <- epr[1:2])
  expect_error(epr[NA] <- epr[1])
  expect_error(epr[, NA] <- epr[1])
  expect_error(epr[1] <- chain(epr[1]))
  epr <- chain(epr)
  expect_error(epr[1] <- epr2[1])
  expect_error(epr[matrix(c(1, 2), 1)] <- unchain(epr[1, 1]))
})

test_that("replacement methods work with a vector", {
  epr[, 1] <- 1
  expect_equal(
    as.matrix(epr),
    matrix(
      c(1, 1, 1, 1, res[5:8]),
      4,
      2,
      dimnames = list(levels = 11:14, time = 1:2)
    )
  )
  expect_equal(
    contrib(epr),
    matrix(
      c(0, 0, 2.5962965079607, 2.88444419044716),
      2,
      2,
      dimnames = list(product = c("11.1", "11.2"), time = 1:2)
    )
  )

  epr[1, c(FALSE, TRUE)] <- "2"
  expect_equal(
    as.matrix(epr),
    matrix(
      c(1, 1, 1, 1, 2, res[6:8]),
      4,
      2,
      dimnames = list(levels = 11:14, time = 1:2)
    )
  )
  expect_equal(
    contrib(epr),
    matrix(0, 0, 2, dimnames = list(product = NULL, time = 1:2))
  )

  # recycling should still happen
  epr2[1:3, ] <- 8:9
  expect_equal(
    epr2[1:3, ],
    as_index(matrix(8:9, 3, 2, dimnames = list(levels = 11:13, time = 1:2)))
  )

  epr[1, c(1, 2, 1)] <- setNames(1:3, letters[1:3])
  expect_equal(
    epr[1, ],
    set_contrib(
      as_index(matrix(3:2, 1, dimnames = list(levels = "11", time = 1:2))),
      value = numeric(0)
    )
  )

  epr["14"] <- 1
  expect_equal(
    as.matrix(epr),
    matrix(
      c(3, 1, 1, 1, 2, res[6:7], 1),
      4,
      2,
      dimnames = list(levels = 11:14, time = 1:2)
    )
  )
  expect_equal(
    contrib(epr, "14"),
    matrix(0, 0, 2, dimnames = list(product = NULL, time = 1:2))
  )

  # Replacing nothing
  epr2 <- epr
  suppressWarnings(epr[FALSE] <- "a")
  expect_identical(epr, epr2)

  suppressWarnings(epr[1, 0] <- "a")
  expect_identical(epr, epr2)

  # Errors
  expect_error(epr[1] <- 1:3)
  expect_error(epr[1:2] <- 1:3)
  expect_error(epr[NA] <- 1)
  expect_error(epr[1, NA] <- 1)
  expect_error(epr[1, c(TRUE, FALSE, FALSE)] <- 1)
  expect_error(epr[5] <- 1)
  expect_error(epr[1] <- numeric(0))
  expect_warning(epr[c(TRUE, FALSE, TRUE), ])
})

test_that("replacement method works with a matrix", {
  epr3 <- replace(epr, is.na(epr), 1:3)
  expect_equal(
    as.matrix(epr3),
    replace(as.matrix(epr), c(4, 6, 7), c(1, 2, 3))
  )

  epr4 <- replace(
    epr3,
    matrix(c("11", "11", "12", "12", "2", "2", "2", "2"), 4),
    1:2
  )
  expect_equal(as.matrix(epr4), replace(as.matrix(epr3), 5:6, 2))

  expect_identical(replace(epr, matrix(FALSE, 4, 2), 0), epr)

  epr5 <- replace(epr3, is.na(epr), epr[1, 1])
  expect_equal(
    as.matrix(epr5),
    replace(as.matrix(epr), c(4, 6, 7), sqrt(2))
  )
  expect_equal(contrib(epr5[4, 1]), contrib(epr[1, 1]))

  # Works when there are no contributions.
  epr5 <- replace(epr2, is.na(epr), epr[1, 1])
  expect_equal(
    as.matrix(epr5),
    replace(as.matrix(epr2), c(4, 6, 7), sqrt(2))
  )
  expect_equal(contrib(epr5[4, 1]), contrib(epr[1, 1]))

  expect_error(replace(epr, matrix(NA, 4, 2), 0))
  expect_error(epr[is.na(epr)] <- numeric(0))
  expect_error(epr[matrix(TRUE, 3, 3)] <- 1)
  expect_error(epr[matrix(1, nrow = 1, ncol = 3)] <- 1)
  expect_warning(epr[is.na(epr)] <- 1:2)
  expect_error(epr[is.na(epr)] <- list(1))
  expect_error(epr[!is.na(epr)] <- epr[1])
})

#' Window a price index
#'
#' Extract and replace index values over a window of time periods.
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param start `[character(1)]` The time period to start the window.
#'   The default in the first period of `x`.
#' @param end `[character(1)]` The time period to end the window. The default
#'   is the last period of `x`.
#' @param ... Not currently used.
#' @param value `[numeric > 0 | piar_index]` A numeric vector or price index
#'   of replacement values.
#'
#' @returns
#' `window()` extracts a price index over a window of time periods that
#' inherits from the same class as `x`. The replacement method replaces these
#' with `value`.
#'
#' @examples
#' x <- as_index(matrix(1:9, 3))
#'
#' window(x, "2")
#'
#' window(x, "2") <- 1
#' x
#'
#' @family index methods
#' @importFrom stats window
#' @export
# Optional arguments are before ... to agree with default in stats::window().
window.piar_index <- function(x, start = NULL, end = NULL, ...) {
  chkDots(...)
  x[, index_window(x, start, end)]
}

#' @rdname window.piar_index
#' @importFrom stats window<-
#' @export
`window<-.piar_index` <- function(x, start = NULL, end = NULL, ..., value) {
  chkDots(...)
  x[, index_window(x, start, end)] <- value
  x
}

#' Get the indexes for a window of time periods
#' @noRd
index_window <- function(x, start, end) {
  start <- if (!is.null(start)) match_time(as.character(start), x) else 1L
  end <- if (!is.null(end)) match_time(as.character(end), x) else ntime(x)

  if (start > end) {
    stop("`start` must refer to a time period before `end`")
  }
  seq.int(start, end)
}

#' Update an aggregation structure
#'
#' Price update the weights in a price index aggregation structure.
#'
#' @importFrom stats update
#' @family aggregation structure methods
#' @export
#'
#' @param object `[piar_aggregation_structure]` A price index aggregation
#'   structure, as made by [aggregation_structure()].
#' @param index `[piar_index]` A fixed-base (direct) price index, or something
#'   that can be coerced into one. Usually an aggregate price index as made by
#'   [`aggregate()`][aggregate.piar_index].
#' @param period `[character(1)]` The time period used to price update the
#'   weights. The default uses the last period in `index`.
#' @param order `[numeric(1)]` Order of the generalized mean to update the
#'   weights. The default is 1 for an arithmetic index.
#' @param ... Not currently used.
#' @param r Deprecated.
#'
#' @returns
#' A copy of `object` with price-updated weights using the index
#' values in `index`.
#'
#' @seealso
#' [`aggregate()`][aggregate.piar_index] to make an aggregated price index.
#'
#' @examples
#' # A simple aggregation structure.
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #  |---+---|       |
#' #  111     112     121
#' #  (1)     (3)     (4)
#' aggregation_weights <- data.frame(
#'   level1 = c("1", "1", "1"),
#'   level2 = c("11", "11", "12"),
#'   ea     = c("111", "112", "121"),
#'   weight = c(1, 3, 4)
#' )
#'
#' pias <- as_aggregation_structure(aggregation_weights)
#'
#' index <- as_index(
#'   matrix(1:9, 3, dimnames = list(c("111", "112", "121"), NULL))
#' )
#'
#' weights(pias, ea_only = FALSE)
#'
#' weights(update(pias, index), ea_only = FALSE)
update.piar_aggregation_structure <- function(
  object,
  index,
  ...,
  period = NULL,
  order = 1,
  r = order
) {
  if ("r" %in% names(sys.call())) {
    warning("`r` is deprecated and will be removed; use `order` instead")
  }
  chkDots(...)
  index <- as_index(index, chainable = FALSE)
  r <- as.numeric(r)
  period <- if (!is.null(period)) {
    match_time(as.character(period), index)
  } else {
    ntime(index)
  }
  eas <- match_eas(object, index)
  if (anyNA(eas)) {
    warning("not all weights in `object` have a corresponding index value")
  }
  weights(object) <- update_weights(
    index$index[, period][eas],
    object$weights,
    order = r
  )
  object
}

#' Back period
#'
#' Offset a vector prices or quantities by computing the position one (or more)
#' period backwards for each product.
#'
#' @export
#'
#' @param period `[factor]` A factor, or something that can be coerced into one,
#'   that
#'   gives the time period for each transaction. The ordering of time periods
#'   follows the levels of `period` to agree with
#'   [`cut()`][cut.Date].
#' @param product `[factor]` A factor, or something that can be coerced into
#'   one, that
#'   gives the product identifier for each transaction. The default is to
#'   assume that all transactions are for the same product.
#' @param match_first `[logical(1)]` Should products in the first period match
#'   with themselves (the default)?
#' @param offset `[integer(1)]` The number of periods to offset. The default
#'   offsets by one period (back period). Setting to `nlevels(period)` gives
#'   the the base period.
#'
#' @returns
#' A numeric vector of indices giving the position of the the back periods.
#'
#' @note
#' By definition, there must be at most one transaction for each product
#' in each time period to determine a back period. If multiple transactions
#' correspond to a period-product pair, then the back period at a point in
#' time is always the first position for that product in the previous period.
#'
#' @seealso
#' [`outliers()`] for common methods to detect outliers for price relatives.
#'
#' `rs_pairs` in the \pkg{rsmatrix} package for making sales pairs.
#'
#' @examples
#' prices <- data.frame(
#'   price = 1:6,
#'   product = factor(c("a", "b")),
#'   period = factor(c(1, 1, 2, 2, 3, 3))
#' )
#'
#' with(prices, back_period(period, product))
#'
#' # Make fixed-base price relatives.
#' with(
#'   prices,
#'   price / price[back_period(period, product, offset = nlevels(period))]
#' )
#'
#' # Change the base period with relevel().
#' with(
#'   prices,
#'   price / price[
#'     back_period(relevel(period, "2"), product, offset = nlevels(period))
#'   ]
#' )
back_period <- function(
  period,
  product = NULL,
  match_first = TRUE,
  offset = 1L
) {
  f <- function(x, offset) {
    x[c(rep.int(1L, offset), seq_len(length(x) - offset))]
  }
  offset <- as.integer(offset)
  period <- as.factor(period)
  if (!is.null(product)) {
    product <- as.factor(product)
    attributes(product) <- NULL # matching is faster on factor codes
  } else {
    product <- rep.int(1L, length(period))
  }

  if (length(period) != length(product)) {
    stop("`period` and `product` must be the same length")
  }
  if (offset < 0L || offset > nlevels(period)) {
    stop("`offset` must be a positive integer less than `nlevels(period)`")
  }

  product <- split(product, period)
  if (duplicate_products(product)) {
    warning("there are duplicated period-product pairs")
  }
  if (offset == 0L) {
    return(seq_along(period))
  }
  m <- Map(match, product, f(product, offset), incomparables = NA)
  if (!match_first) {
    m[seq_len(offset)][] <- NA_integer_
  }
  res <- split(seq_along(period), period)
  unsplit(Map(`[`, f(res, offset), m), period)
}

id <- letters[c(5:1, 1:5)]
period <- rep(1:2, each = 5)

# Offsetting periods works.
local({
  # Length 0 inputs
  expect_identical(back_period(NULL, NULL, offset = 0), integer(0))
  expect_identical(
    back_period(numeric(0), factor(numeric(0), 1:3), offset = 0),
    integer(0)
  )
  expect_identical(
    back_period(factor(numeric(0), 1:3), numeric(0), offset = 0),
    integer(0)
  )
  expect_identical(
    back_period(factor(numeric(0), 1:3), factor(numeric(0), 1:3), offset = 0),
    integer(0)
  )

  # Simple cases
  expect_identical(back_period(1:4), c(1L, 1L, 2L, 3L))
  expect_identical(back_period(1:4, offset = 4), rep(1L, 4))

  # Attributes shouldn't do anything
  expect_identical(back_period(matrix(1:4)), c(1L, 1L, 2L, 3L))

  # Change time periods
  expect_identical(
    back_period(factor(rep(1, 4), levels = 0:1), 1:4),
    rep(NA_integer_, 4)
  )
  expect_identical(back_period(factor(1:4, levels = 4:1)), c(2L, 3L, 4L, 4L))

  # A more interesting case
  expect_identical(
    back_period(period, id),
    c(1L, 2L, 3L, 4L, 5L, 5L, 4L, 3L, 2L, 1L)
  )
  expect_identical(
    back_period(period, id, offset = 2),
    c(1L, 2L, 3L, 4L, 5L, 5L, 4L, 3L, 2L, 1L)
  )
  expect_identical(
    back_period(replace(period, 2, NA), id),
    c(1L, NA, 3L, 4L, 5L, 5L, 4L, 3L, NA, 1L)
  )
  expect_identical(
    back_period(period[-1], id[-1]),
    c(1L, 2L, 3L, 4L, 4L, 3L, 2L, 1L, NA)
  )
  expect_identical(
    back_period(factor(period, levels = NA), id, offset = 0),
    1:10
  )

  # Change time periods again
  expect_identical(
    back_period(factor(period, c(1, 3, 2)), id),
    c(1L, 2L, 3L, 4L, 5L, NA, NA, NA, NA, NA)
  )
  expect_identical(
    back_period(factor(period, c(1, 3, 2)), id, offset = 2),
    c(1L, 2L, 3L, 4L, 5L, 5L, 4L, 3L, 2L, 1L)
  )

  # NA products shouldn't trigger a warning
  expect_identical(
    back_period(period, replace(id, 1:2, NA)),
    c(NA, NA, 3L, 4L, 5L, 5L, 4L, 3L, NA, NA)
  )
})

# Not matching the first period works.
local({
  expect_identical(back_period(period, id, FALSE, offset = 0), 1:10)
  expect_identical(back_period(period, id, FALSE), c(rep(NA, 5), 5:1))
  expect_identical(
    back_period(period, id, FALSE, offset = 2),
    rep(NA_integer_, 10)
  )
})

# Warnings and errors.
local({
  expect_warning(
    back_period(c(1, 1, 2, 3)),
    "there are duplicated period-product pairs"
  )
  expect_error(
    back_period(1:5, 1:4),
    "`period` and `product` must be the same length"
  )
  expect_error(
    back_period(c(1, 1, 2, 2), c(1, 2, 1, 2), offset = 3),
    "`offset` must be a positive integer less than `nlevels\\(period\\)`"
  )
})

#' Scale weights
#'
#' Scale a vector of weights so that they sum to 1.
#'
#' @family math functions
#' @export
#'
#' @param x `[numeric >= 0]` A positive numeric vector with at least one
#'   non-zero element.
#'
#' @returns
#' A numeric vector that sums to 1. If there are `NA`s in `x` then the result
#' sums 1 to if these values are removed.
#'
#' @examples
#' scale_weights(1:5)
#'
#' scale_weights(c(1:5, NA))
scale_weights <- function(x) {
  x / sum(x, na.rm = TRUE)
}

library(gpindex)

ms_prices$rel <- price_relative(ms_prices, price ~ period + product)
ms_prices$w1 <- 1:40
ms_prices$w2 <- 40:1

epr1 <- elementary_index(ms_prices, rel ~ period + business, contrib = TRUE)

epr2 <- elementary_index(
  ms_prices,
  rel ~ period + business,
  r = -1,
  contrib = TRUE,
  na.rm = TRUE
)

# Test a Fisher calculation
fw <- function(x, w1, w2) {
  v1 <- scale_weights(transmute_weights(1, 0)(x, w1))
  v2 <- scale_weights(transmute_weights(-1, 0)(x, w2))
  v1 + v2
}

ms_prices2 <- na.omit(ms_prices)

w <- with(
  ms_prices2,
  grouped(fw)(rel, w1, w2, group = interaction(period, business))
)

epr3 <- elementary_index(
  ms_prices2,
  rel ~ period + business,
  weights = w,
  contrib = TRUE
)

test_that("results agree with an alternate implementation", {
  epr11 <- aggregate(
    rel ~ as.character(business) + period,
    ms_prices,
    function(x) exp(weighted.mean(log(x))),
    na.action = na.pass,
    drop = FALSE
  )
  epr22 <- aggregate(
    rel ~ as.character(business) + period,
    ms_prices,
    function(x) 1 / weighted.mean(1 / x),
    na.action = na.omit,
    drop = FALSE
  )
  names(epr11) <- names(epr22) <- c("level", "period", "value")

  expect_equal(as.data.frame(epr1), epr11[c(2, 1, 3)])
  expect_equal(as.data.frame(epr2), epr22[c(2, 1, 3)])
})

test_that("Fisher calculation agrees with manual calculation", {
  l <- elementary_index(
    ms_prices2,
    rel ~ period + business,
    weights = w1,
    r = 1
  )
  p <- elementary_index(
    ms_prices2,
    rel ~ period + business,
    weights = w2,
    r = -1
  )
  expect_equal(sqrt(as.matrix(l) * as.matrix(p)), as.matrix(epr3))

  # Should work for other kinds of superlative indexes
  fw <- function(x, w1, w2) {
    v1 <- scale_weights(transmute_weights(1.5, 0)(x))
    v2 <- scale_weights(transmute_weights(-1.5, 0)(x, w2))
    v1 + v2
  }

  w <- with(
    ms_prices2,
    grouped(fw)(rel, w1, w2, group = interaction(period, business))
  )

  sepr <- elementary_index(ms_prices2, rel ~ period + business, weights = w)

  l <- elementary_index(ms_prices2, rel ~ period + business, r = 1.5)
  p <- elementary_index(
    ms_prices2,
    rel ~ period + business,
    weights = w2,
    r = -1.5
  )
  expect_equal(sqrt(as.matrix(l) * as.matrix(p)), as.matrix(sepr))
})

test_that("contributions add up", {
  sum_contrib <- \(x, ...) sum(x, ...) / (sum(!is.na(x)) > 0)
  expect_equal(
    as.numeric(epr1$index),
    sapply(epr1$contrib, sum_contrib) + 1
  )

  expect_equal(
    as.numeric(epr2$index),
    sapply(epr2$contrib, sum_contrib, na.rm = TRUE) + 1
  )

  expect_equal(
    as.numeric(epr3$index),
    sapply(epr3$contrib, sum_contrib) + 1
  )
})

test_that("argument checking works", {
  expect_error(elementary_index(1:3, period = 1:2, ea = 1:3))
  expect_error(elementary_index(1:3, period = 1:3, ea = 1:4))
  expect_error(elementary_index(1:3, period = 1:3, ea = 1:3, weights = 1:2))
  expect_error(elementary_index(1:3, period = factor(1:3, levels = numeric(0))))
  expect_error(elementary_index(1:3, ea = factor(1:3, levels = numeric(0))))
  expect_error(elementary_index(setNames(1:3, c("", 1, 2)), contrib = TRUE))
  expect_error(elementary_index(-1:1, period = 1:3, ea = 1:3, r = 1))
  expect_warning(elementary_index(
    setNames(1:3, rep(1, 3)),
    period = gl(1, 3),
    ea = gl(1, 3),
    contrib = TRUE
  ))
})

test_that("nse works", {
  expect_equal(
    elementary_index(ms_prices, rel ~ period:business, contrib = TRUE),
    epr1
  )
  expect_equal(
    elementary_index(ms_prices[c(1:2, 5)], rel ~ ., contrib = TRUE),
    epr1
  )
  expect_error(elementary_index(ms_prices, rel ~ .))
  expect_error(elementary_index(ms_prices, ~ period + business))
})

#' Coerce an aggregation structure into a tabular form
#'
#' Coerce a price index aggregation structure into an aggregation matrix or a
#' data frame.
#'
#' @family aggregation structure methods
#' @export
#'
#' @param x `[piar_agregation_structure]` A price index aggregation structure,
#'   as made by [aggregation_structure()].
#' @param sparse `[logical(1)]` Should the result be a sparse matrix from
#'   \pkg{Matrix}? This is faster for large aggregation structures. The default
#'   returns an ordinary dense matrix.
#' @param row.names `[character]` See [as.data.frame()].
#' @param optional Not currently used.
#' @param ... Not currently used for the matrix method. Extra arguments to
#'   [as.data.frame.list()] for the data frame method.
#'
#' @returns
#' `as.matrix()` represents an aggregation structure as a matrix,
#' such that multiplying with a (column) vector of elementary indexes gives the
#' aggregated index.
#'
#' `as.data.frame()` takes an aggregation structure and returns a data
#' frame that could have generated it.
#'
#' @seealso
#' [as_aggregation_structure()] for coercing into an aggregation structure.
#'
#' [treemap::treemap()] and [data.tree::as.Node()] for visualizing an
#' aggregation structure.
#'
#' @examples
#' # A simple aggregation structure.
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #  |---+---|       |
#' #  111     112     121
#' #  (1)     (3)     (4)
#' aggregation_weights <- data.frame(
#'   level1 = c("1", "1", "1"),
#'   level2 = c("11", "11", "12"),
#'   ea     = c("111", "112", "121"),
#'   weight = c(1, 3, 4)
#' )
#'
#' pias <- as_aggregation_structure(aggregation_weights)
#'
#' as.matrix(pias)
#'
#' all.equal(as.data.frame(pias), aggregation_weights)
#'
#' \dontrun{
#' # Visualize as a treemap.
#' treemap::treemap(
#'   aggregation_weights,
#'   index = names(aggregation_weights)[-4],
#'   vSize = "weight",
#'   title = "aggregation structure"
#' )
#'
#' # Or turn into a more genereal tree object and plot.
#' aggregation_weights$pathString <- do.call(
#'   \(...) paste(..., sep = "/"),
#'   aggregation_weights[-4]
#' )
#' plot(data.tree::as.Node(aggregation_weights))
#' }
as.matrix.piar_aggregation_structure <- function(x, ..., sparse = FALSE) {
  chkDots(...)
  nea <- length(x$weights)
  height <- nlevels(x)
  if (height == 1L) {
    res <- matrix(
      numeric(0L),
      ncol = nea,
      dimnames = list(NULL, x$levels[[1L]])
    )
    if (sparse) {
      return(Matrix::Matrix(res, sparse = TRUE))
    } else {
      return(res)
    }
  }
  cols <- seq_len(nea)
  # Don't need the eas.
  lev <- lapply(as.list(x)[-height], \(z) factor(z, unique(z)))
  res <- vector("list", length(lev))
  # Generate the rows for each level of the matrix and rbind together.
  for (i in seq_along(res)) {
    w <- unsplit(
      lapply(split(x$weights, lev[[i]]), scale_weights),
      lev[[i]]
    )
    if (sparse) {
      mat <- Matrix::sparseMatrix(lev[[i]], cols, x = w)
    } else {
      mat <- matrix(0, nlevels(lev[[i]]), nea)
      mat[cbind(lev[[i]], cols)] <- w
    }
    dimnames(mat) <- list(levels(lev[[i]]), x$levels[[height]])
    res[[i]] <- mat
  }
  res <- do.call(rbind, res)
  names(dimnames(res)) <- c("levels", "")
  res
}

#' @rdname as.matrix.piar_aggregation_structure
#' @export
as.data.frame.piar_aggregation_structure <- function(
  x,
  row.names = NULL,
  optional = FALSE,
  ...
) {
  res <- as.data.frame(
    as.list(x),
    row.names = row.names,
    optional = optional,
    ...
  )
  res$weight <- x$weights
  res
}

#' @export
as.list.piar_aggregation_structure <- function(x, ...) {
  chkDots(...)
  if (nlevels(x) == 1L) {
    return(x$levels[1L])
  }
  res <- vector("list", length(x$parent))
  res[[1L]] <- x$parent[[1L]]
  # Walk up the parent nodes to reconstruct the inputs that generated 'x'.
  for (i in seq_along(x$parent)[-1L]) {
    res[[i]] <- x$parent[[i]][res[[i - 1L]]]
  }
  top <- names(last(x$child))[last(res)]
  res <- c(list(top), lapply(rev(res), names))
  names(res) <- names(x$levels)
  res
}

#' GEKS index
#'
#' Calculate a generalized inter-temporal GEKS price index over a rolling
#' window.
#'
#' @export
#'
#' @param index_formula `[function]` A function giving the index-number formula
#'   in the GEKS
#'   index. Usually a Törnqvist, Fisher (the default), or Walsh index.
#'   It must have arguments `p1`, `p0`, `q1`, and `q0`, and satisfy the
#'   time-reversal test. See `vignette("index-number-formulas")` for details.
#' @param order `[numeric(1)]` A finite number giving the order of the
#'   generalized mean used to
#'   average price indexes over the rolling window. The default uses a
#'   geometric mean.
#' @param price `[numeric > 0]` A numeric vector of prices, the same length
#'   as `quantity`.
#' @param quantity `[numeric >= 0]` A numeric vector of quantities, the same
#'   length as `price`.
#' @param period `[factor]` A factor, or something that can be coerced into one,
#'   that gives the corresponding time period for each element in `price` and
#'   `quantity`. The ordering of time periods follows the levels of `period`
#'   to agree with [`cut()`][cut.Date].
#' @param product `[factor]` A factor, or something that can be coerced into
#'   one, that
#'   gives the corresponding product identifier for each element in `price` and
#'   `quantity`.
#' @param window `[integer(1) > 0]` A positive integer giving the length of the
#'   rolling window.
#'   The default is a window that encompasses all periods in `period`.
#'   Non-integers are truncated towards zero.
#' @param n `[integer(1) > 0]` A positive integer giving the length of the index
#'   series for each
#'   window, starting from the end of the window. For example, if there are 13
#'   periods in `window`, setting `n = 1` gives the index for period 13. The
#'   default gives an index for each period in `window`. Non-integers are
#'   truncated towards zero.
#' @param match_method `[character(1)]` Either `"all"` to match all products
#'   against each other
#'   (the default) or `"back-price"` to match only back prices. The later can be
#'   faster when there is lots of product imbalanced.
#'
#' @returns
#' A list with a named numeric vector giving the value of the respective
#' period-over-period GEKS index for each window.
#'
#' @note
#' Like [back_period()], if multiple prices
#' correspond to a period-product pair, then the back price at a point in time
#' is always the first price for that product in the previous period. Unlike a
#' bilateral index, however, duplicated period-product pairs can have more
#' subtle implications for a multilateral index.
#'
#' @seealso
#' [`splice_index()`] to splice the rolling-window indexes together.
#'
#' `GEKSIndex()` in the \pkg{IndexNumR} package for an implementation of the
#' GEKS index with more options.
#'
#' The \pkg{rsmatrix} package for multilateral repeat-sales indexes.
#'
#' @references
#' Balk, B. M. (2008). *Price and Quantity Index Numbers*.
#' Cambridge University Press.
#'
#' IMF, ILO, Eurostat, UNECE, OECD, and World Bank. (2020).
#' *Consumer Price Index Manual: Concepts and Methods*.
#' International Monetary Fund.
#'
#' Ivancic, L., Diewert, W. E., and Fox, K. J. (2011). Scanner data, time
#' aggregation and the construction of price indexes.
#' *Journal of Econometrics*, 161(1): 24--35.
#'
#' @examples
#' price <- 1:10
#' quantity <- 10:1
#' period <- rep(1:5, 2)
#' product <- rep(letters[1:2], each = 5)
#'
#' cumprod(geks_index(price, quantity, period, product)[[1]])
#'
#' # Calculate the index over a rolling window.
#' (geks <- geks_index(price, quantity, period, product, window = 3))
#'
#' # Use a movement splice to combine the indexes in each window.
#' splice_index(geks, 2)
#'
#' # ... or use a mean splice.
#' splice_index(geks)
#'
#' # Make a Jevons GEKS index.
#' geks_index(
#'   price,
#'   quantity,
#'   period,
#'   product,
#'   index_formula = \(p1, p0, ...) gmean(p1 / p0, na.rm = TRUE, order = 0)
#' )
geks_index <- function(
  price,
  quantity,
  period,
  product,
  index_formula = \(p1, p0, q1, q0) {
    nested_gmean(p1 / p0, list(p0 * q0, p1 * q1), na.rm = TRUE)
  },
  window = nlevels(period),
  n = window - 1L,
  order = 0,
  match_method = c("all", "back-price")
) {
  period <- as.factor(period)
  product <- as.factor(product)
  attributes(product) <- NULL # faster to match on numeric codes

  if (different_length(price, quantity, period, product)) {
    stop(
      "`price`, `quantity`, `period`, and `product` must be the same length"
    )
  }

  match_method <- match.arg(match_method)

  if (nlevels(period) == 0L) {
    return(list())
  }

  window <- as.integer(window)
  if (window < 2L) {
    stop("`window` must be greater than or equal to 2")
  }
  if (window > nlevels(period)) {
    stop(
      "`window` must be less than or equal to the number of levels in",
      " `period`"
    )
  }

  n <- as.integer(n)
  if (n < 1L) {
    stop("`n` must be greater than or equal to 1")
  }
  if (n > window - 1L) {
    stop("`n` must be less than or equal to `window` minus 1")
  }

  mat <- geks_matrix(
    index_formula,
    price,
    quantity,
    period,
    product,
    window,
    n,
    match_method
  )
  rows <- seq_len(window) - 1L
  # Only the last n + 1 indexes in each window need to be kept.
  cols <- seq.int(window - n, window) - 1L
  res <- vector("list", nlevels(period) - window + 1L)
  # Move down the diagonal to make the geks index.
  for (i in seq_along(res)) {
    index <- apply(
      mat[rows + i, cols + i, drop = FALSE],
      2L,
      \(x) gmean(x, order = order, na.rm = TRUE)
    )
    res[[i]] <- index[-1L] / index[-length(index)]
  }
  res
}

#' Make the GEKS matrix
#' @noRd
geks_matrix <- function(
  index_formula,
  price,
  quantity,
  period,
  product,
  window,
  n,
  method
) {
  p <- split(price, period)
  q <- split(quantity, period)

  if (method == "all") {
    product <- balance_products(product, period)
    p <- Map(`[`, p, product)
    q <- Map(`[`, q, product)
  } else {
    product <- split(product, period)
  }

  lt <- vector("list", nlevels(period))
  for (i in seq_along(lt)) {
    if (i < max(window - n, 2L)) {
      # Only the last n + 1 rows are needed for each window,
      # so pad the top rows left of the diagonal with NA.
      ans <- rep_len(NA_real_, i - 1L)
    } else {
      # Matching is only done for the lower-triangular part of the matrix.
      # Match products for window - 1 periods left of the diagonal
      # to minimize the number of back prices to find.
      js <- seq.int(to = i - 1L, length.out = min(window, i) - 1L)
      if (method == "all") {
        ans <- Map(index_formula, p1 = p[js], p0 = p[i], q1 = q[js], q0 = q[i])
      } else {
        m <- Map(match, product[js], product[i])
        bp <- Map(`[`, p[i], m)
        bq <- Map(`[`, q[i], m)
        ans <- Map(index_formula, p1 = p[js], p0 = bp, q1 = q[js], q0 = bq)
      }
    }
    # Add the diagonal at the end and pad with NAs.
    ans <- c(
      unlist(ans, use.names = FALSE),
      index_formula(p[[i]], p[[i]], q[[i]], q[[i]])
    )
    front_pad <- rep_len(NA_real_, max(i - window, 0L))
    back_pad <- rep_len(NA_real_, length(lt) - length(ans) - length(front_pad))
    lt[[i]] <- c(front_pad, ans, back_pad)
  }
  res <- do.call(rbind, lt)
  rownames(res) <- colnames(res) <- names(p) # time periods
  # Exploit time reversal.
  ut <- upper.tri(res)
  res[ut] <- 1 / t(res)[ut]
  res
}

#' Coerce to an aggregation structure
#'
#' Coerce an object into an aggregation structure object.
#'
#' The default method attempts to coerce `x` into a list prior to calling
#' [aggregation_structure()].
#'
#' The data frame and matrix methods treat `x` as a table with a row for
#' each elementary aggregate, a column of labels for each level in the
#' aggregation structure, and a column of weights for the elementary aggregates.
#'
#' @export
#'
#' @param x `[object]` An object to coerce into an aggregation structure.
#' @param weights `[numeric >= 0]` A numeric vector of aggregation weights for
#'   the elementary
#'   aggregates. The default is to give each elementary aggregate the same
#'   weight.
#' @param ... Further arguments passed to or used by methods.
#'
#' @returns
#' A price index aggregation structure that inherits from
#' [`piar_aggregation_structure`].
#'
#' @seealso
#' [`as.matrix()`][as.matrix.piar_aggregation_structure] and
#' [`as.data.frame()`][as.data.frame.piar_aggregation_structure] for
#' coercing an aggregation structure into a tabular form.
#'
#' @examples
#' # A simple aggregation structure
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #  |---+---|       |
#' #  111     112     121
#' #  (1)     (3)     (4)
#' aggregation_weights <- data.frame(
#'   level1 = c("1", "1", "1"),
#'   level2 = c("11", "11", "12"),
#'   ea     = c("111", "112", "121"),
#'   weight = c(1, 3, 4)
#' )
#'
#' pias <- aggregation_structure(
#'   aggregation_weights[1:3],
#'   weights = aggregation_weights[[4]]
#' )
#'
#' all.equal(
#'   pias,
#'   as_aggregation_structure(aggregation_weights)
#' )
#'
#' all.equal(
#'   pias,
#'   as_aggregation_structure(as.matrix(aggregation_weights))
#' )
as_aggregation_structure <- function(x, ...) {
  UseMethod("as_aggregation_structure")
}

#' @rdname as_aggregation_structure
#' @export
as_aggregation_structure.default <- function(x, ..., weights = NULL) {
  chkDots(...)
  aggregation_structure(as.list(x), weights)
}

#' @rdname as_aggregation_structure
#' @export
as_aggregation_structure.data.frame <- function(x, ...) {
  chkDots(...)
  x <- as.list(x)
  aggregation_structure(drop_last(x), last(x))
}

#' @rdname as_aggregation_structure
#' @export
as_aggregation_structure.matrix <- function(x, ...) {
  as_aggregation_structure(as.data.frame(x), ...)
}

#' @export
as_aggregation_structure.piar_aggregation_structure <- function(x, ...) {
  chkDots(...)
  x
}

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")
agg <- aggregation_structure(list(x1, x2, x3))

# Levels.
local({
  expect_identical(
    levels(agg),
    list(
      level1 = c("1", "2"),
      level2 = c("11", "21", "12"),
      ea = c("111", "211", "121", "112")
    )
  )
  expect_error(levels(agg) <- 1:9)

  expect_identical(levels(agg), lapply(weights(agg, ea_only = FALSE), names))
})

# Weights.
local({
  expect_equal(
    weights(agg, ea_only = FALSE),
    list(
      level1 = c("1" = 3, "2" = 1),
      level2 = c("11" = 2, "21" = 1, "12" = 1),
      ea = c("111" = 1, "211" = 1, "121" = 1, "112" = 1)
    )
  )
  expect_equal(
    weights(agg, ea_only = TRUE),
    c("111" = 1, "211" = 1, "121" = 1, "112" = 1)
  )
  expect_equal(
    weights(aggregation_structure(list(x1, x2, x3), 1:4), ea_only = FALSE),
    list(
      level1 = c("1" = 8, "2" = 2),
      level2 = c("11" = 5, "21" = 2, "12" = 3),
      ea = c("111" = 1, "211" = 2, "121" = 3, "112" = 4)
    )
  )

  expect_equal(
    weights(as_aggregation_structure(1:5, weights = 2), ea_only = FALSE),
    list(
      level1 = c("1" = 2),
      level2 = c("2" = 2),
      level3 = c("3" = 2),
      level4 = c("4" = 2),
      ea = c("5" = 2)
    )
  )
  expect_equal(
    weights(as_aggregation_structure(1:5, weights = 2), ea_only = TRUE),
    c("5" = 2)
  )
  expect_equal(
    weights(
      as_aggregation_structure(list(1:5), weights = 1:5),
      ea_only = FALSE
    ),
    list(ea = c("1" = 1, "2" = 2, "3" = 3, "4" = 4, "5" = 5))
  )
  expect_equal(
    weights(as_aggregation_structure(list(1:5), weights = 1:5), ea_only = TRUE),
    c("1" = 1, "2" = 2, "3" = 3, "4" = 4, "5" = 5)
  )
})

# NAs move up the aggregation structure
local({
  expect_equal(
    weights(
      suppressWarnings(aggregation_structure(list(x1, x2, x3), c(NA, 2:4))),
      ea_only = FALSE
    ),
    list(
      level1 = c("1" = NA, "2" = 2),
      level2 = c("11" = NA, "21" = 2, "12" = 3),
      ea = c("111" = NA, "211" = 2, "121" = 3, "112" = 4)
    )
  )

  expect_equal(
    weights(
      suppressWarnings(aggregation_structure(list(x1, x2, x3), c(NA, 2:4))),
      ea_only = FALSE,
      na.rm = TRUE
    ),
    list(
      level1 = c("1" = 7, "2" = 2),
      level2 = c("11" = 4, "21" = 2, "12" = 3),
      ea = c("111" = NA, "211" = 2, "121" = 3, "112" = 4)
    )
  )
})

# Replacement methods.
local({
  weights(agg)[c(1, 3, 4)] <- 10
  expect_equal(weights(agg), c("111" = 10, "211" = 1, "121" = 10, "112" = 10))
})

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")
agg <- aggregation_structure(list(level1 = x1, level2 = x2, ea = x3))

expect_equal(
  as.matrix(agg),
  matrix(
    c(2, 0, 3, 0, 0, 0, 6, 0, 6, 0, 2, 0, 0, 0, 6, 2, 0, 3, 0, 0) / 6,
    5,
    4,
    dimnames = list(
      levels = c("1", "2", "11", "21", "12"),
      c("111", "211", "121", "112")
    )
  )
)
expect_equal(as.matrix(agg), as.matrix(as.matrix(agg, sparse = TRUE)))
expect_equal(
  as.data.frame(agg, row.names = 4:1),
  data.frame(
    level1 = c("1", "2", "1", "1"),
    level2 = c("11", "21", "12", "11"),
    ea = c("111", "211", "121", "112"),
    weight = c(1, 1, 1, 1),
    row.names = 4:1
  )
)
expect_equal(agg, as_aggregation_structure(as.data.frame(agg)))
expect_equal(as.list(agg), list(level1 = x1, level2 = x2, ea = x3))

agg2 <- aggregation_structure(1:3, 2)
expect_equal(
  as.matrix(agg2),
  matrix(c(1, 1), 2, 1, dimnames = list(levels = c("1", "2"), "3"))
)
expect_equal(as.matrix(agg2), as.matrix(as.matrix(agg2, sparse = TRUE)))
expect_equal(
  as.data.frame(agg2),
  data.frame(level1 = "1", level2 = "2", ea = "3", weight = 2)
)
expect_equal(as.list(agg2), list(level1 = "1", level2 = "2", ea = "3"))

agg3 <- aggregation_structure(list(letters[c(2, 1, 3)]), 1:3)
expect_equal(
  as.matrix(agg3),
  matrix(numeric(0), 0, 3, dimnames = list(character(0), c("b", "a", "c")))
)
expect_equal(as.matrix(agg3), as.matrix(as.matrix(agg3, sparse = TRUE)))
expect_equal(
  as.data.frame(agg3),
  data.frame(ea = c("b", "a", "c"), weight = 1:3)
)
expect_equal(as.list(agg3), list(ea = letters[c(2, 1, 3)]))

dat <- data.frame(
  rel = c(1:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

pias <- as_aggregation_structure(
  data.frame(level1 = 1, level2 = c(11, 12, 13, 14), weight = 1)
)

epr <- elementary_index(dat, rel ~ period + ea, contrib = TRUE)
epr2 <- elementary_index(dat, rel ~ period + ea, contrib = FALSE)
index <- aggregate(epr, pias)

res <- c(
  1.41421356237309,
  3.46410161513775,
  5,
  NaN,
  6.48074069840786,
  NaN,
  NaN,
  8
)

# Getter methods.
local({
  expect_identical(levels(epr), as.character(11:14))
  expect_identical(time(epr), as.character(1:2))
  expect_identical(start(epr), "1")
  expect_identical(end(epr), "2")
})

# Setter methods.
local({
  epr3 <- epr
  levels(epr3) <- 1:4
  expect_identical(levels(epr3), as.character(1:4))
  levels(epr3) <- levels(epr)
  expect_identical(epr, epr3)

  time(epr3) <- 2:1
  expect_identical(time(epr3), as.character(2:1))
  time(epr3) <- time(epr)
  expect_identical(epr, epr3)

  expect_error(levels(epr3) <- NULL)
  expect_error(time(epr3) <- NULL)
  expect_error(levels(epr3) <- c(1, 2, 3, 1))
  expect_error(levels(index) <- c(1, 2, 3, 1, 1))
  expect_error(time(epr3) <- c(1, 1))
})

# Group generics.
local({
  m <- as.matrix(epr)
  expect_equal(1 > epr, 1 > m)
  expect_equal(epr > 1, epr > 1)
  expect_error(epr + 1)
  expect_error(sum(epr))
  expect_error(log(epr))
})

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")
agg <- aggregation_structure(list(x1 = x1, x2 = x2, x3 = x3))

expect_identical(agg, as_aggregation_structure(data.frame(x1, x2, x3, 1)))
expect_identical(agg, as_aggregation_structure(cbind(x1, x2, x3, 1)))
expect_identical(agg, as_aggregation_structure(agg))

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")
agg <- aggregation_structure(list(x1 = x1, x2 = x2, x3 = x3))

test_that("as_aggregation_structure methods work", {
  expect_identical(agg, as_aggregation_structure(data.frame(x1, x2, x3, 1)))
  expect_identical(agg, as_aggregation_structure(cbind(x1, x2, x3, 1)))
  expect_identical(agg, as_aggregation_structure(agg))
})

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")
agg <- aggregation_structure(list(x1, x2, x3))

test_that("levels method works", {
  expect_identical(
    levels(agg),
    list(
      level1 = c("1", "2"),
      level2 = c("11", "21", "12"),
      ea = c("111", "211", "121", "112")
    )
  )
  expect_error(levels(agg) <- 1:9)

  expect_identical(levels(agg), lapply(weights(agg, ea_only = FALSE), names))
})

test_that("weights method works", {
  expect_equal(
    weights(agg, ea_only = FALSE),
    list(
      level1 = c("1" = 3, "2" = 1),
      level2 = c("11" = 2, "21" = 1, "12" = 1),
      ea = c("111" = 1, "211" = 1, "121" = 1, "112" = 1)
    )
  )
  expect_equal(
    weights(agg, ea_only = TRUE),
    c("111" = 1, "211" = 1, "121" = 1, "112" = 1)
  )
  expect_equal(
    weights(aggregation_structure(list(x1, x2, x3), 1:4), ea_only = FALSE),
    list(
      level1 = c("1" = 8, "2" = 2),
      level2 = c("11" = 5, "21" = 2, "12" = 3),
      ea = c("111" = 1, "211" = 2, "121" = 3, "112" = 4)
    )
  )

  expect_equal(
    weights(as_aggregation_structure(1:5, weights = 2), ea_only = FALSE),
    list(
      level1 = c("1" = 2),
      level2 = c("2" = 2),
      level3 = c("3" = 2),
      level4 = c("4" = 2),
      ea = c("5" = 2)
    )
  )
  expect_equal(
    weights(as_aggregation_structure(1:5, weights = 2), ea_only = TRUE),
    c("5" = 2)
  )
  expect_equal(
    weights(
      as_aggregation_structure(list(1:5), weights = 1:5),
      ea_only = FALSE
    ),
    list(ea = c("1" = 1, "2" = 2, "3" = 3, "4" = 4, "5" = 5))
  )
  expect_equal(
    weights(as_aggregation_structure(list(1:5), weights = 1:5), ea_only = TRUE),
    c("1" = 1, "2" = 2, "3" = 3, "4" = 4, "5" = 5)
  )
})

test_that("NAs move up the aggregation structure", {
  expect_equal(
    weights(
      suppressWarnings(aggregation_structure(list(x1, x2, x3), c(NA, 2:4))),
      ea_only = FALSE
    ),
    list(
      level1 = c("1" = NA, "2" = 2),
      level2 = c("11" = NA, "21" = 2, "12" = 3),
      ea = c("111" = NA, "211" = 2, "121" = 3, "112" = 4)
    )
  )

  expect_equal(
    weights(
      suppressWarnings(aggregation_structure(list(x1, x2, x3), c(NA, 2:4))),
      ea_only = FALSE,
      na.rm = TRUE
    ),
    list(
      level1 = c("1" = 7, "2" = 2),
      level2 = c("11" = 4, "21" = 2, "12" = 3),
      ea = c("111" = NA, "211" = 2, "121" = 3, "112" = 4)
    )
  )
})

test_that("replacement method works", {
  weights(agg)[c(1, 3, 4)] <- 10
  expect_equal(weights(agg), c("111" = 10, "211" = 1, "121" = 10, "112" = 10))
})

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")
agg <- aggregation_structure(list(level1 = x1, level2 = x2, ea = x3))

test_that("coercion methods works", {
  expect_equal(
    as.matrix(agg),
    matrix(
      c(2, 0, 3, 0, 0, 0, 6, 0, 6, 0, 2, 0, 0, 0, 6, 2, 0, 3, 0, 0) / 6,
      5,
      4,
      dimnames = list(
        levels = c("1", "2", "11", "21", "12"),
        c("111", "211", "121", "112")
      )
    )
  )
  expect_equal(as.matrix(agg), as.matrix(as.matrix(agg, sparse = TRUE)))
  expect_equal(
    as.data.frame(agg, row.names = 4:1),
    data.frame(
      level1 = c("1", "2", "1", "1"),
      level2 = c("11", "21", "12", "11"),
      ea = c("111", "211", "121", "112"),
      weight = c(1, 1, 1, 1),
      row.names = 4:1
    )
  )
  expect_equal(agg, as_aggregation_structure(as.data.frame(agg)))
  expect_equal(as.list(agg), list(level1 = x1, level2 = x2, ea = x3))

  agg2 <- aggregation_structure(1:3, 2)
  expect_equal(
    as.matrix(agg2),
    matrix(c(1, 1), 2, 1, dimnames = list(levels = c("1", "2"), "3"))
  )
  expect_equal(as.matrix(agg2), as.matrix(as.matrix(agg2, sparse = TRUE)))
  expect_equal(
    as.data.frame(agg2),
    data.frame(level1 = "1", level2 = "2", ea = "3", weight = 2)
  )
  expect_equal(as.list(agg2), list(level1 = "1", level2 = "2", ea = "3"))

  agg3 <- aggregation_structure(list(letters[c(2, 1, 3)]), 1:3)
  expect_equal(
    as.matrix(agg3),
    matrix(numeric(0), 0, 3, dimnames = list(character(0), c("b", "a", "c")))
  )
  expect_equal(as.matrix(agg3), as.matrix(as.matrix(agg3, sparse = TRUE)))
  expect_equal(
    as.data.frame(agg3),
    data.frame(ea = c("b", "a", "c"), weight = 1:3)
  )
  expect_equal(as.list(agg3), list(ea = letters[c(2, 1, 3)]))
})

dat <- data.frame(
  rel = c(1:6, NA, 7, 8),
  period = c(1, 1, 1, 1, 1, 2, 2, 2, 2),
  ea = c("11", "11", "12", "12", "13", "11", "12", "11", "14")
)

pias <- as_aggregation_structure(
  data.frame(level1 = 1, level2 = c(11, 12, 13, 14), weight = 1)
)

epr <- elementary_index(dat, rel ~ period + ea, contrib = TRUE)
epr2 <- elementary_index(dat, rel ~ period + ea, contrib = FALSE)
index <- aggregate(epr, pias)

res <- c(
  1.41421356237309,
  3.46410161513775,
  5,
  NaN,
  6.48074069840786,
  NaN,
  NaN,
  8
)

test_that("getter methods work", {
  expect_identical(levels(epr), as.character(11:14))
  expect_identical(time(epr), as.character(1:2))
  expect_identical(start(epr), "1")
  expect_identical(end(epr), "2")
})

test_that("setter methods work", {
  epr3 <- epr
  levels(epr3) <- 1:4
  expect_identical(levels(epr3), as.character(1:4))
  levels(epr3) <- levels(epr)
  expect_identical(epr, epr3)

  time(epr3) <- 2:1
  expect_identical(time(epr3), as.character(2:1))
  time(epr3) <- time(epr)
  expect_identical(epr, epr3)

  expect_error(levels(epr3) <- NULL)
  expect_error(time(epr3) <- NULL)
  expect_error(levels(epr3) <- c(1, 2, 3, 1))
  expect_error(levels(index) <- c(1, 2, 3, 1, 1))
  expect_error(time(epr3) <- c(1, 1))
})

test_that("group generics work", {
  m <- as.matrix(epr)
  expect_equal(1 > epr, 1 > m)
  expect_equal(epr > 1, epr > 1)
  expect_error(epr + 1)
  expect_error(sum(epr))
  expect_error(log(epr))
})

#' Make a price index aggregation structure
#'
#' Create a price index aggregation structure from a hierarchical
#' classification and aggregation weights that can be used to aggregate
#' elementary indexes.
#'
#' @aliases piar_aggregation_structure
#' @export
#'
#' @param x `[list]` A list of character vectors that give the codes/labels for
#'   each
#'   level of the classification, ordered so that moving down the list goes down
#'   the hierarchy. The last vector gives the elementary aggregates, which
#'   should have no duplicates. All vectors should be the same length, without
#'   `NA`s, and there should be no duplicates across different levels of
#'   `x`. Names for `x` are used as level names; otherwise, levels are named
#'   'level1', 'level2', ..., 'ea'.
#' @param weights `[numeric >= 0]` A numeric vector of aggregation weights for
#'   the elementary
#'   aggregates (i.e., the last vector in `x`), or something that can be coerced
#'   into one. The default is to give each elementary aggregate the same weight.
#'
#' @returns
#' A price index aggregation structure of class `piar_aggregation_structure`.
#' This is a list-S3 class with the following components.
#'
#' \item{child}{A nested list that gives the positions of the immediate
#' children for each node in each level of the aggregation structure above the
#' terminal nodes.}
#' \item{parent}{A list that gives the position of the
#' immediate parent for each node of the aggregation structure below the
#' initial nodes.}
#' \item{levels}{A named list of character vectors that give the levels of `x`.}
#' \item{weights}{A vector giving the weight for each elementary
#' aggregate.}
#'
#' @section Warning: The `aggregation_structure()` function does its best
#' to check its arguments, but there should be no expectation that the result
#' of `aggregation_structure()` will make any sense if `x` does not
#' represent a nested hierarchy.
#'
#' @seealso
#' [`aggregate()`][aggregate.piar_index] to aggregate price indexes made
#' with [`elementary_index()`][elementary_index].
#'
#' [expand_classification()] to make `x` from a character
#' representation of a hierarchical aggregation structure.
#'
#' [as_aggregation_structure()] to coerce tabular data into an
#' aggregation structure.
#'
#' [`as.data.frame()`][as.data.frame.piar_aggregation_structure] and
#' [`as.matrix()`][as.matrix.piar_aggregation_structure] to coerce an
#' aggregation structure into a tabular form.
#'
#' [`weights()`][weights.piar_aggregation_structure] to get the
#' weights for an aggregation structure.
#'
#' [`update()`][update.piar_aggregation_structure] for updating a
#' price index aggregation structure with an aggregated index.
#'
#' @examples
#' # A simple aggregation structure.
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #  |---+---|       |
#' #  111     112     121
#' #  (1)     (3)     (4)
#' aggregation_weights <- data.frame(
#'   level1 = c("1", "1", "1"),
#'   level2 = c("11", "11", "12"),
#'   ea     = c("111", "112", "121"),
#'   weight = c(1, 3, 4)
#' )
#'
#' aggregation_structure(
#'   aggregation_weights[1:3],
#'   weights = aggregation_weights[[4]]
#' )
#'
#' # The aggregation structure can also be made by expanding the
#' # elementary aggregates.
#' with(
#'   aggregation_weights,
#'   aggregation_structure(expand_classification(ea), weight)
#' )
aggregation_structure <- function(x, weights = NULL) {
  x <- lapply(x, \(z) factor(z, unique(z)))
  len <- length(x)
  ea <- as.character(unlist(x[len], use.names = FALSE))
  if (length(ea) == 0L) {
    stop("cannot make an aggregation structure with no elementary aggregates")
  }
  if (anyNA(x, recursive = TRUE)) {
    stop("`x` cannot contain NAs")
  }

  if (is.null(weights)) {
    weights <- rep.int(1, length(ea))
  } else {
    weights <- as.numeric(weights)
    if (any(missing_weights(weights))) {
      warning("some elements of `weights` are NA or zero")
    }
  }

  # Basic argument checking to make sure inputs can make an
  # aggregation structure.
  if (any(lengths(x) != length(weights))) {
    stop("all arguments must be the same length")
  }
  if (anyDuplicated(ea)) {
    stop(
      "there are duplicated elementary aggregates; the last vector in `x` ",
      "should not have duplicates"
    )
  }
  if (anyDuplicated(unlist(lapply(x, unique), use.names = FALSE))) {
    stop(
      "there are duplicated nodes in the aggregation structure; the same ",
      "value cannot appear across multiple levels of `x`"
    )
  }
  upper <- x[-len] # nodes above eas
  lower <- x[-1L] # nodes below initial nodes
  child <- parent <- vector("list", length(upper))
  # Produce a list for each level with all the parent and child nodes.
  for (i in seq_along(upper)) {
    child[[i]] <- lapply(
      split(as.integer(lower[[len - i]]), upper[[len - i]]),
      unique
    )
    parent[[i]] <- lapply(
      split(as.integer(upper[[len - i]]), lower[[len - i]]),
      unique
    )
  }
  if (any(lengths(unlist(parent, recursive = FALSE)) > 1L)) {
    stop(
      "some nodes in the price index aggregation structure have ",
      "multiple parent nodes; the aggregation structure does not ",
      "represent a nested hierarchy"
    )
  }
  parent <- lapply(parent, unlist)
  levels <- lapply(x, levels)
  if (is.null(names(levels))) {
    names(levels) <- c(paste0("level", seq_along(child), recycle0 = TRUE), "ea")
  }
  piar_aggregation_structure(child, parent, levels, weights)
}

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")

# Factors do nothing.
local({
  agg1 <- aggregation_structure(list(x1, x2, x3), 1:4)
  f <- factor(x2, levels = c("11", "21", "12"))
  agg2 <- aggregation_structure(list(x1, f, x3), 1:4)
  expect_identical(agg1, agg2)

  f <- factor(x2, levels = c("11", "21", "12", "99"))
  expect_identical(agg2, aggregation_structure(list(x1, f, x3), 1:4))

  f <- factor(replace(x2, 2, NA), levels = c("11", "21", "12"))
  expect_error(aggregation_structure(list(x1, addNA(f), x3), 1:4))
})

# Order is preserved.
local({
  agg <- aggregation_structure(list(x1, x2, x3))
  expect_identical(
    unlist(levels(agg), use.names = FALSE),
    unique(c(x1, x2, x3))
  )

  ord <- c(2, 4, 1, 3)
  agg <- aggregation_structure(list(x1[ord], x2[ord], x3[ord]))
  expect_identical(
    unlist(levels(agg), use.names = FALSE),
    unique(c(x1[ord], x2[ord], x3[ord]))
  )
})

# Names are not required.
local({
  x <- list(a = as.character(1:3), as.character(4:6), c = as.character(7:9))
  agg <- aggregation_structure(x)
  expect_identical(as.list(agg), x)
  expect_identical(
    weights(agg, ea_only = FALSE),
    list(
      a = c("1" = 1, "2" = 1, "3" = 1),
      c("4" = 1, "5" = 1, "6" = 1),
      c = c("7" = 1, "8" = 1, "9" = 1)
    )
  )
  expect_identical(
    as.data.frame(agg),
    as.data.frame(c(x, weight = list(c(1, 1, 1))))
  )
})

# Errors.
local({
  expect_error(aggregation_structure(list()))
  expect_error(aggregation_structure(list(x1, x2, x3), 1:3))
  expect_error(aggregation_structure(list(x1, x2, x3, x3)))
  expect_error(aggregation_structure(list(x1, x2)))
  expect_error(aggregation_structure(list(1:2, c(11, NA))))
  expect_error(aggregation_structure(list(1:2, c(11, 11, 12))))
  expect_error(aggregation_structure(list(1:2, c(3, 3), c(4, 5))))
  expect_error(aggregation_structure(list(1:2, c(3, 3), c(4, ""))))
})

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")
agg <- aggregation_structure(list(x1, x2, x3))

# Updating with a length-0 index makes the weights NA.
local({
  epr <- elementary_index(integer(0), period = gl(1, 0), ea = gl(1, 0))
  index <- aggregate(epr, agg)
  expect_equal(update(agg, index)[-4], agg[-4])

  expect_equal(
    weights(update(agg, index), ea_only = FALSE),
    list(
      level1 = c("1" = NA_real_, "2" = NA),
      level2 = c("11" = NA_real_, "21" = NA, "12" = NA),
      ea = c("111" = NA_real_, "211" = NA, "121" = NA, "112" = NA)
    )
  )
})

# An index that doesn't line up with the pias introduces NA".
local({
  epr <- elementary_index(1:2, ea = c("111", "121"), period = gl(1, 2))
  index <- aggregate(epr, agg)
  expect_equal(
    weights(update(agg, index), ea_only = FALSE),
    list(
      level1 = c("1" = NA_real_, "2" = NA),
      level2 = c("11" = NA_real_, "21" = NA, "12" = 2),
      ea = c("111" = 1, "211" = NA, "121" = 2, "112" = NA)
    )
  )
  expect_equal(
    weights(update(agg, index, order = 0), ea_only = FALSE),
    list(
      level1 = c("1" = NA_real_, "2" = NA),
      level2 = c("11" = NA_real_, "21" = NA, "12" = 1),
      ea = c("111" = 1, "211" = NA, "121" = 1, "112" = NA)
    )
  )
})

# Updating works with only EAs.
local({
  # This used to not work
  epr <- elementary_index(1:2, ea = c("111", "121"), period = gl(1, 2))
  pias <- aggregation_structure(111)
  index <- aggregate(epr, pias)
  expect_equal(update(pias, index), pias)
})

pias <- aggregation_structure(
  list(
    c(1, 1, 1, 2, 2, 1),
    c(11, 11, 12, 21, 22, 12),
    c(112, 111, 121, 211, 221, 121),
    c(1121, 1111, 1211, 2111, 2211, 1212),
    letters[1:6]
  )
)
# Cutting a pias work with upper == TRUE.
local({
  expect_equal(cut(pias, 5), pias)

  expect_identical(
    cut(pias, 4),
    aggregation_structure(
      list(
        level1 = c(1, 1, 1, 2, 2, 1),
        level2 = c(11, 11, 12, 21, 22, 12),
        level3 = c(112, 111, 121, 211, 221, 121),
        level4 = c(1121, 1111, 1211, 2111, 2211, 1212)
      )
    )
  )

  expect_identical(
    cut(pias, 3),
    aggregation_structure(
      list(
        level1 = c(1, 1, 1, 2, 2),
        level2 = c(11, 11, 12, 21, 22),
        level3 = c(112, 111, 121, 211, 221)
      ),
      weights = c(1, 1, 2, 1, 1)
    )
  )

  expect_identical(
    cut(pias, 2),
    aggregation_structure(
      list(
        level1 = c(1, 1, 2, 2),
        level2 = c(11, 12, 21, 22)
      ),
      weights = c(2, 2, 1, 1)
    )
  )

  expect_identical(
    cut(pias, 1),
    aggregation_structure(
      list(
        level1 = c(1, 2)
      ),
      weights = c(4, 2)
    )
  )
})
# Cutting a pias work with upper == FALSE.
local({
  expect_equal(cut(pias, 1, upper = FALSE), pias)

  expect_identical(
    cut(pias, 2, upper = FALSE),
    aggregation_structure(
      list(
        level2 = c(11, 11, 12, 21, 22, 12),
        level3 = c(112, 111, 121, 211, 221, 121),
        level4 = c(1121, 1111, 1211, 2111, 2211, 1212),
        ea = letters[1:6]
      )
    )
  )

  expect_identical(
    cut(pias, 3, upper = FALSE),
    aggregation_structure(
      list(
        level3 = c(112, 111, 121, 211, 221, 121),
        level4 = c(1121, 1111, 1211, 2111, 2211, 1212),
        ea = letters[1:6]
      )
    )
  )

  expect_identical(
    cut(pias, 4, upper = FALSE),
    aggregation_structure(
      list(
        level4 = c(1121, 1111, 1211, 2111, 2211, 1212),
        ea = letters[1:6]
      )
    )
  )

  expect_identical(
    cut(pias, 5, upper = FALSE),
    aggregation_structure(
      list(
        ea = letters[1:6]
      )
    )
  )
})
# Errors work.
local({
  expect_error(cut(pias, 0))
  expect_error(cut(pias, 6))
})
# Length-1 pias works.
local({
  expect_identical(cut(cut(pias, 1), 1, upper = FALSE), cut(pias, 1))
  expect_identical(cut(cut(pias, 1), 1), cut(pias, 1))
})

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")

test_that("factors do nothing", {
  agg1 <- aggregation_structure(list(x1, x2, x3), 1:4)
  f <- factor(x2, levels = c("11", "21", "12"))
  agg2 <- aggregation_structure(list(x1, f, x3), 1:4)
  expect_identical(agg1, agg2)

  f <- factor(x2, levels = c("11", "21", "12", "99"))
  expect_identical(agg2, aggregation_structure(list(x1, f, x3), 1:4))

  f <- factor(replace(x2, 2, NA), levels = c("11", "21", "12"))
  expect_error(aggregation_structure(list(x1, addNA(f), x3), 1:4))
})

test_that("order is preserved", {
  agg <- aggregation_structure(list(x1, x2, x3))
  expect_identical(
    unlist(levels(agg), use.names = FALSE),
    unique(c(x1, x2, x3))
  )

  ord <- c(2, 4, 1, 3)
  agg <- aggregation_structure(list(x1[ord], x2[ord], x3[ord]))
  expect_identical(
    unlist(levels(agg), use.names = FALSE),
    unique(c(x1[ord], x2[ord], x3[ord]))
  )
})

test_that("names are not required", {
  x <- list(a = as.character(1:3), as.character(4:6), c = as.character(7:9))
  agg <- aggregation_structure(x)
  expect_identical(as.list(agg), x)
  expect_identical(
    weights(agg, ea_only = FALSE),
    list(
      a = c("1" = 1, "2" = 1, "3" = 1),
      c("4" = 1, "5" = 1, "6" = 1),
      c = c("7" = 1, "8" = 1, "9" = 1)
    )
  )
  expect_identical(
    as.data.frame(agg),
    as.data.frame(c(x, weight = list(c(1, 1, 1))))
  )
})

test_that("errors work", {
  expect_error(aggregation_structure(list()))
  expect_error(aggregation_structure(list(x1, x2, x3), 1:3))
  expect_error(aggregation_structure(list(x1, x2, x3, x3)))
  expect_error(aggregation_structure(list(x1, x2)))
  expect_error(aggregation_structure(list(1:2, c(11, NA))))
  expect_error(aggregation_structure(list(1:2, c(11, 11, 12))))
  expect_error(aggregation_structure(list(1:2, c(3, 3), c(4, 5))))
  expect_error(aggregation_structure(list(1:2, c(3, 3), c(4, ""))))
})

pias <- aggregation_structure(
  list(
    c(1, 1, 1, 2, 2, 1),
    c(11, 11, 12, 21, 22, 12),
    c(112, 111, 121, 211, 221, 121),
    c(1121, 1111, 1211, 2111, 2211, 1212),
    letters[1:6]
  )
)

test_that("cutting a pias work with upper == TRUE", {
  expect_equal(cut(pias, 5), pias)

  expect_identical(
    cut(pias, 4),
    aggregation_structure(
      list(
        level1 = c(1, 1, 1, 2, 2, 1),
        level2 = c(11, 11, 12, 21, 22, 12),
        level3 = c(112, 111, 121, 211, 221, 121),
        level4 = c(1121, 1111, 1211, 2111, 2211, 1212)
      )
    )
  )

  expect_identical(
    cut(pias, 3),
    aggregation_structure(
      list(
        level1 = c(1, 1, 1, 2, 2),
        level2 = c(11, 11, 12, 21, 22),
        level3 = c(112, 111, 121, 211, 221)
      ),
      weights = c(1, 1, 2, 1, 1)
    )
  )

  expect_identical(
    cut(pias, 2),
    aggregation_structure(
      list(
        level1 = c(1, 1, 2, 2),
        level2 = c(11, 12, 21, 22)
      ),
      weights = c(2, 2, 1, 1)
    )
  )

  expect_identical(
    cut(pias, 1),
    aggregation_structure(
      list(
        level1 = c(1, 2)
      ),
      weights = c(4, 2)
    )
  )
})

test_that("cutting a pias work with upper == FALSE", {
  expect_equal(cut(pias, 1, upper = FALSE), pias)

  expect_identical(
    cut(pias, 2, upper = FALSE),
    aggregation_structure(
      list(
        level2 = c(11, 11, 12, 21, 22, 12),
        level3 = c(112, 111, 121, 211, 221, 121),
        level4 = c(1121, 1111, 1211, 2111, 2211, 1212),
        ea = letters[1:6]
      )
    )
  )

  expect_identical(
    cut(pias, 3, upper = FALSE),
    aggregation_structure(
      list(
        level3 = c(112, 111, 121, 211, 221, 121),
        level4 = c(1121, 1111, 1211, 2111, 2211, 1212),
        ea = letters[1:6]
      )
    )
  )

  expect_identical(
    cut(pias, 4, upper = FALSE),
    aggregation_structure(
      list(
        level4 = c(1121, 1111, 1211, 2111, 2211, 1212),
        ea = letters[1:6]
      )
    )
  )

  expect_identical(
    cut(pias, 5, upper = FALSE),
    aggregation_structure(
      list(
        ea = letters[1:6]
      )
    )
  )
})

test_that("errors work", {
  expect_error(cut(pias, 0))
  expect_error(cut(pias, 6))
})

test_that("length-1 pias works", {
  expect_identical(cut(cut(pias, 1), 1, upper = FALSE), cut(pias, 1))
  expect_identical(cut(cut(pias, 1), 1), cut(pias, 1))
})

x1 <- c("1", "2", "1", "1")
x2 <- c("11", "21", "12", "11")
x3 <- c("111", "211", "121", "112")
agg <- aggregation_structure(list(x1, x2, x3))

test_that("updating with a length-0 index makes the weights NA", {
  epr <- elementary_index(integer(0), period = gl(1, 0), ea = gl(1, 0))
  index <- aggregate(epr, agg)
  expect_equal(update(agg, index)[-4], agg[-4])

  expect_equal(
    weights(update(agg, index), ea_only = FALSE),
    list(
      level1 = c("1" = NA_real_, "2" = NA),
      level2 = c("11" = NA_real_, "21" = NA, "12" = NA),
      ea = c("111" = NA_real_, "211" = NA, "121" = NA, "112" = NA)
    )
  )
})

test_that("an index that doesn't line up with the pias introduces NA", {
  epr <- elementary_index(1:2, ea = c("111", "121"), period = gl(1, 2))
  index <- aggregate(epr, agg)
  expect_equal(
    weights(update(agg, index), ea_only = FALSE),
    list(
      level1 = c("1" = NA_real_, "2" = NA),
      level2 = c("11" = NA_real_, "21" = NA, "12" = 2),
      ea = c("111" = 1, "211" = NA, "121" = 2, "112" = NA)
    )
  )
  expect_equal(
    weights(update(agg, index, r = 0), ea_only = FALSE),
    list(
      level1 = c("1" = NA_real_, "2" = NA),
      level2 = c("11" = NA_real_, "21" = NA, "12" = 1),
      ea = c("111" = 1, "211" = NA, "121" = 1, "112" = NA)
    )
  )
})

test_that("updating works with only EAs", {
  # This used to not work
  epr <- elementary_index(1:2, ea = c("111", "121"), period = gl(1, 2))
  pias <- aggregation_structure(111)
  index <- aggregate(epr, pias)
  expect_equal(update(pias, index), pias)
})

#' Calculate period-over-period price relatives
#'
#' Construct period-over-period price relatives from information on prices and
#' products over time.
#'
#' @export
#'
#' @param x `[object]` Either a numeric vector (or something that can be coerced
#'   into one) or data frame of prices.
#' @param period `[factor]` A factor, or something that can be coerced into one,
#'   that gives the corresponding time period for each element in `x`. The
#'   ordering of time periods follows the levels of `period` to agree with
#'   [`cut()`][cut.Date].
#' @param product `[factor]` A factor, or something that can be coerced into
#'   one, that
#'   gives the corresponding product identifier for each element in `x`.
#' @param ... Further arguments passed to or used by methods.
#' @param formula `[formula]` A two-sided formula, or something that can be
#'   coerced into one, with prices on the left-hand
#'   side and time periods and products (in that order) on the
#'   right-hand side.
#'
#' @returns
#' A numeric vector of price relatives, with `product` as names.
#'
#' @seealso
#' [back_period()] to get only the back price or base price.
#'
#' [impute_prices()] to impute missing prices.
#'
#' [outliers()] for methods to identify outliers with price relatives.
#'
#' @examples
#' price_relative(
#'   1:6,
#'   period = rep(1:2, each = 3),
#'   product = rep(letters[1:3], 2)
#' )
price_relative <- function(x, ...) {
  UseMethod("price_relative")
}

#' @rdname price_relative
#' @export
price_relative.default <- function(x, period, product, ...) {
  chkDots(...)
  x <- as.numeric(x)
  period <- as.factor(period)
  product <- as.factor(product)

  if (different_length(x, period, product)) {
    stop("`x`, `period`, and `product` must be the same length")
  }

  res <- x / x[back_period(period, product)]
  names(res) <- as.character(product)
  res
}

#' @rdname price_relative
#' @export
price_relative.data.frame <- function(x, formula, ...) {
  chkDots(...)
  vars <- formula_vars(formula, x)
  price_relative(vars[[1L]], period = vars[[2L]], product = vars[[3L]])
}

ms_prices$rel <- price_relative(ms_prices, price ~ period + product)
ms_prices$w1 <- 1:40
ms_prices$w2 <- 40:1

epr1 <- elementary_index(ms_prices, rel ~ period + business, contrib = TRUE)

epr2 <- elementary_index(
  ms_prices,
  rel ~ period + business,
  order = -1,
  contrib = TRUE,
  na.rm = TRUE
)

# Test a Fisher calculation
fw <- function(x, w1, w2) {
  v1 <- transmute_weights(x, w1, 1, 0)
  v2 <- transmute_weights(x, w2, -1, 0)
  v1 + v2
}

ms_prices2 <- na.omit(ms_prices)
f <- interaction(ms_prices2$period, ms_prices2$business)
w <- unsplit(lapply(split(ms_prices2, f), \(df) fw(df$rel, df$w1, df$w2)), f)

epr3 <- elementary_index(
  ms_prices2,
  rel ~ period + business,
  weights = w,
  contrib = TRUE
)

# Agrees with alternate implementations.
local({
  epr11 <- aggregate(
    rel ~ as.character(business) + period,
    ms_prices,
    function(x) exp(weighted.mean(log(x))),
    na.action = na.pass,
    drop = FALSE
  )
  epr22 <- aggregate(
    rel ~ as.character(business) + period,
    ms_prices,
    function(x) 1 / weighted.mean(1 / x),
    na.action = na.omit,
    drop = FALSE
  )
  names(epr11) <- names(epr22) <- c("level", "period", "value")

  expect_equal(as.data.frame(epr1), epr11[c(2, 1, 3)])
  expect_equal(as.data.frame(epr2), epr22[c(2, 1, 3)])
})

# Fisher calculation agrees with manual calculation.
local({
  l <- elementary_index(
    ms_prices2,
    rel ~ period + business,
    weights = w1,
    order = 1
  )
  p <- elementary_index(
    ms_prices2,
    rel ~ period + business,
    weights = w2,
    order = -1
  )
  expect_equal(sqrt(as.matrix(l) * as.matrix(p)), as.matrix(epr3))

  # Should work for other kinds of superlative indexes
  fw <- function(x, w1, w2) {
    v1 <- transmute_weights(x, order = 1.5, to = 0)
    v2 <- transmute_weights(x, w2, order = -1.5, to = 0)
    v1 + v2
  }

  f <- interaction(ms_prices2$period, ms_prices2$business)
  w <- unsplit(lapply(split(ms_prices2, f), \(df) fw(df$rel, df$w1, df$w2)), f)

  sepr <- elementary_index(ms_prices2, rel ~ period + business, weights = w)

  l <- elementary_index(ms_prices2, rel ~ period + business, order = 1.5)
  p <- elementary_index(
    ms_prices2,
    rel ~ period + business,
    weights = w2,
    order = -1.5
  )
  expect_equal(sqrt(as.matrix(l) * as.matrix(p)), as.matrix(sepr))
})

# Contributions add up.
local({
  sum_contrib <- \(x, ...) sum(x, ...) / (sum(!is.na(x)) > 0)
  expect_equal(
    as.numeric(epr1$index),
    sapply(epr1$contrib, sum_contrib) + 1
  )

  expect_equal(
    as.numeric(epr2$index),
    sapply(epr2$contrib, sum_contrib, na.rm = TRUE) + 1
  )

  expect_equal(
    as.numeric(epr3$index),
    sapply(epr3$contrib, sum_contrib) + 1
  )
})

# Argument checking.
local({
  expect_error(elementary_index(1:3, period = 1:2, ea = 1:3))
  expect_error(elementary_index(1:3, period = 1:3, ea = 1:4))
  expect_error(elementary_index(1:3, period = 1:3, ea = 1:3, weights = 1:2))
  expect_error(elementary_index(1:3, period = factor(1:3, levels = numeric(0))))
  expect_error(elementary_index(1:3, ea = factor(1:3, levels = numeric(0))))
  expect_error(elementary_index(setNames(1:3, c("", 1, 2)), contrib = TRUE))
  expect_error(elementary_index(-1:1, period = 1:3, ea = 1:3, order = 1))
  expect_warning(elementary_index(
    setNames(1:3, rep(1, 3)),
    period = gl(1, 3),
    ea = gl(1, 3),
    contrib = TRUE
  ))
})

# Non-standard evaluation works.
local({
  expect_equal(
    elementary_index(ms_prices, rel ~ period:business, contrib = TRUE),
    epr1
  )
  expect_equal(
    elementary_index(ms_prices[c(1:2, 5)], rel ~ ., contrib = TRUE),
    epr1
  )
  expect_error(elementary_index(ms_prices, rel ~ .))
  expect_error(elementary_index(ms_prices, ~ period + business))
})

#' Outlier detection for price relatives
#'
#' Standard cutoff-based methods for detecting outliers with price relatives.
#'
#' This function constructs an interval of the form \eqn{[b_l(x) -
#' c_l \times l(x), b_u(x) + c_u \times u(x)]}{[bl(x) - cl * l(x), bu(x) + cu *
#' u(x)]} and assigns a value in `x` as `TRUE` if that value does not
#' belong to the interval, `FALSE` otherwise. The different methods differ in
#' how they construct the values \eqn{b_l(x)}{bl(x)}, \eqn{b_u(x)}{bu(x)},
#' \eqn{l(x)}, and \eqn{u(x)}. Any missing values in `x` are ignored when
#' calculating the interval, but will return `NA`.
#'
#' The quartile method and Tukey algorithm are described in paragraphs 5.113 to
#' 5.135 of the CPI manual (2020).
#' The resistant fences method is an alternative to the quartile method, and is
#' described by Rais (2008) and Hutton (2008). The Kimber method is yet another
#' alternative. Quantile-based methods often
#' identify price relatives as outliers because the distribution is
#' concentrated around 1; setting `scale > 0` puts a floor on the minimum
#' dispersion between quantiles as a fraction of the median. See the references
#' for more details.
# nolint start
#'
#' |                  | \eqn{b_l(x)}{bl(x)} | \eqn{b_u(x)}{bu(x)} | \eqn{l(x)}            | \eqn{u(x)}            |
#' | ---              | --                  | ---                 | ---                   | ---                   |
#' | Quartile         | \eqn{Q_2(x)}        | \eqn{Q_2(x)}        | \eqn{Q_2(x) - Q_1(x)} | \eqn{Q_3(x) - Q_2(x)} |
#' | Resistant fences | \eqn{Q_1(x)}        | \eqn{Q_3(x)}        | \eqn{Q_3(x) - Q_1(x)} | \eqn{Q_3(x) - Q_1(x)} |
#' | Kimber           | \eqn{Q_1(x)}        | \eqn{Q_3(x)}        | \eqn{Q_2(x) - Q_1(x)} | \eqn{Q_3(x) - Q_2(x)} |
#'
# nolint end
#' The robust Z-score is the usual method to identify relatives in the
#' (asymmetric) tails of the distribution, simply replacing the mean with the
#' median, and the standard deviation with the median absolute deviation.
#'
#' These methods often assume that price relatives are symmetrically
#' distributed (if not Gaussian). As the distribution of price relatives often
#' has a long right tail, the natural logarithm can be used to transform price
#' relative before identifying outliers (sometimes under the assumption that
#' price relatives are distributed log-normal). The Hidiroglou-Berthelot
#' transformation is another approach, described in the CPI manual (par.
#' 5.124). (Sometimes the transformed price relatives are multiplied by
#' \eqn{\max(p_1, p_0)^u}{max(p1, p0)^u}, for some
#' \eqn{0 \le u \le 1}{0 <= u <= 1}, so that products with a larger price
#' get flagged as outliers (par. 5.128).)
#'
#' @export
#'
#' @param x `[numeric > 0]` A numeric vector, usually of price relatives.
#'   These can be made with, e.g., [back_period()].
#' @param upper,lower `[numeric(1) > 0]` A number giving the upper and lower
#'   cutoffs for
#'   each element of `x`. By default the lower cutoff is the same as the upper
#'   cutoff
#' @param method `[character(1)]` The outlier detection method, one of
#'   `"quartile"` (the default), `"resistant-fences"`, `"kimber"`, `"robust-z"`,
#'   or `"tukey"`.
#' @param scale `[0 <= numeric(1) <= 1]` A number between 0 and 1 giving the
#'   scale factor for the
#'   median to establish the minimum dispersion between quartiles for each
#'   element of `x`. The default does not set a minimum dispersion.
#' @param quantile_type `[integer(1)]` See [quantile()].
#'
#' @returns
#' A logical vector, the same length as `x`, that is `TRUE` if the
#' corresponding element of `x` is identified as an outlier,
#' `FALSE` otherwise.
#'
#' @seealso
#' [back_period()] for a simple utility function to turn prices
#' in a table into price relatives.
#'
#' The `HBmethod()` function in the \pkg{univOutl} package for the
#' Hidiroglou-Berthelot method for identifying outliers.
#'
#' @references
#' Hutton, H. (2008). Dynamic outlier detection in price index surveys.
#' *Proceedings of the Survey Methods Section: Statistical Society of Canada
#' Annual Meeting*.
#'
#' IMF, ILO, Eurostat, UNECE, OECD, and World Bank. (2020).
#' *Consumer Price Index Manual: Concepts and Methods*.
#' International Monetary Fund.
#'
#' Rais, S. (2008). Outlier detection for the Consumer Price Index.
#' *Proceedings of the Survey Methods Section: Statistical Society of Canada
#' Annual Meeting*.
#'
#' @examples
#' x <- c(1, 10, 15, 100)
#'
#' outliers(x, upper = 2.5, method = "quartile")
#'
#' # Always identifies fewer outliers than above.
#' outliers(x, upper = 2.5, method = "resistant-fences")
outliers <- function(
  x,
  upper,
  lower = upper,
  method = c(
    "quartile",
    "resistant-fences",
    "kimber",
    "robust-z",
    "tukey"
  ),
  scale = 0,
  quantile_type = 7
) {
  method <- match.arg(method)
  x <- as.numeric(x)
  upper <- as.numeric(upper)
  if (upper < 0) {
    stop("`upper` must be greater than 0")
  }
  lower <- as.numeric(lower)
  if (lower < 0) {
    stop("`lower` must be greater than 0")
  }
  scale <- as.numeric(scale)
  if (scale < 0 || scale > 1) {
    stop("`scale` must be between 0 and 1")
  }
  if (method %in% c("quartile", "resistant-fences", "kimber")) {
    q <- stats::quantile(
      x,
      c(0.25, 0.5, 0.75),
      names = FALSE,
      na.rm = TRUE,
      type = quantile_type
    )
    if (method == "quartile") {
      u <- q[2L] + upper * pmax.int(q[3L] - q[2L], abs(scale * q[2L]))
      l <- q[2L] - lower * pmax.int(q[2L] - q[1L], abs(scale * q[2L]))
    } else if (method == "resistant-fences") {
      iqr <- pmax.int(q[3L] - q[1L], abs(scale * q[2L]))
      u <- q[3L] + upper * iqr
      l <- q[1L] - lower * iqr
    } else {
      u <- q[3L] + upper * pmax.int(q[3L] - q[2L], abs(scale * q[2L]))
      l <- q[1L] - lower * pmax.int(q[2L] - q[1L], abs(scale * q[2L]))
    }
  } else if (method == "robust-z") {
    med <- stats::quantile(
      x,
      0.5,
      names = FALSE,
      na.rm = TRUE,
      type = quantile_type
    )
    s <- pmax.int(stats::mad(x, med, na.rm = TRUE), abs(scale * med))
    u <- med + upper * s
    l <- med - lower * s
  } else if (method == "tukey") {
    q <- stats::quantile(
      x,
      c(0.05, 0.95),
      names = FALSE,
      na.rm = TRUE,
      type = quantile_type
    )
    tail <- x < q[1L] | x > q[2L]
    ts <- x[x != 1 & !tail]
    if (length(ts) == 0L) {
      return(tail)
    }
    # In some versions m is the median.
    m <- mean(ts, na.rm = TRUE)
    u <- min(m + upper * (mean(ts[ts >= m], na.rm = TRUE) - m), q[2L])
    l <- max(m - lower * (m - mean(ts[ts <= m], na.rm = TRUE)), q[1L])
  }
  x > u | x < l
}

#' HB transform
#' @rdname outliers
#' @export
hb_transform <- function(x) {
  x <- as.numeric(x)
  med <- stats::median(x, na.rm = TRUE)
  res <- 1 - med / x
  gemed <- which(x >= med)
  res[gemed] <- x[gemed] / med - 1
  res
}

x <- c(1, 2, 1, 0.5, 1, 10, 1, 0.5, 0.2, 0.05)

# Outlier methods work.
local({
  expect_equal(
    outliers(x, upper = 2.5, method = "quartile"),
    x > median(x) + (quantile(x, 0.75) - quantile(x, 0.5)) * 2.5 |
      x < median(x) - (quantile(x, 0.5) - quantile(x, 0.25)) * 2.5
  )
  expect_equal(
    outliers(x, upper = 2.5, method = "resistant-fences"),
    x > quantile(x, 0.75) + (quantile(x, 0.75) - quantile(x, 0.25)) * 2.5 |
      x < quantile(x, 0.25) - (quantile(x, 0.75) - quantile(x, 0.25)) * 2.5
  )
  expect_equal(
    outliers(x, upper = 2.5, method = "kimber"),
    x > quantile(x, 0.75) + (quantile(x, 0.75) - quantile(x, 0.5)) * 2.5 |
      x < quantile(x, 0.25) - (quantile(x, 0.5) - quantile(x, 0.25)) * 2.5
  )
  expect_true(
    sum(outliers(x, upper = 2.5, method = "resistant-fences")) <=
      sum(outliers(x, upper = 2.5, method = "quartile"))
  )
  expect_equal(
    outliers(x, upper = 2.5, method = "robust-z"),
    abs(x - median(x)) / mad(x) > 2.5
  )

  expect_equal(outliers(integer(0), upper = 1, method = "tukey"), logical(0))
  expect_equal(outliers(2, upper = 2.5, method = "tukey"), FALSE)
  expect_equal(
    outliers(seq(0.1, 2, by = 0.2), upper = 2.5, method = "tukey"),
    c(TRUE, rep(FALSE, 8), TRUE)
  )
  expect_equal(
    outliers(c(NA, 1, 2, 3), upper = 2.5, method = "tukey"),
    c(NA, TRUE, FALSE, TRUE)
  )
})

# Outliers work with NAs.
local({
  expect_identical(
    outliers(x, upper = 1, method = "resistant-fences"),
    outliers(c(NA, x), upper = 1, method = "resistant-fences")[-1]
  )
  expect_identical(
    outliers(x, upper = 1, method = "quartile"),
    outliers(c(NA, x), upper = 1, method = "quartile")[-1]
  )
  expect_identical(
    outliers(x, upper = 1, method = "robust-z"),
    outliers(c(NA, x), upper = 1, method = "robust-z")[-1]
  )
  expect_identical(
    outliers(x, upper = 1, method = "tukey"),
    outliers(c(NA, x), upper = 1, method = "tukey")[-1]
  )
  expect_identical(
    outliers(x, upper = 1, method = "kimber"),
    outliers(c(NA, x), upper = 1, method = "kimber")[-1]
  )
})

# HB transform.
local({
  expect_equal(
    hb_transform(x),
    ifelse(x < median(x), 1 - median(x) / x, x / median(x) - 1)
  )
  expect_equal(hb_transform(x), hb_transform(c(NA, x)[-1]))
})

# Recycling gives an error.
local({
  expect_error(outliers(x, upper = 1, lower = rep(2.5, 10)))
  expect_error(outliers(x, upper = rep(2.5, 0)))
  expect_error(outliers(x, upper = 3, scale = rep(0, 11)))
})

#' Merge price indexes
#'
#' @description
#' Combine two price indexes with common time periods, merging together the
#' index values and percent-change contributions for each time period.
#'
#' This is useful for building up an index when different elementary aggregates
#' come from different sources of data, or use different index-number formulas.
#'
#' @name merge.piar_index
#' @aliases merge.piar_index
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param y `[piar_index]` A price index, or something that can coerced into
#'   one. If `x`
#'   is a period-over-period index then `y` is coerced into a chainable
#'   index; otherwise, `y` is coerced into a direct index.
#' @param ... Not currently used.
#'
#' @returns
#' A combined price index that inherits from the same class as `x`.
#'
#' @examples
#' index1 <- as_index(matrix(1:6, 2))
#'
#' index2 <- index1
#' levels(index2) <- 3:4
#'
#' merge(index1, index2)
merge.chainable_piar_index <- function(x, y, ...) {
  y <- as_index(y, chainable = TRUE)
  res <- NextMethod("merge")
  new_piar_index(res$index, res$contrib, res$levels, res$time, chainable = TRUE)
}

#' @rdname merge.piar_index
#' @export
merge.direct_piar_index <- function(x, y, ...) {
  y <- as_index(y, chainable = FALSE)
  res <- NextMethod("merge")
  new_piar_index(
    res$index,
    res$contrib,
    res$levels,
    res$time,
    chainable = FALSE
  )
}

#' @export
merge.piar_index <- function(x, y, ...) {
  chkDots(...)
  if (ntime(x) != ntime(y) || any(x$time != y$time)) {
    stop("`x` and `y` must be indexes for the same time periods")
  }
  if (any(x$levels %in% y$levels)) {
    stop("the same levels cannot appear in both `x` and `y`")
  }
  x$index <- rbind(x$index, y$index)
  if (is.null(x$contrib) && !is.null(y$contrib)) {
    x$contrib <- contrib_skeleton(x$levels, x$time)
  } else if (!is.null(x$contrib) && is.null(y$contrib)) {
    y$contrib <- contrib_skeleton(y$levels, y$time)
  }
  if (!is.null(x$contrib)) {
    x$contrib <- rbind(x$contrib, y$contrib)
  }
  x$levels <- c(x$levels, y$levels)
  # Alternatively call validate_piar_index(x), but this is not necessary.
  x
}

#' Coerce an index into a tabular form
#'
#' Turn an index into a data frame or a matrix.
#'
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param row.names `[character]` See [as.data.frame()].
#' @param stringsAsFactors `[logical(1)]` See [as.data.frame()].
#' @param optional,... Not currently used.
#' @param contrib `[logical(1)]` Include percent-change contributions
#'   (the default does not include them).
#'
#' @returns
#' `as.data.frame()` returns the index values in `x` as a data frame with three
#' columns: `period`, `level`, and `value`. If `contrib = TRUE` then there is
#' a fourth (list) column `contrib` containing percent-change contributions.
#'
#' `as.matrix()` returns the index values in `x` as a matrix with a row for
#' each level and a column for each time period in `x`.
#'
#' @seealso
#' [as_index()] to coerce a matrix/data frame of index values into an index
#' object.
#'
#' @examples
#' index <- as_index(matrix(1:6, 2))
#'
#' as.data.frame(index)
#'
#' as.matrix(index)
as.data.frame.piar_index <- function(
  x,
  row.names = NULL,
  optional = FALSE,
  ...,
  contrib = FALSE,
  stringsAsFactors = FALSE
) {
  chkDots(...)
  value <- x$index
  dim(value) <- NULL
  period <- rep(x$time, each = nlevels(x))
  if (stringsAsFactors) {
    res <- data.frame(
      period = factor(period, x$time),
      level = factor(x$levels, x$levels),
      value,
      row.names = row.names
    )
  } else {
    res <- data.frame(
      period,
      level = x$levels,
      value,
      row.names = row.names,
      stringsAsFactors = FALSE
    )
  }
  if (contrib) {
    contributions <- x$contrib %||% contrib_skeleton(x$levels, x$time)
    dim(contributions) <- NULL
    res$contrib <- contributions
  }
  res
}

#' @rdname as.data.frame.piar_index
#' @export
as.matrix.piar_index <- function(x, ...) {
  chkDots(...)
  res <- x$index
  dimnames(res) <- list(levels = x$levels, time = x$time)
  res
}

#' @export
as.double.piar_index <- function(x, ...) {
  chkDots(...)
  as.double(x$index)
}

#' Coerce an index into a time series
#'
#' Turn an index into a regular time series, represented as a [`ts`] object.
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param ... Additional arguments passed to [`ts()`].
#'
#' @returns A time series object.
#' @examples
#' as.ts(as_index(matrix(1:9, 3)))
#'
#' @importFrom stats as.ts ts
#' @family index methods
#' @export
as.ts.piar_index <- function(x, ...) {
  ts(data = t(x$index), ..., names = x$levels)
}

#' Return the first/last parts of an index
#'
#' Extract the first/last parts of an index as if it were a matrix.
#'
#' @family index methods
#' @importFrom utils head
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param n `[integer]` See [head()]/[tail()]. The default takes the
#'   first/last 6 levels of `x`.
#' @param ... Not currently used.
#'
#' @returns
#' A price index that inherits from the same class as `x`.
#'
#' @examples
#' index <- as_index(matrix(1:9, 3))
#'
#' head(index, 1)
#'
#' tail(index, 1)
# Optional arguments are before ... to agree with default in utils::head().
head.piar_index <- function(x, n = 6L, ...) {
  chkDots(...)
  x$index <- head(x$index, n)
  if (!is.null(x$contrib)) {
    x$contrib <- head(x$contrib, n)
  }
  if (!is.na(n[1L])) {
    x$levels <- head(x$levels, n[1L])
  }
  if (!is.na(n[2L])) {
    x$time <- head(x$time, n[2L])
  }
  validate_index_structure(x)
}

#' @rdname head.piar_index
#' @importFrom utils tail
#' @export
tail.piar_index <- function(x, n = 6L, ...) {
  chkDots(...)
  x$index <- tail(x$index, n, keepnums = FALSE)
  if (!is.null(x$contrib)) {
    x$contrib <- tail(x$contrib, n, keepnums = FALSE)
  }
  if (!is.na(n[1L])) {
    x$levels <- tail(x$levels, n[1L])
  }
  if (!is.na(n[2L])) {
    x$time <- tail(x$time, n[2L])
  }
  validate_index_structure(x)
}

#' Combine hierarchical classifications
#'
#' Combine hierarchical classifications by stacking one classification after
#' another.
#'
#' @export
#'
#' @param ... `[list]` A collection of lists, one for each classification,
#'   each giving
#'   the "digits" that represent each level in the hierarchy, as made
#'   by [expand_classification()] or [split_classification()].
#' @param sep `[character(1)]` A character used to separate the classifications
#'   in `...`. The
#'   default separates levels across classifications by `"."`.
#'
#' @returns
#' A list with a entry for each level in the combined classification.
#'
#' @examples
#' # Combine an unbalanced industry classification with a balanced
#' # geographic classification.
#' industry <- c("111", "112", "12")
#' region <- c("11", "21", "22")
#'
#' combine_classifications(
#'   expand_classification(industry, pad = "0"),
#'   expand_classification(region)
#' )
combine_classifications <- function(..., sep = ".") {
  dots <- lapply(list(...), \(x) lapply(x, as.character))
  Reduce(
    \(x, y) .combine_classifications(x, y, sep = sep),
    dots,
    simplify = FALSE
  )
}

#' Combine classifications (internal)
#' @noRd
.combine_classifications <- function(x, y, sep) {
  lx <- lengths(x)
  if (length(lx) == 0L) {
    return(y)
  }
  ly <- lengths(y)
  if (length(ly) == 0L) {
    return(x)
  }
  if (any(lx != lx[1L]) || any(ly != lx[1L])) {
    stop("each component of `...` must have the same number of elements")
  }

  c(x, lapply(y, \(y) paste(last(x), y, sep = sep)))
}

#' Split a hierarchical classification
#'
#' Expand a character representation of a hierarchical classification to make a
#' price index aggregation structure by splitting along a delimiter.
#'
#' @export
#'
#' @param x `[character]` A character vector, or something that can be coerced
#'   into one, of codes/labels for a specific level in a classification
#'   (e.g., 5-digit COICOP).
#' @param split `[character(1)]` A regular expression to delineate and split the
#'   levels in `x`. See [strsplit()].
#' @param ... Additional argument to pass to [strsplit()].
#' @param sep `[character(1)]` A character used to delineate levels in `x` in
#'   the result. The default separates levels by `"."`.
#' @param pad `[character(1)]` A string used to pad the shorter labels for an
#'   unbalanced classification. The default pads with `NA`.
#'
#' @returns
#' A list with a entry for each level in `x` giving the "digits" that
#' represent each level in the hierarchy.
#'
#' @seealso
#' [aggregation_structure()] to make a price-index aggregation structure.
#'
#' [expand_classification()] to expand a classification by the width of the
#' levels.
#'
#' [combine_classifications()] for combining multiple hierarchical
#' classifications.
#'
#' @examples
#' #' # A simple classification structure.
#' #            1
#' #      |-----+-----|
#' #      11          12
#' #  |---+---|       |
#' #  111     112     121
#'
#' split_classification(c("111", "112", "121"), "")
#'
#' # Useful if there are delimiters in the classification (like COICOP).
#' split_classification(c("01.1.1", "01.1.2", "01.2.1"), ".", fixed = TRUE)
split_classification <- function(x, split, ..., sep = ".", pad = NA) {
  x <- as.character(x)
  if (length(x) == 0L) {
    return(list())
  }
  x <- strsplit(x = x, split = split, ...)
  n <- max(lengths(x))
  if (n == 0L) {
    return(list(rep.int("", length(x))))
  }
  res <- do.call(rbind, lapply(x, \(z) padded_extract(z, n, pad)))
  Reduce(
    \(...) paste(..., sep = sep),
    split(res, gl(ncol(res), nrow(res))),
    accumulate = TRUE,
    simplify = FALSE
  )
}

#' Make elementary/elemental price indexes
#'
#' Compute period-over-period (chainable) or fixed-base (direct) elementary
#' price indexes, with optional percent-change contributions for each
#' product.
#'
#' When supplied with a numeric vector, `elementary_index()` is a simple
#' wrapper that applies [gmean()] and [`transmute_weights()`]
#' (if `contrib = TRUE`) to `x` and `weights` grouped by `ea` and `period`. That
#' is, for every combination of elementary aggregate and time period,
#' `elementary_index()` calculates an index based on a generalized mean of
#' order `order` and, optionally, percent-change contributions. Product names
#' should
#' be unique within each elementary aggregate at each time period when making
#' contributions and, if not, are
#' passed to [make.unique()] with a warning. The default
#' (`order = 0` and no weights) makes Jevons elementary indexes. See chapter 8
#' (pp. 175--190) of the CPI manual (2020) for more detail about making
#' elementary indexes, or chapter 9 of the PPI manual (2004), and chapter 5 of
#' Balk (2008).
#'
#' The default method simply coerces `x` to a numeric vector prior to
#' calling the method above. The data frame method provides a formula interface
#' to specify columns of price relatives, time periods, and elementary
#' aggregates and call the method above.
#'
#' The interpretation of the index depends on how the price relatives in
#' `x` are made. If these are period-over-period relatives, then the
#' result is a collection of period-over-period (chainable) elementary indexes;
#' if these are fixed-base relatives, then the result is a collection of
#' fixed-base (direct) elementary indexes. For the latter, `chainable`
#' should be set to `FALSE` so that no subsequent methods assume that a
#' chained calculation should be used.
#'
#' By default, missing price relatives in `x` will propagate throughout
#' the index calculation. Ignoring missing values with `na.rm = TRUE` is
#' the same as overall-mean imputation, and needs to be explicitly
#' set in the call to `elementary_index()`. Explicit imputation of missing
#' relatives, and especially imputation of missing prices, should be done prior
#' to calling `elementary_index()`.
#'
#' Indexes based on nested generalized means, like the Fisher index (and
#' superlative quadratic mean indexes more generally), can be calculated by
#' supplying the appropriate weights with [transmute_weights2()]; see the
#' example below. It is important to note that there are several ways to
#' make these weights, and this affects how percent-change contributions
#' are calculated.
#'
#' `elemental_index()` is an alias for `elementary_index()`.
#'
#' @export
#'
#' @param x `[object]` Period-over-period or fixed-base price
#'   relatives. Currently there are methods for numeric vectors (which can be
#'   made with [price_relative()]) and data frames.
#' @param period `[factor]` A factor, or something that can be coerced into one,
#'   giving
#'   the time period associated with each price relative in `x`. The
#'   ordering of time periods follows of the levels of `period`, to agree
#'   with [`cut()`][cut.Date]. The default makes an index for one time period.
#' @param ea `[factor]` A factor, or something that can be coerced into one,
#'   giving the
#'   elementary aggregate associated with each price relative in `x`. The
#'   default makes an index for one elementary aggregate.
#' @param weights `[numeric >= 0]` A numeric vector of weights for the price
#'   relatives in `x`,
#'   or something that can be coerced into one. The default is equal weights.
#'   This is evaluated in `x` for the data frame method.
#' @param product `[character]` A character vector of product names, or
#'   something that can
#'   be coerced into one, for each price relative in `x` when making
#'   percent-change contributions. The default uses the names of `x`, if any;
#'   otherwise, elements of `x` are given sequential names within each
#'   elementary aggregate. This is evaluated in `x` for the data frame method.
#' @param contrib `[logical(1)]` Should percent-change contributions be
#'   calculated? The default does not calculate contributions.
#' @param chainable `[logical(1)]` Are the price relatives in `x`
#'   period-over-period
#'   relatives that are suitable for a chained calculation (the default)? This
#'   should be `FALSE` when `x` contains fixed-base relatives.
#' @param na.rm `[logical(1)]` Should missing values be removed? By default,
#'   missing values
#'   are not removed. Setting `na.rm = TRUE` is equivalent to overall-mean
#'   imputation.
#' @param order `[numeric(1)]` Order of the generalized mean to aggregate
#'   price relatives. 0 for a
#'   geometric index (the default for making elementary indexes), 1 for an
#'   arithmetic index (the default for aggregating elementary indexes and
#'   averaging indexes over subperiods), or -1 for a harmonic index (usually for
#'   a Paasche index). Other values are possible; see
#'   [gmean()] for details.
#' @param ... Further arguments passed to or used by methods.
#' @param formula `[formula]` A two-sided formula, or something that can be
#'   coerced into one, with price relatives on the left-hand
#'   side and time periods and elementary aggregates (in that order) on the
#'   right-hand side.
#' @param r Deprecated.
#'
#' @returns
#' A price index that inherits from [`piar_index`]. If
#' `chainable = TRUE` then this is a period-over-period index that also
#' inherits from [`chainable_piar_index`]; otherwise, it is a
#' fixed-based index that inherits from [`direct_piar_index`].
#'
#' @seealso
#' [price_relative()] for making price relatives for the same products over
#' time, and [impute_prices()] for imputation of missing prices.
#'
#' [as_index()] to turn pre-computed (elementary) index values into an
#' index object.
#'
#' [chain()] for chaining period-over-period indexes, and
#' [rebase()] for rebasing an index.
#'
#' [`aggregate()`][aggregate.piar_index] to aggregate elementary indexes
#' according to an aggregation structure.
#'
#' [`as.matrix()`][as.matrix.piar_index] and
#' [`as.data.frame()`][as.data.frame.piar_index] for coercing an index
#' into a tabular form.
#'
#' @references
#' Balk, B. M. (2008). *Price and Quantity Index Numbers*.
#' Cambridge University Press.
#'
#' ILO, IMF, UNECE, OECD, and World Bank. (2004).
#' *Producer Price Index Manual: Theory and Practice*.
#' International Monetary Fund.
#'
#' IMF, ILO, OECD, Eurostat, UNECE, and World Bank. (2020).
#' *Consumer Price Index Manual: Concepts and Methods*.
#' International Monetary Fund.
#'
#' von der Lippe, P. (2007). *Index Theory and Price Statistics*. Peter Lang.
#'
#' @examples
#' prices <- data.frame(
#'   rel = 1:8,
#'   period = rep(1:2, each = 4),
#'   ea = rep(letters[1:2], 4)
#' )
#'
#' # Calculate Jevons elementary indexes.
#' elementary_index(prices, rel ~ period + ea)
#'
#' # Same as using lm() or tapply().
#' exp(coef(lm(log(rel) ~ ea:factor(period) - 1, prices)))
#'
#' with(
#'   prices,
#'   t(tapply(rel, list(period, ea), \(x) gmean(x, order = 0, na.rm = TRUE)))
#' )
#'
#' # Calculate a CSWD index (same as the Jevons in this example)
#' # as an arithmetic index by using the appropriate weights.
#' cswd_weights <- with(
#'   prices,
#'   lapply(split(rel, list(period, ea)), transmute_weights2)
#' )
#'
#' elementary_index(
#'   prices,
#'   rel ~ period + ea,
#'   weights = unsplit(cswd_weights, interaction(period, ea)),
#'   order = 1
#' )
elementary_index <- function(x, ...) {
  if ("r" %in% ...names()) {
    warning("`r` is deprecated and will be removed; use `order` instead")
  }
  UseMethod("elementary_index")
}

#' @rdname elementary_index
#' @export
elementary_index.default <- function(x, ...) {
  elementary_index(as.numeric(x), ...)
}

#' @rdname elementary_index
#' @export
elementary_index.numeric <- function(
  x,
  ...,
  period = NULL,
  ea = NULL,
  weights = NULL,
  product = NULL,
  chainable = TRUE,
  na.rm = FALSE,
  contrib = FALSE,
  order = 0,
  r = order
) {
  chkDots(...)
  if (!is.null(weights)) {
    weights <- as.numeric(weights)
    if (any(weights < 0, na.rm = TRUE)) {
      stop("all elements of `weights` must be non-negative")
    }
  }
  r <- as.numeric(r)
  period <- as.factor(period %||% gl(1, length(x)))
  ea <- as.factor(ea %||% gl(1, length(x)))
  if (!is.null(product)) {
    product <- as.character(product)
  }

  if (different_length(x, period, ea, weights, product)) {
    stop("input vectors must be the same length")
  }
  if (any(x <= 0, na.rm = TRUE)) {
    stop("all elements of `x` must be strictly positive")
  }
  ea_by_period <- period:ea
  time <- levels(period)
  levels <- levels(ea)

  if (contrib) {
    if (!is.null(product)) {
      names(x) <- product
    }
    if (is.null(names(x))) {
      names(x) <- paste(ea, sequential_names(period, ea), sep = ".")
    } else {
      names(x) <- valid_product_names(names(x), ea_by_period)
    }
  }

  x <- split(x, ea_by_period)
  weights <- if (is.null(weights)) list(NULL) else split(weights, ea_by_period)

  index <- mapply(
    gmean,
    x,
    weights,
    r,
    na.rm = na.rm,
    USE.NAMES = FALSE
  )
  dim(index) <- c(nlevels(ea), nlevels(period))

  if (contrib) {
    contributions <- mapply(
      \(x, w, r, m) (x - 1) * transmute_weights(x, w, r, mean = m),
      x,
      weights,
      r,
      index,
      SIMPLIFY = FALSE,
      USE.NAMES = FALSE
    )
    dim(contributions) <- c(nlevels(ea), nlevels(period))
    piar_index(index, contributions, levels, time, chainable = chainable)
  } else {
    piar_index(index, NULL, levels, time, chainable = chainable)
  }
}


#' @rdname elementary_index
#' @export
elementary_index.data.frame <- function(
  x,
  formula,
  ...,
  weights = NULL,
  product = NULL
) {
  vars <- formula_vars(formula, x)
  weights <- eval(substitute(weights), x, parent.frame())
  product <- eval(substitute(product), x, parent.frame())

  elementary_index(
    vars[[1L]],
    period = vars[[2L]],
    ea = vars[[3L]],
    weights = weights,
    product = product,
    ...
  )
}

#' @rdname elementary_index
#' @export
elemental_index <- elementary_index

#' Coerce to a price index
#'
#' Coerce pre-computed index values into an index object.
#'
#' Numeric matrices are coerced into an index object by treating each column as
#' a separate time period, and each row as a separate level of the index (e.g.,
#' an elementary aggregate). Column names
#' are used to denote time periods, and row names are used to denote levels
#' (so they must be unique). This essentially reverses calling
#' [`as.matrix()`][as.matrix.piar_index] on an index object. If a
#' dimension is unnamed, then it is given a sequential label from 1 to the size
#' of that dimension. The default and multiple time series methods coerces `x`
#' to a matrix prior to using the matrix method.
#'
#' The data frame method for `as_index()` is best understood as reversing
#' the effect of [`as.data.frame()`][as.data.frame.piar_index] on an
#' index object. It constructs a matrix by taking the levels of
#' `x[[1]]` as columns and the levels of `x[[2]]` as rows
#' (coercing to a factor if necessary). It then populates this matrix with the
#' corresponding values in `x[[3]]`, and uses the matrix method for
#' `as_index()`. If `contrib = TRUE` and there is a fourth list column of
#' product contributions then these are also included in the resulting index.
#'
#' If `x` is a period-over-period index then it is returned unchanged when
#' `chainable = TRUE` and chained otherwise. Similarly, if `x` is a
#' fixed-base index then it is returned unchanged when
#' `chainable = FALSE` and unchain otherwise.
#'
#' @export
#'
#' @param x `[object]` An object to coerce into a price index.
#' @param chainable `[logical(1)]` Are the index values in `x`
#'   period-over-period
#'   indexes, suitable for a chained calculation (the default)? This should be
#'   `FALSE` when `x` contains fixed-base (direct) index values.
#' @param contrib `[logical(1)]` Should the index values in `x` be used to
#'   construct percent-change contributions? The default does not make
#'   contributions.
#' @param ... Further arguments passed to or used by methods.
#'
#' @returns
#' A price index that inherits from
#' [`piar_index`]. If `chainable = TRUE` then this is a
#' period-over-period price index that also inherits from
#' [`chainable_piar_index`]; otherwise, it is a fixed-base index that
#' inherits from [`direct_piar_index`].
#'
#' @seealso
#' [`as.matrix()`][as.matrix.piar_index] and
#' [`as.data.frame()`][as.data.frame.piar_index] for coercing an index
#' into a tabular form.
#'
#' @examples
#' prices <- data.frame(
#'   rel = 1:8,
#'   period = rep(1:2, each = 4),
#'   ea = rep(letters[1:2], 4)
#' )
#'
#' index <- elementary_index(prices, rel ~ period + ea)
#'
#' all.equal(as_index(as.data.frame(index)), index)
#' all.equal(as_index(as.matrix(index)), index)
as_index <- function(x, ...) {
  UseMethod("as_index")
}

#' @rdname as_index
#' @export
as_index.default <- function(x, ...) {
  as_index(as.matrix(x), ...)
}

#' @rdname as_index
#' @export
as_index.matrix <- function(x, ..., chainable = TRUE, contrib = FALSE) {
  chkDots(...)
  levels <- as.character(rownames(x) %||% seq_len(nrow(x)))
  periods <- as.character(colnames(x) %||% seq_len(ncol(x)))

  index <- as.numeric(x)
  dim(index) <- c(length(levels), length(periods))

  piar_index(
    index,
    if (contrib) index2contrib(index, levels, periods),
    levels,
    periods,
    chainable = chainable
  )
}

#' @rdname as_index
#' @export
as_index.data.frame <- function(x, ..., contrib = FALSE) {
  if (length(x) < 3L) {
    stop(
      "`x` must have a column of time periods, index levels, and index values"
    )
  }
  x[1:2] <- lapply(x[1:2], as.factor)
  time <- levels(x[[1L]])
  levels <- levels(x[[2L]])
  index <- index_skeleton(levels, time)
  dimnames(index) <- list(levels, time)
  index[as.matrix(x[2:1])] <- as.numeric(x[[3L]])

  if (contrib && length(x) > 3L) {
    contributions <- contrib_skeleton(levels, time)
    dimnames(contributions) <- list(levels, time)
    contributions[as.matrix(x[2:1])] <- x[[4L]]

    index <- as_index(index, ...)
    # FIXME: I don't like adding the contributions after making the index.
    index$contrib <- lapply(contributions, valid_contrib)
    dim(index$contrib) <- dim(contributions)
  } else {
    index <- as_index(index, contrib = contrib, ...)
  }
  index
}

#' @rdname as_index
#' @export
as_index.chainable_piar_index <- function(x, ..., chainable = TRUE) {
  chkDots(...)
  if (chainable) x else chain(x)
}

#' @rdname as_index
#' @export
as_index.direct_piar_index <- function(x, ..., chainable = FALSE) {
  chkDots(...)
  if (chainable) unchain(x) else x
}

#' @rdname as_index
#' @export
as_index.mts <- function(x, ...) {
  as_index(t(as.matrix(x)), ...)
}

# as.ts works with an index.
local({
  index <- matrix(1:9, 3, dimnames = list(letters[1:3], NULL))

  expect_equal(as.ts(as_index(index)), as.ts(t(index)))

  expect_equal(
    as.ts(as_index(index), start = 2020, frequency = 12),
    ts(t(index), start = 2020, frequency = 12)
  )
})

# Splitting works.
local({
  x <- as_index(matrix(1:6, 2))
  expect_equal(
    split(x, 1:2),
    list(
      "1" = as_index(matrix(c(1, 3, 5), 1)),
      "2" = as_index(matrix(c(2, 4, 6), 1, dimnames = list(2, 1:3)))
    )
  )
  expect_equal(
    split(x, c(1, 1, 2), along = "time"),
    list(
      "1" = as_index(matrix(1:4, 2)),
      "2" = as_index(matrix(5:6, 2, dimnames = list(1:2, 3)))
    )
  )
})

# Replacement works.
local({
  x <- as_index(matrix(1:6, 2))
  split(x, 1:2) <- 1:2
  expect_equal(x, as_index(matrix(c(1, 2), 2, 3)))

  x <- as_index(matrix(1:6, 2))
  split(x, c(1, 1, 2), along = "time") <- 1:2
  expect_equal(x, as_index(matrix(c(1, 1, 1, 1, 2, 2), 2)))

  y <- x
  split(y, factor(c(1, 1, 2), levels = character(0)), along = "time") <- list()
  expect_identical(x, y)

  expect_warning(split(x, 1:3, along = "time") <- 1:2)
})

# Window works.
local({
  x <- as_index(matrix(1:9, 3))
  expect_equal(window(x), x)
  expect_equal(window(x, "2"), x[, 2:3])
  expect_equal(window(x, end = "2"), x[, 1:2])
  expect_equal(window(x, "2", "2"), x[, 2])
})

# Window replacement works.
local({
  x <- as_index(matrix(1:9, 3))
  window(x, "2") <- 1
  expect_equal(x, as_index(matrix(c(1:3, rep(1, 6)), 3)))
})

# Window errors when expected".
local({
  x <- as_index(matrix(1:9, 3))
  expect_error(window(x, "2", "1"))
  expect_error(window(x, "4"))
  expect_error(window(x, end = "4"))
})

test_that("window works", {
  x <- as_index(matrix(1:9, 3))
  expect_equal(window(x), x)
  expect_equal(window(x, "2"), x[, 2:3])
  expect_equal(window(x, end = "2"), x[, 1:2])
  expect_equal(window(x, "2", "2"), x[, 2])
})

test_that("window replacement works", {
  x <- as_index(matrix(1:9, 3))
  window(x, "2") <- 1
  expect_equal(x, as_index(matrix(c(1:3, rep(1, 6)), 3)))
})

test_that("window errors when expected", {
  x <- as_index(matrix(1:9, 3))
  expect_error(window(x, "2", "1"))
  expect_error(window(x, "4"))
  expect_error(window(x, end = "4"))
})

test_that("splitting works", {
  x <- as_index(matrix(1:6, 2))
  expect_equal(
    split(x, 1:2),
    list(
      "1" = as_index(matrix(c(1, 3, 5), 1)),
      "2" = as_index(matrix(c(2, 4, 6), 1, dimnames = list(2, 1:3)))
    )
  )
  expect_equal(
    split(x, c(1, 1, 2), along = "time"),
    list(
      "1" = as_index(matrix(1:4, 2)),
      "2" = as_index(matrix(5:6, 2, dimnames = list(1:2, 3)))
    )
  )
})

test_that("replacement works", {
  x <- as_index(matrix(1:6, 2))
  split(x, 1:2) <- 1:2
  expect_equal(x, as_index(matrix(c(1, 2), 2, 3)))

  x <- as_index(matrix(1:6, 2))
  split(x, c(1, 1, 2), along = "time") <- 1:2
  expect_equal(x, as_index(matrix(c(1, 1, 1, 1, 2, 2), 2)))

  y <- x
  split(y, factor(c(1, 1, 2), levels = character(0)), along = "time") <- list()
  expect_identical(x, y)

  expect_warning(split(x, 1:3, along = "time") <- 1:2)
})

test_that("as.ts works with an index", {
  index <- matrix(1:9, 3, dimnames = list(letters[1:3], NULL))

  expect_equal(as.ts(as_index(index)), as.ts(t(index)))

  expect_equal(
    as.ts(as_index(index), start = 2020, frequency = 12),
    ts(t(index), start = 2020, frequency = 12)
  )
})

#' Transmute weights for a generalized mean
#'
#' Transmute weight to turn a generalized mean of a given order into a
#' generalized mean of any other order. Useful for calculating additive and
#' multiplicative decompositions for generalized-mean indexes.
#' See `vignette("decomposing-indexes")` for more details.
#'
#' This function generalizes the additive and multiplicative decompositions
#' for arithmetic and geometric indexes by Balk (2008, Chapter 4). It returns
#' a value such that
#'
#' \preformatted{gmean(x, w, r) == gmean(x, transmute_weights(x, w, r, s), s)}
#'
#' Transmuting weights returns a value that is the same length as `x`,
#' so any missing values in `x` or `weights` will return `NA`.
#' Unless all values are `NA`, however, the result will still satisfy
#' the above identity when `na.rm = TRUE`.
#'
#' @family math functions
#' @export
#'
#' @inheritParams gmean
#' @param order `[numeric(1)]` A finite number giving the order of the
#'   generalized mean. The default transmutes the weights for a geometric mean.
#' @param to `[numeric(1)]` A finite number giving the order of the target
#'   generalized mean. The default computes weights for an arithmetic mean.
#' @param mean `[numeric(1)]` A finite number giving the generalized mean of
#'   `x` and `weights`, if known. The default computes this values.
#'
#' @returns
#' A numeric vector, the same length as `x`, that sums to 1.
#'
#' @references
#' Balk, B. M. (2008). *Price and Quantity Index Numbers*.
#' Cambridge University Press.
#'
#' @examples
#' x <- 1:3
#' w <- 3:1
#'
#' # Calculate the geometric mean as an arithmetic mean.
#' gmean(x, order = 0)
#' gmean(x, transmute_weights(x, order = 0, to = 1), order = 1)
transmute_weights <- function(x, weights = NULL, order = 0, to = 1, mean = NA) {
  if (not_finite_scalar(order)) {
    stop("`order` must be a finite number")
  }
  if (not_finite_scalar(to)) {
    stop("`to` must be a finite number")
  }
  if (!is.null(weights) && length(x) != length(weights)) {
    stop("`x` and `weights` must be the same length")
  }
  if (is.na(mean)) {
    mean <- gmean(x, weights, order, na.rm = TRUE)
  }
  .transmute_weights(x, weights, order, to, mean)
}

#' Transmute weights for a nested generalized mean
#'
#' Transmute weights to turn a nested generalized mean of a given order into a
#' generalized mean of any order. Useful for calculating additive and
#' multiplicative decompositions for an index made
#' of nested generalized means (e.g., Fisher index).
#' See `vignette("decomposing-indexes")` for details.
#'
#' This function generalizes the additive and multiplicative decompositions
#' for the Fisher index by Balk (2008, Chapter 4). It returns
#' a value such that
#'
#' \preformatted{nested_gmean(x, list(w1, w2), c(r1, r2)) ==
#'     gmean(x, transmute_weights2(x, list(w1, w2), c(r1, r2), to = s), s)}
#'
#' Transmuting weights returns a value that is the same length as `x`,
#' so any missing values in `x` or `weights` will return `NA`.
#' Unless all values are `NA`, however, the result will still satisfy
#' the above identity when `na.rm = TRUE`.
#'
#' @inheritParams nested_gmean
#' @param to A finite number giving the order of the target generalized mean for
#'   the transmuted weights. The default constructs weights for an arithmetic
#'   mean.
#' @param pivot A finite number giving the pivot value for the transmuted
#'   weights. The default uses the order of the outer generalized mean,
#'   otherwise `to` is common alternative.
#'
#' @returns
#' A numeric vector, the same length as `x`, that sums to 1.
#'
#' @references
#' Balk, B. M. (2008). *Price and Quantity Index Numbers*.
#' Cambridge University Press.
#'
#' @examples
#' x <- 1:3
#' w1 <- 3:1
#' w2 <- c(1, 2, 1)
#'
#' # Calculate the geometric mean of the arithmetic and harmonic means
#' # as an arithmetic mean.
#' nested_gmean(x, list(w1, w2))
#' gmean(x, transmute_weights2(x, list(w1, w2), to = 1))
#' @family math functions
#' @export
transmute_weights2 <- function(
  x,
  weights = list(NULL, NULL),
  order = c(1, -1),
  outer_weights = NULL,
  outer_order = 0,
  to = 1,
  pivot = outer_order
) {
  if (different_length(x, weights[[1]], weights[[2]])) {
    stop("`x` and non-NULL components of `weights` must be the same length")
  }
  if (not_finite_pair(order)) {
    stop("`order` must be a pair of finite numbers")
  }
  if (length(weights) != length(order)) {
    stop("`weights` and `order` must be the same length")
  }
  na_mask <- if (anyNA(x) || anyNA(weights, recursive = TRUE)) {
    stats::complete.cases(x, weights[[1]], weights[[2]])
  }
  if (!is.null(na_mask)) {
    x[!na_mask] <- NA_real_
  }
  inner_mean <- c(
    .gmean(x, weights[[1]], order[1], na_mask),
    .gmean(x, weights[[2]], order[2], na_mask)
  )
  outer_mean <- gmean(inner_mean, outer_weights, outer_order, na.rm = TRUE)
  v1 <- transmute_weights(x, weights[[1]], order[1], pivot, inner_mean[1])
  v2 <- transmute_weights(x, weights[[2]], order[2], pivot, inner_mean[2])
  t <- transmute_weights(
    inner_mean,
    outer_weights,
    outer_order,
    pivot,
    outer_mean
  )
  if (is.na(t[1])) {
    transmute_weights(x, v2 * t[2], pivot, to)
  } else if (is.na(t[2])) {
    transmute_weights(x, v1 * t[1], pivot, to)
  } else {
    transmute_weights(x, v1 * t[1] + v2 * t[2], pivot, to)
  }
}

#' Internal function to transmute weights
#' @noRd
.transmute_weights <- function(x, weights, from, to, mean) {
  if (from == to) {
    if (is.null(weights)) {
      weights <- rep.int(1, length(x))
    }
    if (anyNA(x)) {
      weights[is.na(x)] <- NA_real_
    }
  } else {
    ext_mean <- .extended_mean_pow(x, mean, from, to)
    weights <- if (is.null(weights)) ext_mean else weights * ext_mean
  }
  scale_weights(weights)
}

#' Simplified extended mean for transmuting weights
#' @noRd
.extended_mean_pow <- function(x, m, r, s, tol = .Machine$double.eps^0.5) {
  rdiff <- function(a, b, r) {
    if (r == 0) {
      log(a / b)
    } else if (r == 1) {
      a - b
    } else {
      (a^r - b^r) / r
    }
  }
  res <- rdiff(x, m, r) / rdiff(x, m, s)
  res[abs(x - m) <= tol] <- m^(r - s)
  res
}

#' Extract index values
#'
#' Methods to extract and replace index values like a matrix.
#'
#' The extraction method treats `x` like a matrix of index values with
#' (named) rows for each level and columns for each time period in
#' `x`. Unlike a matrix, dimensions are never dropped as subscripting
#' `x` always returns an index object. The one exception is when subscripting
#' with a matrix, in which case a list of index values is returned. As `x`
#' is not an atomic vector, subscripting with a single index like `x[1]`
#' extracts all time periods for that level.
#'
#' The replacement method similarly treat `x` like a matrix. If `value` is
#' an index object with the same number of time periods as `x[i, j]` and
#' it inherits from the same class as `x`, then the index values and
#' percent-change contributions of `x[i, j]` are replaced with those for the
#' corresponding levels of `value`. If `value` is not an index, then it is
#' coerced to a numeric vector and behaves the same as replacing values in a
#' matrix. Note that replacing the values of an index will remove the
#' corresponding percent-change contributions (if any). When replacing with a
#' matrix, `value` can be a list of index objects.
#'
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, as made by,
#'   e.g., [elementary_index()].
#' @param i,j `[integer | logical | character | matrix]` Indices for the levels
#'   and time periods of a price index. See details.
#' @param value `[numeric > 0 | piar_index | list]` A numeric vector,
#'   price index, or list of price indexes. See details.
#' @param ... Not currently used.
#'
#' @returns
#' A price index that inherits from the same class as `x`.
#'
#' @examples
#' index <- as_index(matrix(1:6, 2))
#'
#' index["1", ]
#'
#' index[, 2]
#'
#' index[1, ] <- 1 # can be useful for doing specific imputations
#'
#' index
`[.piar_index` <- function(x, i, j, ...) {
  chkDots(...)
  if (!missing(i) && is.matrix(i)) {
    if (!missing(j)) {
      warning(
        "indices for time periods do nothing when subscripting with a matrix"
      )
    }
    i <- subscript_index_matrix(x, i)
    lapply(seq_len(nrow(i)), \(j) extract_index(x, i[j, 1L], i[j, 2L]))
  } else {
    levels <- subscript_index(x$levels, i)
    periods <- subscript_index(x$time, j)
    extract_index(x, levels, periods)
  }
}

#' Internal extraction function
#' @noRd
extract_index <- function(x, levels, periods) {
  x$index <- x$index[levels, periods, drop = FALSE]
  if (!is.null(x$contrib)) {
    x$contrib <- x$contrib[levels, periods, drop = FALSE]
  }
  x$levels <- x$levels[levels]
  x$time <- x$time[periods]
  validate_index_structure(x)
}

#' @rdname sub-.piar_index
#' @export
`[<-.piar_index` <- function(x, i, j, ..., value) {
  chkDots(...)
  if (!missing(i) && is.matrix(i)) {
    if (!missing(j)) {
      warning(
        "indices for time periods do nothing when subscripting with a matrix"
      )
    }
    i <- subscript_index_matrix(x, i)
    if (is_index(value)) {
      value <- list(value)
    }
    x <- if (is.list(value)) {
      replace_matrix_list(x, i, value)
    } else {
      replace_matrix_numeric(x, i, value)
    }
  } else {
    levels <- subscript_index(x$levels, i)
    periods <- subscript_index(x$time, j)
    x <- if (is_index(value)) {
      replace_index(x, levels, periods, value)
    } else {
      replace_numeric(x, levels, periods, value)
    }
  }
  # Replacement value should be validated; e.g., x[1] <- -1.
  validate_piar_index(x)
}

#' @export
`[<-.chainable_piar_index` <- function(x, i, j, ..., value) {
  if (is_index(value)) {
    if (!is_chainable_index(value)) {
      stop("`value` must be a period-over-period index")
    }
  } else if (is.list(value)) {
    if (!all(vapply(value, is_chainable_index, logical(1L)))) {
      stop("all elements of `value` must be period-over-period indexes")
    }
  }
  NextMethod("[<-")
}

#' @export
`[<-.direct_piar_index` <- function(x, i, j, ..., value) {
  if (is_index(value)) {
    if (!is_direct_index(value)) {
      stop("`value` must be a fixed-base index")
    }
  } else if (is.list(value)) {
    if (!all(vapply(value, is_direct_index, logical(1L)))) {
      stop("all elements of `value` must be fixed-base indexes")
    }
  }
  NextMethod("[<-")
}

#' Internal replacement functions
#' @noRd
replace_matrix_list <- function(x, i, value) {
  if (nrow(i) == 0L) {
    return(x)
  }
  n <- length(value)
  if (n == 0L) {
    stop("replacement has length zero")
  }
  if (nrow(i) %% n != 0) {
    warning(
      "number of items to replace is not a multiple of replacement length"
    )
  }
  if (any(vapply(value, \(x) ntime(x) * nlevels(x), integer(1L)) > 1L)) {
    stop("`value` must be a list of indexes with one level and time period")
  }
  # Make `value` the same length as replacement to avoid two warnings.
  index <- rep_len(vapply(value, \(x) x$index[[1]], numeric(1L)), nrow(i))
  contributions <- rep_len(lapply(value, \(x) x$contrib[[1]]), nrow(i))
  has_contrib <- any(vapply(contributions, Negate(is.null), logical(1L)))
  # It's possible that only some elements of `value` have contributions.
  if (has_contrib || !is.null(x$contrib)) {
    contributions[lengths(contributions) == 0L] <- list(numeric(0L))
  }
  if (is.null(x$contrib) && has_contrib) {
    x$contrib <- contrib_skeleton(x$levels, x$time)
  }
  x$index[i] <- index
  if (!is.null(x$contrib)) {
    x$contrib[i] <- contributions
  }
  x
}

replace_matrix_numeric <- function(x, i, value) {
  x$index[i] <- as.numeric(value)
  if (!is.null(x$contrib)) {
    x$contrib[i] <- list(numeric(0L))
  }
  x
}

replace_index <- function(x, levels, periods, value) {
  if (length(levels) == 0L || length(periods) == 0L) {
    return(x)
  }

  if (ntime(value) != length(periods)) {
    stop("`x` and `value` must have the same number of time periods")
  }
  if (length(levels) %% nlevels(value) != 0) {
    stop("number of items to replace is not a multiple of replacement length")
  }
  if (is.null(x$contrib) && !is.null(value$contrib)) {
    x$contrib <- contrib_skeleton(x$levels, x$time)
  } else if (!is.null(x$contrib) && is.null(value$contrib)) {
    value$contrib <- contrib_skeleton(value$levels, value$time)
  }
  for (t in seq_along(periods)) {
    x$index[levels, periods[t]] <- value$index[, t]
    if (!is.null(x$contrib)) {
      x$contrib[levels, periods[t]] <- value$contrib[, t]
    }
  }
  x
}

replace_numeric <- function(x, levels, periods, value) {
  x$index[levels, periods] <- as.numeric(value)
  if (!is.null(x$contrib)) {
    x$contrib[levels, periods] <- list(numeric(0L))
  }
  x
}

#' Aggregate elementary price indexes
#'
#' Aggregate elementary price indexes with a price index aggregation structure.
#'
#' The `aggregate()` method loops over each time period in `x` and
#' 1. aggregates the elementary indexes with [`gmean()`] for each level
#' of `pias`;
#' 2. aggregates percent-change contributions for each level of
#' `pias` (if there are any and `contrib = TRUE`);
#' 3. price updates the weights in `pias` with [`update_weights()`] (only for
#' period-over-period elementary indexes).
#'
#' The result is a collection of aggregated period-over-period indexes that
#' can be chained together to get a fixed-base index when `x` are
#' period-over-period elementary indexes. Otherwise, when `x` are fixed-base
#' elementary indexes, the result is a collection of aggregated fixed-base
#' (direct) indexes.
#'
#' By default, missing elementary indexes will propagate when aggregating the
#' index. Missing elementary indexes can be due to both missingness of these
#' values in `x`, and the presence of elementary aggregates in `pias`
#' that are not part of `x`. Setting `na.rm = TRUE` ignores missing
#' values, and is equivalent to overall-mean (or parental) imputation. As an
#' aggregated price index generally cannot have missing values (for otherwise
#' it can't be chained over time and weights can't be price updated), any
#' missing values for a level of `pias` are removed and recursively replaced
#' by the value of its immediate parent.
#'
#' In most cases aggregation is done with an arithmetic mean (the default), and
#' this is detailed in chapter 8 (pp. 190--198) of the CPI manual (2020), with
#' analogous details in chapter 9 of the PPI manual (2004).
#' Aggregating with a non-arithmetic mean follows the same steps, except that
#' the elementary indexes are aggregated with a mean of a different order (e.g.,
#' harmonic for a Paasche index), and the method for price updating the weights
#' is slightly different. Note that, because aggregation is done with a
#' generalized mean, the resulting index is consistent-in-aggregation at each
#' point in time.
#'
#' Aggregating percent-change contributions uses the method in chapter 9 of the
#' CPI manual (equations 9.26 and 9.28) when aggregating with an arithmetic
#' mean. With a non-arithmetic mean, arithmetic weights are constructed using
#' [`transmute_weights()`] in order to apply this method.
#'
#' There may not be contributions for all prices relatives in an elementary
#' aggregate if the elementary indexes are built from several sources (as with
#' [`merge()`][merge.piar_index]). In this case the contribution for
#' a price relative in the aggregated index will be correct, but the sum of all
#' contributions will not equal the change in the value of the index.
#'
#' If two aggregation structures are given then the steps above are done for
#' each aggregation structure, with the aggregation for `pias` done with a
#' generalized mean of order `order` the aggregation for `pias2` done with a
#' generalized mean of order `-order`. The resulting indexes are combined with a
#' geometric mean to make a superlative quadratic mean of order `2*order` index.
#' Percent-change contributions are combined using a generalized van IJzeren
#' decomposition; see [`transmute_weights2()`] for details.
#'
#' @name aggregate.piar_index
#' @aliases aggregate.piar_index
#' @importFrom stats aggregate
#' @family index methods
#' @export
#'
#' @param x `[piar_index]` A price index, usually made by [elementary_index()].
#' @param pias `[piar_aggregation_structure]` A price index
#'   aggregation structure or something that can be coerced into one. This can
#'   be made with [aggregation_structure()].
#' @param pias2 `[piar_aggregation_structure]` An optional secondary
#'   aggregation structure, usually with current-period weights, to make a
#'   superlative index. See details.
#' @param na.rm `[logical(1)]` Should missing values be removed? By default,
#'   missing values are not removed. Setting `na.rm = TRUE` is equivalent to
#'   overall-mean imputation.
#' @param order `[numeric(1)]` Order of the generalized mean to aggregate
#'   index values. 0 for a
#'   geometric index (the default for making elementary indexes), 1 for an
#'   arithmetic index (the default for aggregating elementary indexes and
#'   averaging indexes over subperiods), or -1 for a harmonic index (usually for
#'   a Paasche index). Other values are possible; see
#'   [gmean()] for details. If `pias2` is given then the
#'   index is aggregated with a quadratic mean of order `2*r`.
#' @param contrib `[logical(1)]` Aggregate percent-change contributions in `x`?
#'   By default contributions are aggregated.
#' @param include_ea `[logical(1)]` Should indexes for the elementary aggregates
#'   be included along with the aggregated indexes? By default, all index values
#'   are returned.
#' @param ... Not currently used.
#' @param duplicate_contrib `[character(1)]` The method to deal with duplicate
#'   product contributions. Either `"make.unique"` to treat duplicate
#'   products as distinct products and make their names unique
#'   with [make.unique()] or `"sum"` to add contributions for each product
#'   (the default).
#' @param impute_rules `[function]` (Experimental) A function that applies
#'   imputation rules to the elementary indexes in each time period prior to
#'   aggregation.
#'   It takes two arguments, the elementary indexes for a given time period and
#'   the (price updated) aggregation structure, and must return back the
#'   elementary indexes.
#' @param r Deprecated.
#'
#' @returns
#' An aggregate price index that inherits from the class of `x`.
#'
#' @note
#' For large indexes it can be much faster to turn the aggregation structure
#' into an aggregation matrix with
#' [`as.matrix()`][as.matrix.piar_aggregation_structure], then aggregate
#' elementary indexes as a matrix operation when there are no missing
#' values. See the examples for details.
#'
#' @references
#' Balk, B. M. (2008). *Price and Quantity Index Numbers*.
#' Cambridge University Press.
#'
#' ILO, IMF, UNECE, OECD, and World Bank. (2004).
#' *Producer Price Index Manual: Theory and Practice*.
#' International Monetary Fund.
#'
#' IMF, ILO, OECD, Eurostat, UNECE, and World Bank. (2020).
#' *Consumer Price Index Manual: Concepts and Methods*.
#' International Monetary Fund.
#'
#' von der Lippe, P. (2007). *Index Theory and Price Statistics*. Peter Lang.
#'
#' @examples
#' prices <- data.frame(
#'   rel = 1:8,
#'   period = rep(1:2, each = 4),
#'   ea = rep(letters[1:2], 4)
#' )
#'
#' # A two-level aggregation structure.
#' pias <- aggregation_structure(
#'   list(c("top", "top", "top"), c("a", "b", "c")),
#'   weights = 1:3
#' )
#'
#' # Calculate Jevons elementary indexes.
#' (elementary <- elementary_index(prices, rel ~ period + ea))
#'
#' # Aggregate (note the imputation for elementary index 'c').
#' (index <- aggregate(elementary, pias, na.rm = TRUE))
#'
#' # Aggregation can equivalently be done as matrix multiplication.
#' as.matrix(pias) %*% as.matrix(chain(index[letters[1:3]]))
aggregate.chainable_piar_index <- function(
  x,
  pias,
  ...,
  pias2 = NULL,
  na.rm = FALSE,
  contrib = TRUE,
  order = 1,
  r = order,
  include_ea = TRUE,
  duplicate_contrib = c("sum", "make.unique"),
  impute_rules = NULL
) {
  if ("r" %in% names(sys.call())) {
    warning("`r` is deprecated and will be removed; use `order` instead")
  }
  chkDots(...)
  aggregate_index(
    x,
    pias,
    pias2,
    na.rm = na.rm,
    contrib = contrib,
    order = r,
    include_ea = include_ea,
    chainable = TRUE,
    duplicate_contrib = match.arg(duplicate_contrib),
    impute_rules = impute_rules
  )
}

#' @rdname aggregate.piar_index
#' @export
aggregate.direct_piar_index <- function(
  x,
  pias,
  ...,
  pias2 = NULL,
  na.rm = FALSE,
  contrib = TRUE,
  order = 1,
  r = order,
  include_ea = TRUE,
  duplicate_contrib = c("sum", "make.unique"),
  impute_rules = NULL
) {
  if ("r" %in% names(sys.call())) {
    warning("`r` is deprecated and will be removed; use `order` instead")
  }
  chkDots(...)
  aggregate_index(
    x,
    pias,
    pias2,
    na.rm = na.rm,
    contrib = contrib,
    order = r,
    include_ea = include_ea,
    chainable = FALSE,
    duplicate_contrib = match.arg(duplicate_contrib),
    impute_rules = impute_rules
  )
}

#' Internal functions to aggregate a price index
#' @noRd
aggregate_index <- function(
  x,
  pias,
  pias2,
  na.rm,
  contrib,
  order,
  include_ea,
  chainable,
  duplicate_contrib,
  impute_rules
) {
  pias <- as_aggregation_structure(pias)
  order <- as.numeric(order)
  has_contrib <- !is.null(x$contrib) && contrib
  res <- .aggregate(
    x,
    pias,
    na.rm,
    has_contrib,
    order,
    include_ea,
    chainable,
    duplicate_contrib,
    impute_rules
  )

  if (!is.null(pias2)) {
    pias2 <- as_aggregation_structure(pias2)
    if (!same_hierarchy(pias, pias2)) {
      stop("`pias` and `pias2` must represent the same aggregation structure")
    }
    if (
      contrib &&
        any(missing_weights(pias$weights) != missing_weights(pias2$weights))
    ) {
      stop(
        "any NA or zero weights must appear in both `pias` and `pias2` when",
        " `contrib` is TRUE"
      )
    }
    res2 <- .aggregate(
      x,
      pias2,
      na.rm,
      has_contrib,
      -order,
      include_ea,
      chainable,
      duplicate_contrib,
      impute_rules
    )
    if (has_contrib) {
      res$contrib[] <- Map(
        super_aggregate_contrib,
        res$contrib,
        res2$contrib,
        res$index,
        res2$index,
        order = 0
      )
    }
    res$index[] <- (res$index * res2$index)^0.5
  }

  if (!include_ea && nlevels(pias) > 1L) {
    lev <- unlist(drop_last(pias$levels), use.names = FALSE)
  } else {
    lev <- unlist(pias$levels, use.names = FALSE)
  }

  piar_index(res$index, res$contrib, lev, x$time, chainable = chainable)
}

.aggregate <- function(
  x,
  pias,
  na.rm,
  has_contrib,
  order,
  include_ea,
  chainable,
  duplicate_contrib,
  impute_rules
) {
  eas <- match_eas(pias, x)

  # Loop over each time period.
  index <- contrib <- vector("list", ntime(x))
  for (t in seq_along(x$time)) {
    rel <- con <- vector("list", nlevels(pias))
    if (!is.null(impute_rules)) {
      x[, t] <- impute_rules(x[, t], pias)[x$levels]
    }
    # Align with weights so that positional indexing works.
    rel[[1L]] <- x$index[, t][eas]

    if (has_contrib) {
      con[[1L]] <- x$contrib[, t][eas]
      # Replace NULL contributions from subscripting with empty contributions.
      con[[1L]][lengths(con[[1L]]) == 0L] <- list(numeric(0L))
    }

    # Put the aggregation weights upside down to line up with `pias`.
    w <- rev(weights(pias, ea_only = FALSE, na.rm = na.rm))

    # Loop over each level in `pias` from the bottom up and aggregate.
    for (i in seq_along(rel)[-1L]) {
      nodes <- unname(pias$child[[i - 1L]])
      rel[[i]] <- vapply(
        nodes,
        \(z) gmean(rel[[i - 1L]][z], w[[i - 1L]][z], order, na.rm),
        numeric(1L)
      )
      if (has_contrib) {
        con[[i]] <- lapply(
          seq_along(nodes),
          \(j) {
            aggregate_contrib(
              con[[i - 1L]][nodes[[j]]],
              rel[[i - 1L]][nodes[[j]]],
              w[[i - 1L]][nodes[[j]]],
              order,
              rel[[i]][j],
              duplicate_contrib
            )
          }
        )
      }
    }

    # Parental imputation.
    if (na.rm) {
      for (i in rev(seq_along(rel))[-1L]) {
        impute <- which(is.na(rel[[i]]))
        parent <- pias$parent[[i]][impute]
        rel[[i]][impute] <- rel[[i + 1L]][parent]
        con[[i]][impute] <- con[[i + 1L]][parent]
      }
    }

    # Price update weights for all periods after the first.
    if (chainable) {
      pias$weights <- update_weights(rel[[1L]], w[[1L]], order)
    }

    if (!include_ea && length(rel) > 1L) {
      rel <- rel[-1L]
      con <- con[-1L]
    }
    index[[t]] <- unlist(rev(rel), use.names = FALSE)
    contrib[[t]] <- unlist(rev(con), recursive = FALSE, use.names = FALSE)
  }

  list(index = do.call(cbind, index), contrib = do.call(cbind, contrib))
}

#' Aggregate product contributions
#' @noRd
aggregate_contrib <- function(x, rel, w, order, index, duplicate_contrib) {
  w <- transmute_weights(rel, w, order, to = 1, mean = index)
  res <- Map(`*`, x, w)
  if (all(lengths(res) == 0L)) {
    return(numeric(0L))
  }
  if (duplicate_contrib == "make.unique") {
    res <- unlist(res)
    names(res) <- make.unique(names(res))
  } else {
    products <- unlist(lapply(res, names), use.names = FALSE)
    if (anyDuplicated(products)) {
      products <- unique(products)
      mat <- do.call(cbind, Map(`[`, res, list(products)))
      res <- rowSums(mat, na.rm = TRUE)
      res[apply(is.na(mat), 1L, all)] <- NA_real_
      names(res) <- products
    } else {
      res <- unlist(res)
    }
  }
  res
}

#' Aggregate product contributions for a superlative index
#' @noRd
super_aggregate_contrib <- function(x, y, rel1, rel2, order) {
  w <- transmute_weights(c(rel1, rel2), order = order, to = 1)
  w[1L] * x + w[2L] * y
}

#' Impute missing prices
#'
#' Impute missing prices using the carry forward/backward or the
#' self-correcting overall mean method.
#'
#' The carry forward method replaces a missing price for a product by the price
#' for the same product in the previous period. It tends to push an index value
#' towards 1, and is usually avoided; see paragraph 6.61 in the CPI manual
#' (2020). The carry backwards method does the opposite, but this is rarely
#' used in practice.
#'
#' The self-correcting overall mean method recursively imputes a missing price
#' by the value of
#' the price for the same product in the previous period multiplied by the
#' value of the period-over-period elementary index for the elementary aggregate
#' to which that product belongs. This requires computing and aggregating an
#' index (according to `pias`, unless `pias` is not supplied) for
#' each `period`. The index
#' values used to do the imputations are not returned because the index needs
#' to be recalculated to get correct percent-change contributions. It is
#' identical to simply
#' excluding missing price relatives in the index calculation, except in the
#' period that a missing product returns. For this reason care is needed when
#' using this method. It is sensitive to the assumption that a product does not
#' change over time, and in some cases it is safer to simply omit the missing
#' price relatives instead of imputing the missing prices.
#'
#' Imputation works slightly differently depending on whether data are in a long
#' or wide format. When `x` is a two-column matrix of current and back prices
#' (in that order), then imputation is done separately on the current price
#' at a point in time and the back price at the next point in time. When `x` is
#' a numeric vector then these two prices are necessarily the same.
#'
#' @name impute_prices
#' @export
#'
#' @param x `[object]` Either a numeric vector (or something that can be
#'   coerced into one),
#'   a data frame of prices, or a two-column matrix of current prices
#'   and back prices (in that order).
#' @param period `[factor]` A factor, or something that can be coerced into one,
#'   giving
#'   the time period associated with each price in `x`. The ordering of time
#'   periods follows of the levels of `period`, to agree with
#'   [`cut()`][cut.Date].
#' @param product `[factor]` A factor, or something that can be coerced into
#'   one, giving the product associated with each price in `x`.
#' @param ea `[factor]` A factor, or something that can be coerced into one,
#'   giving the
#'   elementary aggregate associated with each price in `x`. This is evaluated
#'   in `x` for the data frame method. The default pools all data into one
#'   elementary aggregate.
#' @param pias `[piar_aggregation_structure]` A price index aggregation
#'   structure, or something that can be
#'   coerced into one, as made with [aggregation_structure()]. The default
#'   imputes from elementary indexes only (i.e., not recursively).
#' @param weights `[numeric >= 0]` A numeric vector of weights for the prices
#'   in `x` (i.e.,
#'   product weights), or something that can be coerced into one. The default is
#'   to give each price equal weight. This is evaluated in `x` for the data
#'   frame method.
#' @param order `[numeric(2)]` A pair of numeric values. The first gives the
#'   order of the generalized-mean price index used to calculate the
#'   elementary price indexes, defaulting to a geometric index. The second
#'   gives the order of the generalized-mean price index used to aggregate the
#'   elementary price indexes, defaulting to an arithmetic index. Other values
#'   are possible;
#'   see [gmean()] for details.
#' @param formula `[formula]` A two-sided formula, or something that can be
#'   coerced into one, with prices on the left-hand
#'   side and time periods and products on the right-hand side (in that order).
#' @param method `[character(1)]` Name of the imputation method, one
#'   of `"overall-mean"` (the default), `"carry-forward"`,
#'   or `"carry-backward"`.
#' @param ... Further arguments passed to or used by methods.
#' @param impute_rules `[function]` (Experimental) A function that applies
#'   imputation
#'   rules to the elementary indexes in each time period prior to aggregation.
#'   It takes two arguments, the elementary indexes for a given time period and
#'   the (price updated) aggregation structure, and must return back the
#'   elementary indexes.
#' @param r Deprecated.
#'
#' @returns
#' A numeric vector or matrix of prices with missing values replaced
#' (where possible).
#'
#' @seealso
#' [price_relative()] for making price relatives for the
#' same products over time.
#'
#' @references
#' IMF, ILO, OECD, Eurostat, UNECE, and World Bank. (2020).
#' *Consumer Price Index Manual: Concepts and Methods*.
#' International Monetary Fund.
#'
#' @examples
#' prices <- data.frame(
#'   price = c(1:7, NA),
#'   period = rep(1:2, each = 4),
#'   product = 1:4,
#'   ea = rep(letters[1:2], 4)
#' )
#'
#' impute_prices(prices, price ~ period + product, method = "carry-forward")
#'
#' impute_prices(
#'   prices,
#'   price ~ period + product,
#'   ea = ea,
#'   method = "overall-mean"
#' )
#'
#' # Can also be done with current price-back price formulation.
#' prices$back_price <- with(
#'   prices,
#'   price[back_period(period, product)]
#' )
#'
#' impute_prices(
#'   prices,
#'   cbind(price, back_price) ~ period + product,
#'   ea = ea,
#'   method = "overall-mean"
#' )
impute_prices <- function(x, ...) {
  if ("r" %in% ...names()) {
    warning("`r` is deprecated and will be removed; use `order` instead")
  }
  UseMethod("impute_prices")
}

#' @rdname impute_prices
#' @export
impute_prices.default <- function(x, ...) {
  impute_prices(as.numeric(x), ...)
}

#' @rdname impute_prices
#' @export
impute_prices.matrix <- function(
  x,
  period,
  product,
  ...,
  ea = NULL,
  weights = NULL,
  pias = NULL,
  order = c(0, 1),
  r = order,
  method = c("overall-mean", "carry-forward"),
  impute_rules = NULL
) {
  # This is mostly a combination of back_period() and aggregate()
  # it just does it period-by-period and keeps track of prices to impute.
  chkDots(...)
  method <- match.arg(method)
  if (not_finite_pair(r)) {
    stop("`r` must be a pair of finite numbers")
  }
  if (ncol(x) != 2L) {
    stop("`x` must be a two-column matrix")
  }
  period <- as.factor(period)
  product <- as.factor(product)
  attributes(product) <- NULL
  if (!is.null(ea)) {
    ea <- as.factor(ea)
  }
  if (!is.null(weights)) {
    weights <- as.numeric(weights)
  }

  if (different_length(x[, 1L], period, product, ea, weights)) {
    stop("input vectors must be the same length")
  }

  res <- split.data.frame(x, period)
  product <- split(product, period)
  if (duplicate_products(product)) {
    warning("there are duplicated period-product pairs")
  }
  if (!is.null(ea)) {
    ea <- split(ea, period)
  }
  if (!is.null(weights)) {
    weights <- split(weights, period)
  }
  if (!is.null(pias)) {
    pias <- as_aggregation_structure(pias)
  }
  for (t in seq_along(res)) {
    impute <- which(is.na(res[[t]][, 1L]))
    if (method == "overall-mean") {
      index <- elementary_index(
        res[[t]][, 1L] / res[[t]][, 2L],
        ea = ea[[t]],
        weights = weights[[t]],
        na.rm = TRUE,
        order = r[1L]
      )
      time(index) <- names(res)[t]
      if (!is.null(pias)) {
        index <- aggregate(
          index,
          pias,
          na.rm = TRUE,
          order = r[2L],
          impute_rules = impute_rules
        )
        pias <- update(pias, index, order = r[2L])
      }
      eas <- if (!is.null(ea)) {
        match(as.character(ea[[t]][impute]), index$levels)
      } else {
        1L
      }
      res[[t]][impute, 1L] <- as.numeric(index)[eas] * res[[t]][impute, 2L]
    } else {
      res[[t]][impute, 1L] <- res[[t]][impute, 2L]
    }
    # Move imputed prices to next-period back prices.
    if (t < length(res)) {
      impute2 <- which(is.na(res[[t + 1L]][, 2L]))
      matches <- match(
        product[[t + 1L]][impute2],
        product[[t]][impute],
        incomparables = NA
      )
      res[[t + 1L]][impute2, 2L] <- res[[t]][impute, 1L][matches]
    }
  }
  # unsplit() doesn't work nicely with matrices.
  split(x, period) <- res
  x
}

#' @rdname impute_prices
#' @export
impute_prices.numeric <- function(
  x,
  period,
  product,
  ...,
  ea = NULL,
  weights = NULL,
  pias = NULL,
  order = c(0, 1),
  r = order,
  method = c("overall-mean", "carry-forward", "carry-backward"),
  impute_rules = NULL
) {
  # This is mostly a combination of back_period() and aggregate()
  # it just does it period-by-period and keeps track of prices to impute.
  chkDots(...)
  method <- match.arg(method)
  if (not_finite_pair(r)) {
    stop("`r` must be a pair of finite numbers")
  }
  period <- as.factor(period)
  if (method == "carry-backward") {
    period <- factor(period, rev(levels(period)))
  }
  product <- as.factor(product)
  attributes(product) <- NULL
  if (!is.null(ea)) {
    ea <- as.factor(ea)
  }
  if (!is.null(weights)) {
    weights <- as.numeric(weights)
  }

  if (different_length(x, period, product, ea, weights)) {
    stop("input vectors must be the same length")
  }

  res <- split(x, period)
  product <- split(product, period)
  if (duplicate_products(product)) {
    warning("there are duplicated period-product pairs")
  }
  if (!is.null(ea)) {
    ea <- split(ea, period)
  }
  if (!is.null(weights)) {
    weights <- split(weights, period)
  }
  if (!is.null(pias)) {
    pias <- as_aggregation_structure(pias)
  }
  for (t in seq_along(res)[-1L]) {
    impute <- which(is.na(res[[t]]))
    if (method == "overall-mean") {
      matches <- match(product[[t]], product[[t - 1L]], incomparables = NA)
      back_price <- res[[t - 1L]][matches]

      index <- elementary_index(
        res[[t]] / back_price,
        ea = ea[[t]],
        weights = weights[[t]],
        na.rm = TRUE,
        order = r[1L]
      )
      time(index) <- names(res)[t]
      if (!is.null(pias)) {
        index <- aggregate(
          index,
          pias,
          na.rm = TRUE,
          order = r[2L],
          impute_rules = impute_rules
        )
        pias <- update(pias, index, order = r[2L])
      }
      eas <- if (!is.null(ea)) {
        match(as.character(ea[[t]][impute]), index$levels)
      } else {
        1L
      }
      res[[t]][impute] <- as.numeric(index)[eas] * back_price[impute]
    } else {
      matches <- match(
        product[[t]][impute],
        product[[t - 1L]],
        incomparables = NA
      )
      res[[t]][impute] <- res[[t - 1L]][matches]
    }
  }
  split(x, period) <- res
  x
}

#' @rdname impute_prices
#' @export
impute_prices.data.frame <- function(
  x,
  formula,
  ...,
  ea = NULL,
  weights = NULL
) {
  vars <- formula_vars(formula, x, 2L)
  ea <- eval(substitute(ea), x, parent.frame())
  weights <- eval(substitute(weights), x, parent.frame())

  impute_prices(
    vars[[1L]],
    period = vars[[2L]],
    product = vars[[3L]],
    ea = ea,
    weights = weights,
    ...
  )
}

#' @rdname impute_prices
#' @export
carry_forward <- function(x, ...) {
  warning("'carry_forward() is deprecated; use 'impute_prices()' instead")
  impute_prices(x, method = "carry-forward", ...)
}

#' @rdname impute_prices
#' @export
carry_backward <- function(x, ...) {
  warning("'carry_backward() is deprecated; use 'impute_prices()' instead")
  impute_prices(x, method = "carry-backward", ...)
}

#' @rdname impute_prices
#' @export
shadow_price <- function(x, ...) {
  warning("'shadow_price() is deprecated; use 'impute_prices()' instead")
  impute_prices(x, method = "overall-mean", ...)
}

---
title: "Decomposing price indexes"
format: 
  html:
    html-math-method: mathjax
    options:
      number_sections: true
vignette: >
  %\VignetteIndexEntry{Decomposing price indexes}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

*This vignette presents an extended version of Martin (2021), and shows how to use the tools in this package to make these decompositions.*

It is often useful to be able to decompose a price index into an additive or multiplicative form to evaluate how each input to the index affects its value. Following Balk (2008), a price index is said to admit an additive decomposition if there exist weights that allow it to be represented as an arithmetic mean of price relatives, and is said to admit a multiplicative decomposition if there are weights that allow it to be represented as a geometric mean. Switching prices for quantities gives the analogous statements for a quantity index, and nothing is lost by focusing on price indexes.

There are several well-known decompositions for the most common types of bilateral price indexes. Balk (2008, eq. 4.13) gives an additive decomposition for any index based on the geometric mean by transmuting the weights in the geometric mean with the logarithmic mean. This is the same decomposition derived by Reinsdorf et al. (2002, eq. 20) for the Törnqvist index. A similar approach yields a multiplicative decomposition for any index based on the arithmetic mean, again using the logarithmic mean (Balk, 2008, eq. 4.8). Combining these results gives additive and multiplicative decompositions for the Fisher index (Reinsdorf et al., 2002, sec. 6). The van IJzeren additive decomposition for the Fisher index (Balk, 2008, eq. 4.18) is an alternative that does not explicitly use the logarithmic mean. Each of these decompositions results in weights that are positive and sum to one, as required to represent an index as an arithmetic or geometric mean.[^1]

[^1]: There are decompositions with "weights" that do not sum to one, such as those for the Fisher index by Reinsdorf et al. (2002, sec. 2) and Hallerbach (2005), and for other superlative indexes by Diewert (2002). Balk (2008, eq. 4.28) gives the original multiplicative decomposition of the Fisher index, due to Vartia, for which the weights also do not sum to one. Webster and Tarnow-Mordi (2019) consider several decompositions for multilateral indexes, although again the "weights" do not sum to one.

I show how the additive and multiplicative decompositions for geometric, arithmetic, and Fisher indexes can be consolidated and made more general by switching out the logarithmic mean for the more general extended mean. The main result is a function that transmutes the weights in a generalized mean of a given order so that it can be represented as a generalized mean of any other order. This covers additive and multiplicative decompositions for indexes that do not belong to the arithmetic or geometric families, like harmonic indexes or the Lloyd-Moulton index, and allows both additive and multiplicative decompositions to be covered by a single equation, rather than treating them as different cases. Expressing a generalized-mean index as a generalized mean of any other order also allows for the decomposition of indexes that are nested generalized means, like the family of superlative quadratic mean indexes that includes the Fisher index or the AG mean index by Lent and Dorfman (2009). I demonstrate some useful properties of this general decomposition and show how it can be used to decompose indexes based on other types on means, including those that aggregate indexes according to a hierarchical structure, and decompose deflating a change in aggregate values with a price index. See Martin (2026) for an application of these methods to multilateral indexes. Throughout I show how the tools in the **piar** package can be used to decompose a price index.

## Decomposing generalized-mean indexes {#sec-decompose}

A natural extension to the decompositions of indexes based on the arithmetic and geometric means is to derive weights that transform an index based on a generalized mean of order $\rho$ into one based on a generalized mean of order $\varsigma$. To fix notation, let $\mathbf{r} = (r_{1}, r_{2}, \ldots, r_{n}) \in \mathbb{R}^{n}_{++}$ be a vector of price relatives for $n\geq2$ products and let $\mathbf{w} = (w_{1}, w_{2}, \ldots, w_{n}) \in \Delta^{n - 1}$ be the corresponding weights, where $\Delta^{n - 1} = \{\mathbf{w} \in \mathbb{R}_{+}^{n} | \sum_{i = 1}^{n} w_{i} = 1\}$ is the unit simplex. The goal is to find a vector-valued function
$$\mathbf{v}(\mathbf{r}, \mathbf{w}; \rho, \varsigma) = (v_{1}(\mathbf{r}, \mathbf{w}; \rho, \varsigma), v_{2}(\mathbf{r}, \mathbf{w}; \rho, \varsigma),\ldots, v_{n}(\mathbf{r}, \mathbf{w}; \rho, \varsigma))$$
mapping into $\Delta^{n - 1}$ such that
$$\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w}) \equiv \mathfrak{M}_{\varsigma}(\mathbf{r}, \mathbf{v}(\mathbf{r}, \mathbf{w}; \rho, \varsigma)),$$
where $\mathfrak{M}_{\rho}$ is the generalized mean
$$\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w}) = 
\begin{cases}
\left(\sum_{i = 1}^{n} w_{i} r_i^{\rho}\right)^{1 / \rho} & \text{if } \rho \neq 0 \\
\prod_{i = 1}^{n} r_{i}^{w_{i}} & \text{if } \rho = 0.
\end{cases}$$

Setting $\varsigma = 1$ then yields an additive decomposition for any index based on a generalized mean of order $\rho$, such that
$$\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w}) = \sum_{i=1}^{n} r_{i} v_i(\mathbf{r}, \mathbf{w}; \rho, 1),$$
and setting $\varsigma = 0$ yields a multiplicative decomposition, such that
$$\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w}) = \prod_{i=1}^{n} r_{i}^{ v_i(\mathbf{r}, \mathbf{w}; \rho, 0)}.$$
Note that any index that admits an additive decomposition can be used to derive percent-change contributions for each price relative, $v_i(\mathbf{r}, \mathbf{w}; \rho, 1) (r_i - 1)$, that sum up to $\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w}) - 1$, although the converse is not true and it is possible to decompose an index into percent-change contributions that do not allow it to be represented as an arithmetic mean. Note also that other types of decompositions can be had by setting $\varsigma$ to a value other than 0 or 1; for example, setting $\varsigma = -1$ gives the harmonic decomposition (Färe and Karagiannis, 2025),
$$\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w}) = 1 / \left(\sum_{i=1}^{n} v_i(\mathbf{r}, \mathbf{w}; \rho, -1) / r_{i}\right),$$
that will be useful in @sec:harmonic.

Balk (2008) and Reinsdorf et al. (2002) show how to derive $\mathbf{v}(\mathbf{r}, \mathbf{w}; \rho, \varsigma)$ when $\rho = 1$ and $\varsigma = 0$ (multiplicative decomposition of an arithmetic index)
$$v_{i}(\mathbf{r}, \mathbf{w}; 1, 0) = \frac{w_{i} \mathfrak{L}(r_{i}, \mathfrak{M}_{1}(\mathbf{r}, \mathbf{w}))}{\sum_{j=1}^{n} w_{j} \mathfrak{L}(r_{j}, \mathfrak{M}_{1}(\mathbf{r}, \mathbf{w}))}$$
and $\rho = 0$ and $\varsigma = 1$ (additive decomposition of a geometric index)
$$v_{i}(\mathbf{r}, \mathbf{w}; 0, 1) = \frac{w_{i} / \mathfrak{L}(r_{i}, \mathfrak{M}_{0}(\mathbf{r}, \mathbf{w}))}{\sum_{j=1}^{n} w_{j} / \mathfrak{L}(r_{j}, \mathfrak{M}_{0}(\mathbf{r}, \mathbf{w}))},$$
using the logarithmic mean
$$\mathfrak{L}(a, b) =
\begin{cases}
\frac{a - b}{\log(a / b)} & a \neq b\\
a & a = b.
\end{cases}$$

Generalizing these results follows from replacing the logarithmic mean with the more general extended mean (Bullen, 2003, p. 393), defined for any $a,b > 0$ as 
$$\mathfrak{E}_{\rho\varsigma}(a, b) =
\begin{cases}
\left(\frac{\varsigma(a^\rho - b^\rho)}{\rho(a^\varsigma - b^\varsigma)}\right)^{1 / (\rho - \varsigma)} & \rho \neq \varsigma, \rho \neq 0, \varsigma \neq 0, a \neq b \\
\left(\frac{a^\rho - b^\rho}{\rho\log(a / b)}\right)^{1 / \rho} & \rho \neq 0, \varsigma = 0, a \neq b \\
\left(\frac{a^\varsigma - b^\varsigma}{\varsigma\log(a / b)}\right)^{1 / \varsigma} & \rho = 0, \varsigma \neq 0, a \neq b \\
\frac{1}{\exp(1 / \rho)} \left(\frac{a^{a^\rho}}{b^{b^\rho}}\right)^{1 / (a^\rho - b^\rho)} & \rho = \varsigma  \neq 0, a \neq b \\
\sqrt{ab} & \rho = \varsigma = 0, a \neq b \\
a & a = b.
\end{cases}$$
The extended mean reduces to the logarithmic mean when either $\rho = 0$ and $\varsigma = 1$, or $\rho = 1$ and $\varsigma = 0$. But using the extended mean in place of the logarithmic mean allows for decompositions of indexes based on other types of means, like harmonic indexes ($\rho = -1$) and the Lloyd-Moulton index ($\rho = 1 - \sigma$, where $\sigma$ is an elasticity of substitution).

The key to transforming the weights in a generalized mean of order $\rho$ into the weights for a generalized mean of order $\varsigma$ comes from noting that the extended mean is always strictly positive and satisfies the identity
$$\begin{align}
\sum_{i=1}^{n} w_{i} \mathfrak{E}_{\rho\varsigma}(r_i, \mathfrak{M}_{\rho} (\mathbf{r}, \mathbf{w}))^{\rho - \varsigma} \varepsilon_{i}(\mathbf{r}, \mathbf{w}; \rho, \varsigma) \equiv 0,\label{eq:identity}
\end{align}$$
where
$$\varepsilon_{i}(\mathbf{r}, \mathbf{w}; \rho, \varsigma) = 
\begin{cases}
r_i^\varsigma - \mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w})^\varsigma & \text{if } \varsigma \neq 0, \\
\log(r_i) - \log(\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w})) & \text{if } \varsigma = 0.
\end{cases}$$
@eqn:identity uses the extended mean to keep the weighted deviation from the mean constant for each price relative (up to a common factor of proportionality) when changing the order of the mean from $\rho$ to $\varsigma$, without changing the value of the mean. Rearranging then gives that
$$\begin{align}
v_{i}(\mathbf{r}, \mathbf{w}; \rho, \varsigma) =  w_{i} \mathfrak{E}_{\rho\varsigma}(r_i, \mathfrak{M}_{\rho} (\mathbf{r}, \mathbf{w}))^{\rho - \varsigma} \Bigg/ \sum_{j=1}^{n} w_{j} \mathfrak{E}_{\rho\varsigma}(r_j, \mathfrak{M}_{\rho} (\mathbf{r}, \mathbf{w}))^{\rho - \varsigma}\label{eq:res1}
\end{align}$$
is a suitable function to find weights that turn an index based on a generalized mean of order $\rho$ into one based on a generalized mean of order $\varsigma$.

The function given by @eqn:res1 takes on all existing decompositions that I know of as special cases. Setting $\rho = 0$ and $\varsigma = 1$, or $\rho = 1$ and $\varsigma = 0$, gives the special cases by Balk (2008) and Reinsdorf et al. (2002) for decomposing indexes based on arithmetic and geometric means (because the extended mean reduces to the logarithmic mean). Similarly, because $\mathfrak{E}_{\rho\varsigma}(r_i, \mathfrak{M}_{\rho} (\mathbf{r}, \mathbf{w}))^{\rho - \varsigma} \equiv \left(\mathfrak{M}_{\rho} (\mathbf{r}, \mathbf{w}) r_{i}\right)^\rho$ when $\rho = -1$ and $\varsigma = 1$, or $\rho = 1$ and $\varsigma = -1$, setting $\rho = -1$ and $\varsigma = 1$ reduces @eqn:res1 to 
$$v_{i}(\mathbf{r}, \mathbf{w}; -1, 1) = \frac{w_{i} / r_{i}}{\sum_{i=1}^{n}w_{i} / r_{i}} ;$$
if $\mathbf{w}$ is a vector of current-period expenditure/revenue shares then these are the hybrid weights that allow a Paasche index to be calculated as an arithmetic mean of price relatives. Setting $\rho = 1$ and $\varsigma = -1$ reduces @eqn:res1 to
$$v_{i}(\mathbf{r}, \mathbf{w}; 1, -1) = \frac{w_{i} r_{i}}{\sum_{i=1}^{n}w_{i} r_{i}}.$$
If $\mathbf{w}$ is a vector of base-period expenditure/revenue shares then these are the hybrid weights that allow a Laspeyres index to be calculated as a harmonic mean of price relatives. As should be expected, the weights are unchanged if $\rho = \varsigma$ or each element of $\mathbf{r}$ takes on the same value.

### Transitivity, monotonicity, and uniqueness

The extended mean has two properties that make it useful for decomposing price indexes. First, the order of the extended mean has a transitivity property under multiplication: $\mathfrak{E}_{\rho\varsigma}(a, b)^{\rho - \varsigma} = \mathfrak{E}_{\rho\tau}(a, b)^{\rho - \tau}\mathfrak{E}_{\tau\varsigma}(a, b)^{\tau - \varsigma}$. This means the function in @eqn:res1 is transitivity, such that transmuting the weights to turn a generalized mean of order $\rho$ into a generalized mean of order $\tau$, and then transmuting these weights again to turn a generalized mean of order $\tau$ into a generalized mean of order $\varsigma$, is the same as finding weights to make a generalized mean of order $\rho$ into one of $\varsigma$. That is,
$$\mathbf{v}(\mathbf{r}, \mathbf{v}(\mathbf{r}, \mathbf{w}; \rho, \tau); \tau, \varsigma) \equiv \mathbf{v}(\mathbf{r}, \mathbf{w}; \rho, \varsigma).$$
As will be seen later, not all decompositions have this property.

Second, the extended mean is a strictly increasing function in both arguments (Bullen, 2003, p. 395). This means that the function in @eqn:res1 has a monotonicity property whereby the transmuted weights increase (decrease) for large (small) price relatives if and only if $\rho > \varsigma$. That is, assuming $\mathbf{r}$ is ordered from smallest to largest (and does not contain all the same value), if $\rho > \varsigma$ then there is a pair of integer $k,l$, with $k\leq l$, such that $v_{i}(\mathbf{r}, \mathbf{w}; \rho, \varsigma) \geq w_{i}$ for $i\geq l$ and $v_{i}(\mathbf{r}, \mathbf{w}; \rho, \varsigma) \leq w_{i}$ for $i\leq k$, with these equalities reversed if $\rho < \varsigma$.[^2] Again, not all decompositions satisfy this property.

[^2]: These are weak inequalities to account for the case that some weights are 0. If the vector if weights is strictly positive then the inequalities are strict.

The decomposition given by @eqn:res1 is unique when $n = 2$ and $r_{1} \neq r_{2}$, and is the only such function that always returns $\mathbf{w}$ when $r_{1} = r_{2}$. But @eqn:res1 is necessarily not unique when $n\geq3$, and there are infinitely many ways to decompose an index based on the generalized mean.[^3] Nonetheless, the limited case of uniqueness will be useful in @sec:superlative.

[^3]: Take, for example, (1/2, 1, 3/2) with weights (1/6, 1/3, 1/2). The weighted harmonic mean is 1, which can be computed as an arithmetic mean with weights (1/3, 1/3, 1/3) or (1/4, 1/2, 1/4). Hence there are multiple additive decompositions for the harmonic mean.

### Numerical example

Consider the following numerical example for decomposing a quadratic-mean index (von der Lippe, 2007, p. 61) (i.e., a producer price index with an elasticity of substitution of -1) made using the price and quantity data from Balk (2008, tab. 3.1--3.2). The `transmute_weights()` function parameterizes the decomposition in @eqn:res1 to make a function that decomposes any generalized-mean index.

```{r}
library(piar)

p2 <- c(1.2, 3, 1.3, 0.7, 1.4, 0.8)
p1 <- 1
q2 <- c(0.8, 0.9, 1.9, 1.3, 4.7, 0.6)
q1 <- c(1, 1, 2, 1, 4.5, 0.5)

rel <- p2 / p1

s1 <- scale_weights(p1 * q1)
s2 <- scale_weights(p2 * q2)

v <- transmute_weights(rel, s1, order = 2)

all.equal(gmean(rel, s1, order = 2), gmean(rel, v))
```

Because the quadratic mean is larger than the arithmetic mean, the weight from the smaller price relatives is transferred to the largest price relative.

```{r}
(v - s1)[order(rel)]
```

It is also readily seen that turning the arithmetic decomposition into a geometric one is the same as directly constructing the multiplicative decomposition.

```{r}
all.equal(
  transmute_weights(rel, v, order = 1, to = 0),
  transmute_weights(rel, s1, order = 2, to = 0)
)
```

## Extensions

### Decomposing superlative indexes {#sec:superlative}

The additive and multiplicative decompositions for the Fisher index by Reinsdorf et al. (2002, sec. 6) can be generalized in the same way as the decompositions for the arithmetic and geometric indexes by noting that the Fisher index is simply a nested generalized mean of indexes based on the generalized mean. For a pair of generalized means $\mathfrak{N}_{\rho_{1}\rho_{2}}(\mathbf{r}, \mathbf{w}_{1}, \mathbf{w}_{2})=\left(\mathfrak{M}_{\rho_1}(\mathbf{r}, \mathbf{w}_1), \mathfrak{M}_{\rho_2}(\mathbf{r}, \mathbf{w}_2)\right)$ mapping into $\mathbb{R}_2^{++}$ with weights $\mathbf{\omega}=(\omega_1, \omega_2) \in \Delta^1$, an index based on nested generalized means is written as
$$\begin{align}
\mathfrak{M}_{\rho}\left(\mathfrak{N}_{\rho_{1}\rho_{2}}(\mathbf{r}, \mathbf{w}_{1}, \mathbf{w}_{2}), \mathbf{\omega}\right). \label{eq:superlative}
\end{align}$$

The general family of superlative quadratic mean indexes of order $\tau$ comes from setting $\rho = 0$, $\rho_1 = \tau / 2$, and $\rho_2 = -\tau / 2$ when $\omega_1 = \omega_2 = 1 / 2$, $\mathbf{w}_1$ are base-period expenditure/revenue shares, and $\mathbf{w}_2$ are current-period expenditure/revenue shares. In particular, setting $\tau = 2$ gives the Fisher index and setting $\tau = 1$ gives the implicit Walsh index. But @eqn:superlative covers other types of indexes as well; for example, setting each element of $\mathbf{w}_1$ and $\mathbf{w}_2$ to $1 / n$ when $\tau = 2$ gives the Carruthers-Sellwood-Ward-Dalén index that serves as an estimator for the Fisher index, whereas $\tau = 1$ gives the Balk-Walsh index. Setting $\rho = -1$ gives the harmonic analogue of the Fisher index, which is not a superlative quadratic mean indexes of order $\tau$. Finally, setting $\rho = \rho_{1} = \rho_{2}$ and $\mathbf{w}_{1} = \mathbf{w}_{2}$ gives an index based on a generalized mean of order $\rho$, so that the decomposition of an index based on the generalized mean is a special case of the decomposition for @eqn:superlative.

An index of the form in @eqn:superlative can be decomposed into an index based on the generalized mean of order $\rho$ using the weights in @eqn:res1, as it can be written as the generalized mean
$$\mathfrak{M}_{\rho}(\mathbf{r}, \omega_1 \mathbf{v}(\mathbf{r}, \mathbf{w}_1; \rho_1, \rho) + \omega_2\mathbf{v}(\mathbf{r}, \mathbf{w}_2; \rho_2, \rho)).$$
The transformation in @eqn:res1 then applies as before, just replacing $\mathbf{w}$ with the more complicated weights $\omega_1 \mathbf{v}(\mathbf{r}, \mathbf{w}_1; \rho_1, \rho) + \omega_2\mathbf{v}(\mathbf{r}, \mathbf{w}_2, \rho_2, \rho)$, which can be written as
$$\begin{align}
\mathbf{v}(\mathbf{r}, \omega_1 \mathbf{v}(\mathbf{r}, \mathbf{w}_1; \rho_1, \rho) + \omega_2\mathbf{v}(\mathbf{r}, \mathbf{w}_2; \rho_2, \rho); \rho, \varsigma). \label{eq:res2}
\end{align}$$
The idea is to transmute the weights for both inner means to be of the same order as the outer mean ($\rho$) so that they can be added together, and then transmute these weights to represent the outer generalized mean as a mean of order $\varsigma$. Note that this means the decomposition satisfies the same transitivity property as @eqn:res1 with respect to $\rho$ and $\varsigma$.

Continuing with the previous example, the `transmute_weights2()` function parameterizes the decomposition in @eqn:res2 and can be used to decompose the Fisher index.

```{r}
v1 <- transmute_weights2(rel, list(s1, s2))

all.equal(nested_gmean(rel, list(s1, s2)), gmean(rel, v1))

all.equal(
  v1,
  transmute_weights(
    rel,
    transmute_weights2(rel, list(s1, s2), to = 2),
    order = 2
  )
)
```

An alternative approach to decompose @eqn:superlative is to generalize the (additive) van IJzeren decomposition for the Fisher index, which can be written as
$$\begin{align}
v_{1}(\mathfrak{N}_{\rho_{1}\rho_{2}}(\mathbf{r}, \mathbf{w}_{1}, \mathbf{w}_{2}), \mathbf{\omega}; \rho, \varsigma) \mathbf{v}(\mathbf{r}, \mathbf{w}_1; \rho_1, \varsigma) + v_{2}(\mathfrak{N}_{\rho_{1}\rho_{2}}(\mathbf{r}, \mathbf{w}_{1}, \mathbf{w}_{2}), \mathbf{\omega}; \rho, \varsigma) \mathbf{v}(\mathbf{r}, \mathbf{w}_2; \rho_2, \varsigma).\label{eq:res3}
\end{align}$$
Note that for a Fisher index, if $\varsigma=1$ then $\mathbf{v}(\mathbf{r}, \mathbf{w}_1; \rho_1, \varsigma)$ is a vector of base-period expenditure/revenue shares, $\mathbf{v}(\mathbf{r}, \mathbf{w}_2; \rho_2, \varsigma)$ is a vector of hybrid Paasche weights, and $\mathbf{v}(\mathfrak{N}_{\rho_{1}\rho_{2}}(\mathbf{r}, \mathbf{w}_{1}, \mathbf{w}_{2}), \mathbf{\omega}; \rho, \varsigma)$ are the unique weights that decompose the geometric mean of the Laspeyres and Paasche indexes, which equals the van IJzeren decomposition. The idea here is to transmute the weights for both the inner and outer generalized means so that they are means of order $\varsigma$, then take the product of these weights. @eqn:res2 and @eqn:res3 generally give different decompositions, as the latter is not transitive, but reduce to @eqn:res1 when $\rho = \rho_{1} = \rho_{2}$ and $\mathbf{w}_{1} = \mathbf{w}_{2}$.

These two approaches for decomposing @eqn:superlative can be generalized with a pivot value $\psi$ that acts as the link between the inner and outer generalized means. More generally, an index of the form in @eqn:superlative can be decomposed into a generalized mean of order $\rho$ with weights
$$\mathbf{v}(\mathbf{r},v_{1}(\mathfrak{N}_{\rho_{1}\rho_{2}}(\mathbf{r}, \mathbf{w}_{1}, \mathbf{w}_{2}), \mathbf{\omega}; \rho, \psi) \mathbf{v}(\mathbf{r}, \mathbf{w}_1; \rho_1, \psi) + v_{2}(\mathfrak{N}_{\rho_{1}\rho_{2}}(\mathbf{r}, \mathbf{w}_{1}, \mathbf{w}_{2}), \mathbf{\omega}; \rho, \psi) \mathbf{v}(\mathbf{r}, \mathbf{w}_2; \rho_2, \psi); \psi, \varsigma).$$
Setting $\psi=\rho$ gives @eqn:res2, whereas setting $\psi=\varsigma$ gives @eqn:res3.

```{r}
v2 <- transmute_weights2(rel, list(s1, s2), pivot = 1)

all.equal(nested_gmean(rel, list(s1, s2)), gmean(rel, v2))

all.equal(
  v2,
  transmute_weights(
    rel,
    transmute_weights2(rel, list(s1, s2), to = 2, pivot = 1),
    order = 2
  )
)
```

Overall the difference between the two approaches tends to be small.

```{r}
summary(v1 - v2)
```

### Decomposing aggregated indexes

The decomposition in @eqn:res3 can be extended to get a decomposition for an index constructed using the usual two-step procedure to make a price index. If price relatives are partitioned into $k$ groups, and an index is made from $k$ sub-indexes $\mathfrak{N}_{\rho_{1}\ldots\rho_{k}}(\mathbf{r}, \mathbf{w}_{1}, \ldots, \mathbf{w}_{k}) = \left(\mathfrak{M}_{\rho_1}(\mathbf{r}_{1}, \mathbf{w}_{1}), \ldots, \mathfrak{M}_{\rho_k}(\mathbf{r}_{k}, \mathbf{w}_{k})\right)$ with weights $\mathbf{\omega}=(\omega_{1},\ldots,\omega_{k})$ as a generalized mean of order $\rho$, then
$$v_{i}(\mathfrak{N}_{\rho_{1}\ldots\rho_{k}}(\mathbf{r}, \mathbf{w}_{1}, \ldots, \mathbf{w}_{k}), \mathbf{\omega}; \rho, \varsigma) \mathbf{v}(\mathbf{r}_{i}, \mathbf{w}_i; \rho_{i}, \varsigma)$$
decomposes how the components of the $i$-th sub-index contribute towards the total index. Concatenating each of these vectors of weights along with the vector of price relatives represents the nested mean over the partitioning of price relatives as a generalized mean of order $\varsigma$.

```{r}
group <- rep(c("a", "b"), each = 3)

s1_by_group <- split(s1, group)
rel_by_group <- split(rel, group)

index_a <- gmean(rel_by_group$a, s1_by_group$a, order = 2)
index_b <- gmean(rel_by_group$b, order = 0)

gmean(c(index_a, index_b), sapply(s1_by_group, sum), order = 2)

decomp_a <- transmute_weights(rel_by_group$a, s1_by_group$a, order = 2)
decomp_b <- transmute_weights(rel_by_group$b, order = 0)

v <- Map(
  `*`,
  transmute_weights(c(index_a, index_b), sapply(s1_by_group, sum), order = 2),
  list(decomp_a, decomp_b)
) |>
  unlist()

gmean(rel, v)
```

This approach to decomposing aggregate indexes also has the benefit that contributions for different groups can be made using different methods and combined in a way that decomposes the aggregated index. For example, the percent-change decomposition by Diewert (2002) can be used to decompose the Fisher index for one group, whereas percent-change contributions from the additive decomposition of a Jevons index can be used to decompose another group, and combined in a consistent way to get percent-change contributions for the aggregated index.

### Decomposing deflators {#sec:harmonic}

The decomposition in @eqn:res1 can be used to construct a decomposition for how a price index deflates a change in aggregate value over time to get a quantity index. Letting $V$ be the change in total value between two periods, the implicit quantity index for a price index based on the generalized mean of order $\rho$ is

$$
\frac{V}{\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w})}.
$$

Using @eqn:res1 to find weights to represent $\mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w})$ as a harmonic mean means that
$$V/ \mathfrak{M}_{\rho}(\mathbf{r}, \mathbf{w}) \equiv \mathfrak{M}_{\varsigma}(V / \mathbf{r}, \mathbf{v}(\mathbf{r}, \mathbf{w}; \rho, -\varsigma)),$$
where $V / \mathbf{r} = \left(V / r_{1}, \ldots, V / r_{n}\right)$.

Setting $\varsigma=1$ then gives an additive decomposition for how each price relative acts to deflate the change in total value over time, while setting $\varsigma=0$ gives a multiplicative decomposition. The same holds true when deflating with an index based on nested means, using either @eqn:res2 or @eqn:res3.

```{r}
V <- sum(p2 * q2) / sum(p1 * q1)

v <- transmute_weights2(rel, list(s1, s2), to = -1)

all.equal(
  gmean(V / rel, v),
  nested_gmean(q2 / q1, list(s1, s2))
)

V / rel * v
```

### Decomposing contra-harmonic indexes

von der Lippe (2015) discusses using the contra-harmonic mean (or antiharmonic as he calls it), a special case of the Lehmer mean (Bullen, 2003, p. 245) (of order 2), to make price indexes.

```{r}
contraharmonic_mean <- function(x, weights, order) {
  gmean(x, weights * x)
}

# Arithmetic hybrid index
all.equal(
  gmean(p2 / p1, p2 * q1),
  contraharmonic_mean(p2 / p1, p1 * q1)
)

# Palgrave index
all.equal(
  gmean(p2 / p1, p2 * q2),
  contraharmonic_mean(p2 / p1, p1 * q2)
)
```

The Lehmer mean of order $\rho$ is simply the arithmetic mean of the with weights
$$\left(w_{1}r_{1}^{\rho - 1}, \ldots, w_{n}r_{n}^{\rho - 1}\right) / \sum_{i = 1}^{n} w_{i} r_{i}^{\rho - 1},$$
and so indexes based on these means can be decomposed with @eqn:res1 by using the these weights instead of $\mathbf{w}$.

## References {-}

Balk, B. M. (2008). _Price and Quantity Index Numbers_. Cambridge University Press. https://doi.org/10.1017/CBO9780511720758.

Bullen, P. S. (2003). _Handbook of Means and their Inequalities_. Springer Science+Business Media. https://doi.org/10.1007/978-94-017-0399-4.

Diewert, W. E. (2002). The quadratic approximation lemma and decompositions of superlative indexes. _Journal of Economic and Social Measurement_, 28(1-2): 63--88. https://doi.org/10.3233/JEM-2003-0200.

Färe, R., Karagiannis, G. (2025). An alternative derivation of the van IJzeren weights and some new results for the additive decomposition of Fisher indices. _Journal of Productivity Analysis_, 64(3): 363--377. https://doi.org/10.1007/s11123-025-00774-2.

Hallerbach, W. G. (2005). An alternative decomposition of the Fisher index. _Economics Letters_, 86(2): 147--152. https://doi.org/10.1016/j.econlet.2004.07.008.

Lent, J., Dorfman, A. H. (2009). Using a weighted average of base period price indexes to approximate a superlative index. _Journal of Official Statistics_, 25(1): 149--149.

Martin, S. (2021). A note on generalized decompositions for price indexes. _Prices Analytical Series_, Statistics Canada.

Martin, S. (2026). A note on the additive decomposition of GEKS indexes. _Journal of Official Statistics_, 0282423X261451318. https://doi.org/10.1177/0282423X261451318.

Reinsdorf, M. B., Diewert, W. E., Ehemann, C. (2002). Additive decompositions for Fisher, Törnqvist and geometric mean indexes. _Journal of Economic and Social Measurement_, 28(1--2): 51--61. https://doi.org/10.3233/JEM-2003-0194.

von der Lippe, P. (2007). _Index Theory and Price Statistics_. Peter Lang. https://doi.org/10.3726/978-3-653-01120-3.

von der Lippe, P. (2015). _Generalized Statistical Means and New Price Index Formulas, Notes on some unexplored index formulas, their interpretations and generalizations_. Munich Personal RePEc Archive paper no. 64952.

Webster, M. B., Tarnow-Mordi, R. C. (2019). _Decomposing multilateral price indexes into the contributions of individual commodities_. Journal of Official Statistics, 35(2): 461--486. https://doi.org/10.2478/jos-2019-0020.

---
title: "Adjusting annual weights"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Adjusting annual weights}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Making a monthly or quarterly Lowe index with annual expenditure or revenue
weights requires adjusting these weights so that the implicit annual quantity vector
is used as the fixed basket when making the index with the usual two-step procedure. This
can be done by price updating the annual weights to the base period of the index,
and it serves as a good example of extending the functions in this package.

Let's start by making some annual weights and quarterly indexes for a year.

```{r}
set.seed(54321)

library(piar)

# Make an aggregation structure.
#                  1
#      |-----------|-----------|
#      11          12          13
#  |---+---|   |---+---|   |---+---|
# 111     121 121     122 131     132

pias <- data.frame(
  level1 = rep(1, 12),
  level2 = rep(c(11, 12, 13), each = 4),
  level3 = rep(c(111, 112, 121, 122, 131, 132), each = 2),
  ea = sprintf("B%02d", 1:12),
  weight = 1:12
) |>
  as_aggregation_structure()

pias

# Make elementary indexes over 4 quarters.
elementals <- matrix(
  runif(12 * 4, 0.4, 1.2),
  nrow = 12,
  dimnames = list(sprintf("B%02d", 1:12), paste0("Q", 1:4))
) |>
  as_index()

elementals
```

Adjusting the weights is simple when there are no missing elementary indexes: the weight
for each elementary aggregate is just divided by the average (fixed-base) index
for each quarter.

```{r}
weights(pias) / rowMeans(as.matrix(chain(elementals)))
```

The procedure is more complicated with missing elementary indexes
as reaggregating these indexes with the newly adjusted weights will generally result
in different imputations for the missing elementary indexes, which in turn gives a
different adjustment for the weights. In practice this procedure is done a few times
until the index values converge to a fixed point. The following function shows how
to do this adjustment using the tools in this package.

```{r}
# Function to adjust annual weights.
adjust_weights <- function(
  index,
  pias,
  tol = .Machine$double.eps^0.5,
  max_iter = 100
) {
  adj_pias <- pias
  for (i in seq_len(max_iter)) {
    # Parentally impute missing elementary indexes.
    agg_index <- aggregate(index, adj_pias, na.rm = TRUE, contrib = FALSE)
    elementals <- chain(agg_index[levels(pias)[[nlevels(pias)]]])
    # Compute annual elementary indexes.
    pb <- rowMeans(as.matrix(elementals))
    # Stop if average price-update weights are within tolerance of original
    # weights; adjust otherwise.
    if (max(abs(pb * weights(adj_pias) - weights(pias))) < tol) {
      message(gettextf("Converged after %d iterations", i - 1))
      return(adj_pias)
    } else {
      weights(adj_pias) <- weights(pias) / pb
    }
  }
  warning("weights adjustment did not converge")
  adj_pias
}
```

```{r}
elementals[11:12] <- NA

adjust_weights(elementals, pias)
```

---
title: "Index-number formulas"
vignette: >
  %\VignetteIndexEntry{Index-number formulas}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

Price indexes based on a generalized mean of price relatives constitute
a large family of bilateral index-number formulas that are consistent in
aggregation, and any index based on the generalized mean can be used to
make and aggregate elementary indexes. See the last section for a comprehensive list.
To see how to make superlative
indexes as well, let's start with a simple dataset of prices and
quantities for two businesses over three periods. We'll be making
extensive use of the lower-level math functions in this package; see the help pages for these functions to get more detail and
`vignette("decomposing-indexes")` for the theoretical background.

```{r}
library(piar)

prices <- data.frame(
  period = rep(1:3, each = 6),
  product = paste0("P", 1:6),
  business = rep(c("B1", "B2"), each = 3),
  price = 1:18,
  quantity = 18:1
)

prices[c("back_price", "back_quantity")] <-
  prices[back_period(prices$period, prices$product), c("price", "quantity")]

head(prices)
```

## Basic indexes

As seen in `vignette("piar")`, by default the `elementary_index()`
function calculates a Jevons index (equally-weighted geometric mean of
price relatives). Although this is the standard index-number formula for
making elementary indexes, many other types of index-numbers are
possible. Among the unweighted index-number formulas, the Carli index
(equally-weighted arithmetic mean of price relatives) is the historical
competitor to the Jevons, and requires specifying the order of the
generalized mean `order = 1` when calling `elementary_index()`. An order of 1
corresponds to an arithmetic mean.

```{r}
prices |>
  elementary_index(price / back_price ~ period + business, order = 1)
```

The Coggeshall index (equally-weighted harmonic mean of price relatives)
is another competitor to the Jevons, but is seldom used in practice.
Despite it being more exotic, it is just as easy to make by specifying
an `order = -1`.

```{r}
prices |>
  elementary_index(price / back_price ~ period + business, order = -1)
```

Weights can be added to make, for example, elementary indexes using the
geometric Laspeyres formula.

```{r}
prices |>
  elementary_index(
    price / back_price ~ period + business,
    weights = back_price * back_quantity
  )
```

The type of mean used to aggregate elementary indexes can be controlled
in the same way in the call to `aggregate()`. The default makes an
arithmetic index, but any type of generalized-mean index is possible.

## Superlative indexes

Many superlative indexes can be made by supplying unequal and
time-varying weights when making the elementary indexes, usually from
information about quantities. The Törnqvist index is a popular superlative
index-number formula, using average period-over-period value shares as
the weights in a geometric mean. As `elementary_index()` makes a
geometric index by default, all that is needed to make a Törnqvist index
is the weights.

```{r}
period_by_business <- interaction(prices$period, prices$business)

tornqvist_weights <- split(prices, ~period_by_business) |>
  lapply(\(df) {
    0.5 * scale_weights(df$price * df$quantity) +
      0.5 * scale_weights(df$back_price * df$back_quantity)
  }) |>
  unsplit(period_by_business)

prices |>
  elementary_index(
    price / back_price ~ period + business,
    weights = tornqvist_weights
  )
```

Making a Fisher index is more complex because it does not belong to the
generalized-mean family. Despite this, it is possible to make weights to
represent a Fisher index as a generalized mean of any order.

```{r}
fisher_weights <- split(prices, ~period_by_business) |>
  lapply(\(df) {
    transmute_weights2(
      df$price / df$back_price,
      list(df$back_price * df$back_quantity, df$price * df$quantity),
      to = 0
    )
  }) |>
  unsplit(period_by_business)

prices |>
  elementary_index(
    price / back_price ~ period + business,
    weights = fisher_weights
  )
```

Aggregating with a superlative index is more complex, and is the subject
of `vignette("superlative-aggregation")`.

## Product contributions

As shown in `vignette("contributions")`, supplying `contrib = TRUE` in the call
to `elementary_index()` makes percent-change product contributions for each index
value. The method used in this package is flexible and works well for a variety 
of different index-number formulas, but does not include all methods found in the
literature. (See the help page for `elementary_index()`, and the references therein,
for more detail about the exact methods.)

Let's extend the previous example by calculating product contributions for the
Fisher index using the default method.

```{r}
fisher_index <- prices |>
  elementary_index(
    price / back_price ~ period + business,
    weights = fisher_weights,
    contrib = TRUE
  )

contrib(fisher_index, "B1")
```

We can change the method used to make contributions for, say, business `B1` by making
a function to compute contributions according to a different method

```{r}
diewert_contributions <- function(p1, p0, q1, q0) {
  rel <- p1 / p0
  v0 <- scale_weights(p0 * q0)
  v1 <- scale_weights(p0 * q1)
  laspeyres <- gmean(rel, v0)
  fisher <- nested_gmean(rel, list(v0, v1), order = c(1, 1))

  (v0 + laspeyres * v1) / (1 + fisher) * (rel - 1)
}
```

and using this function to replace the contributions for business `B1`.

```{r}
contrib(fisher_index, "B1") <- subset(prices, business == "B1") |>
  split(~period) |>
  sapply(
    \(df) {
      diewert_contributions(
        df$price,
        df$back_price,
        df$quantity,
        df$back_quantity
      )
    }
  )

contrib(fisher_index, "B1")
```

Aggregating the elementary indexes will then consistently aggregate the contributions
for both businesses, even though they use different methods.

## Indexes based on generalized means {#sec-appendix}

```{r}
#| eval: false
# ---- Arithmetic mean ----
# Carli
carli <- \(p1, p0, q1, q0) gmean(p1 / p0)

# Dutot
dutot <- \(p1, p0, q1, q0) gmean(p1 / p0, p0)

# Laspeyres
laspeyres <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * q0)

# Palgrave
palgrave <- \(p1, p0, q1, q0) gmean(p1 / p0, p1 * q1)

# Walsh-1
walsh1 <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * sqrt(q0 * q1))

# Marshall-Edgeworth
marshall_edgeworth <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * (q0 * q1))

# Geay-Khamis
geary_khamis <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 / (1 / q0 + 1 / q1))

# Hybrid CSWD
hybrid_cswd <- \(p1, p0, q1, q0) gmean(p1 / p0, sqrt(p0 / p1))

# Martini
martini <- \(p1, p0, q1, q0, a) gmean(p1 / p0, p0 * q0 * (q1 / p0)^a)

# ---- Geometric mean ----
# Jevons
jevons <- \(p1, p0, q1, q0) gmean(p1 / p0, order = 0)

# Geometric Laspeyres (Jöhr)
geo_laspeyres <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * q0, order = 0)

# Geometric Paasche
geo_paasche <- \(p1, p0, q1, q0) gmean(p1 / p0, p1 * q1, order = 0)

# Walsh-2
walsh2 <- \(p1, p0, q1, q0) gmean(p1 / p0, sqrt(p0 * p1 * q0 * q1), order = 0)

# Theil
theil <- \(p1, p0, q1, q0) {
  w0 <- scale_weights(p0 * q0)
  w1 <- scale_weights(p1 * q1)
  gmean(p1 / p0, ((w0 + w1) / 2 * w0 * w1)^(1 / 3), order = 0)
}

# Rao
rao <- \(p1, p0, q1, q0) {
  w0 <- scale_weights(p0 * q0)
  w1 <- scale_weights(p1 * q1)
  gmean(p1 / p0, w0 * w1 / (w0 + w1), order = 0)
}

# Sato-Vartia
sato_vartia <- function(p1, p0, q1, q0) {
  v0 <- scale_weights(p0 * q0)
  v1 <- scale_weights(p1 * q1)
  gmean(p1 / p0, emean(v0, v1), order = 0)
}

# ---- Harmonic mean ----
# Coggeshall
coggeshall <- \(p1, p0, q1, q0) gmean(p1 / p0, order = -1)

# Paasche
paasche <- \(p1, p0, q1, q0) gmean(p1 / p0, p1 * q1, order = -1)

# Harmonic Laspeyres
harm_laspeyres <- \(p1, p0, q1, q0) gmean(p1 / p0, p0 * q0, order = -1)

# ---- Generalized mean ----
# Lloyd-Moulton
lloyd_moulton <- \(p1, p0, q1, q0, sigma) {
  gmean(p1 / p0, p0 * q0, order = 1 - sigma)
}

# ---- Nested means ----
# Drobisch (Sidgwick)
drobisch <- \(p1, p0, q1, q0) {
  nested_gmean(p1 / p0, list(p0 * q0, p1 * q1), outer_order = 1)
}

# Unnamed
unnamed <- \(p1, p0, q1, q0) {
  nested_gmean(
    p1 / p0,
    list(p0 * q0, p1 * q1),
    order = c(1, 1),
    outer_order = 1
  )
}

# Törnqvist-Theil
tornqvist <- \(p1, p0, q1, q0) {
  nested_gmean(
    p1 / p0,
    list(p0 * q0, p1 * q1),
    order = c(0, 0)
  )
}

# Fisher
fisher <- \(p1, p0, q1, q0) nested_gmean(p1 / p0, list(p0 * q0, p1 * q1))

# CSWD
cswd <- \(p1, p0, q1, q0) nested_gmean(p1 / p0)

# Balk-Walsh
balk_walsh <- \(p1, p0, q1, q0) nested_gmean(p1 / p0, order = c(0.5, -0.5))

# Geometric AG mean
ag_mean <- \(p1, p0, q1, q0, elasticity) {
  nested_gmean(
    p1 / p0,
    list(p0 * q0, p0 * q0),
    order = c(0, 1),
    outer_weights = c(elasticity, 1 - elasticity)
  )
}
```

---
title: "Contributions"
vignette: >
  %\VignetteIndexEntry{Contributions}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

## Product contributions

It's often convenient to decompose an index into the (additive)
contribution of each price relative, also known as the percent-change
contribution. This can be done with the same work flow used in
`vignette("piar")`, specifying `contrib = TRUE` when calling
`elementary_index()`. (See `vignette("decomposing-indexes")` for the underlying
theory.)

```{r}
library(piar)

# Make an aggregation structure.
ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

# Make elementary index with contributions.
elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(
    relative ~ period + business,
    product = product,
    na.rm = TRUE,
    contrib = TRUE
  )
```

As with index values, percent-change contributions for a given level of
the index can be extracted as a matrix.

```{r}
contrib(elementals, level = "B1")
```

Or as a data frame.

```{r}
contrib2DF(elementals, level = "B1")
```

Aggregating the elementary indexes automatically aggregates
percent-change contributions, so no extra steps are needed after the
elementary indexes are made.

```{r}
index <- aggregate(elementals, pias, na.rm = TRUE)

contrib(index)
```

## Index contributions

After an index has been calculated, it's often useful to compute the
contribution of higher-level indexes towards the total index. The
easiest way to do this with a collection of pre-computed index values is
to simply coerce them into an index object with the index values as
contributions and reaggregate with a restricted aggregation structure.

```{r}
index <- as_index(as.matrix(index), contrib = TRUE)
```

If the index values are already an index object, it's also possible to
directly replace the contributions with the `set_contrib_from_index()` function.
We can now cut the aggregation structure to keep only the top two levels and
reaggregate to get the contribution of the second-level indexes to the
top level index.

```{r}
set_contrib_from_index(index) |>
  aggregate(cut(pias, 2)) |>
  contrib()
```

The same approach works with a fixed-base index as well.

```{r}
chain(index) |>
  set_contrib_from_index() |>
  aggregate(cut(pias, 2)) |>
  contrib()
```

---
title: "Matrix aggregation"
vignette: >
  %\VignetteIndexEntry{Matrix aggregation}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

Aggregating a price index can be done as a matrix operation. Although
this approach is less flexible than the `aggregate()` method used in
`vignette("piar")` (i.e., there can be no missing elementary indexes), it
can be considerably faster for larger indexes.

Let's start by building the index in `vignette("piar")` again.

```{r}
library(piar)

# Make an aggregation structure.
ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

# Make a fixed-base index.
elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product),
    business = factor(business, levels = ms_weights$business)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)

index <- elementals |>
  aggregate(pias, na.rm = TRUE) |>
  chain()

index
```

The key to doing this aggregation as a matrix operation is to first turn
the aggregation structure into an aggregation matrix.

```{r}
pias_matrix <- as.matrix(pias)

pias_matrix
```

Multiplying this matrix with a matrix of fixed-base elementary indexes
now computes the aggregate index in each time period.

```{r}
pias_matrix %*% as.matrix(index[levels(pias)$business])
```

## Computing the shadow of an index

It's often useful to determine which higher-level index values are
missing, and subsequently get imputed during aggregation (i.e., compute
the shadow of an index). This is simple to do if there's an elementary
index for each elementary aggregate in the aggregation structure.

The idea is to aggregate an indicator for missingness to get a matrix
that gives the share of missing elementary indexes for each higher-level
index.

```{r}
pias_matrix <- as.matrix(pias) > 0

pias_matrix %*% is.na(elementals) / rowSums(pias_matrix)
```

A value of 1 means that there are no non-missing elementary indexes, and
that the value for this level of the index is imputed from its parent in
the aggregation structure. A value below 1 but above zero means that
some but not all elementary indexes are missing, and the index value for
this level is based on the non-missing elementary indexes. A value of
zero means there's no imputation for this level of the index.

## Sparse matrices

Aggregation structures are naturally sparse. Although using a dense
aggregation matrix does not matter for small indexes, it quickly becomes
inefficient for large indexes---in this case it is better to make a
sparse aggregation matrix.

```{r}
as.matrix(pias, sparse = TRUE)
```

---
title: "Imputation"
vignette: >
  %\VignetteIndexEntry{Imputation}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

The example in `vignette("piar")` used overall-mean imputation to both
impute missing price relatives when calculating the elementary indexes
and to impute missing elementary indexes during aggregation. Although
overall-mean imputation is simple and transparent, it is not the only way to
impute missing prices or index values.

## Imputing missing prices

Instead of implicitly imputing missing price relatives by ignoring
missing values, a common explicit (but methodologically dubious)
imputation strategy when making elementary indexes is to carry forward
the previous price to impute for missing prices. As the
`elementary_index()` function accepts price relatives as its input, any
imputations can be done prior to passing price relatives to this
function. (Missing values still need to be removed in this example
because not all missing prices can be imputed.)

```{r}
library(piar)

elementals <- ms_prices |>
  transform(
    imputed_price = impute_prices(
      price,
      period = period,
      product = product,
      method = "carry-forward"
    )
  ) |>
  elementary_index(
    price_relative(imputed_price, period = period, product = product) ~
      period + business,
    na.rm = TRUE
  )

elementals
```

## Imputation during aggregation

Overall-mean imputation is the usual way to impute missing elementary index
values during aggregation, and it is simple to do with `aggregate()`. In
some cases, however, an elementary index may get imputed with the value
for, say, another elementary aggregate, rather than for an entire group
of elementary aggregates. The simplest way to do this sort of imputation
is to alter the elementary indexes prior to aggregation.

As an example, suppose that missing index values for business B4 should
be imputed as 1, rather than the value for group 12. This replacement
can be done as if the index was a matrix.

```{r}
elementals2 <- elementals
elementals2["B4", 1:3] <- 1

elementals2
```

Aggregating these elementary indexes now incorporates this imputation.

```{r}
ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

aggregate(elementals2, pias, na.rm = TRUE)
```

It is also possible to supply a function with these rules that can be used
by `aggregate()`.

```{r}
impute <- function(x, pias) {
  if (is.na(x["B4"])) x["B4"] <- 1
  x
}

aggregate(elementals, pias, na.rm = TRUE, impute_rules = impute)
```

---
title: "Aggregating across dimensions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Aggregating across dimensions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Price indexes are often aggregated over multiple dimensions. Matched
sample indexes that use sequential Poisson sampling to draw a sample of
businesses are a good example, as there are usually take-all and
take-some strata in addition to, say, an industry classification.

Let's extend `vignette("piar")` by adding another dimension to the
classification to say if a business belongs to the take-all or take-some
sampling stratum.

```{r}
library(piar)

elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)
```

```{r}
ms_weights$stratum <- c("TS", "TA", "TS", "TS", "TS")

ms_weights
```

The easiest way to deal with multiple digit-wise classifications is to
concatenate them into one classification. In this example the "stratum"
dimension comes before the "classification" dimension for the purposes
of parental imputation. This classification can be expanded with the
`expand_classification()` function as before, just with an extra
instruction to say that the last "digit" in the classification is two
characters wide, not one.

```{r}
classification_sps <- paste0(ms_weights$classification, ms_weights$stratum) |>
  expand_classification(width = c(1, 1, 2))

pias_sps <- aggregation_structure(
  c(classification_sps, list(ms_weights$business)),
  ms_weights$weight
)

pias_sps
```

The elementary indexes can now be aggregated according to this new
aggregation structure.

```{r}
index_sps <- aggregate(elementals, pias_sps, na.rm = TRUE)

index_sps
```

When a price index has many dimensions (e.g., industry, sampling
stratum, region), it can be useful to interact the classifications for
these different dimensions to get all possible aggregation structures.
The aggregated index can then be re-aggregated to get index values for
all dimensions.

Continuing with the example, the industry and strata classifications can
be interacted to get two aggregation structures that can be used to
re-aggregate `index_sps`.

```{r}
interacted_hierarchy <- interact_classifications(
  expand_classification(ms_weights$classification),
  expand_classification(ms_weights$stratum)
)

pias_sps2 <- lapply(
  interacted_hierarchy,
  \(x) aggregation_structure(c(x, list(ms_weights$business)), ms_weights$weight)
)

index_sps2 <- lapply(pias_sps2, \(x) {
  aggregate(index_sps, x, include_ea = FALSE)
})
```

The resulting indexes can be merged together to give an index that
includes all combinations of industry and sampling stratum.

```{r}
Reduce(merge, index_sps2)
```

---
title: "Making price indexes"
vignette: >
  %\VignetteIndexEntry{Making price indexes}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

Most price indexes are made with a two-step procedure, where
period-over-period *elementary indexes* are calculated for a collection
of *elementary aggregates* at each point in time, and then aggregated
according to a *price index aggregation structure*. These indexes can
then be chained together to form a time series that gives the evolution
of prices with respect to a fixed base period. This package contains a
collection of functions that revolve around this work flow, making it
easy to build standard price indexes in **R**.

The purpose of this vignette is to give an introductory example for how
to use the core functionality in this package to make a standard price
index. Subsequent vignettes go into more details on advanced topics,
often referencing the example in this vignette.

## Matched-sample index

In this vignette we'll be calculating a matched-sample index where a
fixed set of businesses each provide prices for a collection of products
over time. The products reported by a businesses can change over time,
but the set of businesses is fixed for the duration of the sample. Each
businesses has a weight that is established when the sample is drawn
and represents a particular segment of the economy.

The usual approach for calculating a matched-sample index starts by
computing an elementary index for each business as an equally-weighted
geometric mean of price relatives (i.e., a Jevons index). From there,
index values for different segments of the economy are calculated as an
arithmetic mean of the elementary indexes using the businesses-level
weights (either a Young or Lowe index, depending how the weights are
constructed; see `vignette("adjust-weights")`).

The `ms_prices` dataset has price data for five businesses over four
quarters, and the `ms_weights` dataset has the weight data. Note that
these data have fairly realistic patterns of missing data and are
emblematic of the kinds of survey data used to make price indexes.

```{r}
library(piar)

head(ms_prices)

ms_weights
```

The `elementary_index()` function makes, well, elementary indexes, using
information on price relatives, elementary aggregates (businesses), and
time periods (quarters). By default it makes a Jevons index, but any
bilateral generalized-mean index is possible (see
`vignette("index-number-formulas")` for more details). The only wrinkle
is that price data here are in levels, and not relatives, but the
`price_relative()` function can make the necessary conversion.

```{r}
elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)

elementals
```

As with most functions in **R**, missing values are contagious by
default. Setting `na.rm = TRUE` in `elementary_index()` means that
missing price relatives are ignored, which is equivalent to imputing
these missing relatives with the value of the elementary index for the
respective businesses (i.e., overall-mean imputation). Other
types of imputation are covered in `vignette("imputation")`.

The `elementary_index()` function returns a special index object, and
there are a number of methods for working with these objects. For
example, the resulting indexes to be extracted like a matrix (even
though it's not a matrix).[^1]

[^1]: Note that there are only indexes for four businesses, not five,
    because the fifth business never reports any prices. An elementary
    index can be made for this business by passing a factor with a level
    for all five businesses to `elementary_index()`.

```{r}
elementals[, "202004"]

elementals[c("B1", "B3"), ]
```

With the elementary indexes out of the way, it's time to make a
price-index aggregation structure that maps each business to its
position in the aggregation hierarchy. The only hiccup is unpacking the
digit-wise classification for each businesses that defines the
hierarchy. That's the job of the `expand_classification()` function.

```{r}
ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()
```

It is now simple to aggregate the elementary indexes according to this
aggregation structure with the `aggregate()` function. As with the
elementary indexes, missing values are ignored by setting `na.rm = TRUE`,
which is equivalent to (recursively) imputing a missing value for a point in the
aggregation structure by that of its parent. Note that,
unlike the elementary indexes, missing values are filled in to ensure the
index can be chained over time.

```{r}
index <- aggregate(elementals, pias, na.rm = TRUE)

index
```

## Chaining

The `elementary_index()` function makes period-over-period elementary
indexes by default, which are then aggregated to make a
period-over-period index. Chaining an index is the process of taking the
cumulative product of each of these period-over-period indexes to make a
time series that compares prices to a fixed base period.

The `chain()` function can be used to chain the values in an index
object.

```{r}
chained_index <- chain(index)

chained_index
```

This gives almost the same result as directly manipulating the index as
a matrix, except that the former returns an index object (not a matrix).

Chained indexes often need be to rebased, and this can be done with the
`rebase()` function. For example, rebasing the index so that 202004 is
the base period just requires dividing the chained index by the slice
for 202004.

```{r}
rebase(chained_index, base = chained_index[, "202004"])
```

## Working with indexes

Once an index has been calculated, it usually needs to be turned into a
table of index values. This can be done by either coercing an index into
a matrix

```{r}
as.matrix(chained_index)
```

or a data frame

```{r}
as.data.frame(chained_index)
```

It is also sometimes useful to get the price-updated weights used to
aggregate the index; these can be calculated by first updating the
aggregation structure with the aggregated index, then made into a table.

```{r}
update(pias, index) |>
  as.data.frame()
```

---
title: "Aggregating across baskets"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Aggregating across baskets}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Most prices indexes use the same set of weights to aggregate elementary
indexes for several years, periodically updating these weights and
adding or removing elementary aggregates. The general approach to keep
the time series going across baskets is to "chain" the index across
baskets.

Let's start by making some quarterly elementary indexes and weights for
two baskets. Each basket has index values for two years, and there is a
one year overlap between the baskets.

```{r}
library(piar)

set.seed(12345)

# Make quarterly elementary indexes for two baskets with a 1 year overlap.

elementals1 <- matrix(
  runif(5 * 8, 0.8, 1.2),
  nrow = 5,
  dimnames = list(paste0("B", 1:5), NULL)
) |>
  as_index()

elementals2 <- matrix(
  runif(6 * 8, 0.8, 1.2),
  nrow = 6,
  dimnames = list(paste0("B", 1:6), 5:12)
) |>
  as_index()

# Make aggregation weights for basket 1.
#            1
#      |-----+-----|
#      11          12
#  |---+---|   |---+---|
#  B1  B2  B3  B4      B5

weights1 <- data.frame(
  level1 = 1,
  level2 = c(11, 11, 11, 12, 12),
  ea = levels(elementals1),
  weights = runif(5, 100, 200)
)

# Make aggregation weights for basket 2.
#            1
#      |-----+-----|
#      11          12
#  |---+---|   |---+---|
#  B1  B2  B3  B4  B5  B6

weights2 <- data.frame(
  level1 = 1,
  level2 = c(11, 11, 11, 12, 12, 12),
  ea = levels(elementals2),
  weights = runif(6, 100, 200)
)
```

Aggregating the indexes for both baskets involves simply mapping the
usual workflow across both baskets, keeping only the higher-level
indexes because the elementary indexes change across baskets.

```{r}
index <- Map(
  aggregate,
  list(elementals1, elementals2),
  list(weights1, weights2),
  include_ea = FALSE
)
```

There are now several ways to combine these index series to form a
single time series. Assuming that we don't want to revise any index
values from the old basket, the easiest method is to simply ignore the
overlap period and stack the upper-level period-over-period indexes for
the new basket onto the those in the old basket.

```{r}
stack(index[[1]], window(index[[2]], start = "9")) |>
  chain()
```

Alternatively, the index values from the new basket can be chained using
the values in the last period of the first basket as a link factor and
combined with the chained index values from the old basket.

```{r}
link_factor <- chain(index[[1]]) |>
  window(start = end(index[[1]])) |>
  as.numeric()

stack(
  chain(index[[1]]),
  chain(window(index[[2]], start = "9"), link = link_factor)
)
```

The index values for the overlap period can be used to make more complex
link factors when the index series have an annual base period.

```{r}
index <- index |>
  lapply(chain) |>
  lapply(\(x) rebase(x, base = mean(x[, 1:4])))
```

Keeping the base year of the old basket can be done by simply rebasing
the index series for the new basket.

```{r}
link_factor <- window(index[[1]], start = "5") |>
  mean() |>
  as.numeric()

stack(
  index[[1]],
  rebase(window(index[[2]], start = "9"), base = 1 / link_factor)
)
```

Updating the reference year to that of the new basket involves rebasing
the series for both the old basket and the new basket.

```{r}
index[[1]] <- rebase(index[[1]], base = mean(window(index[[1]], start = "5")))

link_factor <- as.numeric(index[[1]][, "8"]) / as.numeric(index[[2]][, "8"])

stack(
  index[[1]],
  rebase(window(index[[2]], start = "9"), base = 1 / link_factor)
)
```

---
title: "Multiple sources of data"
vignette: >
  %\VignetteIndexEntry{Multiple sources of data}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

Price indexes are usually made from several sources of data. An
important benefit of the usual two-step workflow to make price indexes
is that the elementary indexes can be built piecemeal---using different
sources of data and different index-number formulas---and then
aggregated with a consistent structure.

Let's extend the example in `vignette("piar")` by having an alternate
source of data for business `B5` that is always missing in the
`ms_prices` dataset.

```{r}
library(piar)

# Make an aggregation structure.
ms_weights[c("level1", "level2")] <-
  expand_classification(ms_weights$classification)

pias <- ms_weights[c("level1", "level2", "business", "weight")] |>
  as_aggregation_structure()

# Make elementary index.
elementals <- ms_prices |>
  transform(
    relative = price_relative(price, period = period, product = product)
  ) |>
  elementary_index(relative ~ period + business, na.rm = TRUE)

elementals
```

Instead of using survey-like data for the other businesses, `B5` is made
from scanner-like data with many price and quantity observations at each
point in time.

```{r}
set.seed(12345)

scanner_prices <- data.frame(
  period = rep(c("201904", time(elementals)), each = 200),
  product = 1:200,
  price = round(rlnorm(5 * 200) * 10, 1),
  quantity = round(runif(5 * 200, 100, 1000))
)

head(scanner_prices)
```

These type of data often require the use of a multilateral index like
the GEKS. For the sake of illustration, we'll make a Fisher GEKS index
over a 3 quarter rolling window and use a mean splice to make a single
time series.

```{r}
geks_elementals <- with(
  scanner_prices,
  geks_index(price, quantity, period, product, window = 3)
) |>
  splice_index() |>
  t() |>
  as_index(chainable = FALSE) |>
  set_levels("B5") |>
  rebase(base = "202001")

geks_elementals
```

These values can now be merged with the other elementary indexes, getting
turned into a period-over-period index in the process, and then
aggregated.

```{r}
merge(elementals, geks_elementals) |>
  aggregate(pias, na.rm = TRUE)
```

---
title: "Spatial price indexes"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Spatial price indexes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

It is possible to rework the functions in this package to make a spatial, rather
than temporal, price index using a star model by simply replacing "time" with "space".
Although this is not a suitable
approach for making a system of purchasing power parities, is it useful for making indexes
that adjust pay for cost-of-living differences relative to a fixed location. These
sorts of indexes are made by the United Nations International Civil Service Commission, Eurostat, and various governments
and private-sector organizations for administering cost-of-living allowances.

The basic workflow is the same as with a temporal index, just treating each
country as a different "time period" in a fixed-based index.

```{r}
library(piar)

set.seed(12345)

# Make indexes for 6 basic headings for 4 countries.
bh_index <- matrix(
  c(rep(1, 6), runif(6 * 3, 0.8, 1.2)),
  nrow = 6,
  dimnames = list(
    paste0("BH", 1:6),
    paste("Country", 1:4)
  )
) |>
  as_index(chainable = FALSE)

head(bh_index)

# Make fixed aggregation weights.
#            1
#      |-----+-----|
#      11          12
#  |---+---|   |---+---|
#  B1  B2  B3  B4  B5  B6

weights <- data.frame(
  level1 = 1,
  level2 = rep(11:12, each = 3),
  bh = levels(bh_index),
  weights = runif(6, 100, 200)
)

head(weights)
```

The indexes at the basic-heading level can be aggregated as usual to get a
collection of indexes that give the difference in purchasing power relative to
the base country (country 1 in this example). Different choices for weights make
different assumptions about how people change their spending patterns between
each country and the base country.

```{r}
index <- aggregate(bh_index, weights)
```

As it can be costly to collect price data from various countries on a regular
basis, these sorts of indexes are often updated over time by changes in exchange
rates and inflation rates for each country.

```{r}
as.matrix(index[1, -1])

update_factors <- runif(3, 0.8, 1.2)

as.matrix(index[1, -1]) * update_factors
```

---
title: "Superlative aggregation"
vignette: >
  %\VignetteIndexEntry{Superlative aggregation}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

All of the examples so far have used a single set of weights to
aggregate an index. Although this is by far the most common case, it is
not suitable for aggregating a superlative index where there is more
than one set of aggregation weights that change every period.

## Aggregating a Paasche index

Let's start by making a chained Paasche index from 6 elementary indexes
over 4 time periods to see how to deal with time-varying aggregation
weights. This is similar to the example of building an index across
multiple basket in `vignette("multiple-baskets")`, except that the
weights change each period.

```{r}
library(piar)

set.seed(12345)

# Make 6 elementary indexes over 4 time periods.
elementals <- matrix(c(rep(1, 6), runif(6 * 3, 0.8, 1.2)), nrow = 6) |>
  as_index() |>
  set_levels(paste0("B", 1:6))

head(elementals)

# Make aggregation weights over 4 time periods.
#            1
#      |-----+-----|
#      11          12
#  |---+---|   |---+---|
#  B1  B2  B3  B4  B5  B6

weights <- data.frame(
  level1 = 1,
  level2 = rep(11:12, each = 3),
  ea = levels(elementals),
  weights = runif(4 * 6, 100, 200),
  period = rep(1:4, each = 6)
)

head(weights)
```

The key tools to deal with time-varying aggregation weights are the
`stack()` and `unstack()` functions. `stack()` appends a later index
series onto an earlier one for the same levels, whereas `unstack()`
pulls apart an index series for many periods into a collection of
one-period indexes.^[`split()` can also be used for this.] These functions allow
the aggregation to be done with a map-reduce.

The first step to making a Paasche index is to unstack the elementary
indexes into a list of elementary indexes for each period. (Trying to
make the elementary indexes period-by-period can be dangerous when there
are missing values.)

```{r}
elementals <- unstack(elementals)
```

This is followed by making a sequence of aggregation structures for each
set of weights.

```{r}
paasche_pias <- split(
  weights[c("level1", "level2", "ea", "weights")],
  weights[["period"]]
) |>
  lapply(as_aggregation_structure)
```

Computing the Paasche index for each period is now just a case of
mapping the `aggregate()` function to each elementary index and
aggregation structure, and then reducing the result with the `stack()`
function.

```{r}
paasche <- Map(
  aggregate,
  elementals,
  paasche_pias,
  na.rm = TRUE,
  include_ea = FALSE,
  order = -1
) |>
  Reduce(stack, x = _)

paasche
```

## Making a Fisher index

Making a chained Fisher index requires two sets of weights: base-period
weights to make the Laspeyres index and current-period weights to make
the Paasche index. As with the Paasche index this can be done with a
map-reduce, except now by passing two aggregation structures to the
`aggregate()` function instead of one.

```{r}
laspeyres_pias <- paasche_pias[c(1, 1, 2, 3)]

fisher <- Map(
  aggregate,
  elementals,
  pias = laspeyres_pias,
  pias2 = paasche_pias,
  na.rm = TRUE,
  include_ea = FALSE
) |>
  Reduce(stack, x = _)

fisher
```

This gives the same result as calculating the Laspeyres and Paasche
indexes individually, and then calculating the Fisher index manually.

```{r}
laspeyres <- Map(
  aggregate,
  elementals,
  pias = laspeyres_pias,
  na.rm = TRUE,
  include_ea = FALSE
) |>
  Reduce(stack, x = _)

sqrt(as.matrix(laspeyres) * as.matrix(paasche))
```

---
title: "Aggregating over subperiods"
vignette: >
  %\VignetteIndexEntry{Aggregating over subperiods}
  %\VignetteEngine{litedown::vignette}
  %\VignetteEncoding{UTF-8}
---

With the exception of `vignette("spatial-price-index")`, all the examples so far
have revolved around aggregating over the levels of a price index for each
time period. Although this is the core workflow in this package, it is also
useful to be able to aggregate over the time periods for each level of an index
to turn a monthly or quarterly index into an annual index.

Let's modify the example in `vignette("adjust-weights")` by adding an additional
eight quarters to make three years of data.

```{r}
set.seed(54321)

library(piar)

# Make an aggregation structure.
#                  1
#      |-----------|-----------|
#      11          12          13
#  |---+---|   |---+---|   |---+---|
# 111     121 121     122 131     132

pias <- data.frame(
  level1 = rep(1, 12),
  level2 = rep(c(11, 12, 13), each = 4),
  level3 = rep(c(111, 112, 121, 122, 131, 132), each = 2),
  ea = sprintf("B%02d", 1:12),
  weight = 1:12
) |>
  as_aggregation_structure()

pias

# Make elementary indexes over 3 years and aggregate.
quarterly_index <- matrix(
  runif(12 * 12, 0.4, 1.2),
  nrow = 12,
  dimnames = list(sprintf("B%02d", 1:12), paste0("Q", 1:12))
) |>
  as_index() |>
  aggregate(pias)

head(quarterly_index)
```

The conventional way to turn a quarterly arithmetic index into an annual one is
to take the (unweighted) arithmetic mean of the index values over each year and
rebase to a new base year.

```{r}
annual_index <- chain(quarterly_index) |>
  mean(window = 4)

annual_index |>
  rebase(base = "Q1") |>
  head()
```

It's worth noting that, at least with an arithmetic index, the aggregation
properties of the index continue to hold after price-updating the weights.

```{r}
annual_pias <- pias |>
  update(annual_index, period = "Q1")

annual_index <- annual_index |>
  rebase(base = "Q1")

annual_index |>
  aggregate(annual_pias) |>
  all.equal(annual_index)
```

This means that, for example, the workflow in `vignette("contributions")` for
determining, say, the quarter-over-quarter contribution of the level 2 indexes
to top-level index remains the same.

```{r}
annual_index |>
  unchain() |>
  set_contrib_from_index() |>
  aggregate(cut(annual_pias, 2)) |>
  contrib()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split.piar_index.R
\name{split.piar_index}
\alias{split.piar_index}
\alias{split<-.piar_index}
\title{Split an index into groups}
\usage{
\method{split}{piar_index}(x, f, drop = FALSE, ..., along = c("levels", "time"))

\method{split}{piar_index}(x, f, drop = FALSE, ..., along = c("levels", "time")) <- value
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{f}{\verb{[factor | list]} A factor or list of factors to group elements
of \code{x}.}

\item{drop}{\verb{[logical(1)]} Should levels that do not occur in \code{f} be dropped?
By default all levels are kept.}

\item{...}{Further arguments passed to \code{\link[=split.default]{split.default()}}.}

\item{along}{\verb{[character(1)]} Either \code{"levels"} to split over the levels
of \code{x} (the default), or \code{"time"} to split over the time periods of \code{x}.}

\item{value}{\verb{[list]} A list of values compatible with the splitting of \code{x},
or something that can be coerced into one, recycled if necessary.}
}
\value{
\code{split()} returns a list of index objects for each level in \code{f}. The
replacement method replaces these values with the corresponding element of
\code{value}.
}
\description{
Split an index into groups of indexes according to a factor, along either
the levels or time periods of the index.
}
\examples{
index <- as_index(matrix(1:6, 2))

split(index, 1:2)

split(index, c(1, 1, 2), along = "time")
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/emean.R
\name{emean}
\alias{emean}
\title{Extended mean}
\usage{
emean(x, y, order = c(0, 1), tol = .Machine$double.eps^0.5)
}
\arguments{
\item{x, y}{\verb{[numeric > 0]} A strictly positive numeric vector.}

\item{order}{\verb{[numeric(2)]} A pair of finite numbers giving the order of the
extended mean.
The default calculates the ordinary logarithmic mean. Setting either the
first or second element to 1 gives the generalized logarithmic mean.}

\item{tol}{\verb{[numeric > 0]} The tolerance used to determine if \code{x == y}.
The default value is the same as \code{\link[=all.equal]{all.equal()}}.}
}
\value{
A numeric vector, the same length as
\code{max(length(x), length(y))}, giving the component-wise extended mean
of \code{x} and \code{y}.
}
\description{
Calculate the component-wise extended mean.
}
\details{
The extended mean is also called the difference mean, Stolarsky mean, or
extended mean-value mean; see Bullen (2003, p. 393) for details.

Both \code{x} and \code{y} should be strictly positive. This is not enforced, but the
results may not make sense when the extended mean is not defined. The usual
recycling rules apply when \code{x} and \code{y} are not the same length.

By definition, the extended mean of \code{x}
and \code{y} is \code{x} when \code{x == y}. The \code{tol} argument is used
to test equality by checking if \code{abs(x - y) <= tol}.  In some cases it's
useful to multiply
\code{tol} by a scale factor, such as \code{max(abs(x), abs(y))}. This often
doesn't matter when making price indexes, however, as \code{x} and \code{y}
are usually around 1.
}
\examples{
x <- 8:5
y <- 1:4

# The arithmetic and geometric means are special cases of the
# generalized logarithmic mean.
all.equal(emean(x, y, c(2, 1)), (x + y) / 2)
all.equal(emean(x, y, c(-1, 1)), sqrt(x * y))

# The harmonic mean cannot be expressed as a logarithmic mean, but can
# be expressed as an extended mean.
all.equal(emean(x, y, c(-2, -1)), 2 / (1 / x + 1 / y))

# The quadratic mean is also a type of extended mean.
all.equal(emean(x, y, c(2, 4)), sqrt(x^2 / 2 + y^2 / 2))

# As are heronian and centroidal means.
all.equal(
  emean(x, y, c(0.5, 1.5)),
  (x + sqrt(x * y) + y) / 3
)
all.equal(
  emean(x, y, c(2, 3)),
  2 / 3 * (x^2 + x * y + y^2) / (x + y)
)
}
\references{
Bullen, P. S. (2003). \emph{Handbook of Means and Their Inequalities}.
Springer Science+Business Media.
}
\seealso{
Other math functions:
\code{\link[=gmean]{gmean()}},
\code{\link[=nested_gmean]{nested_gmean()}},
\code{\link[=scale_weights]{scale_weights()}},
\code{\link[=transmute_weights]{transmute_weights()}},
\code{\link[=transmute_weights2]{transmute_weights2()}},
\code{\link[=update_weights]{update_weights()}}
}
\concept{math functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gmean.R
\name{gmean}
\alias{gmean}
\title{Generalized mean}
\usage{
gmean(x, weights = NULL, order = 1, na.rm = FALSE)
}
\arguments{
\item{x}{\verb{[numeric > 0]} A strictly positive numeric vector.}

\item{weights}{\verb{[numeric >= 0]} A positive numeric vector of weights, the
same length as \code{x}. The default is to equally weight each element of \code{x}.}

\item{order}{\verb{[numeric(1)]} A finite number giving the order (or exponent) of
the generalized mean. The default calculates an arithmetic mean.}

\item{na.rm}{\verb{[logical(1)]} Should missing values be removed? By default,
missing values are not removed.}
}
\value{
A numeric value for the generalized mean.
}
\description{
Calculated a weighted generalized mean.
}
\details{
The generalized mean is also called the power mean, Hölder mean, or \eqn{l_p}
mean; see Bullen (2003, p. 175) for details.

Both \code{x} and \code{weights} are usually strictly positive
(and finite), especially for the purpose of making a price index. This is not
enforced, but the results may not make sense if the generalized mean is not
defined. There are two exceptions to this.
\enumerate{
\item The convention by Hardy et al. (1952, p. 13) is used in cases where \code{x}
has zeros: the generalized mean is 0 whenever the weights are strictly
positive and \code{order < 0}. The analogous convention holds whenever at least
one element of \code{x} is \code{Inf}: the generalized mean is \code{Inf} whenever the
weights are strictly positive and \code{order > 0}.
\item Some authors let the weights be non-negative and sum to 1. If there are
zero weights then the corresponding element
of \code{x} has no impact on the result whenever \code{x} is strictly
positive. Unlike \code{\link[=weighted.mean]{weighted.mean()}}, however,
zero weights are not strong zeros, so infinite values in \code{x} will
propagate.
}

The weights are scaled to sum to 1 to satisfy the definition of a
generalized mean.
}
\note{
The generalized mean can be defined on the extended real line, so
that \code{order = -Inf / Inf} returns \code{\link[=min]{min()}}/\code{\link[=max]{max()}}, to agree with the
definition by Bullen (2003). This is not implemented, and the order of the
generalized mean must be finite.
}
\examples{
x <- 1:3
w <- c(0.25, 0.25, 0.5)

# Arithmetic mean.
gmean(x, w)

# Geometric mean.
gmean(x, w, order = 0)

# The Lehmer mean is a generalized mean with specific weights.
gmean(x, w * x)
}
\references{
Bullen, P. S. (2003). \emph{Handbook of Means and Their Inequalities}.
Springer Science+Business Media.

Hardy, G., Littlewood, J. E., and Polya, G. (1952). \emph{Inequalities} (2nd
edition). Cambridge University Press.
}
\seealso{
Other math functions:
\code{\link[=emean]{emean()}},
\code{\link[=nested_gmean]{nested_gmean()}},
\code{\link[=scale_weights]{scale_weights()}},
\code{\link[=transmute_weights]{transmute_weights()}},
\code{\link[=transmute_weights2]{transmute_weights2()}},
\code{\link[=update_weights]{update_weights()}}
}
\concept{math functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/price_data.R
\docType{data}
\name{price_data}
\alias{price_data}
\alias{ms_prices}
\alias{ms_weights}
\alias{fs_prices}
\alias{fs_weights}
\title{Price data}
\description{
Sample price and weight data for both a match sample and fixed sample type
index.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/head.piar_index.R
\name{head.piar_index}
\alias{head.piar_index}
\alias{tail.piar_index}
\title{Return the first/last parts of an index}
\usage{
\method{head}{piar_index}(x, n = 6L, ...)

\method{tail}{piar_index}(x, n = 6L, ...)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{n}{\verb{[integer]} See \code{\link[=head]{head()}}/\code{\link[=tail]{tail()}}. The default takes the
first/last 6 levels of \code{x}.}

\item{...}{Not currently used.}
}
\value{
A price index that inherits from the same class as \code{x}.
}
\description{
Extract the first/last parts of an index as if it were a matrix.
}
\examples{
index <- as_index(matrix(1:9, 3))

head(index, 1)

tail(index, 1)
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand_classification.R
\name{expand_classification}
\alias{expand_classification}
\alias{interact_classifications}
\title{Expand a hierarchical classification}
\usage{
expand_classification(x, width = 1L, pad = NA)

interact_classifications(..., sep = ":")
}
\arguments{
\item{x}{\verb{[character]} A character vector, or something that can be coerced
into one, of
codes/labels for a specific level in a classification (e.g., 5-digit
COICOP, 5-digit NAICS, 4-digit SIC).}

\item{width}{\verb{[integer > 0]} An integer vector that gives the width of each
digit in
\code{x}. A single value is recycled to span the longest element in
\code{x}. This cannot contain NAs. The default assumes each digit has a
width of 1, as in the NAICS, NAPCS, and SIC classifications.}

\item{pad}{\verb{[character(1)]} A string used to pad the shorter labels for an
unbalanced classification. The default pads with NA.}

\item{...}{\verb{[list]} Lists of character vectors that give the codes/labels for
each
level of the classification, ordered so that moving down the list goes down
the hierarchy (as made by \code{expand_classification()}).}

\item{sep}{\verb{[character(1)]} A character used to combine codes/labels across
elements of \code{...}. The default uses \code{":"}.}
}
\value{
\code{expand_classification()} returns a list with a entry for each level
in \code{x} giving the "digits" that represent each level in the hierarchy.

\code{interact_classfications()} returns a list of lists with the same structure
as \code{expand_classification()}.
}
\description{
Expand a character representation of a hierarchical classification to make a
price index aggregation structure. Expanded classifications be interacted
together to get all combinations of aggregation structures.
}
\examples{
# A simple classification structure.
#            1
#      |-----+-----|
#      11          12
#  |---+---|       |
#  111     112     121

expand_classification(c("111", "112", "121"))

# Expanding more complex classifications:
# ... if last 'digit' is either TA or TS.
expand_classification(
  c("111TA", "112TA", "121TS"),
  width = c(1, 1, 1, 2)
)

# ... if first 'digit' is either 11 or 12.
expand_classification(c("111", "112", "121"), width = c(2, 1))

# ...if there are delimiters in the classification (like COICOP).
expand_classification(c("01.1.1", "01.1.2", "01.2.1"), width = 2)
}
\seealso{
\code{\link[=aggregation_structure]{aggregation_structure()}} to make a price-index aggregation structure.

\code{\link[=split_classification]{split_classification()}} to expand a classification by splitting along
a delimiter.

\code{\link[=combine_classifications]{combine_classifications()}} for combining multiple hierarchical
classifications.

\code{csh_from_digits()} in the \pkg{accumulate} package for different handling
of unbalanced classifications.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splice_index.R
\name{splice_index}
\alias{splice_index}
\title{Splice an index series}
\usage{
splice_index(x, periods = NULL, initial = NULL, published = FALSE)
}
\arguments{
\item{x}{\verb{[list]} A list of equal-length numeric vectors giving the
period-over-period indexes for each window.}

\item{periods}{\verb{[integer]} An integer vector giving the splice points for
each window. The default splices on each point in the window.}

\item{initial}{\verb{[numeric]} A numeric vector giving an initial
period-over-period index
series onto which the elements of \code{x} are spliced. The default uses the
first element of \code{x}.}

\item{published}{\verb{[logical(1)]} Should the splice be done against the
published series? The default splices using the recalculated index series.}
}
\value{
A numeric vector giving the spliced (fixed-base) index series.
}
\description{
Splice a collection of index series computed over a rolling window into one
index series. Splicing on multiple points combines the results with a
geometric mean.
}
\examples{
# Make an index series over a rolling window.
x <- list(c(1.1, 0.9, 1.2), c(0.8, 1.3, 1.4), c(1.3, 1.3, 0.8))

# Mean splice.
splice_index(x)

# Movement splice.
splice_index(x, 3)

# Window splice.
splice_index(x, 1)

# Splicing on the published series preserves the within-window
# movement of the index series.
splice_index(x, 1, published = TRUE)
}
\references{
Chessa, A. G. (2019).
\emph{A Comparison of Index Extension Methods for Multilateral Methods.} Paper
presented at the 16th Meeting of the Ottawa Group on Price Indices,
8-10 May 2019, Rio de Janeiro, Brazil.

Krsinich, F. (2016). The FEWS index: Fixed effects with a window splice.
\emph{Journal of Official Statistics}, 32(2), 375-404.
}
\concept{price index functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute_prices.R
\name{impute_prices}
\alias{impute_prices}
\alias{impute_prices.default}
\alias{impute_prices.matrix}
\alias{impute_prices.numeric}
\alias{impute_prices.data.frame}
\alias{carry_forward}
\alias{carry_backward}
\alias{shadow_price}
\title{Impute missing prices}
\usage{
impute_prices(x, ...)

\method{impute_prices}{default}(x, ...)

\method{impute_prices}{matrix}(
  x,
  period,
  product,
  ...,
  ea = NULL,
  weights = NULL,
  pias = NULL,
  order = c(0, 1),
  r = order,
  method = c("overall-mean", "carry-forward"),
  impute_rules = NULL
)

\method{impute_prices}{numeric}(
  x,
  period,
  product,
  ...,
  ea = NULL,
  weights = NULL,
  pias = NULL,
  order = c(0, 1),
  r = order,
  method = c("overall-mean", "carry-forward", "carry-backward"),
  impute_rules = NULL
)

\method{impute_prices}{data.frame}(x, formula, ..., ea = NULL, weights = NULL)

carry_forward(x, ...)

carry_backward(x, ...)

shadow_price(x, ...)
}
\arguments{
\item{x}{\verb{[object]} Either a numeric vector (or something that can be
coerced into one),
a data frame of prices, or a two-column matrix of current prices
and back prices (in that order).}

\item{...}{Further arguments passed to or used by methods.}

\item{period}{\verb{[factor]} A factor, or something that can be coerced into one,
giving
the time period associated with each price in \code{x}. The ordering of time
periods follows of the levels of \code{period}, to agree with
\code{\link[=cut.Date]{cut()}}.}

\item{product}{\verb{[factor]} A factor, or something that can be coerced into
one, giving the product associated with each price in \code{x}.}

\item{ea}{\verb{[factor]} A factor, or something that can be coerced into one,
giving the
elementary aggregate associated with each price in \code{x}. This is evaluated
in \code{x} for the data frame method. The default pools all data into one
elementary aggregate.}

\item{weights}{\verb{[numeric >= 0]} A numeric vector of weights for the prices
in \code{x} (i.e.,
product weights), or something that can be coerced into one. The default is
to give each price equal weight. This is evaluated in \code{x} for the data
frame method.}

\item{pias}{\verb{[piar_aggregation_structure]} A price index aggregation
structure, or something that can be
coerced into one, as made with \code{\link[=aggregation_structure]{aggregation_structure()}}. The default
imputes from elementary indexes only (i.e., not recursively).}

\item{order}{\verb{[numeric(2)]} A pair of numeric values. The first gives the
order of the generalized-mean price index used to calculate the
elementary price indexes, defaulting to a geometric index. The second
gives the order of the generalized-mean price index used to aggregate the
elementary price indexes, defaulting to an arithmetic index. Other values
are possible;
see \code{\link[=gmean]{gmean()}} for details.}

\item{r}{Deprecated.}

\item{method}{\verb{[character(1)]} Name of the imputation method, one
of \code{"overall-mean"} (the default), \code{"carry-forward"},
or \code{"carry-backward"}.}

\item{impute_rules}{\verb{[function]} (Experimental) A function that applies
imputation
rules to the elementary indexes in each time period prior to aggregation.
It takes two arguments, the elementary indexes for a given time period and
the (price updated) aggregation structure, and must return back the
elementary indexes.}

\item{formula}{\verb{[formula]} A two-sided formula, or something that can be
coerced into one, with prices on the left-hand
side and time periods and products on the right-hand side (in that order).}
}
\value{
A numeric vector or matrix of prices with missing values replaced
(where possible).
}
\description{
Impute missing prices using the carry forward/backward or the
self-correcting overall mean method.
}
\details{
The carry forward method replaces a missing price for a product by the price
for the same product in the previous period. It tends to push an index value
towards 1, and is usually avoided; see paragraph 6.61 in the CPI manual
(2020). The carry backwards method does the opposite, but this is rarely
used in practice.

The self-correcting overall mean method recursively imputes a missing price
by the value of
the price for the same product in the previous period multiplied by the
value of the period-over-period elementary index for the elementary aggregate
to which that product belongs. This requires computing and aggregating an
index (according to \code{pias}, unless \code{pias} is not supplied) for
each \code{period}. The index
values used to do the imputations are not returned because the index needs
to be recalculated to get correct percent-change contributions. It is
identical to simply
excluding missing price relatives in the index calculation, except in the
period that a missing product returns. For this reason care is needed when
using this method. It is sensitive to the assumption that a product does not
change over time, and in some cases it is safer to simply omit the missing
price relatives instead of imputing the missing prices.

Imputation works slightly differently depending on whether data are in a long
or wide format. When \code{x} is a two-column matrix of current and back prices
(in that order), then imputation is done separately on the current price
at a point in time and the back price at the next point in time. When \code{x} is
a numeric vector then these two prices are necessarily the same.
}
\examples{
prices <- data.frame(
  price = c(1:7, NA),
  period = rep(1:2, each = 4),
  product = 1:4,
  ea = rep(letters[1:2], 4)
)

impute_prices(prices, price ~ period + product, method = "carry-forward")

impute_prices(
  prices,
  price ~ period + product,
  ea = ea,
  method = "overall-mean"
)

# Can also be done with current price-back price formulation.
prices$back_price <- with(
  prices,
  price[back_period(period, product)]
)

impute_prices(
  prices,
  cbind(price, back_price) ~ period + product,
  ea = ea,
  method = "overall-mean"
)
}
\references{
IMF, ILO, OECD, Eurostat, UNECE, and World Bank. (2020).
\emph{Consumer Price Index Manual: Concepts and Methods}.
International Monetary Fund.
}
\seealso{
\code{\link[=price_relative]{price_relative()}} for making price relatives for the
same products over time.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piar-package.R
\docType{package}
\name{piar-package}
\alias{piar}
\alias{piar-package}
\title{piar: Price Index Aggregation}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Most price indexes are made with a two-step procedure, where period-over-period elementary indexes are first calculated for a collection of elementary aggregates at each point in time, and then aggregated according to a price index aggregation structure. These indexes can then be chained together to form a time series that gives the evolution of prices with respect to a fixed base period. This package contains a collection of functions that revolve around this work flow, making it easy to build standard price indexes, and implement the methods described by Balk (2008, \doi{10.1017/CBO9780511720758}), von der Lippe (2007, \doi{10.3726/978-3-653-01120-3}), and the CPI manual (2020, \doi{10.5089/9781484354841.069}) and PPI manual (2004, \doi{10.5089/9781589063044.069}) for bilateral price indexes.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://marberts.github.io/piar/}
  \item \url{https://github.com/marberts/piar}
  \item Report bugs at \url{https://github.com/marberts/piar/issues}
}

}
\author{
\strong{Maintainer}: Steve Martin \email{marberts@protonmail.com} (\href{https://orcid.org/0000-0003-2544-9480}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Steve Martin \email{marberts@protonmail.com} (\href{https://orcid.org/0000-0003-2544-9480}{ORCID}) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geks_index.R
\name{geks_index}
\alias{geks_index}
\title{GEKS index}
\usage{
geks_index(
  price,
  quantity,
  period,
  product,
  index_formula = function(p1, p0, q1, q0) {
     nested_gmean(p1/p0, list(p0 * q0, p1 *
    q1), na.rm = TRUE)
 },
  window = nlevels(period),
  n = window - 1L,
  order = 0,
  match_method = c("all", "back-price")
)
}
\arguments{
\item{price}{\verb{[numeric > 0]} A numeric vector of prices, the same length
as \code{quantity}.}

\item{quantity}{\verb{[numeric >= 0]} A numeric vector of quantities, the same
length as \code{price}.}

\item{period}{\verb{[factor]} A factor, or something that can be coerced into one,
that gives the corresponding time period for each element in \code{price} and
\code{quantity}. The ordering of time periods follows the levels of \code{period}
to agree with \code{\link[=cut.Date]{cut()}}.}

\item{product}{\verb{[factor]} A factor, or something that can be coerced into
one, that
gives the corresponding product identifier for each element in \code{price} and
\code{quantity}.}

\item{index_formula}{\verb{[function]} A function giving the index-number formula
in the GEKS
index. Usually a Törnqvist, Fisher (the default), or Walsh index.
It must have arguments \code{p1}, \code{p0}, \code{q1}, and \code{q0}, and satisfy the
time-reversal test. See \code{vignette("index-number-formulas")} for details.}

\item{window}{\verb{[integer(1) > 0]} A positive integer giving the length of the
rolling window.
The default is a window that encompasses all periods in \code{period}.
Non-integers are truncated towards zero.}

\item{n}{\verb{[integer(1) > 0]} A positive integer giving the length of the index
series for each
window, starting from the end of the window. For example, if there are 13
periods in \code{window}, setting \code{n = 1} gives the index for period 13. The
default gives an index for each period in \code{window}. Non-integers are
truncated towards zero.}

\item{order}{\verb{[numeric(1)]} A finite number giving the order of the
generalized mean used to
average price indexes over the rolling window. The default uses a
geometric mean.}

\item{match_method}{\verb{[character(1)]} Either \code{"all"} to match all products
against each other
(the default) or \code{"back-price"} to match only back prices. The later can be
faster when there is lots of product imbalanced.}
}
\value{
A list with a named numeric vector giving the value of the respective
period-over-period GEKS index for each window.
}
\description{
Calculate a generalized inter-temporal GEKS price index over a rolling
window.
}
\note{
Like \code{\link[=back_period]{back_period()}}, if multiple prices
correspond to a period-product pair, then the back price at a point in time
is always the first price for that product in the previous period. Unlike a
bilateral index, however, duplicated period-product pairs can have more
subtle implications for a multilateral index.
}
\examples{
price <- 1:10
quantity <- 10:1
period <- rep(1:5, 2)
product <- rep(letters[1:2], each = 5)

cumprod(geks_index(price, quantity, period, product)[[1]])

# Calculate the index over a rolling window.
(geks <- geks_index(price, quantity, period, product, window = 3))

# Use a movement splice to combine the indexes in each window.
splice_index(geks, 2)

# ... or use a mean splice.
splice_index(geks)

# Make a Jevons GEKS index.
geks_index(
  price,
  quantity,
  period,
  product,
  index_formula = \(p1, p0, ...) gmean(p1 / p0, na.rm = TRUE, order = 0)
)
}
\references{
Balk, B. M. (2008). \emph{Price and Quantity Index Numbers}.
Cambridge University Press.

IMF, ILO, Eurostat, UNECE, OECD, and World Bank. (2020).
\emph{Consumer Price Index Manual: Concepts and Methods}.
International Monetary Fund.

Ivancic, L., Diewert, W. E., and Fox, K. J. (2011). Scanner data, time
aggregation and the construction of price indexes.
\emph{Journal of Econometrics}, 161(1): 24--35.
}
\seealso{
\code{\link[=splice_index]{splice_index()}} to splice the rolling-window indexes together.

\code{GEKSIndex()} in the \pkg{IndexNumR} package for an implementation of the
GEKS index with more options.

The \pkg{rsmatrix} package for multilateral repeat-sales indexes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gmean.R
\name{nested_gmean}
\alias{nested_gmean}
\title{Nested generalized means}
\usage{
nested_gmean(
  x,
  weights = list(NULL, NULL),
  order = c(1, -1),
  outer_weights = NULL,
  outer_order = 0,
  na.rm = FALSE
)
}
\arguments{
\item{x}{\verb{[numeric > 0]} A strictly positive numeric vector.}

\item{weights}{\verb{[list]} A list of positive numeric vector of
weights, each
the same length as \code{x}, for both of the inner generalized means. \code{NULL}
elements of \code{weights} equally weight each element of \code{x}. The default
uses equal weights for both inner generalized mean.}

\item{order}{\verb{[numeric(2)]} A finite numeric vector giving the order of each
of the inner
generalized means. The default computes an arithmetic mean and a harmonic
mean.}

\item{outer_weights}{\verb{[numeric(2)]} A strictly positive numeric vector
weights for each of
the inner generalized means as used in the outer generalized mean. The
default weights each inner generalized mean equally.}

\item{outer_order}{\verb{[numeric(1)]} A finite number giving the order of the
outer generalized mean. The default uses a geometric mean.}

\item{na.rm}{\verb{[logical(1)]} Should missing values in \code{x} and \code{weights} be
removed? By
default missing values are not removed. Note that removal of missing values
is balanced across \code{x} and both elements of \code{weights}.}
}
\value{
A numeric value for the nested generalized mean.
}
\description{
Calculate a weighted (outer) generalized mean of two (inner) generalized
means (i.e., crossing means).
}
\examples{
x <- 1:3
w1 <- c(0.25, 0.25, 0.5)
w2 <- c(0.3, 0.3, 0.4)
# Calculate the geometric mean of the arithmetic and harmonic means
# to make a Fisher index.
nested_gmean(x, list(w1, w2))
}
\seealso{
Other math functions:
\code{\link[=emean]{emean()}},
\code{\link[=gmean]{gmean()}},
\code{\link[=scale_weights]{scale_weights()}},
\code{\link[=transmute_weights]{transmute_weights()}},
\code{\link[=transmute_weights2]{transmute_weights2()}},
\code{\link[=update_weights]{update_weights()}}
}
\concept{math functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights.piar_aggregation_structure.R
\name{weights.piar_aggregation_structure}
\alias{weights.piar_aggregation_structure}
\alias{weights<-}
\alias{weights<-.piar_aggregation_structure}
\alias{set_weights}
\title{Get the weights for an aggregation structure}
\usage{
\method{weights}{piar_aggregation_structure}(object, ..., ea_only = TRUE, na.rm = FALSE)

weights(object) <- value

\method{weights}{piar_aggregation_structure}(object) <- value

set_weights(object, value)
}
\arguments{
\item{object}{\verb{[piar_aggregation_structure]} A price index aggregation
structure, as made by \code{\link[=aggregation_structure]{aggregation_structure()}}.}

\item{...}{Not currently used.}

\item{ea_only}{\verb{[logical(1)]} Should weights be returned for only the
elementary aggregates
(the default)? Setting to \code{FALSE} gives the weights for the entire
aggregation structure.}

\item{na.rm}{\verb{[logical(1)]} Should missing values be removed when aggregating
the weights (i.e., when \code{ea_only = FALSE})? By default, missing values are
not removed.}

\item{value}{\verb{[numeric >= 0]} A numeric vector of weights for the elementary
aggregates of \code{object}.}
}
\value{
\code{weights()} returns a named vector of weights for the elementary aggregates.
The replacement method replaces these values without changing the
aggregation structure. (\code{set_weights()} is an alias that's easier to use with
pipes.)

If \code{ea_only = FALSE} then the return value is a list
with a named vector of weights for each level in the aggregation structure.
}
\description{
Get and set the weights for a price index aggregation structure.
}
\examples{
# A simple aggregation structure.
#            1
#      |-----+-----|
#      11          12
#  |---+---|       |
#  111     112     121
#  (1)     (3)     (4)
aggregation_weights <- data.frame(
  level1 = c("1", "1", "1"),
  level2 = c("11", "11", "12"),
  ea     = c("111", "112", "121"),
  weight = c(1, 3, 4)
)

pias <- as_aggregation_structure(aggregation_weights)

# Extract the weights.
weights(pias)

# ... or update them.
weights(pias) <- 1:3
weights(pias)
}
\seealso{
Other aggregation structure methods:
\code{\link[=as.matrix.piar_aggregation_structure]{as.matrix.piar_aggregation_structure()}},
\code{\link[=cut.piar_aggregation_structure]{cut.piar_aggregation_structure()}},
\code{\link[=levels.piar_aggregation_structure]{levels.piar_aggregation_structure()}},
\code{\link[=update.piar_aggregation_structure]{update.piar_aggregation_structure()}}
}
\concept{aggregation structure methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce-index.R
\name{as.ts.piar_index}
\alias{as.ts.piar_index}
\title{Coerce an index into a time series}
\usage{
\method{as.ts}{piar_index}(x, ...)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{...}{Additional arguments passed to \code{\link[=ts]{ts()}}.}
}
\value{
A time series object.
}
\description{
Turn an index into a regular time series, represented as a \code{\link{ts}} object.
}
\examples{
as.ts(as_index(matrix(1:9, 3)))

}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update.aggregation_structure.R
\name{update.piar_aggregation_structure}
\alias{update.piar_aggregation_structure}
\title{Update an aggregation structure}
\usage{
\method{update}{piar_aggregation_structure}(object, index, ..., period = NULL, order = 1, r = order)
}
\arguments{
\item{object}{\verb{[piar_aggregation_structure]} A price index aggregation
structure, as made by \code{\link[=aggregation_structure]{aggregation_structure()}}.}

\item{index}{\verb{[piar_index]} A fixed-base (direct) price index, or something
that can be coerced into one. Usually an aggregate price index as made by
\code{\link[=aggregate.piar_index]{aggregate()}}.}

\item{...}{Not currently used.}

\item{period}{\verb{[character(1)]} The time period used to price update the
weights. The default uses the last period in \code{index}.}

\item{order}{\verb{[numeric(1)]} Order of the generalized mean to update the
weights. The default is 1 for an arithmetic index.}

\item{r}{Deprecated.}
}
\value{
A copy of \code{object} with price-updated weights using the index
values in \code{index}.
}
\description{
Price update the weights in a price index aggregation structure.
}
\examples{
# A simple aggregation structure.
#            1
#      |-----+-----|
#      11          12
#  |---+---|       |
#  111     112     121
#  (1)     (3)     (4)
aggregation_weights <- data.frame(
  level1 = c("1", "1", "1"),
  level2 = c("11", "11", "12"),
  ea     = c("111", "112", "121"),
  weight = c(1, 3, 4)
)

pias <- as_aggregation_structure(aggregation_weights)

index <- as_index(
  matrix(1:9, 3, dimnames = list(c("111", "112", "121"), NULL))
)

weights(pias, ea_only = FALSE)

weights(update(pias, index), ea_only = FALSE)
}
\seealso{
\code{\link[=aggregate.piar_index]{aggregate()}} to make an aggregated price index.

Other aggregation structure methods:
\code{\link[=as.matrix.piar_aggregation_structure]{as.matrix.piar_aggregation_structure()}},
\code{\link[=cut.piar_aggregation_structure]{cut.piar_aggregation_structure()}},
\code{\link[=levels.piar_aggregation_structure]{levels.piar_aggregation_structure()}},
\code{\link[=weights.piar_aggregation_structure]{weights.piar_aggregation_structure()}}
}
\concept{aggregation structure methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/levels.piar_aggregation_structure.R
\name{levels.piar_aggregation_structure}
\alias{levels.piar_aggregation_structure}
\title{Get the levels for an aggregation structure}
\usage{
\method{levels}{piar_aggregation_structure}(x)
}
\arguments{
\item{x}{\verb{[piar_aggregation_structure]} A price index aggregation structure,
as made by \code{\link[=aggregation_structure]{aggregation_structure()}}.}
}
\value{
A list of character vectors giving the levels for each position in the
aggregation structure.
}
\description{
Get the hierarchical list of levels for an aggregation structure. It is
an error to try and replace these values.
}
\seealso{
Other aggregation structure methods:
\code{\link[=as.matrix.piar_aggregation_structure]{as.matrix.piar_aggregation_structure()}},
\code{\link[=cut.piar_aggregation_structure]{cut.piar_aggregation_structure()}},
\code{\link[=update.piar_aggregation_structure]{update.piar_aggregation_structure()}},
\code{\link[=weights.piar_aggregation_structure]{weights.piar_aggregation_structure()}}
}
\concept{aggregation structure methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.piar_index.R
\name{[.piar_index}
\alias{[.piar_index}
\alias{[<-.piar_index}
\title{Extract index values}
\usage{
\method{[}{piar_index}(x, i, j, ...)

\method{[}{piar_index}(x, i, j, ...) <- value
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{i, j}{\verb{[integer | logical | character | matrix]} Indices for the levels
and time periods of a price index. See details.}

\item{...}{Not currently used.}

\item{value}{\verb{[numeric > 0 | piar_index | list]} A numeric vector,
price index, or list of price indexes. See details.}
}
\value{
A price index that inherits from the same class as \code{x}.
}
\description{
Methods to extract and replace index values like a matrix.
}
\details{
The extraction method treats \code{x} like a matrix of index values with
(named) rows for each level and columns for each time period in
\code{x}. Unlike a matrix, dimensions are never dropped as subscripting
\code{x} always returns an index object. The one exception is when subscripting
with a matrix, in which case a list of index values is returned. As \code{x}
is not an atomic vector, subscripting with a single index like \code{x[1]}
extracts all time periods for that level.

The replacement method similarly treat \code{x} like a matrix. If \code{value} is
an index object with the same number of time periods as \code{x[i, j]} and
it inherits from the same class as \code{x}, then the index values and
percent-change contributions of \code{x[i, j]} are replaced with those for the
corresponding levels of \code{value}. If \code{value} is not an index, then it is
coerced to a numeric vector and behaves the same as replacing values in a
matrix. Note that replacing the values of an index will remove the
corresponding percent-change contributions (if any). When replacing with a
matrix, \code{value} can be a list of index objects.
}
\examples{
index <- as_index(matrix(1:6, 2))

index["1", ]

index[, 2]

index[1, ] <- 1 # can be useful for doing specific imputations

index
}
\seealso{
Other index methods:
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregation_structure.R
\name{aggregation_structure}
\alias{aggregation_structure}
\alias{piar_aggregation_structure}
\title{Make a price index aggregation structure}
\usage{
aggregation_structure(x, weights = NULL)
}
\arguments{
\item{x}{\verb{[list]} A list of character vectors that give the codes/labels for
each
level of the classification, ordered so that moving down the list goes down
the hierarchy. The last vector gives the elementary aggregates, which
should have no duplicates. All vectors should be the same length, without
\code{NA}s, and there should be no duplicates across different levels of
\code{x}. Names for \code{x} are used as level names; otherwise, levels are named
'level1', 'level2', ..., 'ea'.}

\item{weights}{\verb{[numeric >= 0]} A numeric vector of aggregation weights for
the elementary
aggregates (i.e., the last vector in \code{x}), or something that can be coerced
into one. The default is to give each elementary aggregate the same weight.}
}
\value{
A price index aggregation structure of class \code{piar_aggregation_structure}.
This is a list-S3 class with the following components.

\item{child}{A nested list that gives the positions of the immediate
children for each node in each level of the aggregation structure above the
terminal nodes.}
\item{parent}{A list that gives the position of the
immediate parent for each node of the aggregation structure below the
initial nodes.}
\item{levels}{A named list of character vectors that give the levels of \code{x}.}
\item{weights}{A vector giving the weight for each elementary
aggregate.}
}
\description{
Create a price index aggregation structure from a hierarchical
classification and aggregation weights that can be used to aggregate
elementary indexes.
}
\section{Warning}{
 The \code{aggregation_structure()} function does its best
to check its arguments, but there should be no expectation that the result
of \code{aggregation_structure()} will make any sense if \code{x} does not
represent a nested hierarchy.
}

\examples{
# A simple aggregation structure.
#            1
#      |-----+-----|
#      11          12
#  |---+---|       |
#  111     112     121
#  (1)     (3)     (4)
aggregation_weights <- data.frame(
  level1 = c("1", "1", "1"),
  level2 = c("11", "11", "12"),
  ea     = c("111", "112", "121"),
  weight = c(1, 3, 4)
)

aggregation_structure(
  aggregation_weights[1:3],
  weights = aggregation_weights[[4]]
)

# The aggregation structure can also be made by expanding the
# elementary aggregates.
with(
  aggregation_weights,
  aggregation_structure(expand_classification(ea), weight)
)
}
\seealso{
\code{\link[=aggregate.piar_index]{aggregate()}} to aggregate price indexes made
with \code{\link[=elementary_index]{elementary_index()}}.

\code{\link[=expand_classification]{expand_classification()}} to make \code{x} from a character
representation of a hierarchical aggregation structure.

\code{\link[=as_aggregation_structure]{as_aggregation_structure()}} to coerce tabular data into an
aggregation structure.

\code{\link[=as.data.frame.piar_aggregation_structure]{as.data.frame()}} and
\code{\link[=as.matrix.piar_aggregation_structure]{as.matrix()}} to coerce an
aggregation structure into a tabular form.

\code{\link[=weights.piar_aggregation_structure]{weights()}} to get the
weights for an aggregation structure.

\code{\link[=update.piar_aggregation_structure]{update()}} for updating a
price index aggregation structure with an aggregated index.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_classifications.R
\name{combine_classifications}
\alias{combine_classifications}
\title{Combine hierarchical classifications}
\usage{
combine_classifications(..., sep = ".")
}
\arguments{
\item{...}{\verb{[list]} A collection of lists, one for each classification,
each giving
the "digits" that represent each level in the hierarchy, as made
by \code{\link[=expand_classification]{expand_classification()}} or \code{\link[=split_classification]{split_classification()}}.}

\item{sep}{\verb{[character(1)]} A character used to separate the classifications
in \code{...}. The
default separates levels across classifications by \code{"."}.}
}
\value{
A list with a entry for each level in the combined classification.
}
\description{
Combine hierarchical classifications by stacking one classification after
another.
}
\examples{
# Combine an unbalanced industry classification with a balanced
# geographic classification.
industry <- c("111", "112", "12")
region <- c("11", "21", "22")

combine_classifications(
  expand_classification(industry, pad = "0"),
  expand_classification(region)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.piar_index.R
\name{aggregate.piar_index}
\alias{aggregate.piar_index}
\alias{aggregate.chainable_piar_index}
\alias{aggregate.direct_piar_index}
\title{Aggregate elementary price indexes}
\usage{
\method{aggregate}{chainable_piar_index}(
  x,
  pias,
  ...,
  pias2 = NULL,
  na.rm = FALSE,
  contrib = TRUE,
  order = 1,
  r = order,
  include_ea = TRUE,
  duplicate_contrib = c("sum", "make.unique"),
  impute_rules = NULL
)

\method{aggregate}{direct_piar_index}(
  x,
  pias,
  ...,
  pias2 = NULL,
  na.rm = FALSE,
  contrib = TRUE,
  order = 1,
  r = order,
  include_ea = TRUE,
  duplicate_contrib = c("sum", "make.unique"),
  impute_rules = NULL
)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, usually made by \code{\link[=elementary_index]{elementary_index()}}.}

\item{pias}{\verb{[piar_aggregation_structure]} A price index
aggregation structure or something that can be coerced into one. This can
be made with \code{\link[=aggregation_structure]{aggregation_structure()}}.}

\item{...}{Not currently used.}

\item{pias2}{\verb{[piar_aggregation_structure]} An optional secondary
aggregation structure, usually with current-period weights, to make a
superlative index. See details.}

\item{na.rm}{\verb{[logical(1)]} Should missing values be removed? By default,
missing values are not removed. Setting \code{na.rm = TRUE} is equivalent to
overall-mean imputation.}

\item{contrib}{\verb{[logical(1)]} Aggregate percent-change contributions in \code{x}?
By default contributions are aggregated.}

\item{order}{\verb{[numeric(1)]} Order of the generalized mean to aggregate
index values. 0 for a
geometric index (the default for making elementary indexes), 1 for an
arithmetic index (the default for aggregating elementary indexes and
averaging indexes over subperiods), or -1 for a harmonic index (usually for
a Paasche index). Other values are possible; see
\code{\link[=gmean]{gmean()}} for details. If \code{pias2} is given then the
index is aggregated with a quadratic mean of order \code{2*r}.}

\item{r}{Deprecated.}

\item{include_ea}{\verb{[logical(1)]} Should indexes for the elementary aggregates
be included along with the aggregated indexes? By default, all index values
are returned.}

\item{duplicate_contrib}{\verb{[character(1)]} The method to deal with duplicate
product contributions. Either \code{"make.unique"} to treat duplicate
products as distinct products and make their names unique
with \code{\link[=make.unique]{make.unique()}} or \code{"sum"} to add contributions for each product
(the default).}

\item{impute_rules}{\verb{[function]} (Experimental) A function that applies
imputation rules to the elementary indexes in each time period prior to
aggregation.
It takes two arguments, the elementary indexes for a given time period and
the (price updated) aggregation structure, and must return back the
elementary indexes.}
}
\value{
An aggregate price index that inherits from the class of \code{x}.
}
\description{
Aggregate elementary price indexes with a price index aggregation structure.
}
\details{
The \code{aggregate()} method loops over each time period in \code{x} and
\enumerate{
\item aggregates the elementary indexes with \code{\link[=gmean]{gmean()}} for each level
of \code{pias};
\item aggregates percent-change contributions for each level of
\code{pias} (if there are any and \code{contrib = TRUE});
\item price updates the weights in \code{pias} with \code{\link[=update_weights]{update_weights()}} (only for
period-over-period elementary indexes).
}

The result is a collection of aggregated period-over-period indexes that
can be chained together to get a fixed-base index when \code{x} are
period-over-period elementary indexes. Otherwise, when \code{x} are fixed-base
elementary indexes, the result is a collection of aggregated fixed-base
(direct) indexes.

By default, missing elementary indexes will propagate when aggregating the
index. Missing elementary indexes can be due to both missingness of these
values in \code{x}, and the presence of elementary aggregates in \code{pias}
that are not part of \code{x}. Setting \code{na.rm = TRUE} ignores missing
values, and is equivalent to overall-mean (or parental) imputation. As an
aggregated price index generally cannot have missing values (for otherwise
it can't be chained over time and weights can't be price updated), any
missing values for a level of \code{pias} are removed and recursively replaced
by the value of its immediate parent.

In most cases aggregation is done with an arithmetic mean (the default), and
this is detailed in chapter 8 (pp. 190--198) of the CPI manual (2020), with
analogous details in chapter 9 of the PPI manual (2004).
Aggregating with a non-arithmetic mean follows the same steps, except that
the elementary indexes are aggregated with a mean of a different order (e.g.,
harmonic for a Paasche index), and the method for price updating the weights
is slightly different. Note that, because aggregation is done with a
generalized mean, the resulting index is consistent-in-aggregation at each
point in time.

Aggregating percent-change contributions uses the method in chapter 9 of the
CPI manual (equations 9.26 and 9.28) when aggregating with an arithmetic
mean. With a non-arithmetic mean, arithmetic weights are constructed using
\code{\link[=transmute_weights]{transmute_weights()}} in order to apply this method.

There may not be contributions for all prices relatives in an elementary
aggregate if the elementary indexes are built from several sources (as with
\code{\link[=merge.piar_index]{merge()}}). In this case the contribution for
a price relative in the aggregated index will be correct, but the sum of all
contributions will not equal the change in the value of the index.

If two aggregation structures are given then the steps above are done for
each aggregation structure, with the aggregation for \code{pias} done with a
generalized mean of order \code{order} the aggregation for \code{pias2} done with a
generalized mean of order \code{-order}. The resulting indexes are combined with a
geometric mean to make a superlative quadratic mean of order \code{2*order} index.
Percent-change contributions are combined using a generalized van IJzeren
decomposition; see \code{\link[=transmute_weights2]{transmute_weights2()}} for details.
}
\note{
For large indexes it can be much faster to turn the aggregation structure
into an aggregation matrix with
\code{\link[=as.matrix.piar_aggregation_structure]{as.matrix()}}, then aggregate
elementary indexes as a matrix operation when there are no missing
values. See the examples for details.
}
\examples{
prices <- data.frame(
  rel = 1:8,
  period = rep(1:2, each = 4),
  ea = rep(letters[1:2], 4)
)

# A two-level aggregation structure.
pias <- aggregation_structure(
  list(c("top", "top", "top"), c("a", "b", "c")),
  weights = 1:3
)

# Calculate Jevons elementary indexes.
(elementary <- elementary_index(prices, rel ~ period + ea))

# Aggregate (note the imputation for elementary index 'c').
(index <- aggregate(elementary, pias, na.rm = TRUE))

# Aggregation can equivalently be done as matrix multiplication.
as.matrix(pias) \%*\% as.matrix(chain(index[letters[1:3]]))
}
\references{
Balk, B. M. (2008). \emph{Price and Quantity Index Numbers}.
Cambridge University Press.

ILO, IMF, UNECE, OECD, and World Bank. (2004).
\emph{Producer Price Index Manual: Theory and Practice}.
International Monetary Fund.

IMF, ILO, OECD, Eurostat, UNECE, and World Bank. (2020).
\emph{Consumer Price Index Manual: Concepts and Methods}.
International Monetary Fund.

von der Lippe, P. (2007). \emph{Index Theory and Price Statistics}. Peter Lang.
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.piar_index.R
\name{merge.piar_index}
\alias{merge.piar_index}
\alias{merge.chainable_piar_index}
\alias{merge.direct_piar_index}
\title{Merge price indexes}
\usage{
\method{merge}{chainable_piar_index}(x, y, ...)

\method{merge}{direct_piar_index}(x, y, ...)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{y}{\verb{[piar_index]} A price index, or something that can coerced into
one. If \code{x}
is a period-over-period index then \code{y} is coerced into a chainable
index; otherwise, \code{y} is coerced into a direct index.}

\item{...}{Not currently used.}
}
\value{
A combined price index that inherits from the same class as \code{x}.
}
\description{
Combine two price indexes with common time periods, merging together the
index values and percent-change contributions for each time period.

This is useful for building up an index when different elementary aggregates
come from different sources of data, or use different index-number formulas.
}
\examples{
index1 <- as_index(matrix(1:6, 2))

index2 <- index1
levels(index2) <- 3:4

merge(index1, index2)
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chain.R
\name{chain}
\alias{chain}
\alias{chain.default}
\alias{chain.chainable_piar_index}
\alias{unchain}
\alias{unchain.default}
\alias{unchain.direct_piar_index}
\alias{rebase}
\alias{rebase.default}
\alias{rebase.direct_piar_index}
\title{Chain and rebase a price index}
\usage{
chain(x, ...)

\method{chain}{default}(x, ...)

\method{chain}{chainable_piar_index}(x, ..., link = NULL)

unchain(x, ...)

\method{unchain}{default}(x, ...)

\method{unchain}{direct_piar_index}(x, ..., base = NULL)

rebase(x, ...)

\method{rebase}{default}(x, ...)

\method{rebase}{direct_piar_index}(x, ..., base = NULL)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{...}{Further arguments passed to or used by methods.}

\item{link}{\verb{[numeric > 0]} A numeric vector, or something that can coerced
into one, of
link values for each level in \code{x}. The default is equivalent to a vector
of 1s so that no linking is done.}

\item{base}{\verb{[numeric > 0 | character(1)]} A numeric vector, or something
that can coerced into one, of
base-period index values for each level in \code{x}. The default is a equivalent
to a vector
of 1s so that the base period remains the same. If \code{base} is a length-one
character vector giving a time period of \code{x} then the index values for this
time period are used as the base-period values.}
}
\value{
\code{chain()} and \code{rebase()} return a fixed-base index that inherits
from \code{\link{direct_piar_index}}.

\code{unchain()} returns a period-over-period index that inherits from
\code{\link{chainable_piar_index}}.
}
\description{
Chain a period-over-period index by taking the cumulative product of its
values to turn it into a fixed-base (direct) index.

Unchain a fixed-base index by dividing its values for successive periods to
get a period-over-period index.

Rebase a fixed-base index by dividing its values with the value of the index
in the new base period.
}
\details{
The default methods attempt to coerce \code{x} into an index with
\code{\link[=as_index]{as_index()}} prior to chaining/unchaining/rebasing.

Chaining an index takes the cumulative product of the index values for each
level; this is roughly the same as
\code{t(apply(as.matrix(x), 1, cumprod)) * link}. Unchaining does the opposite,
so these are inverse operations. Note that unchaining a period-over-period
index does nothing, as does chaining a fixed-base index.

Rebasing a fixed-base index divides the values for each level of this index
by the corresponding values for each level in the new base period. It's
roughly the same as \code{as.matrix(x) / base}. Like unchaining, rebasing a
period-over-period index does nothing.

Percent-change contributions are removed when chaining/unchaining/rebasing
an index as it's not usually possible to update them correctly.
}
\examples{
index <- as_index(matrix(1:9, 3))

# Make period 0 the fixed base period.
chain(index)

# Chaining and unchaining reverse each other.
all.equal(index, unchain(chain(index)))

# Change the base period to period 2 (note the
# loss of information for period 0).
index <- chain(index)
rebase(index, base = index[, 2])
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outliers.R
\name{outliers}
\alias{outliers}
\alias{hb_transform}
\title{Outlier detection for price relatives}
\usage{
outliers(
  x,
  upper,
  lower = upper,
  method = c("quartile", "resistant-fences", "kimber", "robust-z", "tukey"),
  scale = 0,
  quantile_type = 7
)

hb_transform(x)
}
\arguments{
\item{x}{\verb{[numeric > 0]} A numeric vector, usually of price relatives.
These can be made with, e.g., \code{\link[=back_period]{back_period()}}.}

\item{upper, lower}{\verb{[numeric(1) > 0]} A number giving the upper and lower
cutoffs for
each element of \code{x}. By default the lower cutoff is the same as the upper
cutoff}

\item{method}{\verb{[character(1)]} The outlier detection method, one of
\code{"quartile"} (the default), \code{"resistant-fences"}, \code{"kimber"}, \code{"robust-z"},
or \code{"tukey"}.}

\item{scale}{\verb{[0 <= numeric(1) <= 1]} A number between 0 and 1 giving the
scale factor for the
median to establish the minimum dispersion between quartiles for each
element of \code{x}. The default does not set a minimum dispersion.}

\item{quantile_type}{\verb{[integer(1)]} See \code{\link[=quantile]{quantile()}}.}
}
\value{
A logical vector, the same length as \code{x}, that is \code{TRUE} if the
corresponding element of \code{x} is identified as an outlier,
\code{FALSE} otherwise.
}
\description{
Standard cutoff-based methods for detecting outliers with price relatives.
}
\details{
This function constructs an interval of the form \eqn{[b_l(x) -
c_l \times l(x), b_u(x) + c_u \times u(x)]}{[bl(x) - cl * l(x), bu(x) + cu *
u(x)]} and assigns a value in \code{x} as \code{TRUE} if that value does not
belong to the interval, \code{FALSE} otherwise. The different methods differ in
how they construct the values \eqn{b_l(x)}{bl(x)}, \eqn{b_u(x)}{bu(x)},
\eqn{l(x)}, and \eqn{u(x)}. Any missing values in \code{x} are ignored when
calculating the interval, but will return \code{NA}.

The quartile method and Tukey algorithm are described in paragraphs 5.113 to
5.135 of the CPI manual (2020).
The resistant fences method is an alternative to the quartile method, and is
described by Rais (2008) and Hutton (2008). The Kimber method is yet another
alternative. Quantile-based methods often
identify price relatives as outliers because the distribution is
concentrated around 1; setting \code{scale > 0} puts a floor on the minimum
dispersion between quantiles as a fraction of the median. See the references
for more details.\tabular{lllll}{
    \tab \eqn{b_l(x)}{bl(x)} \tab \eqn{b_u(x)}{bu(x)} \tab \eqn{l(x)} \tab \eqn{u(x)} \cr
   Quartile \tab \eqn{Q_2(x)} \tab \eqn{Q_2(x)} \tab \eqn{Q_2(x) - Q_1(x)} \tab \eqn{Q_3(x) - Q_2(x)} \cr
   Resistant fences \tab \eqn{Q_1(x)} \tab \eqn{Q_3(x)} \tab \eqn{Q_3(x) - Q_1(x)} \tab \eqn{Q_3(x) - Q_1(x)} \cr
   Kimber \tab \eqn{Q_1(x)} \tab \eqn{Q_3(x)} \tab \eqn{Q_2(x) - Q_1(x)} \tab \eqn{Q_3(x) - Q_2(x)} \cr
}


The robust Z-score is the usual method to identify relatives in the
(asymmetric) tails of the distribution, simply replacing the mean with the
median, and the standard deviation with the median absolute deviation.

These methods often assume that price relatives are symmetrically
distributed (if not Gaussian). As the distribution of price relatives often
has a long right tail, the natural logarithm can be used to transform price
relative before identifying outliers (sometimes under the assumption that
price relatives are distributed log-normal). The Hidiroglou-Berthelot
transformation is another approach, described in the CPI manual (par.
5.124). (Sometimes the transformed price relatives are multiplied by
\eqn{\max(p_1, p_0)^u}{max(p1, p0)^u}, for some
\eqn{0 \le u \le 1}{0 <= u <= 1}, so that products with a larger price
get flagged as outliers (par. 5.128).)
}
\examples{
x <- c(1, 10, 15, 100)

outliers(x, upper = 2.5, method = "quartile")

# Always identifies fewer outliers than above.
outliers(x, upper = 2.5, method = "resistant-fences")
}
\references{
Hutton, H. (2008). Dynamic outlier detection in price index surveys.
\emph{Proceedings of the Survey Methods Section: Statistical Society of Canada
Annual Meeting}.

IMF, ILO, Eurostat, UNECE, OECD, and World Bank. (2020).
\emph{Consumer Price Index Manual: Concepts and Methods}.
International Monetary Fund.

Rais, S. (2008). Outlier detection for the Consumer Price Index.
\emph{Proceedings of the Survey Methods Section: Statistical Society of Canada
Annual Meeting}.
}
\seealso{
\code{\link[=back_period]{back_period()}} for a simple utility function to turn prices
in a table into price relatives.

The \code{HBmethod()} function in the \pkg{univOutl} package for the
Hidiroglou-Berthelot method for identifying outliers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mean.piar_index.R
\name{mean.piar_index}
\alias{mean.piar_index}
\alias{mean.chainable_piar_index}
\alias{mean.direct_piar_index}
\title{Aggregate a price index over subperiods}
\usage{
\method{mean}{chainable_piar_index}(
  x,
  ...,
  weights = NULL,
  window = NULL,
  na.rm = FALSE,
  contrib = TRUE,
  order = 1,
  r = order,
  duplicate_contrib = c("sum", "make.unique")
)

\method{mean}{direct_piar_index}(
  x,
  ...,
  weights = NULL,
  window = NULL,
  na.rm = FALSE,
  contrib = TRUE,
  order = 1,
  r = order,
  duplicate_contrib = c("sum", "make.unique")
)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{...}{Not currently used.}

\item{weights}{\verb{[numeric >= 0]} A numeric vector of weights for the index
values in \code{x}, or
something that can be coerced into one. The
default is equal weights. It is usually easiest to specify these weights as
a matrix with a row for each index value in \code{x} and a column for each
time period.}

\item{window}{\verb{[integer(1) > 0]} A positive integer giving the size of the
window used to
average index values across subperiods. The default averages over all
periods in \code{x}. Non-integers are truncated towards 0.}

\item{na.rm}{\verb{[logical(1)]} Should missing values be removed? By default,
missing values
are not removed. Setting \code{na.rm = TRUE} is equivalent to overall-mean
imputation.}

\item{contrib}{\verb{[logical(1)]} Aggregate percent-change contributions in \code{x}?
By default contributions are aggregated.}

\item{order}{\verb{[numeric(1)]} Order of the generalized mean to aggregate
index values. 0 for a
geometric index (the default for making elementary indexes), 1 for an
arithmetic index (the default for aggregating elementary indexes and
averaging indexes over subperiods), or -1 for a harmonic index (usually for
a Paasche index). Other values are possible; see
\code{\link[=gmean]{gmean()}} for details.}

\item{r}{Deprecated.}

\item{duplicate_contrib}{\verb{[character(1)]} The method to deal with duplicate
product contributions. Either \code{"make.unique"} to make duplicate product
names unique with \code{\link[=make.unique]{make.unique()}} or \code{"sum"} to add contributions for the
same products across subperiods (the default).}
}
\value{
A price index, averaged over subperiods, that inherits from the same
class as \code{x}.
}
\description{
Aggregate an index over subperiods by taking the (usually arithmetic) mean
of index values over consecutive windows of subperiods.
}
\details{
The \code{mean()} method constructs a set of non-overlapping windows of
length \code{window}, starting in the first period of the index, and takes
the mean of each index value in these windows for each level of the index.
The last window is discarded if it is incomplete (with a warning), so that
index values are
always averaged over \code{window} periods. The names for the first time
period in each window form the new names for the aggregated time periods.

Percent-change contributions are aggregated if \code{contrib = TRUE} following the
same approach as \code{\link[=aggregate.piar_index]{aggregate()}}.

An optional vector of weights can be specified when aggregating index values
over subperiods, which is often useful when aggregating a Paasche index; see
section 4.3 of Balk (2008) for details.
}
\examples{
index <- as_index(matrix(c(1:12, 12:1), 2, byrow = TRUE), chainable = FALSE)

# Turn a monthly index into a quarterly index.
mean(index, window = 3)
}
\references{
Balk, B. M. (2008). \emph{Price and Quantity Index Numbers}.
Cambridge University Press.
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elementary_index.R
\name{elementary_index}
\alias{elementary_index}
\alias{elementary_index.default}
\alias{elementary_index.numeric}
\alias{elementary_index.data.frame}
\alias{elemental_index}
\title{Make elementary/elemental price indexes}
\usage{
elementary_index(x, ...)

\method{elementary_index}{default}(x, ...)

\method{elementary_index}{numeric}(
  x,
  ...,
  period = NULL,
  ea = NULL,
  weights = NULL,
  product = NULL,
  chainable = TRUE,
  na.rm = FALSE,
  contrib = FALSE,
  order = 0,
  r = order
)

\method{elementary_index}{data.frame}(x, formula, ..., weights = NULL, product = NULL)

elemental_index(x, ...)
}
\arguments{
\item{x}{\verb{[object]} Period-over-period or fixed-base price
relatives. Currently there are methods for numeric vectors (which can be
made with \code{\link[=price_relative]{price_relative()}}) and data frames.}

\item{...}{Further arguments passed to or used by methods.}

\item{period}{\verb{[factor]} A factor, or something that can be coerced into one,
giving
the time period associated with each price relative in \code{x}. The
ordering of time periods follows of the levels of \code{period}, to agree
with \code{\link[=cut.Date]{cut()}}. The default makes an index for one time period.}

\item{ea}{\verb{[factor]} A factor, or something that can be coerced into one,
giving the
elementary aggregate associated with each price relative in \code{x}. The
default makes an index for one elementary aggregate.}

\item{weights}{\verb{[numeric >= 0]} A numeric vector of weights for the price
relatives in \code{x},
or something that can be coerced into one. The default is equal weights.
This is evaluated in \code{x} for the data frame method.}

\item{product}{\verb{[character]} A character vector of product names, or
something that can
be coerced into one, for each price relative in \code{x} when making
percent-change contributions. The default uses the names of \code{x}, if any;
otherwise, elements of \code{x} are given sequential names within each
elementary aggregate. This is evaluated in \code{x} for the data frame method.}

\item{chainable}{\verb{[logical(1)]} Are the price relatives in \code{x}
period-over-period
relatives that are suitable for a chained calculation (the default)? This
should be \code{FALSE} when \code{x} contains fixed-base relatives.}

\item{na.rm}{\verb{[logical(1)]} Should missing values be removed? By default,
missing values
are not removed. Setting \code{na.rm = TRUE} is equivalent to overall-mean
imputation.}

\item{contrib}{\verb{[logical(1)]} Should percent-change contributions be
calculated? The default does not calculate contributions.}

\item{order}{\verb{[numeric(1)]} Order of the generalized mean to aggregate
price relatives. 0 for a
geometric index (the default for making elementary indexes), 1 for an
arithmetic index (the default for aggregating elementary indexes and
averaging indexes over subperiods), or -1 for a harmonic index (usually for
a Paasche index). Other values are possible; see
\code{\link[=gmean]{gmean()}} for details.}

\item{r}{Deprecated.}

\item{formula}{\verb{[formula]} A two-sided formula, or something that can be
coerced into one, with price relatives on the left-hand
side and time periods and elementary aggregates (in that order) on the
right-hand side.}
}
\value{
A price index that inherits from \code{\link{piar_index}}. If
\code{chainable = TRUE} then this is a period-over-period index that also
inherits from \code{\link{chainable_piar_index}}; otherwise, it is a
fixed-based index that inherits from \code{\link{direct_piar_index}}.
}
\description{
Compute period-over-period (chainable) or fixed-base (direct) elementary
price indexes, with optional percent-change contributions for each
product.
}
\details{
When supplied with a numeric vector, \code{elementary_index()} is a simple
wrapper that applies \code{\link[=gmean]{gmean()}} and \code{\link[=transmute_weights]{transmute_weights()}}
(if \code{contrib = TRUE}) to \code{x} and \code{weights} grouped by \code{ea} and \code{period}. That
is, for every combination of elementary aggregate and time period,
\code{elementary_index()} calculates an index based on a generalized mean of
order \code{order} and, optionally, percent-change contributions. Product names
should
be unique within each elementary aggregate at each time period when making
contributions and, if not, are
passed to \code{\link[=make.unique]{make.unique()}} with a warning. The default
(\code{order = 0} and no weights) makes Jevons elementary indexes. See chapter 8
(pp. 175--190) of the CPI manual (2020) for more detail about making
elementary indexes, or chapter 9 of the PPI manual (2004), and chapter 5 of
Balk (2008).

The default method simply coerces \code{x} to a numeric vector prior to
calling the method above. The data frame method provides a formula interface
to specify columns of price relatives, time periods, and elementary
aggregates and call the method above.

The interpretation of the index depends on how the price relatives in
\code{x} are made. If these are period-over-period relatives, then the
result is a collection of period-over-period (chainable) elementary indexes;
if these are fixed-base relatives, then the result is a collection of
fixed-base (direct) elementary indexes. For the latter, \code{chainable}
should be set to \code{FALSE} so that no subsequent methods assume that a
chained calculation should be used.

By default, missing price relatives in \code{x} will propagate throughout
the index calculation. Ignoring missing values with \code{na.rm = TRUE} is
the same as overall-mean imputation, and needs to be explicitly
set in the call to \code{elementary_index()}. Explicit imputation of missing
relatives, and especially imputation of missing prices, should be done prior
to calling \code{elementary_index()}.

Indexes based on nested generalized means, like the Fisher index (and
superlative quadratic mean indexes more generally), can be calculated by
supplying the appropriate weights with \code{\link[=transmute_weights2]{transmute_weights2()}}; see the
example below. It is important to note that there are several ways to
make these weights, and this affects how percent-change contributions
are calculated.

\code{elemental_index()} is an alias for \code{elementary_index()}.
}
\examples{
prices <- data.frame(
  rel = 1:8,
  period = rep(1:2, each = 4),
  ea = rep(letters[1:2], 4)
)

# Calculate Jevons elementary indexes.
elementary_index(prices, rel ~ period + ea)

# Same as using lm() or tapply().
exp(coef(lm(log(rel) ~ ea:factor(period) - 1, prices)))

with(
  prices,
  t(tapply(rel, list(period, ea), \(x) gmean(x, order = 0, na.rm = TRUE)))
)

# Calculate a CSWD index (same as the Jevons in this example)
# as an arithmetic index by using the appropriate weights.
cswd_weights <- with(
  prices,
  lapply(split(rel, list(period, ea)), transmute_weights2)
)

elementary_index(
  prices,
  rel ~ period + ea,
  weights = unsplit(cswd_weights, interaction(period, ea)),
  order = 1
)
}
\references{
Balk, B. M. (2008). \emph{Price and Quantity Index Numbers}.
Cambridge University Press.

ILO, IMF, UNECE, OECD, and World Bank. (2004).
\emph{Producer Price Index Manual: Theory and Practice}.
International Monetary Fund.

IMF, ILO, OECD, Eurostat, UNECE, and World Bank. (2020).
\emph{Consumer Price Index Manual: Concepts and Methods}.
International Monetary Fund.

von der Lippe, P. (2007). \emph{Index Theory and Price Statistics}. Peter Lang.
}
\seealso{
\code{\link[=price_relative]{price_relative()}} for making price relatives for the same products over
time, and \code{\link[=impute_prices]{impute_prices()}} for imputation of missing prices.

\code{\link[=as_index]{as_index()}} to turn pre-computed (elementary) index values into an
index object.

\code{\link[=chain]{chain()}} for chaining period-over-period indexes, and
\code{\link[=rebase]{rebase()}} for rebasing an index.

\code{\link[=aggregate.piar_index]{aggregate()}} to aggregate elementary indexes
according to an aggregation structure.

\code{\link[=as.matrix.piar_index]{as.matrix()}} and
\code{\link[=as.data.frame.piar_index]{as.data.frame()}} for coercing an index
into a tabular form.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index-class.R
\name{piar_index}
\alias{piar_index}
\alias{chainable_piar_index}
\alias{direct_piar_index}
\title{Price index objects}
\description{
There are several classes to represent price indexes.
\itemize{
\item All indexes inherit from the \code{piar_index} virtual class.
\item Period-over-period indexes that can be chained over time inherit from
\code{chainable_piar_index}.
\item Fixed-base indexes inherit from \code{direct_piar_index}.
}
}
\details{
The \code{piar_index} object is a list-S3 class with the following
components:
\describe{
\item{index}{A matrix of index values with a column for each period in \code{time}
and a row for each level in \code{levels}.}
\item{contrib}{A list-matrix containing named vectors that give the
percent-change contributions for each price relative with a column for each
time period in \code{time} and a row for each level in \code{levels}, or \code{NULL}.}
\item{levels}{A character vector giving the levels of the index.}
\item{time}{A character vector giving the time periods for the index.}
}

The \code{chainable_piar_index} and \code{direct_piar_index} subclasses have
the same structure as the \code{piar_index} class, but differ in the methods
used to manipulate the indexes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack.piar_index.R
\name{stack.piar_index}
\alias{stack.piar_index}
\alias{stack.chainable_piar_index}
\alias{stack.direct_piar_index}
\alias{unstack.chainable_piar_index}
\alias{unstack.direct_piar_index}
\title{Stack price indexes}
\usage{
\method{stack}{chainable_piar_index}(x, y, ...)

\method{stack}{direct_piar_index}(x, y, ...)

\method{unstack}{chainable_piar_index}(x, ...)

\method{unstack}{direct_piar_index}(x, ...)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{y}{\verb{[piar_index]} A price index, or something that can coerced into
one. If \code{x}
is a period-over-period index then \code{y} is coerced into a chainable
index; otherwise, \code{y} is coerced into a direct index.}

\item{...}{Not currently used.}
}
\value{
\code{stack()} returns a combined price index that inherits from the same class
as \code{x}.

\code{unstack()} returns a list of price indexes with the same class as \code{x}.
}
\description{
\code{stack()} combines two price indexes with common levels, stacking index
values and percent-change contributions for one index after the other.

\code{unstack()} breaks up a price index into a list of indexes for each
time period.

These methods can be used in a map-reduce to make an index with multiple
aggregation structures (like a Paasche index).
}
\note{
It may be necessary to use \code{rebase()} prior to stacking fixed-based price
indexes to ensure they have the same base period.
}
\examples{
index1 <- as_index(matrix(1:6, 2))

index2 <- index1
time(index2) <- 4:6

stack(index1, index2)

# Unstack does the reverse.
all.equal(
  c(unstack(index1), unstack(index2)),
  unstack(stack(index1, index2))
)
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregation_structure-class.R
\name{is_aggregation_structure}
\alias{is_aggregation_structure}
\title{Test if an object is an aggregation structure}
\usage{
is_aggregation_structure(x)
}
\arguments{
\item{x}{\verb{[object]} An object to test.}
}
\value{
Returns \code{TRUE} if \code{x} inherits from \code{\link{piar_aggregation_structure}}.
}
\description{
Test if an object is a price index aggregation structure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce-index.R
\name{as.data.frame.piar_index}
\alias{as.data.frame.piar_index}
\alias{as.matrix.piar_index}
\title{Coerce an index into a tabular form}
\usage{
\method{as.data.frame}{piar_index}(
  x,
  row.names = NULL,
  optional = FALSE,
  ...,
  contrib = FALSE,
  stringsAsFactors = FALSE
)

\method{as.matrix}{piar_index}(x, ...)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{row.names}{\verb{[character]} See \code{\link[=as.data.frame]{as.data.frame()}}.}

\item{optional, ...}{Not currently used.}

\item{contrib}{\verb{[logical(1)]} Include percent-change contributions
(the default does not include them).}

\item{stringsAsFactors}{\verb{[logical(1)]} See \code{\link[=as.data.frame]{as.data.frame()}}.}
}
\value{
\code{as.data.frame()} returns the index values in \code{x} as a data frame with three
columns: \code{period}, \code{level}, and \code{value}. If \code{contrib = TRUE} then there is
a fourth (list) column \code{contrib} containing percent-change contributions.

\code{as.matrix()} returns the index values in \code{x} as a matrix with a row for
each level and a column for each time period in \code{x}.
}
\description{
Turn an index into a data frame or a matrix.
}
\examples{
index <- as_index(matrix(1:6, 2))

as.data.frame(index)

as.matrix(index)
}
\seealso{
\code{\link[=as_index]{as_index()}} to coerce a matrix/data frame of index values into an index
object.

Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce-aggregation_structure.R
\name{as.matrix.piar_aggregation_structure}
\alias{as.matrix.piar_aggregation_structure}
\alias{as.data.frame.piar_aggregation_structure}
\title{Coerce an aggregation structure into a tabular form}
\usage{
\method{as.matrix}{piar_aggregation_structure}(x, ..., sparse = FALSE)

\method{as.data.frame}{piar_aggregation_structure}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{\verb{[piar_agregation_structure]} A price index aggregation structure,
as made by \code{\link[=aggregation_structure]{aggregation_structure()}}.}

\item{...}{Not currently used for the matrix method. Extra arguments to
\code{\link[=as.data.frame.list]{as.data.frame.list()}} for the data frame method.}

\item{sparse}{\verb{[logical(1)]} Should the result be a sparse matrix from
\pkg{Matrix}? This is faster for large aggregation structures. The default
returns an ordinary dense matrix.}

\item{row.names}{\verb{[character]} See \code{\link[=as.data.frame]{as.data.frame()}}.}

\item{optional}{Not currently used.}
}
\value{
\code{as.matrix()} represents an aggregation structure as a matrix,
such that multiplying with a (column) vector of elementary indexes gives the
aggregated index.

\code{as.data.frame()} takes an aggregation structure and returns a data
frame that could have generated it.
}
\description{
Coerce a price index aggregation structure into an aggregation matrix or a
data frame.
}
\examples{
# A simple aggregation structure.
#            1
#      |-----+-----|
#      11          12
#  |---+---|       |
#  111     112     121
#  (1)     (3)     (4)
aggregation_weights <- data.frame(
  level1 = c("1", "1", "1"),
  level2 = c("11", "11", "12"),
  ea     = c("111", "112", "121"),
  weight = c(1, 3, 4)
)

pias <- as_aggregation_structure(aggregation_weights)

as.matrix(pias)

all.equal(as.data.frame(pias), aggregation_weights)

\dontrun{
# Visualize as a treemap.
treemap::treemap(
  aggregation_weights,
  index = names(aggregation_weights)[-4],
  vSize = "weight",
  title = "aggregation structure"
)

# Or turn into a more genereal tree object and plot.
aggregation_weights$pathString <- do.call(
  \(...) paste(..., sep = "/"),
  aggregation_weights[-4]
)
plot(data.tree::as.Node(aggregation_weights))
}
}
\seealso{
\code{\link[=as_aggregation_structure]{as_aggregation_structure()}} for coercing into an aggregation structure.

\code{\link[treemap:treemap]{treemap::treemap()}} and \code{\link[data.tree:as.Node]{data.tree::as.Node()}} for visualizing an
aggregation structure.

Other aggregation structure methods:
\code{\link[=cut.piar_aggregation_structure]{cut.piar_aggregation_structure()}},
\code{\link[=levels.piar_aggregation_structure]{levels.piar_aggregation_structure()}},
\code{\link[=update.piar_aggregation_structure]{update.piar_aggregation_structure()}},
\code{\link[=weights.piar_aggregation_structure]{weights.piar_aggregation_structure()}}
}
\concept{aggregation structure methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cut.piar_aggregation_structure.R
\name{cut.piar_aggregation_structure}
\alias{cut.piar_aggregation_structure}
\title{Cut an aggregation structure}
\usage{
\method{cut}{piar_aggregation_structure}(x, level, ..., na.rm = FALSE, upper = TRUE)
}
\arguments{
\item{x}{\verb{[piar_aggregation_structure]} A price index aggregation structure,
as made by \code{\link[=aggregation_structure]{aggregation_structure()}}.}

\item{level}{\verb{[integer(1) > 0]} A positive integer, or something that can be
coerced into one, giving the level at which to cut \code{x}.}

\item{...}{Not currently used.}

\item{na.rm}{\verb{[logical(1)]} Should missing values be removed when aggregating
the weights? By default, missing values are not removed.}

\item{upper}{\verb{[logical(1)]} Keep only the part of \code{x} above \code{level}
(the default); otherwise, return the part of \code{x} below \code{level}.}
}
\value{
A price index aggregation structure of class \code{\link{piar_aggregation_structure}}.
}
\description{
Keep only the part of an aggregation structure above or below a certain
level.
}
\examples{
# A simple aggregation structure.
#            1
#      |-----+-----|
#      11          12
#  |---+---|       |
#  111     112     121
#  (1)     (3)     (4)
aggregation_weights <- data.frame(
  level1 = c("1", "1", "1"),
  level2 = c("11", "11", "12"),
  ea     = c("111", "112", "121"),
  weight = c(1, 3, 4)
)

pias <- aggregation_structure(
  aggregation_weights[1:3],
  weights = aggregation_weights[[4]]
)

# Turn it into
#            1
#      |-----+-----|
#      11          12
#     (4)         (4)
cut(pias, 2)
}
\seealso{
Other aggregation structure methods:
\code{\link[=as.matrix.piar_aggregation_structure]{as.matrix.piar_aggregation_structure()}},
\code{\link[=levels.piar_aggregation_structure]{levels.piar_aggregation_structure()}},
\code{\link[=update.piar_aggregation_structure]{update.piar_aggregation_structure()}},
\code{\link[=weights.piar_aggregation_structure]{weights.piar_aggregation_structure()}}
}
\concept{aggregation structure methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/back_period.R
\name{back_period}
\alias{back_period}
\title{Back period}
\usage{
back_period(period, product = NULL, match_first = TRUE, offset = 1L)
}
\arguments{
\item{period}{\verb{[factor]} A factor, or something that can be coerced into one,
that
gives the time period for each transaction. The ordering of time periods
follows the levels of \code{period} to agree with
\code{\link[=cut.Date]{cut()}}.}

\item{product}{\verb{[factor]} A factor, or something that can be coerced into
one, that
gives the product identifier for each transaction. The default is to
assume that all transactions are for the same product.}

\item{match_first}{\verb{[logical(1)]} Should products in the first period match
with themselves (the default)?}

\item{offset}{\verb{[integer(1)]} The number of periods to offset. The default
offsets by one period (back period). Setting to \code{nlevels(period)} gives
the the base period.}
}
\value{
A numeric vector of indices giving the position of the the back periods.
}
\description{
Offset a vector prices or quantities by computing the position one (or more)
period backwards for each product.
}
\note{
By definition, there must be at most one transaction for each product
in each time period to determine a back period. If multiple transactions
correspond to a period-product pair, then the back period at a point in
time is always the first position for that product in the previous period.
}
\examples{
prices <- data.frame(
  price = 1:6,
  product = factor(c("a", "b")),
  period = factor(c(1, 1, 2, 2, 3, 3))
)

with(prices, back_period(period, product))

# Make fixed-base price relatives.
with(
  prices,
  price / price[back_period(period, product, offset = nlevels(period))]
)

# Change the base period with relevel().
with(
  prices,
  price / price[
    back_period(relevel(period, "2"), product, offset = nlevels(period))
  ]
)
}
\seealso{
\code{\link[=outliers]{outliers()}} for common methods to detect outliers for price relatives.

\code{rs_pairs} in the \pkg{rsmatrix} package for making sales pairs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transmute_weights.R
\name{transmute_weights2}
\alias{transmute_weights2}
\title{Transmute weights for a nested generalized mean}
\usage{
transmute_weights2(
  x,
  weights = list(NULL, NULL),
  order = c(1, -1),
  outer_weights = NULL,
  outer_order = 0,
  to = 1,
  pivot = outer_order
)
}
\arguments{
\item{x}{\verb{[numeric > 0]} A strictly positive numeric vector.}

\item{weights}{\verb{[list]} A list of positive numeric vector of
weights, each
the same length as \code{x}, for both of the inner generalized means. \code{NULL}
elements of \code{weights} equally weight each element of \code{x}. The default
uses equal weights for both inner generalized mean.}

\item{order}{\verb{[numeric(2)]} A finite numeric vector giving the order of each
of the inner
generalized means. The default computes an arithmetic mean and a harmonic
mean.}

\item{outer_weights}{\verb{[numeric(2)]} A strictly positive numeric vector
weights for each of
the inner generalized means as used in the outer generalized mean. The
default weights each inner generalized mean equally.}

\item{outer_order}{\verb{[numeric(1)]} A finite number giving the order of the
outer generalized mean. The default uses a geometric mean.}

\item{to}{A finite number giving the order of the target generalized mean for
the transmuted weights. The default constructs weights for an arithmetic
mean.}

\item{pivot}{A finite number giving the pivot value for the transmuted
weights. The default uses the order of the outer generalized mean,
otherwise \code{to} is common alternative.}
}
\value{
A numeric vector, the same length as \code{x}, that sums to 1.
}
\description{
Transmute weights to turn a nested generalized mean of a given order into a
generalized mean of any order. Useful for calculating additive and
multiplicative decompositions for an index made
of nested generalized means (e.g., Fisher index).
See \code{vignette("decomposing-indexes")} for details.
}
\details{
This function generalizes the additive and multiplicative decompositions
for the Fisher index by Balk (2008, Chapter 4). It returns
a value such that

\preformatted{nested_gmean(x, list(w1, w2), c(r1, r2)) ==
    gmean(x, transmute_weights2(x, list(w1, w2), c(r1, r2), to = s), s)}

Transmuting weights returns a value that is the same length as \code{x},
so any missing values in \code{x} or \code{weights} will return \code{NA}.
Unless all values are \code{NA}, however, the result will still satisfy
the above identity when \code{na.rm = TRUE}.
}
\examples{
x <- 1:3
w1 <- 3:1
w2 <- c(1, 2, 1)

# Calculate the geometric mean of the arithmetic and harmonic means
# as an arithmetic mean.
nested_gmean(x, list(w1, w2))
gmean(x, transmute_weights2(x, list(w1, w2), to = 1))
}
\references{
Balk, B. M. (2008). \emph{Price and Quantity Index Numbers}.
Cambridge University Press.
}
\seealso{
Other math functions:
\code{\link[=emean]{emean()}},
\code{\link[=gmean]{gmean()}},
\code{\link[=nested_gmean]{nested_gmean()}},
\code{\link[=scale_weights]{scale_weights()}},
\code{\link[=transmute_weights]{transmute_weights()}},
\code{\link[=update_weights]{update_weights()}}
}
\concept{math functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/price_relative.R
\name{price_relative}
\alias{price_relative}
\alias{price_relative.default}
\alias{price_relative.data.frame}
\title{Calculate period-over-period price relatives}
\usage{
price_relative(x, ...)

\method{price_relative}{default}(x, period, product, ...)

\method{price_relative}{data.frame}(x, formula, ...)
}
\arguments{
\item{x}{\verb{[object]} Either a numeric vector (or something that can be coerced
into one) or data frame of prices.}

\item{...}{Further arguments passed to or used by methods.}

\item{period}{\verb{[factor]} A factor, or something that can be coerced into one,
that gives the corresponding time period for each element in \code{x}. The
ordering of time periods follows the levels of \code{period} to agree with
\code{\link[=cut.Date]{cut()}}.}

\item{product}{\verb{[factor]} A factor, or something that can be coerced into
one, that
gives the corresponding product identifier for each element in \code{x}.}

\item{formula}{\verb{[formula]} A two-sided formula, or something that can be
coerced into one, with prices on the left-hand
side and time periods and products (in that order) on the
right-hand side.}
}
\value{
A numeric vector of price relatives, with \code{product} as names.
}
\description{
Construct period-over-period price relatives from information on prices and
products over time.
}
\examples{
price_relative(
  1:6,
  period = rep(1:2, each = 3),
  product = rep(letters[1:3], 2)
)
}
\seealso{
\code{\link[=back_period]{back_period()}} to get only the back price or base price.

\code{\link[=impute_prices]{impute_prices()}} to impute missing prices.

\code{\link[=outliers]{outliers()}} for methods to identify outliers with price relatives.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contrib.R
\name{contrib}
\alias{contrib}
\alias{contrib2DF}
\alias{contrib<-}
\alias{set_contrib}
\alias{set_contrib_from_index}
\title{Extract percent-change contributions}
\usage{
contrib(x, level = NULL, period = NULL, pad = 0)

contrib2DF(x, levels = NULL, period = NULL)

contrib(x, level = NULL, period = NULL) <- value

set_contrib(x, level = NULL, period = NULL, value)

set_contrib_from_index(x)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{level, levels}{\verb{[character]} The level of an index for which
percent-change contributions
are desired, defaulting to the first level (usually the top-level for an
aggregate index). \code{contrib2DF()} can accept multiple levels.}

\item{period}{\verb{[character]} The time periods for which percent-change
contributions are desired, defaulting to all time periods.}

\item{pad}{\verb{[numeric(1)]} A numeric value to pad contributions so that they
fit into a
rectangular array when products differ over time. The default is 0.}

\item{value}{\verb{[matrix]} A numeric matrix of replacement contributions with a
row for each product and a column for each time period. Recycling occurs
along time periods.}
}
\value{
\code{contrib()} returns a matrix of percent-change contributions with a column
for each \code{period} and a row for each product (sorted) for which there are
contributions in \code{level}. Contributions are padded with \code{pad} to fit into a
rectangular array when products differ over time. The replacement methods
returns a copy of \code{x} with contributions given by the matrix \code{value}.
(\code{set_contrib()} is an alias that's easier to use with pipes.)
\code{set_contrib_from_index()} is a helper to return a copy of \code{x} with all
contributions set to the corresponding index value minus 1.

\code{contrib2DF()} returns a data frame of contributions with four
columns: \code{period}, \code{level}, \code{product}, and \code{value}.
}
\description{
Extract a matrix or data frame of percent-change contributions from a price
index.
}
\examples{
prices <- data.frame(
  rel = 1:8,
  period = rep(1:2, each = 4),
  ea = rep(letters[1:2], 4)
)

index <- elementary_index(prices, rel ~ period + ea, contrib = TRUE)

pias <- aggregation_structure(
  list(c("top", "top", "top"), c("a", "b", "c")),
  weights = 1:3
)

index <- aggregate(index, pias, na.rm = TRUE)

# Percent-change contributions for the top-level index.
contrib(index)

contrib2DF(index)

# Calculate EA contributions for the chained index.
arithmetic_contributions <- function(x, w, order = 1) {
  (x - 1) * transmute_weights(x, w, order, to = 1)
}

arithmetic_contributions(
  as.matrix(chain(index))[c("a", "b", "c"), 2],
  weights(pias)
)
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/levels.piar_index.R
\name{levels.piar_index}
\alias{levels.piar_index}
\alias{levels<-.piar_index}
\alias{set_levels}
\title{Get the levels for a price index}
\usage{
\method{levels}{piar_index}(x)

\method{levels}{piar_index}(x) <- value

set_levels(x, value)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{value}{\verb{[character]} A character vector, or something that can be
coerced into one, giving the replacement levels for \code{x}.}
}
\value{
\code{levels()} returns a character vector with the levels for a price index.

The replacement method returns a copy of \code{x} with the levels in \code{value}.
(\code{set_levels()} is an alias that's easier to use with pipes.)
}
\description{
Methods to get and set the levels for a price index.
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale_weights.R
\name{scale_weights}
\alias{scale_weights}
\title{Scale weights}
\usage{
scale_weights(x)
}
\arguments{
\item{x}{\verb{[numeric >= 0]} A positive numeric vector with at least one
non-zero element.}
}
\value{
A numeric vector that sums to 1. If there are \code{NA}s in \code{x} then the result
sums 1 to if these values are removed.
}
\description{
Scale a vector of weights so that they sum to 1.
}
\examples{
scale_weights(1:5)

scale_weights(c(1:5, NA))
}
\seealso{
Other math functions:
\code{\link[=emean]{emean()}},
\code{\link[=gmean]{gmean()}},
\code{\link[=nested_gmean]{nested_gmean()}},
\code{\link[=transmute_weights]{transmute_weights()}},
\code{\link[=transmute_weights2]{transmute_weights2()}},
\code{\link[=update_weights]{update_weights()}}
}
\concept{math functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_aggregation_structure.R
\name{as_aggregation_structure}
\alias{as_aggregation_structure}
\alias{as_aggregation_structure.default}
\alias{as_aggregation_structure.data.frame}
\alias{as_aggregation_structure.matrix}
\title{Coerce to an aggregation structure}
\usage{
as_aggregation_structure(x, ...)

\method{as_aggregation_structure}{default}(x, ..., weights = NULL)

\method{as_aggregation_structure}{data.frame}(x, ...)

\method{as_aggregation_structure}{matrix}(x, ...)
}
\arguments{
\item{x}{\verb{[object]} An object to coerce into an aggregation structure.}

\item{...}{Further arguments passed to or used by methods.}

\item{weights}{\verb{[numeric >= 0]} A numeric vector of aggregation weights for
the elementary
aggregates. The default is to give each elementary aggregate the same
weight.}
}
\value{
A price index aggregation structure that inherits from
\code{\link{piar_aggregation_structure}}.
}
\description{
Coerce an object into an aggregation structure object.
}
\details{
The default method attempts to coerce \code{x} into a list prior to calling
\code{\link[=aggregation_structure]{aggregation_structure()}}.

The data frame and matrix methods treat \code{x} as a table with a row for
each elementary aggregate, a column of labels for each level in the
aggregation structure, and a column of weights for the elementary aggregates.
}
\examples{
# A simple aggregation structure
#            1
#      |-----+-----|
#      11          12
#  |---+---|       |
#  111     112     121
#  (1)     (3)     (4)
aggregation_weights <- data.frame(
  level1 = c("1", "1", "1"),
  level2 = c("11", "11", "12"),
  ea     = c("111", "112", "121"),
  weight = c(1, 3, 4)
)

pias <- aggregation_structure(
  aggregation_weights[1:3],
  weights = aggregation_weights[[4]]
)

all.equal(
  pias,
  as_aggregation_structure(aggregation_weights)
)

all.equal(
  pias,
  as_aggregation_structure(as.matrix(aggregation_weights))
)
}
\seealso{
\code{\link[=as.matrix.piar_aggregation_structure]{as.matrix()}} and
\code{\link[=as.data.frame.piar_aggregation_structure]{as.data.frame()}} for
coercing an aggregation structure into a tabular form.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transmute_weights.R
\name{transmute_weights}
\alias{transmute_weights}
\title{Transmute weights for a generalized mean}
\usage{
transmute_weights(x, weights = NULL, order = 0, to = 1, mean = NA)
}
\arguments{
\item{x}{\verb{[numeric > 0]} A strictly positive numeric vector.}

\item{weights}{\verb{[numeric >= 0]} A positive numeric vector of weights, the
same length as \code{x}. The default is to equally weight each element of \code{x}.}

\item{order}{\verb{[numeric(1)]} A finite number giving the order of the
generalized mean. The default transmutes the weights for a geometric mean.}

\item{to}{\verb{[numeric(1)]} A finite number giving the order of the target
generalized mean. The default computes weights for an arithmetic mean.}

\item{mean}{\verb{[numeric(1)]} A finite number giving the generalized mean of
\code{x} and \code{weights}, if known. The default computes this values.}
}
\value{
A numeric vector, the same length as \code{x}, that sums to 1.
}
\description{
Transmute weight to turn a generalized mean of a given order into a
generalized mean of any other order. Useful for calculating additive and
multiplicative decompositions for generalized-mean indexes.
See \code{vignette("decomposing-indexes")} for more details.
}
\details{
This function generalizes the additive and multiplicative decompositions
for arithmetic and geometric indexes by Balk (2008, Chapter 4). It returns
a value such that

\preformatted{gmean(x, w, r) == gmean(x, transmute_weights(x, w, r, s), s)}

Transmuting weights returns a value that is the same length as \code{x},
so any missing values in \code{x} or \code{weights} will return \code{NA}.
Unless all values are \code{NA}, however, the result will still satisfy
the above identity when \code{na.rm = TRUE}.
}
\examples{
x <- 1:3
w <- 3:1

# Calculate the geometric mean as an arithmetic mean.
gmean(x, order = 0)
gmean(x, transmute_weights(x, order = 0, to = 1), order = 1)
}
\references{
Balk, B. M. (2008). \emph{Price and Quantity Index Numbers}.
Cambridge University Press.
}
\seealso{
Other math functions:
\code{\link[=emean]{emean()}},
\code{\link[=gmean]{gmean()}},
\code{\link[=nested_gmean]{nested_gmean()}},
\code{\link[=scale_weights]{scale_weights()}},
\code{\link[=transmute_weights2]{transmute_weights2()}},
\code{\link[=update_weights]{update_weights()}}
}
\concept{math functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.na.piar_index.R
\name{is.na.piar_index}
\alias{is.na.piar_index}
\alias{anyNA.piar_index}
\title{Missing values in a price index}
\usage{
\method{is.na}{piar_index}(x)

\method{anyNA}{piar_index}(x, recursive = FALSE)
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{recursive}{\verb{[logical(1)]} Check if \code{x} also has missing percent-change
contributions. By default only index values are checked for missingness.}
}
\value{
\code{is.na()} returns a logical matrix, with a row for each level of \code{x} and a
columns for each time period, that indicates which index values are missing.

\code{anyNA()} returns \code{TRUE} if any index values are missing, or percent-change
contributions (if \code{recursive = TRUE}).
}
\description{
Identify missing values in a price index.
}
\examples{
index <- as_index(matrix(c(1, 2, 3, NA, 5, NA), 2))

anyNA(index)
is.na(index)

# Carry forward imputation.
index[is.na(index)] <- 1
index
}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}},
\code{\link[=window.piar_index]{window.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_index.R
\name{as_index}
\alias{as_index}
\alias{as_index.default}
\alias{as_index.matrix}
\alias{as_index.data.frame}
\alias{as_index.chainable_piar_index}
\alias{as_index.direct_piar_index}
\alias{as_index.mts}
\title{Coerce to a price index}
\usage{
as_index(x, ...)

\method{as_index}{default}(x, ...)

\method{as_index}{matrix}(x, ..., chainable = TRUE, contrib = FALSE)

\method{as_index}{data.frame}(x, ..., contrib = FALSE)

\method{as_index}{chainable_piar_index}(x, ..., chainable = TRUE)

\method{as_index}{direct_piar_index}(x, ..., chainable = FALSE)

\method{as_index}{mts}(x, ...)
}
\arguments{
\item{x}{\verb{[object]} An object to coerce into a price index.}

\item{...}{Further arguments passed to or used by methods.}

\item{chainable}{\verb{[logical(1)]} Are the index values in \code{x}
period-over-period
indexes, suitable for a chained calculation (the default)? This should be
\code{FALSE} when \code{x} contains fixed-base (direct) index values.}

\item{contrib}{\verb{[logical(1)]} Should the index values in \code{x} be used to
construct percent-change contributions? The default does not make
contributions.}
}
\value{
A price index that inherits from
\code{\link{piar_index}}. If \code{chainable = TRUE} then this is a
period-over-period price index that also inherits from
\code{\link{chainable_piar_index}}; otherwise, it is a fixed-base index that
inherits from \code{\link{direct_piar_index}}.
}
\description{
Coerce pre-computed index values into an index object.
}
\details{
Numeric matrices are coerced into an index object by treating each column as
a separate time period, and each row as a separate level of the index (e.g.,
an elementary aggregate). Column names
are used to denote time periods, and row names are used to denote levels
(so they must be unique). This essentially reverses calling
\code{\link[=as.matrix.piar_index]{as.matrix()}} on an index object. If a
dimension is unnamed, then it is given a sequential label from 1 to the size
of that dimension. The default and multiple time series methods coerces \code{x}
to a matrix prior to using the matrix method.

The data frame method for \code{as_index()} is best understood as reversing
the effect of \code{\link[=as.data.frame.piar_index]{as.data.frame()}} on an
index object. It constructs a matrix by taking the levels of
\code{x[[1]]} as columns and the levels of \code{x[[2]]} as rows
(coercing to a factor if necessary). It then populates this matrix with the
corresponding values in \code{x[[3]]}, and uses the matrix method for
\code{as_index()}. If \code{contrib = TRUE} and there is a fourth list column of
product contributions then these are also included in the resulting index.

If \code{x} is a period-over-period index then it is returned unchanged when
\code{chainable = TRUE} and chained otherwise. Similarly, if \code{x} is a
fixed-base index then it is returned unchanged when
\code{chainable = FALSE} and unchain otherwise.
}
\examples{
prices <- data.frame(
  rel = 1:8,
  period = rep(1:2, each = 4),
  ea = rep(letters[1:2], 4)
)

index <- elementary_index(prices, rel ~ period + ea)

all.equal(as_index(as.data.frame(index)), index)
all.equal(as_index(as.matrix(index)), index)
}
\seealso{
\code{\link[=as.matrix.piar_index]{as.matrix()}} and
\code{\link[=as.data.frame.piar_index]{as.data.frame()}} for coercing an index
into a tabular form.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index-class.R
\name{is_index}
\alias{is_index}
\alias{is_chainable_index}
\alias{is_direct_index}
\title{Test if an object is a price index}
\usage{
is_index(x)

is_chainable_index(x)

is_direct_index(x)
}
\arguments{
\item{x}{\verb{[object]} An object to test.}
}
\value{
\code{is_index()} returns \code{TRUE} if \code{x} inherits from \code{\link{piar_index}}.

\code{is_chainable_index()} returns \code{TRUE} if \code{x} inherits from
\code{\link{chainable_piar_index}}.

\code{is_direct_index()} returns \code{TRUE} if \code{x} inherits from
\code{\link{direct_piar_index}}.
}
\description{
Test if an object is a index object or a subclass of an index object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update_weights.R
\name{update_weights}
\alias{update_weights}
\title{Update weights}
\usage{
update_weights(x, weights = NULL, order = 1)
}
\arguments{
\item{x}{\verb{[numeric > 0]} A strictly positive numeric vector.}

\item{weights}{\verb{[numeric >= 0]} A positive numeric vector of weights, the
same length as \code{x}. The default is to equally weight each element of \code{x}.}

\item{order}{\verb{[numeric(1)]} A finite number giving the order (or exponent) of
the generalized mean. The default calculates an arithmetic mean.}
}
\value{
A numeric vector the same length as \code{x}.
}
\description{
Factor weights to turn the generalized mean of a product into the product
of generalized means. Useful for price-updating the weights in a
generalized-mean index.
}
\details{
This function generalizes the result in section C.5 of Chapter 9 of the PPI
Manual for chaining the Young index, and gives a way to chain
generalized-mean price indexes over time. It returns a value such that

\preformatted{gmean(x * y, w) ==
    gmean(x, w) * gmean(y, update_weights(x, w))}

Factoring weights returns a value that is the same length as \code{x},
so any missing values in \code{x} or \code{weights} will return \code{NA}.
Unless all values are \code{NA}, however, the result will still satisfy
the above identity when \code{na.rm = TRUE}.
}
\examples{
x <- 1:3
y <- 4:6
w <- 3:1

# Factor the arithmetic mean by chaining the calculation.
gmean(x * y, w)
gmean(x, w) * gmean(y, update_weights(x, w))

# In cases where x and y have the same order, Chebyshev's
# inequality implies that the chained calculation is too small.
gmean(x * y, w) > gmean(x, w) * gmean(y, w)
}
\references{
ILO, IMF, OECD, UNECE, and World Bank. (2004).
\emph{Producer Price Index Manual: Theory and Practice}.
International Monetary Fund.
}
\seealso{
Other math functions:
\code{\link[=emean]{emean()}},
\code{\link[=gmean]{gmean()}},
\code{\link[=nested_gmean]{nested_gmean()}},
\code{\link[=scale_weights]{scale_weights()}},
\code{\link[=transmute_weights]{transmute_weights()}},
\code{\link[=transmute_weights2]{transmute_weights2()}}
}
\concept{math functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/window.piar_index.R
\name{window.piar_index}
\alias{window.piar_index}
\alias{window<-.piar_index}
\title{Window a price index}
\usage{
\method{window}{piar_index}(x, start = NULL, end = NULL, ...)

\method{window}{piar_index}(x, start = NULL, end = NULL, ...) <- value
}
\arguments{
\item{x}{\verb{[piar_index]} A price index, as made by,
e.g., \code{\link[=elementary_index]{elementary_index()}}.}

\item{start}{\verb{[character(1)]} The time period to start the window.
The default in the first period of \code{x}.}

\item{end}{\verb{[character(1)]} The time period to end the window. The default
is the last period of \code{x}.}

\item{...}{Not currently used.}

\item{value}{\verb{[numeric > 0 | piar_index]} A numeric vector or price index
of replacement values.}
}
\value{
\code{window()} extracts a price index over a window of time periods that
inherits from the same class as \code{x}. The replacement method replaces these
with \code{value}.
}
\description{
Extract and replace index values over a window of time periods.
}
\examples{
x <- as_index(matrix(1:9, 3))

window(x, "2")

window(x, "2") <- 1
x

}
\seealso{
Other index methods:
\code{\link[=[.piar_index]{[.piar_index()}},
\code{\link{aggregate.piar_index}},
\code{\link[=as.data.frame.piar_index]{as.data.frame.piar_index()}},
\code{\link[=as.ts.piar_index]{as.ts.piar_index()}},
\code{\link[=chain]{chain()}},
\code{\link[=contrib]{contrib()}},
\code{\link[=head.piar_index]{head.piar_index()}},
\code{\link[=is.na.piar_index]{is.na.piar_index()}},
\code{\link[=levels.piar_index]{levels.piar_index()}},
\code{\link{mean.piar_index}},
\code{\link[=merge.piar_index]{merge.piar_index()}},
\code{\link[=split.piar_index]{split.piar_index()}},
\code{\link[=stack.piar_index]{stack.piar_index()}},
\code{\link[=time.piar_index]{time.piar_index()}}
}
\concept{index methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split_classification.R
\name{split_classification}
\alias{split_classification}
\title{Split a hierarchical classification}
\usage{
split_classification(x, split, ..., sep = ".", pad = NA)
}
\arguments{
\item{x}{\verb{[character]} A character vector, or something that can be coerced
into one, of codes/labels for a specific level in a classification
(e.g., 5-digit COICOP).}

\item{split}{\verb{[character(1)]} A regular expression to delineate and split the
levels in \code{x}. See \code{\link[=strsplit]{strsplit()}}.}

\item{...}{Additional argument to pass to \code{\link[=strsplit]{strsplit()}}.}

\item{sep}{\verb{[character(1)]} A character used to delineate levels in \code{x} in
the result. The default separates levels by \code{"."}.}

\item{pad}{\verb{[character(1)]} A string used to pad the shorter labels for an
unbalanced classification. The default pads with \code{NA}.}
}
\value{
A list with a entry for each level in \code{x} giving the "digits" that
represent each level in the hierarchy.
}
\description{
Expand a character representation of a hierarchical classification to make a
price index aggregation structure by splitting along a delimiter.
}
\examples{
#' # A simple classification structure.
#            1
#      |-----+-----|
#      11          12
#  |---+---|       |
#  111     112     121

split_classification(c("111", "112", "121"), "")

# Useful if there are delimiters in the classification (like COICOP).
split_classification(c("01.1.1", "01.1.2", "01.2.1"), ".", fixed = TRUE)
}
\seealso{
\code{\link[=aggregation_structure]{aggregation_structure()}} to make a price-index aggregation structure.

\code{\link[=expand_classification]{expand_classification()}} to expand a classification by the width of the
levels.

\code{\link[=combine_classifications]{combine_classifications()}} for combining multiple hierarchical
classifications.
}
